//=============================================================================
//
//		Copyright (C) 1995,1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// DirArray.m
//
//		An extensible array of directory entries.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id$
// $Log$
//-----------------------------------------------------------------------------
#import "DirArray.h"
#import <AppKit/AppKit.h>
#import <assert.h>
#import <libc.h>


@implementation DirArray

//-----------------------------------------------------------------------------
// dirEntryCons
//-----------------------------------------------------------------------------
- (void) dirEntryCons: (DirEntry*) de
	name: (NSString*) n
	path: (NSString*) path
	{
	int rc;
	char const* ppath = [path cString];
	memset( de, 0, sizeof(*de) );

	de->dir = self;
	de->name = [n copy];

	if ((rc = lstat( ppath, &(de->st) )) == 0)
		{
		if ((de->st.st_mode & S_IFMT) == S_IFLNK)		// soft-link?
			{
			char symbuf[ FILENAME_MAX + 1 ];
			int cc;
			cc = readlink( ppath, symbuf, FILENAME_MAX );
			if (cc >= 0)
				{
				symbuf[ cc ] = '\0';
				de->soft_link = [[NSString stringWithCString:symbuf] retain];
				rc = stat( ppath, &(de->st) );
				}
			else
				rc = -1;
			}
		}

	if (rc != 0)
		de->err = errno;
	}


//-----------------------------------------------------------------------------
// dirEntryDestroy:
//-----------------------------------------------------------------------------
- (void) dirEntryDestroy:(DirEntry*)de
	{
	[de->name release];
	if (de->image != 0)
		[de->image release];
	if (de->soft_link != 0)
		[de->soft_link release];
	}


//-----------------------------------------------------------------------------
// entryAt:
//-----------------------------------------------------------------------------
- (DirEntry const*)entryAt:(int)n
	{
	assert( 0 <=n && n < num_entries );
	return (entries + n);
	}


//-----------------------------------------------------------------------------
// isDir:
//-----------------------------------------------------------------------------
- (BOOL) isDir:(DirEntry const*)de
	{
	return ((de->st.st_mode & S_IFMT) == S_IFDIR);
	}


//-----------------------------------------------------------------------------
// isDirAt:
//-----------------------------------------------------------------------------
- (BOOL) isDirAt:(int)n
	{
	return [self isDir:[self entryAt:n]];
	}


//-----------------------------------------------------------------------------
// loadImage:
//
//		NOTE *1*: -iconForFile: under OpenStep 4.0 for Mach PR2 no longer 
//				returns the icon for the executable inside of application 
//				wrappers (.app).  Instead if some application associates an
//				icon with the extension ".app" then that icon is returned.
//				In PR2, MallocDebug does just this with the result that all
//				".app" directories display the exact *same* icon (that of
//				MallocDebug), so browsing a directory such as /NextApps is
//				pretty pointless (so far as the iconic display goes).  We hack
//				around the problem by explicitly checking for the ".app"
//				extension and fetching the icon of the executable inside the
//				application wrapper.
//-----------------------------------------------------------------------------
- (void) loadImage:(DirEntry*)de
	{
	NSString* s = [de->dir->name stringByAppendingPathComponent:de->name];
	if ([[s pathExtension] isEqualToString:@"app"])				// NOTE *1*
		s = [s stringByAppendingPathComponent:
				[[s lastPathComponent] stringByDeletingPathExtension]];
	de->image = [[[NSWorkspace sharedWorkspace] iconForFile:s] retain];
	[de->image setScalesWhenResized:YES];
	}


//-----------------------------------------------------------------------------
// getImage:
//-----------------------------------------------------------------------------
- (NSImage*) getImage:(DirEntry const*)de
	{
	if (de->image == 0)
		[self loadImage:(DirEntry*)de];
	return de->image;
	}


//-----------------------------------------------------------------------------
// getImageAt:
//-----------------------------------------------------------------------------
- (NSImage*) getImageAt:(int)n
	{
	return [self getImage:[self entryAt:n]];
	}


//-----------------------------------------------------------------------------
// init
//-----------------------------------------------------------------------------
- init
	{
	[super init];
	name = 0;
	total_bytes = 0;
	num_entries = 0;
	max_entries = 0;
	entries = 0;
	return self;
	}


//-----------------------------------------------------------------------------
// empty
//-----------------------------------------------------------------------------
- (void) empty
	{
	DirEntry* p;
	for (p = entries + num_entries; p != entries; )
		[self dirEntryDestroy:--p];
	num_entries = 0;
	total_bytes = 0;
	}


//-----------------------------------------------------------------------------
// dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
	[self empty];
	[name release];
	if (entries != 0)
		NSZoneFree( [self zone], entries );
	[super dealloc];
	}


//-----------------------------------------------------------------------------
// count
//-----------------------------------------------------------------------------
- (int) count
	{
	return num_entries;
	}


//-----------------------------------------------------------------------------
// totalBytes
//-----------------------------------------------------------------------------
- (int) totalBytes
	{
	return total_bytes;
	}


//-----------------------------------------------------------------------------
// remove:
//-----------------------------------------------------------------------------
- (void) remove:(int)n
	{
	if (0 <= n && n < num_entries)
		{
		DirEntry* p = entries + n;
		total_bytes -= p->st.st_size;
		num_entries--;
		[self dirEntryDestroy:p];
		if (n < num_entries)
			memmove( p, p + 1, (num_entries - n) * sizeof(*p) );
		}
	}


//-----------------------------------------------------------------------------
// expand
//-----------------------------------------------------------------------------
- (void) expand
	{
	DirEntry* p = entries;
	int N = max_entries;
	if (N == 0)
		{
		N = 16;
		p = (DirEntry*) NSZoneMalloc( [self zone], N * sizeof(*p) );
		}
	else
		{
		N += N;
		p = (DirEntry*) NSZoneRealloc( [self zone], p, N * sizeof(*p) );
		}
	max_entries = N;
	entries = p;
	}


//-----------------------------------------------------------------------------
// addName:path:
//-----------------------------------------------------------------------------
- (void) addName:(NSString*)n path:(NSString*)fullpath
	{
	DirEntry* p;
	if (num_entries >= max_entries)
		[self expand];
	p = entries + num_entries++;
	[self dirEntryCons:p name:n path:fullpath];
	total_bytes += p->st.st_size;
	}


//-----------------------------------------------------------------------------
// loadPath:showHidden:
//-----------------------------------------------------------------------------
- (int) loadPath:(NSString*)path showHidden:(BOOL)showHidden
	{
	int rc = 0;
	DIR* dirp;

	[self empty];
	if (name != 0)
		[name autorelease];
	name = [path copy];

	if ((dirp = opendir( [path cString] )) != 0)
		{
		struct direct const* dp;

		while ((dp = readdir(dirp)) != 0)
			{			// Do not include the "." (self) entry.
			int const len = dp->d_namlen;
			char const* const nam = dp->d_name;
			BOOL const dot_name = (nam[0] == '.');
			if ((len > 1 || !dot_name) &&		// exclude "."
				(showHidden || !dot_name ||
				(len == 2 && nam[1] == '.')))	// include ".."
				{
				NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
				NSString* file = [NSString stringWithCString:nam length:len];
				NSString* s = [path stringByAppendingPathComponent:file];
				[self addName:file path:s];
				[pool release];
				}
			}

		closedir( dirp );
		}
	else
		{
		rc = errno;
		}

	return rc;
	}

@end
