//=============================================================================
//
//		Copyright (C) 1995,1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// DirWindow.m
//
//		Manages window which displays directory listing.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id$
// $Log$
//-----------------------------------------------------------------------------
#import "DirWindow.h"
#import "NameCache.h"
#import "Defaults.h"
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableCell.h>
#import <AppKit/NSButton.h>
#import <AppKit/NSFont.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSNibLoading.h>
#import <AppKit/NSPanel.h>
#import <AppKit/NSPasteboard.h>
#import <AppKit/NSTextField.h>
#import <AppKit/NSWorkspace.h>
#import <assert.h>
#import <errno.h>

#ifdef _POSIX_SOURCE
# define GET_CURR_DIR(_BF_,_SZ_)		getcwd(_BF_,_SZ_)
#else
# define GET_CURR_DIR(_BF_,_SZ_)		getwd(_BF_)
#endif

#define USE_PRIVATE_ZONE		1

#define CASCADE_MAX 10
static int cascadeCounter = 0;
static float cascadeOriginX;
static float cascadeOriginY;
static float cascadeDeltaX;
static float cascadeDeltaY;

static BOOL defaultAutoSort;
static BOOL defaultShowHidden;
static BOOL defaultHighlightDirs;
static NSColor* defaultColor;
static NSSize defaultWinSize;
static NSFont* defaultFont;
static NSMutableArray* openDirs = 0;
static OwnerCache* owners = 0;
static GroupCache* groups = 0;

#define COLOR_DEF		@"DirColor"
#define SIZE_DEF		@"DirSize"
#define FONT_DEF		@"DirFont"
#define SORT_DEF		@"AutoSort"
#define HIDDEN_DEF		@"ShowHidden"
#define HLIGHT_DEF		@"HighlightDirs"
#define COL_SIZES_DEF	@"ColSizes"
#define COL_ORDER_DEF	@"ColOrder"


//-----------------------------------------------------------------------------
// normalize_path
//-----------------------------------------------------------------------------
static inline NSString* normalize_path( NSString* buff )
	{
	return [buff stringByStandardizingPath];
	}


//-----------------------------------------------------------------------------
// str_int
//-----------------------------------------------------------------------------
static NSString* str_int( int n )
	{
	return [[NSNumber numberWithInt:n] stringValue];
	}


//-----------------------------------------------------------------------------
// str_date
//-----------------------------------------------------------------------------
static NSString* str_date( time_t t )
	{
	NSCalendarDate* d = [NSCalendarDate dateWithTimeIntervalSince1970:t];
	return [d descriptionWithCalendarFormat:@"%m/%d/%y %H:%M"];
	}


//-----------------------------------------------------------------------------
// str_perms
//-----------------------------------------------------------------------------
static NSString* str_perms( unsigned int mode )
	{
	char buff[ 16 ];
	char* s;
	int ftype;
	int c;

	s = buff;

	ftype = (mode & S_IFMT);
	switch (ftype)
		{
		case S_IFDIR:	c = 'd'; break;
		case S_IFCHR:	c = 'c'; break;
		case S_IFBLK:	c = 'b'; break;
		case S_IFREG:	c = '-'; break;
		case S_IFLNK:	c = 'l'; break;
		case S_IFSOCK:	c = 's'; break;
		default:		c = '-'; break;
		}
	*s++ = c;
	*s++ = (mode & 0400) ? 'r' : '-';
	*s++ = (mode & 0200) ? 'w' : '-';
	*s++ = (mode & S_ISUID) ? 's' : (mode & 0100) ? 'x' : '-';
	*s++ = (mode & 0040) ? 'r' : '-';
	*s++ = (mode & 0020) ? 'w' : '-';
	*s++ = (mode & S_ISGID) ? 's' : (mode & 0010) ? 'x' : '-';
	*s++ = (mode & 0004) ? 'r' : '-';
	*s++ = (mode & 0002) ? 'w' : '-';
	*s++ = (mode & S_ISVTX) ? 't' : (mode & 0001) ? 'x' : '-';
	*s = '\0';

	return [NSString stringWithCString:buff];
	}


//-----------------------------------------------------------------------------
// fmt_icon
//-----------------------------------------------------------------------------
static void fmt_icon( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	float h,w,s;
	NSSize sz;
	NSImage* img;

	w = [ts colSize:ICON_SLOT]; if (w == 0) w = 18;
	h = [ts uniformSizeRows];	if (h == 0) h = 18;
	s = (w < h ? w : h) - 1.0;
	sz.width = s;
	sz.height = s;

	img = [de->dir getImage:de];
	[img setSize:sz];
	[cell setImage:img];
	}


//-----------------------------------------------------------------------------
// fmt_name
//-----------------------------------------------------------------------------
static void fmt_name( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setStringValue:de->name];
	}


//-----------------------------------------------------------------------------
// fmt_size
//-----------------------------------------------------------------------------
static void fmt_size( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setAlignment:NSRightTextAlignment];
	[cell setIntValue:de->st.st_size];
	}


//-----------------------------------------------------------------------------
// fmt_modified
//-----------------------------------------------------------------------------
static void fmt_modified( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	int const t = (int) de->st.st_mtime;
	[cell setStringValue:str_date( t )];
	[cell setTag:t];
	}


//-----------------------------------------------------------------------------
// fmt_perms
//-----------------------------------------------------------------------------
static void fmt_perms( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	unsigned int const mode = de->st.st_mode;
	[cell setStringValue:str_perms(mode)];
	[cell setTag:mode];
	}


//-----------------------------------------------------------------------------
// fmt_owner
//-----------------------------------------------------------------------------
static void fmt_owner( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setStringValue:[owners lookup:de->st.st_uid]];
	}


//-----------------------------------------------------------------------------
// fmt_group
//-----------------------------------------------------------------------------
static void fmt_group( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setStringValue:[groups lookup:de->st.st_gid]];
	}


//-----------------------------------------------------------------------------
// fmt_hardlinks
//-----------------------------------------------------------------------------
static void fmt_hardlinks( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setIntValue:de->st.st_nlink];
	}


//-----------------------------------------------------------------------------
// fmt_softlink
//-----------------------------------------------------------------------------
static void fmt_softlink( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	NSString* s = de->soft_link;
	[cell setStringValue:(s ? s : @"")];
	}


//-----------------------------------------------------------------------------
// fmt_message
//-----------------------------------------------------------------------------
static void fmt_message( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setStringValue:[NSString stringWithCString:strerror(de->err)]];
	}


//-----------------------------------------------------------------------------
// fmt_blank
//-----------------------------------------------------------------------------
static void fmt_blank( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setStringValue:@""];
	}


//-----------------------------------------------------------------------------
// fmt_unknown
//-----------------------------------------------------------------------------
static void fmt_unknown( MiscTableScroll* ts, DirEntry const* de, id cell )
	{
	[cell setStringValue:@"?"];
	}



//-----------------------------------------------------------------------------
// FORMAT_FUNC
//-----------------------------------------------------------------------------

typedef void (*FormatFunc)( MiscTableScroll* ts, DirEntry const* de, id cell );

static FormatFunc FORMAT_FUNC[ MAX_SLOT ] =
		{
		fmt_icon,				// ICON_SLOT,
		fmt_name,				// NAME_SLOT,
		fmt_size,				// SIZE_SLOT,
		fmt_modified,			// MODIFIED_SLOT,
		fmt_perms,				// PERMS_SLOT,
		fmt_owner,				// OWNER_SLOT,
		fmt_group,				// GROUP_SLOT,
		fmt_hardlinks,			// HARDLINKS_SLOT,
		fmt_softlink,			// SOFTLINK_SLOT,
		fmt_blank,				// MESSAGE_SLOT
		};

static FormatFunc FORMAT_ERR_FUNC[ MAX_SLOT ] =
		{
		fmt_icon,				// ICON_SLOT,
		fmt_name,				// NAME_SLOT,
		fmt_unknown,			// SIZE_SLOT,
		fmt_unknown,			// MODIFIED_SLOT,
		fmt_unknown,			// PERMS_SLOT,
		fmt_unknown,			// OWNER_SLOT,
		fmt_unknown,			// GROUP_SLOT,
		fmt_unknown,			// HARDLINKS_SLOT,
		fmt_softlink,			// SOFTLINK_SLOT,
		fmt_message,			// MESSAGE_SLOT
		};


//-----------------------------------------------------------------------------
// format_cell
//-----------------------------------------------------------------------------
static inline void format_cell(
		MiscTableScroll* ts,
		DirEntry const* de,
		id cell,
		unsigned int col )
	{
	if (de->err == 0)
		(*(FORMAT_FUNC[col]))( ts, de, cell );
	else
		(*(FORMAT_ERR_FUNC[col]))( ts, de, cell );
	}


//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation DirWindow

//-----------------------------------------------------------------------------
// + initCascader
//-----------------------------------------------------------------------------
+ (void) initCascader
	{
	NSSize screenSize;
	screenSize = [[NSScreen mainScreen] frame].size;
	cascadeOriginX = screenSize.width / 4;
	cascadeOriginY = screenSize.height - 20;
	cascadeDeltaX = 20;
	cascadeDeltaY = 20;
	}


//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void) initialize
	{
	if (self == [DirWindow class])
		{
		[self initCascader];
		openDirs = [[NSMutableArray alloc] init];
		owners = [OwnerCache commonInstance];
		groups = [GroupCache commonInstance];
		defaultColor = [[Defaults getColor:COLOR_DEF
						fallback:[NSColor lightGrayColor]] retain];
		defaultAutoSort = [Defaults getBool:SORT_DEF fallback:NO];
		defaultShowHidden = [Defaults getBool:HIDDEN_DEF fallback:NO];
		defaultHighlightDirs = [Defaults getBool:HLIGHT_DEF fallback:NO];
		}
	}


//-----------------------------------------------------------------------------
// - cascade
//-----------------------------------------------------------------------------
- (void) cascade
	{
	float top,left;
	left = cascadeOriginX + (cascadeDeltaX * cascadeCounter);
	top	 = cascadeOriginY - (cascadeDeltaY * cascadeCounter);
	[window setFrameTopLeftPoint:NSMakePoint(left, top)];
	if (++cascadeCounter >= CASCADE_MAX)
		cascadeCounter = 0;
	}


//-----------------------------------------------------------------------------
// - updateButtons
//-----------------------------------------------------------------------------
- (void) updateButtons
	{
	BOOL const enable = [scroll numSelectedRows] == 1 &&
				  [dir_array isDirAt:[scroll selectedRow]];
	if (enable != [cdButton isEnabled])
		[cdButton setEnabled:enable];
	}


//-----------------------------------------------------------------------------
// - tableScroll:cellAtRow:column:
//-----------------------------------------------------------------------------
- (id) tableScroll:(MiscTableScroll*)ts cellAtRow:(int)row column:(int)col
	{
	id cell = lazy_row[col];
	DirEntry const* const de = [dir_array entryAt:row];
	format_cell( ts, de, cell, col );
	if ([cell respondsToSelector:@selector(setBackgroundColor:)])
		{
		if (highlightDirs && [dir_array isDir:de])
			[cell setBackgroundColor:[NSColor cyanColor]];
		else
			[cell setBackgroundColor:[ts backgroundColor]];
		}
	return cell;
	}


//-----------------------------------------------------------------------------
// - tableScroll:stringValueAtRow:column:
//-----------------------------------------------------------------------------
- (NSString*) tableScroll:(MiscTableScroll*)ts
	stringValueAtRow:(int)r column:(int)c
	{
	NSString* rc = 0;
	if (r < [dir_array count])
		{
		DirEntry const* const de = [dir_array entryAt:r];
		if (de->err == 0)
			{
			switch (c)
				{
			case SIZE_SLOT:		rc = str_int(de->st.st_size);			break;
			case MODIFIED_SLOT: rc = str_date(de->st.st_mtime);			break;
			case PERMS_SLOT:	rc = str_perms(de->st.st_mode);			break;
			case OWNER_SLOT:	rc = [owners lookup:de->st.st_uid];		break;
			case GROUP_SLOT:	rc = [groups lookup:de->st.st_gid];		break;
			case HARDLINKS_SLOT:rc = str_int(de->st.st_nlink);			break;
			case NAME_SLOT:		rc = de->name;							break;
			case SOFTLINK_SLOT: rc = de->soft_link;						break;
			case MESSAGE_SLOT:
			case ICON_SLOT:
			default:			rc = @"";								break;
				}
			}
		else
			{
			switch (c)
				{
			case NAME_SLOT:		rc = de->name;							break;
			case SOFTLINK_SLOT: rc = de->soft_link;						break;
			case MESSAGE_SLOT:	rc = [NSString stringWithCString:
												strerror(de->err)];		break;
			case SIZE_SLOT:
			case MODIFIED_SLOT:
			case PERMS_SLOT:
			case OWNER_SLOT:
			case GROUP_SLOT:
			case HARDLINKS_SLOT:
			case ICON_SLOT:
			default:			rc = @"?";								break;
				}
			}
		}
	return (rc ? rc : @"");
	}


//-----------------------------------------------------------------------------
// -intValueAtRow:column:
//-----------------------------------------------------------------------------
- (int) intValueAtRow:(int)r column:(int)c
	{
	int rc = 0;
	if (r < [dir_array count])
		{
		DirEntry const* const de = [dir_array entryAt:r];
		if (de->err == 0)
			{
			switch (c)
				{
			case SIZE_SLOT:				rc = de->st.st_size;	break;
			case MODIFIED_SLOT:			rc = de->st.st_mtime;	break;
			case PERMS_SLOT:			rc = de->st.st_mode;	break;
			case OWNER_SLOT:			rc = de->st.st_uid;		break;
			case GROUP_SLOT:			rc = de->st.st_gid;		break;
			case HARDLINKS_SLOT:		rc = de->st.st_nlink;	break;
			case ICON_SLOT:
			case NAME_SLOT:
			case SOFTLINK_SLOT:
			case MESSAGE_SLOT:
			default:											break;
				}
			}
		}
	return rc;
	}


//-----------------------------------------------------------------------------
// - tableScroll:intValueAtRow:column:
//-----------------------------------------------------------------------------
- (int) tableScroll:(MiscTableScroll*)ts intValueAtRow:(int)r column:(int)c
	{
	return [self intValueAtRow:r column:c];
	}


//-----------------------------------------------------------------------------
// - tableScroll:tagAtRow:column:
//-----------------------------------------------------------------------------
- (int) tableScroll:(MiscTableScroll*)ts tagAtRow:(int)r column:(int)c
	{
	return [self intValueAtRow:r column:c];
	}


//-----------------------------------------------------------------------------
// - tableScroll:fontChangedFrom:to:
//-----------------------------------------------------------------------------
- (void) tableScroll:(MiscTableScroll*)ts
		fontChangedFrom:(NSFont*)oldFont
		to:(NSFont*)newFont
	{
	int col;
	for (col = 0; col < MAX_SLOT; col++)
		[lazy_row[col] setFont:newFont];

	[defaultFont autorelease];
	defaultFont = [newFont retain];
	[Defaults set:FONT_DEF font:defaultFont];
	}


//-----------------------------------------------------------------------------
// - tableScroll:border:slotResized:
//-----------------------------------------------------------------------------
- (void) tableScroll:(MiscTableScroll*)ts
		border:(MiscBorderType)b
		slotResized:(int)n
	{
	NSString* s;
	assert( b == MISC_COL_BORDER );
	s = [ts colSizesAsString];
	[Defaults set:COL_SIZES_DEF str:s];
	}


//-----------------------------------------------------------------------------
// - tableScroll:border:slotDraggedFrom:to:
//-----------------------------------------------------------------------------
- (void) tableScroll:(MiscTableScroll*)ts
		border:(MiscBorderType)b
		slotDraggedFrom:(int)fromPos
		to:(int)toPos;
	{
	NSString* s;
	assert( b == MISC_COL_BORDER );
	s = [ts colOrderAsString];
	[Defaults set:COL_ORDER_DEF str:s];
	}


//-----------------------------------------------------------------------------
// - fillScroll
//-----------------------------------------------------------------------------
- (void) fillScroll
	{
	NSString* buff;

	[scroll renewRows: [dir_array count]];

	if ([scroll autoSortRows])
		[scroll sortRows];

	buff = [NSString stringWithFormat:@"%d files   %d bytes",
						[dir_array count], [dir_array totalBytes]];
	[countField setStringValue:buff];

	[self updateButtons];
	}


//-----------------------------------------------------------------------------
// - setPath:
//-----------------------------------------------------------------------------
- (void) setPath:(NSString*)dirname
	{
	[path autorelease];
	if (dirname == 0) dirname = NSHomeDirectory();
	if ([dirname length] == 0) dirname = @"/";
	path = [dirname retain];
	[window setTitle:path];
	}


//-----------------------------------------------------------------------------
// - loadDirectory:
//-----------------------------------------------------------------------------
- (void) loadDirectory:(NSString*)dirname
	{
	int rc = [dir_array loadPath:dirname showHidden:showHidden];
	if (rc != 0)
		{
		NSRunAlertPanel( @"Can't Open", @"Cannot open directory, %@\n%d:%s",
								@"OK", nil, nil, path, rc, strerror(rc));
		}
	}


//-----------------------------------------------------------------------------
// - reload
//-----------------------------------------------------------------------------
- (void) reload
	{
	[self loadDirectory:path];
	[self fillScroll];
	}


//-----------------------------------------------------------------------------
// - load:
//-----------------------------------------------------------------------------
- (void) load:(NSString*)dirname
	{
	[self setPath:dirname];
	[self reload];
	}


//-----------------------------------------------------------------------------
// - open:
//-----------------------------------------------------------------------------
- (void) open:(id)sender
	{
	if ([scroll hasRowSelection])
		{
		int i;
		NSArray* list = [scroll selectedRows];
		for (i = [list count]; i-- > 0; )
			{
			int const row = [[list objectAtIndex:i] intValue];
			DirEntry const* de = [dir_array entryAt:row];

			NSString* s = [path stringByAppendingPathComponent:de->name];
			if ([dir_array isDir:de])
				[[self class] launchDir:s];
			else
				[[NSWorkspace sharedWorkspace] openFile:s];
			}
		}
	}


//-----------------------------------------------------------------------------
// - destroy:
//-----------------------------------------------------------------------------
- (void) destroy:(id)sender
	{
	if ([scroll hasRowSelection])
		{
		int i;
		NSArray* list = [scroll selectedRows];
		NSMutableArray* file_list = [NSMutableArray array];

		for (i = [list count]; i-- > 0; )
			{
			int const row = [[list objectAtIndex:i] intValue];
			DirEntry const* de = [dir_array entryAt:row];
			NSString* s = de->name;
			if (s != 0)
				[file_list addObject:s];
			}

		[[NSWorkspace sharedWorkspace]
				performFileOperation:NSWorkspaceDestroyOperation
				source:path destination:@"" files:file_list tag:0];

		[self reload];
		}
	}


//-----------------------------------------------------------------------------
// - updatePressed:
//-----------------------------------------------------------------------------
- (void) updatePressed:(id)sender
	{
	[self reload];
	}


//-----------------------------------------------------------------------------
// - cdPressed:
//-----------------------------------------------------------------------------
- (void) cdPressed:(id)sender
	{
	if ([scroll numSelectedRows] == 1)
		{
		DirEntry const* const de = [dir_array entryAt:[scroll selectedRow]];
		if ([dir_array isDir:de])
			[self load:normalize_path( [path stringByAppendingPathComponent:
						de->name] )];
		}
	}


//-----------------------------------------------------------------------------
// - autoSortClick:
//-----------------------------------------------------------------------------
- (void) autoSortClick:(id)sender
	{
	BOOL const switchState = [autoSortSwitch state];
	if (autoSort != switchState)
		{
		defaultAutoSort = autoSort = switchState;
		[Defaults set:SORT_DEF bool:defaultAutoSort];
		[scroll setAutoSortRows:switchState];
		if (switchState)
			[scroll sortRows];
		}
	}


//-----------------------------------------------------------------------------
// - hiddenFilesClick:
//
//		FIXME: Should not have to rescan the whole directory just to show/hide
//		hidden files.  A more careful implementation would not be this sloppy.
//-----------------------------------------------------------------------------
- (void) hiddenFilesClick:(id)sender
	{
	BOOL const switchState = [hiddenFilesSwitch state];
	if (showHidden != switchState)
		{
		defaultShowHidden = showHidden = switchState;
		[Defaults set:HIDDEN_DEF bool:defaultShowHidden];
		[self reload];
		}
	}


//-----------------------------------------------------------------------------
// - highlightClick:
//-----------------------------------------------------------------------------
- (void) highlightClick:(id)sender
	{
	BOOL const switchState = [highlightSwitch state];
	if (highlightDirs != switchState)
		{
		defaultHighlightDirs = highlightDirs = switchState;
		[Defaults set:HLIGHT_DEF bool:defaultHighlightDirs];
		[scroll setNeedsDisplay:YES];
		}
	}


//-----------------------------------------------------------------------------
// - didClick:
//-----------------------------------------------------------------------------
- (void) didClick:(id)sender
	{
	[self updateButtons];
	}


//-----------------------------------------------------------------------------
// - didDoubleClick:
//-----------------------------------------------------------------------------
- (void) didDoubleClick:(id)sender
	{
	[self open:sender];
	}


//-----------------------------------------------------------------------------
// - makeKeyAndOrderFront:
//-----------------------------------------------------------------------------
- (void) makeKeyAndOrderFront:(id)sender
	{
	[window makeKeyAndOrderFront:sender];
	}


//-----------------------------------------------------------------------------
// - windowShouldClose:
//-----------------------------------------------------------------------------
- (BOOL) windowShouldClose:(id)sender
	{
	[openDirs removeObject:self];
	[self autorelease];
	return YES;
	}


//-----------------------------------------------------------------------------
// - windowDidResize:
//-----------------------------------------------------------------------------
- (void) windowDidResize:(NSNotification*)notification
	{
	NSWindow* theWindow = [notification object];
	NSRect r = [theWindow frame];
	if (NSWidth (r) != defaultWinSize.width ||
		NSHeight(r) != defaultWinSize.height)
		{
		defaultWinSize = r.size;
		[Defaults set:SIZE_DEF size:defaultWinSize];
		}
	}


//-----------------------------------------------------------------------------
// - setDefaultColor:
//-----------------------------------------------------------------------------
- (void) setDefaultColor: (NSColor*)c
	{
	[defaultColor autorelease];
	defaultColor = [c retain];
	[Defaults set:COLOR_DEF color:c];
	}


//-----------------------------------------------------------------------------
// - setDaColors:
//-----------------------------------------------------------------------------
- (void) setDaColors: (NSColor*)c
	{
	[window setBackgroundColor:c];
	[scroll setBackgroundColor:c];
	[window display];
	}


//-----------------------------------------------------------------------------
// - draggingEntered:
//-----------------------------------------------------------------------------
- (unsigned int) draggingEntered:(id<NSDraggingInfo>)sender
	{
	return ([sender draggingSourceOperationMask] & NSDragOperationGeneric);
	}


//-----------------------------------------------------------------------------
// - performDragOperation:
//-----------------------------------------------------------------------------
- (BOOL) performDragOperation:(id<NSDraggingInfo>)sender
	{
	[self setDefaultColor:
				[NSColor colorFromPasteboard:[sender draggingPasteboard]]];
	[self setDaColors: defaultColor];
	return YES;
	}


//-----------------------------------------------------------------------------
// - initLazyRow
//-----------------------------------------------------------------------------
- (void) initLazyRow
	{
	NSFont* font = [scroll font];
	NSZone* const z = [self zone];
	int i;
	for (i = 0; i < MAX_SLOT; i++)
		{
		id cell = [[scroll colCellPrototype:i] copyWithZone:z];
		if ([cell respondsToSelector:@selector(setFont:)])
			[cell setFont:font];
		lazy_row[i] = cell;
		}
	}


//-----------------------------------------------------------------------------
// - initDefaults
//-----------------------------------------------------------------------------
- (void) initDefaults
	{
	static BOOL initialized = NO;
	if (!initialized)
		{
		NSRect r = [window frame];
		defaultWinSize = r.size;
		defaultFont = 
				[[Defaults getFont:FONT_DEF fallback:[scroll font]] retain];
		initialized = YES;
		}
	}


//-----------------------------------------------------------------------------
// - loadDefaults
//-----------------------------------------------------------------------------
- (void) loadDefaults
	{
	NSString* s;

	NSRect r = [window frame];
	r.size = [Defaults getSize:SIZE_DEF fallback:defaultWinSize];
	[window setFrame:r display:NO];

	autoSort = defaultAutoSort;
	showHidden = defaultShowHidden;
	highlightDirs = defaultHighlightDirs;

	[autoSortSwitch setState:autoSort];
	[hiddenFilesSwitch setState:showHidden];
	[highlightSwitch setState:highlightDirs];

	[scroll setAutoSortRows:autoSort];
	[scroll setFont:defaultFont];
	[self setDaColors:defaultColor];

	s = [Defaults getStr:COL_SIZES_DEF fallback:0];
	if (s)
		[scroll setColSizesFromString:s];

	s = [Defaults getStr:COL_ORDER_DEF fallback:0];
	if (s)
		[scroll setColOrderFromString:s];
	}


//-----------------------------------------------------------------------------
// - initWithDir:
//-----------------------------------------------------------------------------
- (id) initWithDir:(NSString*)dirname
	{
	[super init];
	[NSBundle loadNibNamed:@"DirWindow" owner:self];
	[window registerForDraggedTypes:
				[NSArray arrayWithObject:NSColorPboardType]];
	path = [[NSString alloc] init];
	dir_array = [[DirArray allocWithZone:[self zone]] init];
	[self initDefaults];
	[self loadDefaults];
	[self initLazyRow];
	[self load:dirname];
	[openDirs addObject:self];
	[self cascade];
	[window makeKeyAndOrderFront:self];
	return self;
	}


//-----------------------------------------------------------------------------
// - init
//-----------------------------------------------------------------------------
- (id) init
	{
	return [self initWithDir:NSHomeDirectory()];
	}


//-----------------------------------------------------------------------------
// - dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
	int i;
#if(USE_PRIVATE_ZONE)
	NSZone* const z = [self zone];
#endif
	[window setDelegate:0];
	[window close];
	[window release];
	[path release];
	[dir_array release];
	for (i = 0; i < MAX_SLOT; i++)
		[lazy_row[i] release];
	[super dealloc];
#if(USE_PRIVATE_ZONE)
	NSRecycleZone( z );
#endif
	}


//-----------------------------------------------------------------------------
// - path
//-----------------------------------------------------------------------------
- (NSString*) path
	{
	return path;
	}


//-----------------------------------------------------------------------------
// + findDir:
//-----------------------------------------------------------------------------
+ (DirWindow*) findDir:(NSString*)normalized_path
	{
	if (normalized_path != 0)
		{
		unsigned int i;
		unsigned int const lim = [openDirs count];
		for (i = 0;	 i < lim;  i++)
			{
			DirWindow* p = (DirWindow*)[openDirs objectAtIndex:i];
			NSString* s = [p path];
			if (s != 0 && [s isEqualToString:normalized_path])
				return p;
			}
		}
	return 0;
	}


//-----------------------------------------------------------------------------
// + launchDir:
//-----------------------------------------------------------------------------
+ (void) launchDir:(NSString*)dirname
	{
	DirWindow* p = 0;
	if (dirname == 0) dirname = NSHomeDirectory();
	if (dirname == 0) dirname = @"/";
	dirname = normalize_path( dirname );
	if ((p = [self findDir:dirname]) != 0)
		[p makeKeyAndOrderFront:self];
	else
		{
#if(USE_PRIVATE_ZONE)
		NSZone* z = NSCreateZone( NSPageSize(), NSPageSize(), YES );
		p = [[self allocWithZone:z] initWithDir:dirname];
#else
		p = [[self alloc] initDir:dirname];
#endif
		}
	}


@end
