//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscBorderCell.M
//
//		Cell class used by MiscBorderView to manage column headings and row 
//		labels for MiscTableScroll.	 
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscBorderCell.M,v 1.4 96/05/07 02:09:40 sunshine Exp $
// $Log:		MiscBorderCell.M,v $
//	Revision 1.4  96/05/07	02:09:40  sunshine
//	Rewrote -drawSelf:inView: so that it makes only 1 trip to Window Server
//	instead of 8, and is now table driven.
//	
//	Revision 1.3  96/04/30	05:38:06  sunshine
//	Ported to OpenStep 4.0 for Mach PR2.
//-----------------------------------------------------------------------------
#import "MiscBorderCell.h"

extern "Objective-C" {
#import <AppKit/NSFont.h>
}

@implementation MiscBorderCell

//-----------------------------------------------------------------------------
// initTextCell:
//-----------------------------------------------------------------------------
- (id) initTextCell:(NSString*)s
	{
	[super initTextCell:s];
	[self setBordered:NO];
	[self setWraps:NO];
	[self setAlignment:NSCenterTextAlignment];
	[self setFont:[NSFont boldSystemFontOfSize:12.0]];
	return self;
	}


//-----------------------------------------------------------------------------
// backgroundColor
//-----------------------------------------------------------------------------
- (NSColor*) backgroundColor
	{
	NSColor* c = [NSColor darkGrayColor];
	if ([self state])
		c = [c blendedColorWithFraction:0.5 ofColor:[NSColor blackColor]];
	return c;
	}


//-----------------------------------------------------------------------------
// setUpFieldEditorAttributes:
//-----------------------------------------------------------------------------
- (NSText*) setUpFieldEditorAttributes:(NSText*)textObject
	{
	[super setUpFieldEditorAttributes:textObject];
	[textObject setTextColor:[NSColor whiteColor]];
	[textObject setBackgroundColor:[self backgroundColor]];
	return textObject;
	}


//-----------------------------------------------------------------------------
// -cellSizeForBounds:
//-----------------------------------------------------------------------------
- (NSSize)cellSizeForBounds:(NSRect)aRect
	{
	NSSize s = [super cellSizeForBounds:aRect];
	s.width += 3;
	s.height += 3;
	return s;
	}


//-----------------------------------------------------------------------------
// drawInteriorWithFrame:inView:
//-----------------------------------------------------------------------------
- (void) drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView*)controlView
	{
	[[self backgroundColor] set];
	NSRect r = NSMakeRect(	NSMinX(cellFrame) + 2, NSMinY(cellFrame) + 2,
							NSWidth(cellFrame) - 3, NSHeight(cellFrame) - 3 );
	NSRectFill( r );
	[super drawInteriorWithFrame:r inView:controlView];
	}


//-----------------------------------------------------------------------------
// drawWithFRame:inView:
//
//		Draw light gray lines at top and left edges.  Draw white lines just
//		inside light gray lines at top and left edges.	Draw black lines at
//		bottom and right edges.	 Draw dark gray pixel at junction of white
//		and black lines (in lower-left and upper-right corners).
//
//		The lines are derived from a table of scalers and constants which are
//		applied to the cellFrame(cf) using the following formulas:
//				x = cf.x * xScaler + cf.maxx * maxxScaler + xConst
//				y = cf.y * yScaler + cf.maxy * maxyScaler + yConst
//				w = cf.w * wScaler + wConst
//				h = cf.h * hScaler + hConst
//
//-----------------------------------------------------------------------------
- (void) drawWithFrame:(NSRect)cellFrame inView:(NSView*)aView
	{
	int const NUM_RECTS = 8;
	float const grays[ NUM_RECTS ] =
						{ NSLightGray, NSLightGray, NSWhite, NSWhite,
						  NSBlack, NSBlack, NSDarkGray, NSDarkGray };
	struct BC_Lines { char xs, maxxs, xc, ys, maxys, yc, ws, wc, hs, hc; };
	static BC_Lines const LINES[ NUM_RECTS ] =
		{
			{ 1, 0,	 0, 1, 0,  0, 0,  1, 1,	 0 }, // ltgray, left edge
			{ 1, 0,	 1, 1, 0,  0, 1, -1, 0,	 1 }, // ltgray, top edge
			{ 1, 0,	 1, 1, 0,  1, 0,  1, 1, -2 }, // white,	 left edge + 1
			{ 1, 0,	 2, 1, 0,  1, 1, -3, 0,	 1 }, // white,	 top edge + 1
			{ 0, 1, -1, 1, 0,  2, 0,  1, 1, -2 }, // black,	 right edge
			{ 1, 0,	 2, 0, 1, -1, 1, -3, 0,	 1 }, // black,	 bottom edge
			{ 1, 0,	 1, 0, 1, -1, 0,  1, 0,	 1 }, // dkgray, lower left corner
			{ 0, 1, -1, 1, 0,  1, 0,  1, 0,	 1 }  // dkgray, upper right corner
		};

	NSRect rects[ NUM_RECTS ];
	for (int i = 0; i < NUM_RECTS; i++)
		{
		BC_Lines const& r = LINES[i];
		rects[i] = NSMakeRect(
				r.xs * NSMinX(cellFrame) + r.maxxs * NSMaxX(cellFrame) + r.xc,
				r.ys * NSMinY(cellFrame) + r.maxys * NSMaxY(cellFrame) + r.yc,
				r.ws * NSWidth(cellFrame) + r.wc,
				r.hs * NSHeight(cellFrame) + r.hc );
		}

	NSRectFillListWithGrays( rects, grays, NUM_RECTS );
	[self drawInteriorWithFrame:cellFrame inView:aView];
	}

@end
