//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscBorderView.M
//
//		View class for the row/column borders on a MiscTableView.
//		Supports resizing, dragging.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscBorderView.M,v 1.15 96/05/07 09:36:04 sunshine Exp $
// $Log:		MiscBorderView.M,v $
//	Revision 1.15  96/05/07	 09:36:04  sunshine
//	Ditched -reflectSelection; replaced with -selectionChanged.
//	-setCursor: --> -setCursorSlot:
//	
//	Revision 1.14  96/05/07	 02:10:58  sunshine
//	Uses new MiscGeometry instead of mounds "type variation" methods.
//	Ripped out all "type variation" methods (which were also duplicated in
//	TableView).	 -frameHeight now returns MiscPixels rather than NXCoord.
//	Fixed bug where modal mouse tracking loop for resizing slots was not
//	adjusting for border orientation when checking whether auto-scrolling was
//	needed.	 As a result, autoscrolling frequently failed to occur when
//	resizing rows.
//	
//	Revision 1.13  96/04/30	 05:38:15  sunshine
//	Ported to OpenStep 4.0 for Mach PR2.
//-----------------------------------------------------------------------------
#import "MiscBorderView.h"
#import "MiscBorderCell.h"
#import "MiscGeometry.h"
#import "MiscHighlightTracker.h"
#import "MiscListTracker.h"
#import "MiscRadioTracker.h"
#import "MiscTableBorder.h"
#import "MiscTableView.h"
#import "MiscTableWell.h"
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableTypes.h>

#import <new.h>

extern "Objective-C" {
#import <AppKit/NSApplication.h>
#import <AppKit/NSClipView.h>
#import <AppKit/NSFont.h>
#import <AppKit/NSCursor.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSText.h>
#import <AppKit/NSWindow.h>
#import <AppKit/psopsNeXT.h>
#import <Foundation/NSBundle.h>
}

extern "C" {
#import <assert.h>
#import <float.h>
#import <limits.h>
#import <math.h>
#import <string.h>		// memset().
#import <stdio.h>		// FILENAME_MAX
}

#define MISC_RESIZE_EPSILON		5
#define MISC_FRAME_HEIGHT		18
#define MISC_FRAME_WIDTH		46

static NSCursor* horzCursor = 0;
static NSCursor* vertCursor = 0;
static BOOL sendingPeriodicEvents = NO;

static inline MiscPixels dmin( MiscPixels a, MiscPixels b )
										{ return (a < b ? a : b); }
static inline MiscPixels dmax( MiscPixels a, MiscPixels b )
										{ return (a > b ? a : b); }

@interface MiscBorderView(InlineImages)
+ (NSImage*) inlineImageNamed:(NSString*)name;
@end

//----------------------------------------------------------------------------
// stopTimer
//----------------------------------------------------------------------------
static inline void stopTimer()
	{
	if (sendingPeriodicEvents)
		{
		[NSEvent stopPeriodicEvents];
		sendingPeriodicEvents = NO;
		}
	}


//----------------------------------------------------------------------------
// startTimer
//----------------------------------------------------------------------------
static inline void startTimer()
	{
	if (!sendingPeriodicEvents)
		{
		sendingPeriodicEvents = YES;
		[NSEvent startPeriodicEventsAfterDelay:0.1 withPeriod:0.1];
		}
	}



//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation MiscBorderView

//-----------------------------------------------------------------------------
// + cursorFromImageName:
//-----------------------------------------------------------------------------
+ (NSCursor*) cursorFromImageName:(NSString*)name
	{
	NSImage* image = [NSImage imageNamed:name];
	if (image == 0)
		{
		NSString* path = [[NSBundle bundleForClass:self]
						pathForImageResource:name];
		if (path != 0)
			{
			image = [[NSImage alloc] initByReferencingFile:path];
			[image setName:name];
			}
		else
			image = [self inlineImageNamed:name];
		}
	NSPoint const hot = { 8, 8 };
	NSCursor* cursor = [[NSCursor alloc] initWithImage:image];
	[cursor setHotSpot:hot];
	return cursor;
	}


//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void) initialize
	{
	if (horzCursor == 0)
		{
		horzCursor = [self cursorFromImageName:@"MiscHorzResizeCursor"];
		vertCursor = [self cursorFromImageName:@"MiscVertResizeCursor"];
		}
	}


//-----------------------------------------------------------------------------
// TYPE VARIATIONS
//-----------------------------------------------------------------------------

- (MiscPixels) frameHeight
		{ return (isHorz ? MISC_FRAME_HEIGHT : MISC_FRAME_WIDTH); }
- (NSCursor*) cursor
		{ return isHorz ? horzCursor : vertCursor; }
- (MiscBorderType) borderType
		{ return isHorz ? MISC_COL_BORDER : MISC_ROW_BORDER; }
- (MiscBorderType) otherBorderType
		{ return isHorz ? MISC_ROW_BORDER : MISC_COL_BORDER; }
- (MiscTableBorder*) otherBorder
		{ return [scroll border: [self otherBorderType]]; }
- (void) clearOtherBorder { [self otherBorder]->selectionSet().empty(); }
- (id) cellAtRow:(MiscCoord_P)row column:(MiscCoord_P)col
	{ return isHorz ?	[scroll cellAtRow:row column:col] :
						[scroll cellAtRow:col column:row]; }

- (void) changeFrameIfNeeded
	{
	MiscRect_O r( isHorz, [self frame] );
	MiscPixels const my_width = r.getWidth_O();
	MiscPixels const i_width = info->totalSize();
	if (my_width != i_width)
		{
		[[self window] invalidateCursorRectsForView:self];
		r.setWidth_O( i_width );
		[self setFrameSize:r];
		[[scroll documentView] adjustSize];
		}
	}

- (void) setPos:(MiscCoord_V)pos width:(MiscPixels)size
	{
	MiscCoord_P pPos = info->visualToPhysical(pos);
	if (isHorz)
		[scroll setCol:pPos size:(float)size];
	else
		[scroll setRow:pPos size:(float)size];
	[self changeFrameIfNeeded];
	[[self window] invalidateCursorRectsForView:self];
	[scroll setNeedsDisplay:YES];
	}

- (void) moveFrom:(MiscCoord_V)from to:(MiscCoord_V)to
	{
	if (isHorz)
		[scroll moveColFrom:from to:to];
	else
		[scroll moveRowFrom:from to:to];
	[[self window] invalidateCursorRectsForView:self];
	[self setNeedsDisplay:YES];
	}


//-----------------------------------------------------------------------------
// - initWithFrame:scroll:info:type:
//-----------------------------------------------------------------------------
- initWithFrame: (NSRect) frameRect
	 scroll: (MiscTableScroll*) i_scroll
	   info: (MiscTableBorder*) i_info
	   type: (MiscBorderViewType) type
	{
	NSZone* const z = [self zone];

	[super initWithFrame:NSZeroRect];

	isHorz = (type == MISC_COL_BORDER_VIEW);
	scroll = i_scroll;
	info = i_info;
	theCell = [[MiscBorderCell allocWithZone:z] initTextCell:@"Kilroy"];

	oldSel = new( NSZoneMalloc(z,sizeof(*oldSel)) ) MiscSparseSet;
	[self setSelectionMode: [scroll selectionMode]];

	MiscRect_O myFrame( isHorz, frameRect );
	myFrame.setWidth_O( i_info->totalSize() );
	myFrame.setHeight_O( [self frameHeight] );
	[self setFrame:myFrame];
	return self;
	}


//-----------------------------------------------------------------------------
// - dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
	[theCell release];
	[tracker release];
	if (oldSel != 0)
		{
		oldSel->MiscSparseSet::~MiscSparseSet();
		NSZoneFree( [self zone], oldSel );
		}
	[super dealloc];
	}


//-----------------------------------------------------------------------------
// - isFlipped
//-----------------------------------------------------------------------------
- (BOOL) isFlipped
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - isOpaque
//-----------------------------------------------------------------------------
- (BOOL) isOpaque
	{
	return YES;
	}



//=============================================================================
// CONVERSIONS
//=============================================================================
//-----------------------------------------------------------------------------
// - range::fromRect:
//
//		Returns by reference the range of visual slots contained in rect.
//		rMin is inclusive, rMax is exclusive.
//-----------------------------------------------------------------------------
- (void) range: (MiscCoord_V*)rMin : (MiscCoord_V*)rMax fromRect: (NSRect)rect
	{
	MiscRect_O r( isHorz, rect );
	*rMin = info->visualForOffset( r.getX_O() );
	*rMax = info->visualForOffset( r.getMaxX_O() ) + 1;
	}


//-----------------------------------------------------------------------------
// - rectForPos:
//-----------------------------------------------------------------------------
- (NSRect) rectForPos:(MiscCoord_V)pos
	{
	MiscRect_O r( isHorz, [self bounds] );
	r.setX_O( info->getOffset(pos) );
	r.setWidth_O( info->effectiveSize(pos) );
	return r;
	}



//=============================================================================
// DRAWING
//=============================================================================
//-----------------------------------------------------------------------------
// - getVisibleRange::
//
//		Returns by reference the range of visual slots contained in the visible
//		rectangle.	rMin is inclusive, rMax is exclusive.
//-----------------------------------------------------------------------------
- (void) getVisibleRange: (MiscCoord_V*)rMin : (MiscCoord_V*)rMax
	{
	id v = [self superview];
	NSRect visRect = [v documentVisibleRect];
	[self range:rMin:rMax fromRect:visRect];
	}


//-----------------------------------------------------------------------------
// drewPos:
//
//		Keeps the old selection sets up to date whenever we draw a cell.  This
//		way -selectionChanged has valid data to work from.
//-----------------------------------------------------------------------------
- (void) drewPos: (MiscCoord_V)slot
	{
	BOOL isSelected = info->selectionSet().contains( slot );
	if (isSelected != oldSel->contains(slot))
		{
		if (isSelected)
			oldSel->add( slot );
		else
			oldSel->remove( slot );
		}
	}


//-----------------------------------------------------------------------------
// - drawPos:inRect:controlView:
//-----------------------------------------------------------------------------
- (void) drawPos:(MiscCoord_V)p inRect:(NSRect)r controlView:(NSView*)v
	{
	[theCell setStringValue:info->getTitle(p)];
	[theCell setState:info->selectionSet().contains(p)];
	[theCell drawWithFrame:r inView:v];
	}


//-----------------------------------------------------------------------------
// - drawPos:updateSel:
//-----------------------------------------------------------------------------
- (void) drawPos:(MiscCoord_V)pos updateSel:(BOOL)updateSel
	{
	if (0 <= pos && pos < info->count())
		{
		NSRect rect = [self rectForPos:pos];
		[self drawPos:pos inRect:rect controlView:self];
		if (updateSel)
			[self drewPos:pos];
		}
	}


//-----------------------------------------------------------------------------
// - drawPos:
//-----------------------------------------------------------------------------
- (void) drawPos: (MiscCoord_V) pos
	{
	[self drawPos:pos updateSel:YES];
	}


//-----------------------------------------------------------------------------
// - drawRect:
//-----------------------------------------------------------------------------
- (void) drawRect:(NSRect)rect
	{
	MiscCoord_V pos_min;
	MiscCoord_V pos_max;
	[self range:&pos_min:&pos_max fromRect:rect];
	for (MiscCoord_V pos = pos_min;	 pos < pos_max;	 pos++)
		[self drawPos:pos];
	}


//-----------------------------------------------------------------------------
// - drawSlot:	[public]
//-----------------------------------------------------------------------------
- (void) drawSlot: (MiscCoord_V) n
	{
	MiscCoord_V rMin, rMax;
	[self getVisibleRange:&rMin:&rMax];
	if (rMin <= n && n < rMax)
		[self setNeedsDisplayInRect:[self rectForPos:n]];
	}



//=============================================================================
// SELECTION
//=============================================================================
//-----------------------------------------------------------------------------
// - setSelectionMode:
//-----------------------------------------------------------------------------
- (void) setSelectionMode: (MiscSelectionMode) mode
	{
	NSZone* const z = [self zone];
	if (tracker)
		[tracker release];
	switch (mode)
		{
		case MISC_LIST_MODE:
			tracker = [MiscListTracker allocWithZone:z];
			break;
		case MISC_RADIO_MODE:
			tracker = [MiscRadioTracker allocWithZone:z];
			break;
		case MISC_HIGHLIGHT_MODE:
			tracker = [MiscHighlightTracker allocWithZone:z];
			break;
		}
	[tracker initBorder: info];
	}


//-----------------------------------------------------------------------------
// - selectionChanged
//-----------------------------------------------------------------------------
- (void) selectionChanged
	{
	MiscSparseSet const& newSel = info->selectionSet();

	MiscCoord_V rMin, rMax;
	[self getVisibleRange:&rMin:&rMax];

	for (MiscCoord_V i = rMin;	i < rMax;  i++)
		if (oldSel->contains(i) != newSel.contains(i))
			[self setNeedsDisplayInRect:[self rectForPos:i]];

	*oldSel = newSel;
	}


//-----------------------------------------------------------------------------
// - selectPos:
//-----------------------------------------------------------------------------
- (void) selectPos: (MiscCoord_V) pos
	{
	MiscSparseSet& s = info->selectionSet();
	s.empty();
	s.add( pos );
	[scroll selectionChanged];
	}



//=============================================================================
// CURSOR MANAGEMENT
//=============================================================================
//-----------------------------------------------------------------------------
// - cursorRect:forPos:
//-----------------------------------------------------------------------------
- (BOOL) cursorRect:(NSRect*)nsrect forPos:(MiscCoord_V)pos;
	{
	MiscPixels const pos_width = info->effectiveSize(pos);
	if (pos_width > 0)
		{
		MiscPixels const pos_offset = info->getOffset(pos);
		MiscPixels const min_x = pos_offset + pos_width - MISC_RESIZE_EPSILON;
		MiscRect_O r( isHorz, [self bounds] );
		r.setX_O( min_x );
		r.setWidth_O( MISC_RESIZE_EPSILON );
		*nsrect = r;
		return YES;
		}
	return NO;
	}


//-----------------------------------------------------------------------------
// - resetCursorRects
//-----------------------------------------------------------------------------
- (void) resetCursorRects
	{
	if (info->isSizeable())
		{
		NSCursor* const cursor = [self cursor];
		MiscCoord_V min_pos, max_pos;
		[self getVisibleRange:&min_pos:&max_pos];
		int const count = info->count();
		if (0 <= min_pos && min_pos < count)
			{
			NSRect r;
			int const lim = (max_pos <= count) ? max_pos : count;
			for (MiscCoord_V pos = min_pos;	 pos < lim;	 pos++)
				if (info->isSizeable(pos) && [self cursorRect:&r forPos:pos])
					[self addCursorRect:r cursor:cursor];
			}
		}
	}



//=============================================================================
// MOUSE-TRACKING
//=============================================================================
//-----------------------------------------------------------------------------
// - startTracking
//-----------------------------------------------------------------------------
- (void) startTracking
	{
	[self clearOtherBorder];
	if ([[self window] firstResponder] == [scroll documentView])
		[[self window] makeFirstResponder:[self window]];
	}


//-----------------------------------------------------------------------------
// - endTracking:
//-----------------------------------------------------------------------------
- (void) endTracking: (MiscCoord_V)pos
	{
	if (pos < 0)
		pos = 0;
	else if (pos >= info->count())
		pos = info->count() - 1;
	[scroll border:[self borderType] setCursorSlot:info->visualToPhysical(pos)];
	[[self window] selectKeyViewFollowingView:[scroll previousKeyView]];
	}


//-----------------------------------------------------------------------------
// - posForMousePt:
//-----------------------------------------------------------------------------
- (MiscCoord_V) posForMousePt:(NSPoint)p
	{
	MiscPixels pix = MiscPoint_O( isHorz, p ).getX_O();
	MiscRect_O r( isHorz, [self bounds] );
	return (pix < r.getMaxX_O() ? info->visualForOffset(pix) : info->count());
	}



//=============================================================================
// RESIZING
//=============================================================================
//-----------------------------------------------------------------------------
// - resizeEvent:x:deltaX:minX:maxX:
//-----------------------------------------------------------------------------
- (int) resizeEvent:(NSEvent*)p
		x:(MiscPixels)x
		deltaX:(MiscPixels)deltaX
		minX:(MiscPixels)minX
		maxX:(MiscPixels)maxX
	{
	NSRect nsDocFrame = [scroll getDocClipFrame];
	NSRect nsClipFrame = [[self superview] frame];
	MiscRect_O docFrame( isHorz, nsDocFrame );
	MiscRect_O clipFrame( isHorz, nsClipFrame );
	MiscPixels const minDrawX = clipFrame.getX_O();		// Scroll coords.
	MiscPixels const maxDrawX = clipFrame.getMaxX_O();	// Scroll coords.
	unsigned int const WANTED =
				(NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSPeriodicMask);

	x += deltaX;

	MiscPixels draw_x;									// Scroll coords.
	MiscRect_O line( isHorz );
	line.setX_O( x - 1 );
	line.setY_O( 0 );
	line.setWidth_O( 2 );
	line.setHeight_O( clipFrame.getHeight_O() + docFrame.getHeight_O() - 1 );
	line = [scroll convertPoint:line fromView:self];
	draw_x = line.getX_O();

	BOOL did_scroll = NO;
	BOOL in_bounds = (minDrawX <= draw_x && draw_x <= maxDrawX);

	[scroll lockFocus];
	PSsetinstance(YES);
	[[NSColor blackColor] set];

	if (in_bounds)
		NSRectFill( line );

	startTimer();
	NSEvent* lastEvent = [p copy];

	for (;;)
		{
		p = [[self window] nextEventMatchingMask:WANTED];

		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		else if ([p type] == NSPeriodic)
			{
			NSPoint mousePt =
				[scroll convertPoint:[lastEvent locationInWindow] fromView:0];
			MiscPixels const mousePtX = MiscPoint_O( isHorz,mousePt ).getX_O();
			if (mousePtX < minDrawX || maxDrawX < mousePtX)
				{
				PSsetinstance(NO);
				[self autoscroll:lastEvent];
				PSsetinstance(YES);
				did_scroll = YES;
				}
			}
		else
			{
			[lastEvent release];
			lastEvent = [p copy];
			}

		NSPoint new_loc =
				[self convertPoint:[lastEvent locationInWindow] fromView:0];
		MiscPixels mouseX = MiscPoint_O( isHorz, new_loc ).getX_O();
		MiscPixels new_x = (mouseX + deltaX);
		if (new_x < minX)
			new_x = minX;
		else if (new_x >= maxX)
			new_x = maxX - 1;

		BOOL const did_move = (new_x != x);
		if (did_move || did_scroll)
			{
			if (in_bounds && !did_scroll)
				PShideinstance( line.getX(), line.getY(),
								line.getWidth(), line.getHeight() );
			if (did_move)
				{
				x = new_x;
				line.setX_O( x - 1 );
				line.setY_O( 0 );
				line = [scroll convertPoint:line fromView:self];
				draw_x = line.getX_O();
				in_bounds = (minDrawX <= draw_x && draw_x <= maxDrawX);
				}

			if (in_bounds)
				NSRectFill( line );

			did_scroll = NO;
			}
		}

	[lastEvent release];
	stopTimer();

	if (in_bounds)
		PShideinstance( line.getX(), line.getY(),
								line.getWidth(), line.getHeight() );

	PSsetinstance(NO);
	[scroll unlockFocus];

	x -= deltaX;
	return x;
	}


//-----------------------------------------------------------------------------
// - resizeEvent:inPos:atX:deltaX:finalWidth:
//-----------------------------------------------------------------------------
- (BOOL) resizeEvent:(NSEvent*)p
		 inPos: (MiscCoord_V) pos
		 atX: (MiscPixels) x_origin
		 deltaX: (MiscPixels) delta_x
		 finalWidth: (MiscPixels*) finalWidth
	{
	MiscPixels const org_x = info->getOffset(pos);
	MiscPixels const min_x = org_x + info->effectiveMinSize(pos);
	MiscPixels const max_x = org_x + info->effectiveMaxSize(pos) + 1;
	MiscPixels const curr_x =
		[self resizeEvent:p x:x_origin deltaX:delta_x minX:min_x maxX:max_x];
	MiscPixels final_delta = curr_x - x_origin;
	if (final_delta != 0)
		{
		*finalWidth = info->effectiveSize(pos) + final_delta;
		return YES;
		}

	return NO;
	}


//-----------------------------------------------------------------------------
// - inHotZone:forPos:atX:deltaX:
//-----------------------------------------------------------------------------
- (BOOL) inHotZone: (NSPoint) pt
		forPos: (MiscCoord_V*) pos
		atX: (MiscPixels*) pos_x
		deltaX: (MiscPixels*) delta_x
	{
	if (info->isSizeable())
		{
		MiscCoord_V const plim = info->count();
		MiscPixels x = MiscPoint_O( isHorz, pt ).getX_O();
		MiscCoord_V p = info->visualForOffset(x);
		if (0 <= p && p < plim)
			{
			MiscPixels max_x = info->getOffset(p) + info->effectiveSize(p);
			MiscPixels delta = max_x - x;
		
			if (0 <= delta && delta <= MISC_RESIZE_EPSILON)
				{
				do { p++; } while (p < plim && info->effectiveSize(p) <= 0);
				p--;
				if (0 <= p && p < plim && info->isSizeable(p))
					{
					*pos = p;
					*pos_x = x;
					*delta_x = delta;
					return YES;
					}
				}
			}
		}

	return NO;
	}


//-----------------------------------------------------------------------------
// - adjustSize
//-----------------------------------------------------------------------------
- (void) adjustSize
	{
	[self changeFrameIfNeeded];
	}



//=============================================================================
// DRAGGING
//=============================================================================
//-----------------------------------------------------------------------------
// - drawDocCells:at:controlView
//-----------------------------------------------------------------------------
- (void) drawDocCells: (MiscCoord_V) pos at: (NSPoint) pt
	controlView: (NSView*)controlView
	{
	MiscRect_O rDoc( isHorz, [[scroll documentView] visibleRect] );
	MiscTableBorder* b = [self otherBorder];
	MiscCoord_V sMin = b->visualForOffset( rDoc.getY_O() );
	MiscCoord_V sMax = b->visualForOffset( rDoc.getMaxY_O() );
	if (sMin >= 0 && sMax >= 0)
		{
		MiscCoord_P p_pos = info->visualToPhysical( pos );
		MiscPixels delta = rDoc.getY_O() - b->getOffset( sMin );
	
		MiscRect_O r( isHorz );
		r.setY_O( MiscPoint_O(isHorz,pt).getY_O() - delta );
		r.setWidth_O( info->effectiveSize(pos) );

		for (int i = sMin; i <= sMax; i++)
			{
			id cell = [self cellAtRow:b->visualToPhysical(i) column:p_pos];
			r.setHeight_O( b->effectiveSize(i) );
			[cell drawWithFrame:r inView:controlView];
			r.setY_O( r.getY_O() + r.getHeight_O() );
			}
		}
	}


//-----------------------------------------------------------------------------
// - setDragCache:forPos:
//-----------------------------------------------------------------------------
- (void) setDragCache: (NSImage*) cache forPos: (MiscCoord_V) pos
	{
	id docView = [scroll documentView];
	MiscRect_O rClip( isHorz, [[self superview] frame] );		// scroll coord
	MiscRect_O rDoc( isHorz,									// scroll coord
				[scroll convertRect:[docView visibleRect] fromView:docView] );

	MiscRect_O r( isHorz, 0, 0 );
	r.setWidth_O( info->effectiveSize(pos) );
	r.setHeight_O( rClip.getHeight_O() + rDoc.getHeight_O() );

	MiscRect_O rBorder(r);
	rBorder.setHeight_O( rClip.getHeight_O() );

	MiscRect_O rContent(r);
	rContent.setY_O( rClip.getHeight_O() );

	[cache setFlipped:YES];
	[cache setSize:r];
	[cache lockFocus];
	[self drawDocCells:pos at:rContent controlView:docView];
	[self drawPos:pos inRect:rBorder controlView:self];
	[cache unlockFocus];
	}


//-----------------------------------------------------------------------------
// - getVisibleCacheMin:max:
//-----------------------------------------------------------------------------
- (NSImage*) getVisibleCacheMin:(MiscPixels*)pMin max:(MiscPixels*)pMax
	{
	MiscRect_O rDoc( isHorz, [[scroll documentView] visibleRect] );
	MiscRect_O rVis( isHorz, [self visibleRect] );
	*pMin = rVis.getX_O();
	*pMax = rVis.getMaxX_O();
	rDoc = [scroll convertRect:rDoc fromView:[scroll documentView]];
	rVis = [scroll convertRect:rVis fromView:self];

	MiscRect_O r( isHorz );
	r.setX_O( rVis.getX_O() );
	r.setY_O( rVis.getY_O() );
	r.setWidth_O( rVis.getWidth_O() );
	r.setHeight_O( rVis.getHeight_O() + rDoc.getHeight_O() );

	[scroll lockFocus];
	NSBitmapImageRep* rep = [[NSBitmapImageRep allocWithZone:[self zone]]
								initWithFocusedViewRect:r];
	[scroll unlockFocus];

	NSImage* cache = [[NSImage allocWithZone:[self zone]] initWithSize:r];
	[cache addRepresentation:rep];
	return [cache autorelease];
	}


//-----------------------------------------------------------------------------
// - setWells::forPos:
//-----------------------------------------------------------------------------
- (void) setWells: (MiscTableWell**) w1 : (MiscTableWell**) w2
	forPos: (MiscCoord_V) pos
	{
	NSZone* const z = [self zone];
	MiscTableView* doc = [scroll documentView];
	MiscRect_O rDoc( isHorz, [doc visibleRect] );
	MiscRect_O rClip( isHorz, [[self superview] frame] );

	MiscRect_O r( isHorz );
	r.setX_O( info->getOffset(pos) );
	r.setWidth_O( info->effectiveSize(pos) );
	r.setHeight_O( rClip.getHeight_O() );

	*w1 = [[MiscTableWell allocWithZone:z] initWithFrame:r];
	[self addSubview: *w1];

	r.setY_O( rDoc.getY_O() );
	r.setHeight_O( rDoc.getHeight_O() );
	*w2 = [[MiscTableWell allocWithZone:z] initWithFrame:r];
	[doc addSubview: *w2];

	[*w1 display];
	[*w2 display];
	}


//-----------------------------------------------------------------------------
// - clearWells::
//-----------------------------------------------------------------------------
- (void) clearWells: (MiscTableWell**) w1 : (MiscTableWell**) w2
	{
	[*w1 removeFromSuperview];
	[*w2 removeFromSuperview];
	[*w1 release];
	[*w2 release];
	*w1 = 0;
	*w2 = 0;
	// NOTE: Does not need display here.  Everything will get displayed later.
	}


//-----------------------------------------------------------------------------
// - offsetFromEvent:
//-----------------------------------------------------------------------------
- (float) offsetFromEvent:(NSEvent*)ev 
	{
	NSPoint mLoc = [self convertPoint:[ev locationInWindow] fromView:0];
	return MiscPoint_O( isHorz, mLoc ).getX_O();
	}


//-----------------------------------------------------------------------------
// - calcDrop:::
//-----------------------------------------------------------------------------
- (MiscCoord_V) calcDrop:(MiscCoord_V)fromPos
						:(NSPoint)mouseDownPt
						:(NSPoint)mouseUpPt
	{
	MiscCoord_V toPos = fromPos;
	MiscPixels const start_pos = MiscPoint_O( isHorz, mouseDownPt ).getX_O();
	MiscPixels const end_pos   = MiscPoint_O( isHorz, mouseUpPt	  ).getX_O();
	MiscPixels const delta_pos = (end_pos - start_pos);

	MiscPixels const SLOP = 4;
	if (delta_pos < -SLOP || SLOP < delta_pos)
		{
		MiscPixels const start_ofs = info->getOffset( fromPos );
		MiscPixels drop_pos = start_ofs + delta_pos;
		if (delta_pos < 0)
			drop_pos += SLOP;
		else
			drop_pos += info->effectiveSize(fromPos) - SLOP;
		toPos = info->visualForOffset( drop_pos );
		if (toPos < 0)
			toPos = 0;
		}

	return toPos;
	}


//-----------------------------------------------------------------------------
// - dragEvent:inPos:
//-----------------------------------------------------------------------------
- (MiscCoord_V) dragEvent:(NSEvent*)event inPos:(MiscCoord_V) pos
	{
	int const WANTED =
				(NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSPeriodicMask);

	NSPoint mouseDownPt =
				[self convertPoint:[event locationInWindow] fromView:0];
	NSPoint mouseUpPt = mouseDownPt;

	NSWindow* win = [self window];
	[win disableFlushWindow];

	MiscTableWell *w1, *w2;
	[self setWells: &w1 : &w2 forPos: pos];

	NSZone* const z = [self zone];
	MiscPixels pMin,pMax;
	NSImage* visCache = [[self getVisibleCacheMin:&pMin max:&pMax] retain];
	NSImage* dragCache = [[NSImage allocWithZone:z] init];
	[self setDragCache: dragCache forPos: pos];

	MiscSize_O size( isHorz, [dragCache size] );
	MiscPixels pLoc = info->getOffset( pos );
	MiscPixels delta = MiscPixels( [self offsetFromEvent: event] - pLoc );

	NSEvent* lastEvent = [event copy];
	[scroll lockFocus];
	for (;;)
		{
		MiscPoint_O pt( isHorz );
		MiscRect_O rDrag( isHorz );
		MiscPixels const w = size.getWidth_O();
		MiscPixels const dw =
			dmin( dmin( dmin(w, pMax - pMin), pMax - pLoc), pLoc + w - pMin );
		BOOL const shouldDraw = (dw > 0);
		if (shouldDraw)
			{
			if (isHorz)
				{
				pt.setX( dmax( pLoc, pMin ) );
				pt.setY( size.getHeight_O() );
				if (pLoc < pMin)
					rDrag.setX( pMin - pLoc );
				}
			else
				{
				pt.setX( 0 );
				pt.setY( dmax( pLoc, pMin ) + dw );
				if (pLoc + w >= pMax)
					rDrag.setY( pLoc + w - pMax );
				}
			pt = [scroll convertPoint:pt fromView:self];
			rDrag.setWidth_O( dw );
			rDrag.setHeight_O( size.getHeight_O() );
			[dragCache compositeToPoint:pt fromRect:rDrag
						operation:NSCompositeCopy];
			}
		[win enableFlushWindow];
		[win flushWindow];

		event = [[self window] nextEventMatchingMask:WANTED];

		[win disableFlushWindow];
		if (shouldDraw)
			{
			NSSize s = [visCache size];
			MiscPixels xTarg;
			if (isHorz)
				{
				xTarg = (pLoc < pMin ? 0 : pLoc - pMin);
				}
			else
				{
				if (pLoc < pMin)
					xTarg = MiscPixels(s.height) - dw;
				else if (pLoc < pMax)
					xTarg = pMax - pLoc - dw;
				else
					xTarg = 0;
				}
			MiscRect_O rVis( isHorz );
			rVis.setX_O( xTarg );
			rVis.setWidth_O( rDrag.getWidth_O() );
			rVis.setHeight_O( rDrag.getHeight_O() );
			[visCache compositeToPoint:pt fromRect:rVis
						operation:NSCompositeCopy];
			}

		if (event == 0)
			break;
		else if ([event type] == NSLeftMouseUp)
			{
			mouseUpPt = [self convertPoint:[event locationInWindow] fromView:0];
			break;
			}
		else if ([event type] != NSPeriodic)
			{
			[lastEvent release];
			lastEvent = [event copy];
			}

		MiscPixels mLoc = MiscPixels( [self offsetFromEvent: lastEvent] );
		if ((mLoc < pMin && pMin > 0.0) || (mLoc > pMax &&
			pMax < MiscRect_O( isHorz, [self bounds] ).getMaxX_O()))
			{
			[self autoscroll:lastEvent];
			[visCache release];
			visCache = [[self getVisibleCacheMin:&pMin max:&pMax] retain];
			mLoc = MiscPixels( [self offsetFromEvent: lastEvent] );
			startTimer();
			}
		else
			{
			stopTimer();
			}

		pLoc = mLoc - delta;
		if (pLoc < pMin - size.getWidth_O())
			pLoc = pMin - size.getWidth_O();
		else if (pLoc > pMax)
			pLoc = pMax;
		}

	[lastEvent release];
	stopTimer();
	[scroll unlockFocus];
	[self clearWells: &w1 : &w2];
	[visCache release];
	[dragCache release];

	MiscCoord_V const toPos = [self calcDrop:pos :mouseDownPt :mouseUpPt];

	if (toPos != pos)
		{
		[self moveFrom: pos to: toPos];
		[self selectPos: toPos];
		[scroll border:[self borderType] slotDraggedFrom:pos to:toPos];
		}
	else
		{
		// Need to redisplay the slot that the wells were covering.
		MiscBorderType const b = [self borderType];
		int const phys_pos = [scroll border:b slotAtPosition:toPos];
		[scroll border:b drawSlotTitle:phys_pos];
		[scroll border:b drawSlot:phys_pos];
		}

	[scroll displayIfNeeded];
	[win enableFlushWindow];
	[win flushWindow];
	return toPos;
	}


//-----------------------------------------------------------------------------
// - awaitDragEvent:inPos:
//-----------------------------------------------------------------------------
- (MiscCoord_V) awaitDragEvent:(NSEvent*)event inPos:(MiscCoord_V) pos
	{
	MiscCoord_V toPos = pos;
	int const WANTED = (NSLeftMouseUpMask | NSLeftMouseDraggedMask);
	float const SLOP = 2.0;
	NSWindow* win = [self window];
	NSEvent* mouseDown = [event copy];
	NSPoint origLoc = [mouseDown locationInWindow];
	for (;;)
		{
		NSEvent* p = [win nextEventMatchingMask:WANTED
						untilDate:[NSDate distantFuture]
						inMode:NSEventTrackingRunLoopMode
						dequeue:NO];
		if (p == 0)
			break;
		else if ([p type] == NSLeftMouseUp)
			{
			[win nextEventMatchingMask:NSLeftMouseUpMask];
			break;
			}
		else // ([p type] == NSLeftMouseDragged)
			{
			NSPoint newLoc = [p locationInWindow];
			if (newLoc.x > origLoc.x + SLOP || newLoc.y > origLoc.y + SLOP ||
				newLoc.x < origLoc.x - SLOP || newLoc.y < origLoc.y - SLOP)
				{
				toPos = [self dragEvent: mouseDown inPos: pos];
				break;
				}
			else
				[win nextEventMatchingMask:NSLeftMouseDraggedMask];
			}
		}
	[mouseDown release];
	return toPos;
	}
	


//=============================================================================
// SELECTION
//=============================================================================
//-----------------------------------------------------------------------------
// - selectionEvent:fromPos:
//-----------------------------------------------------------------------------
- (void) selectionEvent:(NSEvent*)p fromPos:(MiscCoord_V)pos
	{
	unsigned int const WANTED =
				(NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSPeriodicMask);
	BOOL doubleClicked = ([p clickCount] > 1);

	[tracker mouseDown: p atPos: pos];
	[scroll selectionChanged];
	[scroll displayIfNeeded];

	startTimer();
	NSEvent* lastEvent = [p copy];

	for(;;)
		{
		p = [[self window] nextEventMatchingMask:WANTED];

		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		else if ([p type] == NSPeriodic)
			[self autoscroll:lastEvent];
		else
			{
			[lastEvent release];
			lastEvent = [p copy];
			}

		NSPoint new_loc = [lastEvent locationInWindow];
		new_loc = [self convertPoint:new_loc fromView:0];
		MiscCoord_V new_pos = [self posForMousePt:new_loc];
		if (new_pos != pos)
			{
			pos = new_pos;
			[tracker mouseDragged: p atPos: pos];
			[scroll selectionChanged];
			[scroll displayIfNeeded];
			}
		}

	[lastEvent release];
	stopTimer();

	[tracker mouseUp: p atPos: pos];
	[scroll selectionChanged];
	[self endTracking: pos];

	[scroll displayIfNeeded];
	if ([scroll isEnabled])
		{
		[scroll sendAction];
		if (doubleClicked)
			[scroll sendDoubleAction];
		}
	}



//=============================================================================
// MOUSE-EVENTS
//=============================================================================
//-----------------------------------------------------------------------------
// - acceptsFirstMouse:
//-----------------------------------------------------------------------------
- (BOOL) acceptsFirstMouse:(NSEvent*)theEvent
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - mouseDown:
//-----------------------------------------------------------------------------
- (void) mouseDown:(NSEvent*)p 
	{
	NSPoint evpt = [self convertPoint:[p locationInWindow] fromView:0];
	MiscCoord_V pos = [self posForMousePt:evpt];
	MiscPixels x, delta_x;

	if ([self inHotZone:evpt forPos:&pos atX:&x deltaX:&delta_x])
		{
		[self startTracking];
		[self selectPos:pos];
		[scroll displayIfNeeded];
		MiscPixels finalWidth;
		BOOL doit;
		doit = [self resizeEvent:p inPos:pos atX:x deltaX:delta_x
						finalWidth:&finalWidth];
		if (doit)
			{
			[self setPos:pos width:finalWidth];
			[scroll border:[self borderType] slotResized:pos];
			}
		[self endTracking:pos];
		[scroll displayIfNeeded];
		[scroll sendActionIfEnabled];
		}
	else if (info->isDraggable() && (info->isModifierDrag() ==
		(([p modifierFlags] & NSCommandKeyMask) != 0)))
		{
		[self startTracking];
		[self selectPos:pos];
		[scroll displayIfNeeded];
		MiscCoord_V toPos = [self awaitDragEvent:p inPos:pos];
		[self endTracking:toPos];
		[scroll displayIfNeeded];
		[scroll sendActionIfEnabled];
		}
	else
		{
		assert( tracker != 0 );
		[self startTracking];
		[self selectionEvent: p fromPos: pos];
		}
	}

@end


//=============================================================================
// INLINE IMAGES
//=============================================================================
@implementation MiscBorderView(InlineImages)

static unsigned char const HORZ_CURSOR_DATA[] =
	{
	0x4d, 0x4d, 0x00, 0x2a, 0x00, 0x00, 0x00, 0x88, 0x00, 0x0f, 
	0xc0, 0x00, 0x00, 0x0c, 0xc0, 0x00, 0x00, 0x0c, 0xc0, 0x00, 
	0x00, 0xfc, 0xfc, 0x00, 0x03, 0x3c, 0xf3, 0x00, 0x0c, 0x3c, 
	0xf0, 0xc0, 0x30, 0x3c, 0xf0, 0x30, 0xc0, 0x00, 0x00, 0x0c, 
	0xc0, 0x00, 0x00, 0x0c, 0x30, 0x3c, 0xf0, 0x30, 0x0c, 0x3c, 
	0xf0, 0xc0, 0x03, 0x3c, 0xf3, 0x00, 0x00, 0xfc, 0xfc, 0x00, 
	0x00, 0x0c, 0xc0, 0x00, 0x00, 0x0c, 0xc0, 0x00, 0x00, 0x0f, 
	0xc0, 0x00, 0x00, 0x0f, 0xc0, 0x00, 0x00, 0x0f, 0xc0, 0x00, 
	0x00, 0x0f, 0xc0, 0x00, 0x00, 0xff, 0xfc, 0x00, 0x03, 0xff, 
	0xff, 0x00, 0x0f, 0xff, 0xff, 0xc0, 0x3f, 0xff, 0xff, 0xf0, 
	0xff, 0xff, 0xff, 0xfc, 0xff, 0xff, 0xff, 0xfc, 0x3f, 0xff, 
	0xff, 0xf0, 0x0f, 0xff, 0xff, 0xc0, 0x03, 0xff, 0xff, 0x00, 
	0x00, 0xff, 0xfc, 0x00, 0x00, 0x0f, 0xc0, 0x00, 0x00, 0x0f, 
	0xc0, 0x00, 0x00, 0x0f, 0xc0, 0x00, 0x00, 0x0d, 0x01, 0x00, 
	0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x10, 0x00, 0x00, 
	0x01, 0x01, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x10, 
	0x00, 0x00, 0x01, 0x02, 0x00, 0x03, 0x00, 0x00, 0x00, 0x02, 
	0x00, 0x02, 0x00, 0x02, 0x01, 0x03, 0x00, 0x03, 0x00, 0x00, 
	0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x06, 0x00, 0x03, 
	0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x11, 
	0x00, 0x04, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x01, 0x2a, 
	0x01, 0x15, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 
	0x00, 0x00, 0x01, 0x17, 0x00, 0x04, 0x00, 0x00, 0x00, 0x02, 
	0x00, 0x00, 0x01, 0x32, 0x01, 0x1a, 0x00, 0x05, 0x00, 0x00, 
	0x00, 0x01, 0x00, 0x00, 0x01, 0x3a, 0x01, 0x1b, 0x00, 0x05, 
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x42, 0x01, 0x1c, 
	0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x00, 
	0x01, 0x28, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 
	0x00, 0x00, 0x01, 0x52, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x08, 0x00, 0x00, 0x00, 0x48, 0x00, 0x00, 0x00, 0x40, 
	0x00, 0x00, 0x00, 0x40, 0x00, 0x0a, 0xfc, 0x80, 0x00, 0x00, 
	0x27, 0x10, 0x00, 0x0a, 0xfc, 0x80, 0x00, 0x00, 0x27, 0x10 
	};
int const HORZ_CURSOR_LEN =
		sizeof(HORZ_CURSOR_DATA)/sizeof(HORZ_CURSOR_DATA[0]);

static unsigned char const VERT_CURSOR_DATA[] =
	{
	0x4d, 0x4d, 0x00, 0x2a, 0x00, 0x00, 0x00, 0x88, 0x00, 0x03, 
	0xc0, 0x00, 0x00, 0x0c, 0x30, 0x00, 0x00, 0x30, 0x0c, 0x00, 
	0x00, 0xc0, 0x03, 0x00, 0x03, 0x00, 0x00, 0xc0, 0x03, 0xfc, 
	0x3f, 0xc0, 0xff, 0xfc, 0x3f, 0xff, 0xc0, 0x00, 0x00, 0x03, 
	0xff, 0xfc, 0x3f, 0xff, 0x03, 0xfc, 0x3f, 0xc0, 0x03, 0x00, 
	0x00, 0xc0, 0x00, 0xc0, 0x03, 0x00, 0x00, 0x30, 0x0c, 0x00, 
	0x00, 0x0c, 0x30, 0x00, 0x00, 0x03, 0xc0, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x03, 0xc0, 0x00, 0x00, 0x0f, 0xf0, 0x00, 
	0x00, 0x3f, 0xfc, 0x00, 0x00, 0xff, 0xff, 0x00, 0x03, 0xff, 
	0xff, 0xc0, 0x03, 0xff, 0xff, 0xc0, 0xff, 0xff, 0xff, 0xff, 
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x03, 0xff, 
	0xff, 0xc0, 0x03, 0xff, 0xff, 0xc0, 0x00, 0xff, 0xff, 0x00, 
	0x00, 0x3f, 0xfc, 0x00, 0x00, 0x0f, 0xf0, 0x00, 0x00, 0x03, 
	0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0d, 0x01, 0x00, 
	0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x10, 0x00, 0x00, 
	0x01, 0x01, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x10, 
	0x00, 0x00, 0x01, 0x02, 0x00, 0x03, 0x00, 0x00, 0x00, 0x02, 
	0x00, 0x02, 0x00, 0x02, 0x01, 0x03, 0x00, 0x03, 0x00, 0x00, 
	0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x06, 0x00, 0x03, 
	0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x01, 0x11, 
	0x00, 0x04, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x01, 0x2a, 
	0x01, 0x15, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 
	0x00, 0x00, 0x01, 0x17, 0x00, 0x04, 0x00, 0x00, 0x00, 0x02, 
	0x00, 0x00, 0x01, 0x32, 0x01, 0x1a, 0x00, 0x05, 0x00, 0x00, 
	0x00, 0x01, 0x00, 0x00, 0x01, 0x3a, 0x01, 0x1b, 0x00, 0x05, 
	0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x42, 0x01, 0x1c, 
	0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x00, 
	0x01, 0x28, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 
	0x00, 0x00, 0x01, 0x52, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 
	0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x08, 0x00, 0x00, 0x00, 0x48, 0x00, 0x00, 0x00, 0x40, 
	0x00, 0x00, 0x00, 0x40, 0x00, 0x0a, 0xfc, 0x80, 0x00, 0x00, 
	0x27, 0x10, 0x00, 0x0a, 0xfc, 0x80, 0x00, 0x00, 0x27, 0x10
	};
int const VERT_CURSOR_LEN =
		sizeof(VERT_CURSOR_DATA)/sizeof(VERT_CURSOR_DATA[0]);


//-----------------------------------------------------------------------------
// + inlineImageNamed:
//-----------------------------------------------------------------------------
+ (NSImage*) inlineImageNamed:(NSString*)name
	{
	typedef struct
		{
		NSString* name;
		unsigned char const* data;
		unsigned int len;
		} MISC_CURSOR_INFO;
	static MISC_CURSOR_INFO const DATA[] =
		{
			{ @"MiscHorzResizeCursor", HORZ_CURSOR_DATA, HORZ_CURSOR_LEN },
			{ @"MiscVertResizeCursor", VERT_CURSOR_DATA, VERT_CURSOR_LEN }
		};
	NSImage* image = 0;
	for (int i = sizeof(DATA)/sizeof(DATA[0]); i-- > 0; )
		{
		MISC_CURSOR_INFO const& p = DATA[i];
		if ([p.name isEqualToString:name])
			{
			NSData* data = [NSData dataWithBytes:p.data length:p.len];
			image = [[[NSImage alloc] initWithData:data] autorelease];
			[image setName:name];
			break;
			}
		}
	return image;
	}

@end
