//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableCell.M
//
//		Default cell class used by MiscTableScroll to display text.
//
// NOTE *OPENSTEP-4.0-PR2*
//		OpenStep 4.0 PR2 converts NSImageCellType cells to NSTextCellType 
//		cells upon receipt of -setFont: which is not the behavior that we want
//		(nor expect since in previous NEXTSTEP versions -setFont: left the 
//		cell alone if it was not NX_TEXTCELL).
//
// NOTE *SET-OWNER-VALUE*
//		We must implement these methods so that MiscTableScroll will call 
//		these methods instead of the explicit instance -setXxx methods.	 
//		However, we don't need to do any work.	We will ask the owner for 
//		it's value whenever we need it.	 
//
// FIXME: optional-allocation stuff does not address alignment requirements.
// FIXME: automate most of the optional-allocation stuff.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableCell.M,v 1.9 96/05/07 09:38:14 sunshine Exp $
// $Log:		MiscTableCell.M,v $
// Revision 1.9	 96/05/07  09:38:14	 sunshine
// Bumped OpenStep class version numbers way up (to 1000) to give NEXTSTEP 3.3
// TableScroll room to grow.
// Fixed bug: Was retaining unarchived NSColors one more time than necesary.
// 
// Revision 1.8	 96/04/30  05:38:46	 sunshine
// Ported to OpenStep 4.0 for Mach PR2.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableCell.h>

extern "Objective-C" {
#import <AppKit/NSFont.h>
}

extern "C" {
#import <string.h>
}

#define MISC_TC_VERSION_0		0		// Conditional tag.
#define MISC_TC_VERSION_1		1		// Tag became unconditional.
#define MISC_TC_VERSION_1000	1000	// First OpenStep version (4.0 PR2).
#define MISC_TC_VERSION			MISC_TC_VERSION_1000

@implementation MiscTableCell

//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void) initialize
	{
	if (self == [MiscTableCell class])
		{
		[self setVersion: MISC_TC_VERSION];
		}
	}


//-----------------------------------------------------------------------------
// + defaultFont
//-----------------------------------------------------------------------------
+ (NSFont*) defaultFont
	{
	return [NSFont userFontOfSize:12.0];
	}


//-----------------------------------------------------------------------------
// initTextCell:
//		NOTE *1*  -[Cell initTextCell:] invokes -setStringValue: which
//		invokes -setFont:.	This means that this cell will get an explicit
//		-setFont call before we finish initialization, and we need to 
//		restore the useOwnerFont setting.
//-----------------------------------------------------------------------------
- (id) initTextCell:(NSString*)s
	{
	[super initTextCell:s];								// NOTE *1*
	[self setBordered:NO];
	[self setWraps:NO];
	[self setAlignment:NSLeftTextAlignment];
	[self setUseOwnerFont:YES];							// NOTE *1*
	[self setOwnerFont:[[self class] defaultFont]];
	return self;
	}


//-----------------------------------------------------------------------------
// initImageCell:
//-----------------------------------------------------------------------------
- (id) initImageCell:(NSImage*)s
	{
	[super initImageCell:s];
	[self setBordered:NO];
	return self;
	}


//-----------------------------------------------------------------------------
// tag
//-----------------------------------------------------------------------------
- (int) tag
	{
	return tag;
	}


//-----------------------------------------------------------------------------
// setTag:
//-----------------------------------------------------------------------------
- (void) setTag:(int)x
	{
	tag = x;
	}


//-----------------------------------------------------------------------------
// owner
//-----------------------------------------------------------------------------
- (id) owner
	{
	return owner;
	}


//-----------------------------------------------------------------------------
// setOwner:
//-----------------------------------------------------------------------------
- (void) setOwner:(id)obj
	{
	if (obj != owner)
		{
		owner = obj;
		if (owner != 0)
			{
			if ([owner respondsToSelector:@selector(font)])
				[self setOwnerFont:[owner font]];
			if ([owner respondsToSelector:@selector(textColor)])
				[self setOwnerTextColor:[owner textColor]];
			if ([owner respondsToSelector:@selector(backgroundColor)])
				[self setOwnerBackgroundColor:[owner backgroundColor]];
			if ([owner respondsToSelector:@selector(highlightTextColor)])
				[self setOwnerTextColor:[owner highlightTextColor]];
			if ([owner respondsToSelector:@selector(highlightBackgroundColor)])
				[self setOwnerBackgroundColor:[owner highlightBackgroundColor]];
			}
		} 
	}


//-----------------------------------------------------------------------------
// font
//-----------------------------------------------------------------------------
- font
	{
	return [super font];
	}


//-----------------------------------------------------------------------------
// setFont:
//-----------------------------------------------------------------------------
- (void) setFont:(NSFont*)fontObj
	{
	if ([self type] == NSTextCellType)			// NOTE *OPENSTEP-4.0-PR2*
		{
		[self setUseOwnerFont:NO];
		[super setFont:fontObj];
		}
	}


//-----------------------------------------------------------------------------
// tc1Data
//-----------------------------------------------------------------------------
- (void*) tc1Data
	{
	return tc1_data;
	}


//-----------------------------------------------------------------------------
// tc1Flags
//-----------------------------------------------------------------------------
- (unsigned int) tc1Flags
	{
	return tc1_flags;
	}


//-----------------------------------------------------------------------------
// tc1DataSize
//-----------------------------------------------------------------------------
- (unsigned int) tc1DataSize
	{
	unsigned int size = 0;
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR)
		size += [self tc1TextColorLen];
	if (tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR)
		size += [self tc1BackgroundColorLen];
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR_H)
		size += [self tc1HighlightTextColorLen];
	if (tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR_H)
		size += [self tc1HighlightBackgroundColorLen];
	return size;
	}


//-----------------------------------------------------------------------------
// tc1TextColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1TextColorPos
	{
	return 0;
	}


//-----------------------------------------------------------------------------
// tc1BackgroundColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1BackgroundColorPos
	{
	unsigned int pos = [self tc1TextColorPos];
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR)
		pos += [self tc1TextColorLen];
	return pos;
	}


//-----------------------------------------------------------------------------
// tc1HighlightTextColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightTextColorPos
	{
	unsigned int pos = [self tc1BackgroundColorPos];
	if (tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR)
		pos += [self tc1BackgroundColorLen];
	return pos;
	}


//-----------------------------------------------------------------------------
// tc1HighlightBackgroundColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightBackgroundColorPos
	{
	unsigned int pos = [self tc1HighlightTextColorPos];
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR_H)
		pos += [self tc1HighlightTextColorLen];
	return pos;
	}


//-----------------------------------------------------------------------------
// tc1TextColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1TextColorLen
	{
	return sizeof(NSColor*);
	}


//-----------------------------------------------------------------------------
// tc1BackgroundColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1BackgroundColorLen
	{
	return sizeof(NSColor*);
	}


//-----------------------------------------------------------------------------
// tc1HighlightTextColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightTextColorLen
	{
	return sizeof(NSColor*);
	}


//-----------------------------------------------------------------------------
// tc1HighlightBackgroundColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightBackgroundColorLen
	{
	return sizeof(NSColor*);
	}


//-----------------------------------------------------------------------------
// tc1TextColorPtr
//-----------------------------------------------------------------------------
- (NSColor**) tc1TextColorPtr
	{
	return (NSColor**)((char*)tc1_data + [self tc1TextColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1BackgroundColorPtr
//-----------------------------------------------------------------------------
- (NSColor**) tc1BackgroundColorPtr
	{
	return (NSColor**)((char*)tc1_data + [self tc1BackgroundColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1HighlightTextColorPtr
//-----------------------------------------------------------------------------
- (NSColor**) tc1HighlightTextColorPtr
	{
	return (NSColor**)((char*)tc1_data + [self tc1HighlightTextColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1HighlightBackgroundColorPtr
//-----------------------------------------------------------------------------
- (NSColor**) tc1HighlightBackgroundColorPtr
	{
	return (NSColor**)((char*)tc1_data + [self tc1HighlightBackgroundColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1InsertData:len:pos:
//-----------------------------------------------------------------------------
- (void*) tc1InsertData:(void const*)data
		pos:(unsigned int)pos
		len:(unsigned int)len
	{
	if (len > 0)
		{
		unsigned int const old_size = [self tc1DataSize];
		unsigned int const new_size = old_size + len;
		if (old_size == 0)
			tc1_data = NSZoneMalloc( [self zone], new_size );
		else
			{
			tc1_data = NSZoneRealloc( [self zone], tc1_data, new_size );
			if (pos < old_size)
				memmove( (char*) tc1_data + pos + len,			// Destination
						(char*) tc1_data + pos,					// Source
						old_size - pos );						// # bytes.
			}
		if (data != 0)
			memcpy( (char*) tc1_data + pos, data, len );
		else
			memset( (char*) tc1_data + pos, 0, len );
		return (void*)((char*) tc1_data + pos );
		}
	return 0;
	}


//-----------------------------------------------------------------------------
// tc1DeleteDataPos:len:
//-----------------------------------------------------------------------------
- (void) tc1DeleteDataPos:(unsigned int)pos len:(unsigned int)len
	{
	if (len > 0)
		{
		unsigned int const old_size = [self tc1DataSize];
		unsigned int const new_size = old_size - len;
		if (new_size == 0)
			{
			NSZoneFree( [self zone], tc1_data );
			tc1_data = 0;
			}
		else
			{
			if (pos < new_size)
				memmove( (char*) tc1_data + pos,				// Destination
						(char*) tc1_data + pos + len,			// Source
						new_size - pos );						// # bytes
			tc1_data = NSZoneRealloc( [self zone], tc1_data, new_size );
			}
		}
	}


//-----------------------------------------------------------------------------
// useOwnerFont
//-----------------------------------------------------------------------------
- (BOOL) useOwnerFont
	{
	return ((tc1_flags & MISC_TC1_SELF_FONT) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerTextColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerTextColor
	{
	return ((tc1_flags & MISC_TC1_SELF_TEXT_COLOR) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerBackgroundColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerBackgroundColor
	{
	return ((tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerHighlightTextColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerHighlightTextColor
	{
	return ((tc1_flags & MISC_TC1_SELF_TEXT_COLOR_H) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerHighlightBackgroundColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerHighlightBackgroundColor
	{
	return ((tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR_H) == 0);
	}


//-----------------------------------------------------------------------------
// setUseOwnerFont:
//-----------------------------------------------------------------------------
- (void) setUseOwnerFont:(BOOL)flag
	{
	if ([self useOwnerFont] != flag)
		{
		if (flag)
			{
			tc1_flags &= ~(MISC_TC1_SELF_FONT);
			if (owner != 0 && [owner respondsToSelector:@selector(font)])
				[self setOwnerFont:[owner font]];
			}
		else // (!flag)
			{
			tc1_flags |= MISC_TC1_SELF_FONT;
			}
		} 
	}


//-----------------------------------------------------------------------------
// setUseOwnerTextColor:
//-----------------------------------------------------------------------------
- (void) setUseOwnerTextColor:(BOOL)flag
	{
	if ([self useOwnerTextColor] != flag)
		{
		unsigned int const pos = [self tc1TextColorPos];
		unsigned int const len = [self tc1TextColorLen];
		if (flag)
			{
			[[self textColor] release];
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_TEXT_COLOR);
			}
		else // (!flag)
			{
			NSColor* color = [[[self class] defaultTextColor] retain];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_TEXT_COLOR;
			}
		} 
	}


//-----------------------------------------------------------------------------
// setUseOwnerBackgroundColor:
//-----------------------------------------------------------------------------
- (void) setUseOwnerBackgroundColor:(BOOL)flag
	{
	if ([self useOwnerBackgroundColor] != flag)
		{
		unsigned int const pos = [self tc1BackgroundColorPos];
		unsigned int const len = [self tc1BackgroundColorLen];
		if (flag)
			{
			[[self backgroundColor] release];
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_BACKGROUND_COLOR);
			}
		else // (!flag)
			{
			NSColor* color = [[[self class] defaultBackgroundColor] retain];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_BACKGROUND_COLOR;
			}
		} 
	}


//-----------------------------------------------------------------------------
// setUseOwnerHighlightTextColor:
//-----------------------------------------------------------------------------
- (void) setUseOwnerHighlightTextColor:(BOOL)flag
	{
	if ([self useOwnerHighlightTextColor] != flag)
		{
		unsigned int const pos = [self tc1HighlightTextColorPos];
		unsigned int const len = [self tc1HighlightTextColorLen];
		if (flag)
			{
			[[self highlightTextColor] release];
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_TEXT_COLOR_H);
			}
		else // (!flag)
			{
			NSColor* color = [[[self class] defaultHighlightTextColor] retain];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_TEXT_COLOR_H;
			}
		} 
	}


//-----------------------------------------------------------------------------
// setUseOwnerHighlightBackgroundColor:
//-----------------------------------------------------------------------------
- (void) setUseOwnerHighlightBackgroundColor:(BOOL)flag
	{
	if ([self useOwnerHighlightBackgroundColor] != flag)
		{
		unsigned int const pos = [self tc1HighlightBackgroundColorPos];
		unsigned int const len = [self tc1HighlightBackgroundColorLen];
		if (flag)
			{
			[[self highlightBackgroundColor] release];
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_BACKGROUND_COLOR_H);
			}
		else // (!flag)
			{
			NSColor* color =
				[[[self class] defaultHighlightBackgroundColor] retain];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_BACKGROUND_COLOR_H;
			}
		} 
	}


//-----------------------------------------------------------------------------
// setOwnerFont:
//-----------------------------------------------------------------------------
- (void) setOwnerFont:fontObj
	{
	if ([self type] == NSTextCellType)			// NOTE *OPENSTEP-4.0-PR2*
		if ([self useOwnerFont])
			[super setFont:fontObj]; 
	}


//-----------------------------------------------------------------------------
// setOwner values...	NOTE *SET-OWNER-VALUE*
//-----------------------------------------------------------------------------
- (void) setOwnerTextColor:(NSColor*)pcolor {}
- (void) setOwnerBackgroundColor:(NSColor*)pcolor {}
- (void) setOwnerHighlightTextColor:(NSColor*)pcolor {}
- (void) setOwnerHighlightBackgroundColor:(NSColor*)pcolor {}


//-----------------------------------------------------------------------------
// +defaultBackgroundColor
//-----------------------------------------------------------------------------
+ (NSColor*) defaultBackgroundColor
	{
	return [NSColor lightGrayColor];
	}


//-----------------------------------------------------------------------------
// backgroundColor
//-----------------------------------------------------------------------------
- (NSColor*) backgroundColor
	{
	if ([self useOwnerBackgroundColor])
		{
		if (owner != 0 &&
			[owner respondsToSelector:@selector(backgroundColor)])
			return [owner backgroundColor];
		return [[self class] defaultBackgroundColor];
		}
	return *[self tc1BackgroundColorPtr];
	}


//-----------------------------------------------------------------------------
// setBackgroundColor
//-----------------------------------------------------------------------------
- (void) setBackgroundColor:(NSColor*)c
	{
	NSColor** p;
	[self setUseOwnerBackgroundColor:NO];
	p = [self tc1BackgroundColorPtr];
	[*p autorelease];
	*p = [c retain];
	}


//-----------------------------------------------------------------------------
// +defaultHighlightBackgroundColor
//-----------------------------------------------------------------------------
+ (NSColor*) defaultHighlightBackgroundColor
	{
	return [NSColor whiteColor];
	}


//-----------------------------------------------------------------------------
// highlightBackgroundColor
//-----------------------------------------------------------------------------
- (NSColor*) highlightBackgroundColor
	{
	if ([self useOwnerHighlightBackgroundColor])
		{
		if (owner &&
			[owner respondsToSelector:@selector(highlightBackgroundColor)])
			return [owner highlightBackgroundColor];
		return [[self class] defaultHighlightBackgroundColor];
		}
	return *[self tc1HighlightBackgroundColorPtr];
	}


//-----------------------------------------------------------------------------
// setHighlightBackgroundColor
//-----------------------------------------------------------------------------
- (void) setHighlightBackgroundColor:(NSColor*)c
	{
	NSColor** p;
	[self setUseOwnerHighlightBackgroundColor:NO];
	p = [self tc1HighlightBackgroundColorPtr];
	[*p autorelease];
	*p = [c retain]; 
	}


//-----------------------------------------------------------------------------
// +defaultTextColor
//-----------------------------------------------------------------------------
+ (NSColor*) defaultTextColor
	{
	return [NSColor blackColor];
	}


//-----------------------------------------------------------------------------
// textColor
//-----------------------------------------------------------------------------
- (NSColor*) textColor
	{
	if ([self useOwnerTextColor])
		{
		if (owner != 0 && [owner respondsToSelector:@selector(textColor)])
			return [owner textColor];
		return [[self class] defaultTextColor];
		}
	return *[self tc1TextColorPtr];
	}


//-----------------------------------------------------------------------------
// setTextColor
//-----------------------------------------------------------------------------
- (void) setTextColor:(NSColor*)c
	{
	NSColor** p;
	[self setUseOwnerTextColor:NO];
	p = [self tc1TextColorPtr];
	[*p autorelease];
	*p = [c retain];
	}


//-----------------------------------------------------------------------------
// +defaultHighlightTextColor
//-----------------------------------------------------------------------------
+ (NSColor*) defaultHighlightTextColor
	{
	return [NSColor blackColor];
	}


//-----------------------------------------------------------------------------
// highlightTextColor
//-----------------------------------------------------------------------------
- (NSColor*) highlightTextColor
	{
	if ([self useOwnerHighlightTextColor])
		{
		if (owner != 0 &&
			[owner respondsToSelector:@selector(highlightTextColor)])
			return [owner highlightTextColor];
		return [[self class] defaultHighlightTextColor];
		}
	return *[self tc1HighlightTextColorPtr];
	}


//-----------------------------------------------------------------------------
// setHighlightTextColor
//-----------------------------------------------------------------------------
- (void) setHighlightTextColor:(NSColor*)c
	{
	NSColor** p;
	[self setUseOwnerHighlightTextColor:NO];
	p = [self tc1HighlightTextColorPtr];
	[*p autorelease];
	*p = [c retain]; 
	}


//-----------------------------------------------------------------------------
// bgColor
//-----------------------------------------------------------------------------
- (NSColor*) bgColor
	{
	// FIXME: Use [self isHighlighted] when we figure out how to support
	// this for lazy mode.
	return [self state] ?
				[self highlightBackgroundColor] : [self backgroundColor];
	}


//-----------------------------------------------------------------------------
// fgColor
//-----------------------------------------------------------------------------
- (NSColor*) fgColor
	{
	// FIXME: Use [self isHighlighted] when we figure out how to support
	// this for lazy mode.
	return [self state] ? [self highlightTextColor] : [self textColor];
	}


//-----------------------------------------------------------------------------
// calcCellSize:inRect:
//-----------------------------------------------------------------------------
- (NSSize) cellSizeForBounds:(NSRect)aRect
	{
	
	NSSize theSize = [super cellSizeForBounds:aRect];
	theSize.width += 1;
	theSize.height += 1;
	return theSize;
	}


//-----------------------------------------------------------------------------
// setUpFieldEditorAttributes:
//-----------------------------------------------------------------------------
- (NSText*) setUpFieldEditorAttributes:(NSText*)textObject
	{
	[super setUpFieldEditorAttributes:textObject];
	[textObject setTextColor:[self fgColor]];
	[textObject setBackgroundColor:[self bgColor]];
	return textObject;
	}


//-----------------------------------------------------------------------------
// drawInteriorWithFrame:inView:
//-----------------------------------------------------------------------------
- (void) drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView*)controlView
	{
	NSRect rect = NSMakeRect( NSMinX(cellFrame), NSMinY(cellFrame),
							  NSWidth(cellFrame) - 1, NSHeight(cellFrame) - 1 );
	[[self bgColor] set];
	NSRectFill( rect );
	[super drawInteriorWithFrame:rect inView:controlView];
	}


//-----------------------------------------------------------------------------
// drawWithFrame:inView:
//-----------------------------------------------------------------------------
- (void) drawWithFrame:(NSRect)cellFrame inView:(NSView*)aView
	{
	NSRect r[2];
	(r[0]) = NSMakeRect( NSMaxX(cellFrame) - 1, NSMinY(cellFrame),
						1, NSHeight(cellFrame) );
	(r[1]) = NSMakeRect( NSMinX(cellFrame), NSMaxY(cellFrame) - 1,
						NSWidth(cellFrame), 1 );

	[[NSColor grayColor] set];
	NSRectFillList( r, 2 );
	[self drawInteriorWithFrame:cellFrame inView:aView];
	}


//-----------------------------------------------------------------------------
// tc1DestroyData
//-----------------------------------------------------------------------------
- (void) tc1DestroyData
	{
	if (![self useOwnerTextColor])
		[[self textColor] release];
	if (![self useOwnerBackgroundColor])
		[[self backgroundColor] release];
	if (![self useOwnerHighlightTextColor])
		[[self highlightTextColor] release];
	if (![self useOwnerHighlightBackgroundColor])
		[[self highlightBackgroundColor] release];

	tc1_flags = 0;
	}


//-----------------------------------------------------------------------------
// tc1FreeData
//-----------------------------------------------------------------------------
- (void) tc1FreeData
	{
	if (tc1_data != 0)
		{
		NSZoneFree( [self zone], tc1_data );
		tc1_data = 0;
		}
	}


//-----------------------------------------------------------------------------
// dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
	[self tc1DestroyData];
	[self tc1FreeData];
	[super dealloc];
	}


//-----------------------------------------------------------------------------
// copyWithZone:
//-----------------------------------------------------------------------------
- copyWithZone:(NSZone*)zone;
	{
	MiscTableCell* the_clone = [super copyWithZone:zone];
	if (tc1_data != 0)
		{
		unsigned int const size = [self tc1DataSize];
		void* p = NSZoneMalloc( [self zone], size );
		memcpy( p, tc1_data, size );
		the_clone->tc1_data = p;

		if (![the_clone useOwnerTextColor])
			*[the_clone tc1TextColorPtr] =
						[[self textColor] retain];
		if (![the_clone useOwnerBackgroundColor])
			*[the_clone tc1BackgroundColorPtr] =
						[[self backgroundColor] retain];
		if (![the_clone useOwnerHighlightTextColor])
			*[the_clone tc1HighlightTextColorPtr] =
						[[self highlightTextColor] retain];
		if (![the_clone useOwnerHighlightBackgroundColor])
			*[the_clone tc1HighlightBackgroundColorPtr] =
						[[self highlightBackgroundColor] retain];
		}
	return the_clone;
	}


//-----------------------------------------------------------------------------
// initWithCoder_v0:
//-----------------------------------------------------------------------------
- (void) initWithCoder_v0:(NSCoder*)aDecoder
	{
	unsigned int x;

	owner = [[aDecoder decodeObject] retain];

	[aDecoder decodeValueOfObjCType:@encode(unsigned int) at:&x];
	if (x & MISC_TC1_HAS_TAG)
		[aDecoder decodeValueOfObjCType:@encode(int) at:&tag];
	else
		tag = 0;

	if (x & MISC_TC1_SELF_TEXT_COLOR)
		[self setTextColor:[aDecoder decodeNXColor]];
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR)
		[self setBackgroundColor:[aDecoder decodeNXColor]];
	if (x & MISC_TC1_SELF_TEXT_COLOR_H)
		[self setHighlightTextColor:[aDecoder decodeNXColor]];
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR_H)
		[self setHighlightBackgroundColor:[aDecoder decodeNXColor]];
	}


//-----------------------------------------------------------------------------
// initWithCoder_v1:
//-----------------------------------------------------------------------------
- (void) initWithCoder_v1:(NSCoder*)aDecoder
	{
	unsigned int x;

	owner = [[aDecoder decodeObject] retain];

	[aDecoder decodeValueOfObjCType:@encode(int) at:&tag];
	[aDecoder decodeValueOfObjCType:@encode(unsigned int) at:&x];

	if (x & MISC_TC1_SELF_TEXT_COLOR)
		[self setTextColor:[aDecoder decodeNXColor]];
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR)
		[self setBackgroundColor:[aDecoder decodeNXColor]];
	if (x & MISC_TC1_SELF_TEXT_COLOR_H)
		[self setHighlightTextColor:[aDecoder decodeNXColor]];
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR_H)
		[self setHighlightBackgroundColor:[aDecoder decodeNXColor]];
	}


//-----------------------------------------------------------------------------
// initWithCoder_v1000:
//-----------------------------------------------------------------------------
- (void) initWithCoder_v1000:(NSCoder*)aDecoder
	{
	unsigned int x;

	owner = [[aDecoder decodeObject] retain];

	[aDecoder decodeValueOfObjCType:@encode(int) at:&tag];
	[aDecoder decodeValueOfObjCType:@encode(unsigned int) at:&x];

	if (x & MISC_TC1_SELF_TEXT_COLOR)
		[self setTextColor:[aDecoder decodeObject]];
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR)
		[self setBackgroundColor:[aDecoder decodeObject]];
	if (x & MISC_TC1_SELF_TEXT_COLOR_H)
		[self setHighlightTextColor:[aDecoder decodeObject]];
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR_H)
		[self setHighlightBackgroundColor:[aDecoder decodeObject]];
	}


//-----------------------------------------------------------------------------
// initWithCoder:
//-----------------------------------------------------------------------------
- (id) initWithCoder:(NSCoder*)aDecoder
	{
	[super initWithCoder:aDecoder];

	[self tc1DestroyData];
	[self tc1FreeData];

	unsigned int const ver =
			[aDecoder versionForClassName:[[MiscTableCell class] description]];

	switch (ver)
		{
		case MISC_TC_VERSION_0:	   [self initWithCoder_v0:	 aDecoder]; break;
		case MISC_TC_VERSION_1:	   [self initWithCoder_v1:	 aDecoder]; break;
		case MISC_TC_VERSION_1000: [self initWithCoder_v1000:aDecoder]; break;
		default:
			[NSException raise:NSGenericException
					format:@"Cannot read: unknown version %d", ver];
			break;
		}

	return self;
	}


//-----------------------------------------------------------------------------
// encodeWithCoder:
//-----------------------------------------------------------------------------
- (void) encodeWithCoder:(NSCoder*)aCoder
	{
	[super encodeWithCoder:aCoder];
	[aCoder encodeConditionalObject:owner];
	[aCoder encodeValueOfObjCType:@encode(int) at:&tag];
	[aCoder encodeValueOfObjCType:@encode(unsigned int) at:&tc1_flags];
	if (![self useOwnerTextColor])
		[aCoder encodeObject:[self textColor]];
	if (![self useOwnerBackgroundColor])
		[aCoder encodeObject:[self backgroundColor]];
	if (![self useOwnerHighlightTextColor])
		[aCoder encodeObject:[self highlightTextColor]];
	if (![self useOwnerHighlightBackgroundColor])
		[aCoder encodeObject:[self highlightBackgroundColor]];
	}

@end
