//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableScroll.M
//
//		ScrollView class that displays a 2-D table of cells.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableScroll.M,v 1.25 96/05/13 07:46:26 sunshine Exp $
// $Log:		MiscTableScroll.M,v $
// Revision 1.25  96/05/13	07:46:26  sunshine
// Header listed -font as returning (NSFont*) but it was listed as (id) here,
// so corrected it here.
// 
// Revision 1.24  96/05/07	09:43:28  sunshine
// Ditched -reflectSelection; replaced with -selectionChanged (since they
// are identical under OpenStep).
// Bumped OpenStep class version numbers way up (to 1000) to give NEXTSTEP 3.3
// TableScroll room to grow.
// Added some missing explicit (id) return types and argument types.
// Was incorrectly retaining objects which it never releases: delegate,
// dataDelegate, target, doubleTarget (and old-styld nextText, previousText).
// Slot-sort-vector changed from int array (int*) to NSArray.
// Renamed the following methods for consistency with other method names:
//		border:selectTags:		->>		border:selectSlotTags:
//		border:setCursor:		->>		border:setCursorSlot:
//		border:setTitlesOn:		->>		border:setSlotTitlesOn:
//		border:setTitleMode:	->>		border:setSlotTitleMode:
// 
// Revision 1.23  96/05/06	04:33:49  sunshine
// slotDragged and slotResized notification is now sent to dataDelegate too.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableCell.h>
#import "MiscTableBorder.h"
#import "MiscTableView.h"
#import "MiscColView.h"
#import "MiscRowView.h"
#import "MiscNullView.h"

#import <new.h>

extern "Objective-C" {
#import <AppKit/NSApplication.h>
#import <AppKit/NSCell.h>
#import <AppKit/NSClipView.h>
#import <AppKit/NSFont.h>
#import <AppKit/NSFontManager.h>
}

extern "C" {
#import <assert.h>
#import <math.h>
#import <strings.h>
}


@implementation MiscTableScroll

//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void) initialize
	{
	if (self == [MiscTableScroll class])
		{
		[self setVersion: MISC_TS_VERSION];
		}
	}


//-----------------------------------------------------------------------------
// Tag Stuff
//-----------------------------------------------------------------------------
- (int) tag				{ return tag; }
- (void) setTag:(int)x	{ tag = x; }


//-----------------------------------------------------------------------------
// Multicast Messages
//-----------------------------------------------------------------------------
- (void) sendAction:(SEL)aSel to:(id)obj forAllCells:(BOOL)flag
	{
	int const rlim = num_rows;
	int const clim = num_cols;
	for (int r = 0;	 r < rlim;	r++)
		for (int c = 0;	 c < clim;	c++)
			if (flag || [self cellIsSelected:r:c])
				if (![obj performSelector:aSel
					withObject:[self cellAtRow:r column:c]])
					break;
	}

- (int) makeCellsPerform:(SEL)aSel with:(id)p1 with:(id)p2
	selectedOnly:(BOOL)flag
	{
	int count = 0;
	int const rlim = num_rows;
	int const clim = num_cols;
	for (int r = 0;	 r < rlim;	r++)
		for (int c = 0;	 c < clim;	c++)
			if (!flag || [self cellIsSelected:r:c])
				{
				id cell = [self cellAtRow:r column:c];
				if ([cell respondsToSelector:aSel])
					if ([cell performSelector:aSel
						withObject:p1 withObject:p2])
						count++;
					else
						return count;
				}
	return count;
	}

- (int) makeCellsPerform:(SEL)aSel with:(id)p1 selectedOnly:(BOOL)flag
	{ return [self makeCellsPerform:aSel with:p1 with:0 selectedOnly:flag]; }
- (int) makeCellsPerform:(SEL)aSel selectedOnly:(BOOL)flag
	{ return [self makeCellsPerform:aSel with:0 with:0 selectedOnly:flag]; }

- (int) makeCellsPerform:(SEL)aSel
	{ return [self makeCellsPerform:aSel selectedOnly:NO]; }
- (int) makeCellsPerform:(SEL)aSel with:(id)p1
	{ return [self makeCellsPerform:aSel with:p1 selectedOnly:NO]; }
- (int) makeCellsPerform:(SEL)aSel with:(id)p1 with:(id)p2
	{ return [self makeCellsPerform:aSel with:p1 with:p2 selectedOnly:NO];}


//-----------------------------------------------------------------------------
// FINDING CELLS
//-----------------------------------------------------------------------------

- (int) border:(MiscBorderType)b findSlotWithTag:(int)x
		{
		int const lim = (int) [self numberOfSlots:b];
		for (int i = 0;	 i < lim;  i++)
			if ([self border:b slotTag:i] == x)
				return i;
		return -1;
		}
- (int) findColWithTag:(int)x
		{ return [self border:MISC_COL_BORDER findSlotWithTag:x]; }
- (int) findRowWithTag:(int)x
		{ return [self border:MISC_ROW_BORDER findSlotWithTag:x]; }


- (BOOL) findCell:(id)cell row:(int*)row col:(int*)col
	{
	int const NRows = [self numberOfRows];
	int const NCols = [self numberOfColumns];
	for (int r = 0;	 r < NRows;	 r++)
		for (int c = 0;	 c < NCols;	 c++)
			if ([self cellAtRow:r column:c] == cell)
				{
				*row = r;
				*col = c;
				return YES;
				}
	*row = -1;
	*col = -1;
	return NO;
	}

- (id) cellWithTag:(int)x row:(int*)row col:(int*)col
	{
	int const NRows = [self numberOfRows];
	int const NCols = [self numberOfColumns];
	for (int r = 0;	 r < NRows;	 r++)
		for (int c = 0;	 c < NCols;	 c++)
			{
			id const cell = [self cellAtRow:r column:c];
			if (cell && [cell respondsToSelector:@selector(tag)] &&
				[cell tag] == x)
				{
				*row = r;
				*col = c;
				return cell;
				}
			}
	*row = -1;
	*col = -1;
	return 0;
	}

- (id) cellWithTag:(int)x
	{
	int r, c;
	return [self cellWithTag:x row:&r col:&c];
	}




//-----------------------------------------------------------------------------
// Selection stuff
//-----------------------------------------------------------------------------
- (MiscSelectionMode) selectionMode { return mode; }
- (void) setSelectionMode:(MiscSelectionMode) x
	{
	if (x != mode)
		{
		[self clearSelection];
		mode = x;
		[colInfo.view setSelectionMode: mode];
		[rowInfo.view setSelectionMode: mode];
		[tableView setSelectionMode: mode];
		}
	}

- (id) selectedCell { return 0; }		// FIXME: write this

- (BOOL) hasSlotSelection:(MiscBorderType)b
		{ return info[b]->border->hasSelection(); }
- (BOOL) hasRowSelection { return [self hasSlotSelection:MISC_ROW_BORDER]; }
- (BOOL) hasColSelection { return [self hasSlotSelection:MISC_COL_BORDER]; }

- (BOOL) hasMultipleSlotSelection:(MiscBorderType)b
		{ return info[b]->border->hasMultipleSelection(); }
- (BOOL) hasMultipleRowSelection
		{ return [self hasMultipleSlotSelection:MISC_ROW_BORDER]; }
- (BOOL) hasMultipleColSelection
		{ return [self hasMultipleSlotSelection:MISC_COL_BORDER]; }

- (unsigned int) numSelectedSlots:(MiscBorderType)b
		{ return info[b]->border->numSelected(); }
- (unsigned int) numSelectedRows
		{ return [self numSelectedSlots:MISC_ROW_BORDER]; }
- (unsigned int) numSelectedCols
		{ return [self numSelectedSlots:MISC_COL_BORDER]; }

- (BOOL) border: (MiscBorderType)b slotIsSelected:(MiscCoord_P)slot
		{ return info[b]->border->slotIsSelected( slot ); }
- (BOOL) rowIsSelected:(MiscCoord_P)row
		{ return [self border:MISC_ROW_BORDER slotIsSelected:row]; }
- (BOOL) colIsSelected:(MiscCoord_P)col
		{ return [self border:MISC_COL_BORDER slotIsSelected:col]; }
- (BOOL) cellIsSelected:(MiscCoord_P)r :(MiscCoord_P)c
		{ return ([self rowIsSelected:r] || [self colIsSelected:c]); }

- (MiscCoord_P) selectedSlot:(MiscBorderType)b
		{ return info[b]->border->selectedSlot(); }
- (MiscCoord_P) selectedRow
		{ return [self selectedSlot:MISC_ROW_BORDER]; }
- (MiscCoord_P) selectedColumn
		{ return [self selectedSlot:MISC_COL_BORDER]; }

- (NSArray*) selectedSlotTags:(MiscBorderType)b
		{ return info[b]->border->selectedTags(); }
- (NSArray*) selectedRowTags
		{ return [self selectedSlotTags:MISC_ROW_BORDER]; }
- (NSArray*) selectedColTags
		{ [self selectedSlotTags:MISC_COL_BORDER]; }

- (NSArray*) selectedSlots:(MiscBorderType)b
		{ return info[b]->border->selectedSlots(); }
- (NSArray*) selectedRows { [self selectedSlots:MISC_ROW_BORDER]; }
- (NSArray*) selectedCols { [self selectedSlots:MISC_COL_BORDER]; }

- (void) border: (MiscBorderType)b selectSlot:(MiscCoord_P)slot
		{
		BOOL changed = NO;
		MiscTableBorder* const ob = info[ MISC_OTHER_BORDER(b) ]->border;
		if (ob->hasSelection())
			{
			ob->selectNone();
			changed = YES;
			}
		MiscTableBorder* const bp = info[b]->border;
		unsigned int const n = bp->numSelected();
		if (0 <= slot && slot < bp->count())
			{
			if (n != 1 || bp->selectedSlot() != slot)
				{
				if (mode == MISC_RADIO_MODE)
					bp->selectNone();
				bp->selectSlot( slot );
				changed = YES;
				}
			}
		else if (n > 0)
			{
			bp->selectNone();
			changed = YES;
			}
		if (changed)
			[self selectionChanged];
		}
- (void) selectRow:(MiscCoord_P)row
		{ [self border:MISC_ROW_BORDER selectSlot:row]; }
- (void) selectCol:(MiscCoord_P)col
		{ [self border:MISC_COL_BORDER selectSlot:col]; }

- (void) border:(MiscBorderType)b selectSlotTags:(NSArray*)tags
		{
		info[b]->border->selectTags( tags );
		[self selectionChanged];
		}
- (void) selectRowTags: (NSArray*) tags
		{ [self border:MISC_ROW_BORDER selectSlotTags:tags]; }
- (void) selectColTags: (NSArray*) tags
		{ [self border:MISC_COL_BORDER selectSlotTags:tags]; }

- (void) border: (MiscBorderType)b selectSlots: (NSArray*)slots
		{
		info[b]->border->selectSlots( slots );
		[self selectionChanged];
		}
- (void) selectRows: (NSArray*) rows
		{ [self border:MISC_ROW_BORDER selectSlots:rows]; }
- (void) selectCols: (NSArray*) cols;
		{ [self border:MISC_COL_BORDER selectSlots:cols]; }

- (void) selectAllSlots:(MiscBorderType)b
		{
		if (mode == MISC_LIST_MODE || mode == MISC_HIGHLIGHT_MODE)
			{
			info[b]->border->selectAll();
			[self selectionChanged];
			}
		}
- (void) selectAllRows { [self selectAllSlots:MISC_ROW_BORDER]; }
- (void) selectAllCols { [self selectAllSlots:MISC_COL_BORDER]; }
- (void) selectAll:(id)sender
		{
		[self selectAllRows];
		[self sendActionIfEnabled];
		}

- (void) clearSlotSelection:(MiscBorderType)b
		{
		info[b]->border->selectNone();
		[self selectionChanged];
		}
- (void) clearRowSelection { [self clearSlotSelection:MISC_ROW_BORDER]; }
- (void) clearColSelection { [self clearSlotSelection:MISC_COL_BORDER]; }
- (void) clearSelection { [self clearRowSelection]; [self clearColSelection]; }


//-----------------------------------------------------------------------------
// Keyboard cursor stuff
//-----------------------------------------------------------------------------
- (void) reflectCursor { [tableView reflectCursor]; }
- (MiscCoord_P) cursorSlot:(MiscBorderType)b
		{ return info[b]->border->getCursor_P(); }
- (MiscCoord_P) cursorRow { return [self cursorSlot:MISC_ROW_BORDER]; }
- (MiscCoord_P) cursorCol { return [self cursorSlot:MISC_COL_BORDER]; }

- (void) border:(MiscBorderType)b setCursorSlot:(MiscCoord_P)slot
		{ info[b]->border->setCursor_P( slot ); [self reflectCursor]; }
- (void) setCursorRow: (MiscCoord_P) row
		{ [self border:MISC_ROW_BORDER setCursorSlot:row]; }
- (void) setCursorCol: (MiscCoord_P) col
		{ [self border:MISC_COL_BORDER setCursorSlot:col]; }

- (void) clearCursorSlot:(MiscBorderType)b
		{ info[b]->border->clearCursor(); [self reflectCursor]; }
- (void) clearCursorRow { [self clearCursorSlot:MISC_ROW_BORDER]; }
- (void) clearCursorCol { [self clearCursorSlot:MISC_COL_BORDER]; }
- (void) clearCursor { [self clearCursorRow]; [self clearCursorCol]; }

- (BOOL) hasValidCursorSlot:(MiscBorderType)b
		{ return info[b]->border->hasValidCursor(); }
- (BOOL) hasValidCursorRow
		{ return [self hasValidCursorSlot:MISC_ROW_BORDER]; }
- (BOOL) hasValidCursorCol
		{ return [self hasValidCursorSlot:MISC_COL_BORDER]; }


//-----------------------------------------------------------------------------
// Overrides...	 Scrollers are always required.
//-----------------------------------------------------------------------------
- (void) setHasHorizontalScroller:(BOOL)flag {}
- (void) setHasVerticalScroller:(BOOL)flag {}


//-----------------------------------------------------------------------------
// - getDocClipFrame
//-----------------------------------------------------------------------------
- (NSRect) getDocClipFrame
	{
	return [[self contentView] frame];
	}


//-----------------------------------------------------------------------------
// - freeBorder:
//-----------------------------------------------------------------------------
- (void) freeBorder:(MiscBorderInfo*)p
	{
	[p->view removeFromSuperview];
	[p->view release];
	[p->clip removeFromSuperview];
	[p->clip release];
	delete p->border;
	}


//-----------------------------------------------------------------------------
// - dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
	NSWindow* win = [self window];
	if ([win firstResponder] == tableView)
		[win makeFirstResponder:win];
	[self emptyAndReleaseCells];
	[nullView removeFromSuperview];
	[nullView release];
	[tableView removeFromSuperview];
	[tableView release];
	[self freeBorder:&colInfo];
	[self freeBorder:&rowInfo];
	[super dealloc];
	}


//-----------------------------------------------------------------------------
// get/set min/max total size limits.
//-----------------------------------------------------------------------------
- (float) totalSize:(MiscBorderType)b
		{ return (float) info[b]->border->totalSize(); }
- (float) totalWidth
		{ return [self totalSize:MISC_COL_BORDER]; }
- (float) totalHeight
		{ return [self totalSize:MISC_ROW_BORDER]; }

- (float) minTotalSize:(MiscBorderType)b
		{ return (float) info[b]->border->getMinTotalSize(); }
- (float) maxTotalSize:(MiscBorderType)b
		{ return (float) info[b]->border->getMaxTotalSize(); }
- (float) minTotalWidth
		{ return [self minTotalSize:MISC_COL_BORDER]; }
- (float) maxTotalWidth
		{ return [self maxTotalSize:MISC_COL_BORDER]; }
- (float) minTotalHeight
		{ return [self minTotalSize:MISC_ROW_BORDER]; }
- (float) maxTotalHeight
		{ return [self maxTotalSize:MISC_ROW_BORDER]; }

// FIXME: Need to decide whether or not to update the views here.
// FIXME: Should check current value before setting.
- (void) border:(MiscBorderType)b setMinTotalSize:(float)size
		{ info[b]->border->setMinTotalSize( (MiscPixels)floor(size) ); }
- (void) border:(MiscBorderType)b setMaxTotalSize:(float)size
		{ info[b]->border->setMaxTotalSize( (MiscPixels)floor(size) ); }
- (void) setMinTotalWidth:(float)size
		{ [self border:MISC_COL_BORDER setMinTotalSize:size]; }
- (void) setMaxTotalWidth:(float)size
		{ [self border:MISC_COL_BORDER setMaxTotalSize:size]; }
- (void) setMinTotalHeight:(float)size
		{ [self border:MISC_ROW_BORDER setMinTotalSize:size]; }
- (void) setMaxTotalHeight:(float)size
		{ [self border:MISC_ROW_BORDER setMaxTotalSize:size]; }


//-----------------------------------------------------------------------------
// - constrainSize
//-----------------------------------------------------------------------------
- (void) constrainSize
	{
	NSRect r = [self getDocClipFrame];
	if ([self minTotalSizeIsConstrained:MISC_COL_BORDER])
		[self border:MISC_COL_BORDER setMinTotalSize: r.size.width];
	if ([self maxTotalSizeIsConstrained:MISC_COL_BORDER])
		[self border:MISC_COL_BORDER setMaxTotalSize: r.size.width];
	if ([self minTotalSizeIsConstrained:MISC_ROW_BORDER])
		[self border:MISC_ROW_BORDER setMinTotalSize: r.size.height];
	if ([self maxTotalSizeIsConstrained:MISC_ROW_BORDER])
		[self border:MISC_ROW_BORDER setMaxTotalSize: r.size.height];
	[rowInfo.view adjustSize];
	[colInfo.view adjustSize];
	[tableView adjustSize];
	}


//-----------------------------------------------------------------------------
// - sizeToCells
//-----------------------------------------------------------------------------
- (void) sizeToCells
	{
	[self constrainSize];
	}


//-----------------------------------------------------------------------------
// get/set min/max constrained
//-----------------------------------------------------------------------------
- (BOOL) minTotalSizeIsConstrained:(MiscBorderType)b
		{ return info[b]->constrain_min; }
- (BOOL) maxTotalSizeIsConstrained:(MiscBorderType)b
		{ return info[b]->constrain_max; }
- (BOOL) minTotalWidthIsConstrained
		{ return [self minTotalSizeIsConstrained:MISC_COL_BORDER]; }
- (BOOL) maxTotalWidthIsConstrained
		{ return [self maxTotalSizeIsConstrained:MISC_COL_BORDER]; }
- (BOOL) minTotalHeightIsConstrained
		{ return [self minTotalSizeIsConstrained:MISC_ROW_BORDER]; }
- (BOOL) maxTotalHeightIsConstrained
		{ return [self maxTotalSizeIsConstrained:MISC_ROW_BORDER]; }

- (void) border:(MiscBorderType)b constrainMinTotalSize:(BOOL)flag;
		{
		MiscBorderInfo* ip = info[b];
		if (ip->constrain_min != flag)
			{
			ip->constrain_min = flag;
			if (!flag)
				[self border:MISC_COL_BORDER setMinTotalSize:0];
			[self constrainSize];
			}
		}

- (void) border:(MiscBorderType)b constrainMaxTotalSize:(BOOL)flag
		{
		MiscBorderInfo* ip = info[b];
		if (ip->constrain_max != flag)
			{
			ip->constrain_max = flag;
			if (!flag)
				[self border:MISC_COL_BORDER
						setMaxTotalSize: (float) MISC_MAX_PIXELS_SIZE ];
			[self constrainSize];
			}
		}

- (void) constrainMinTotalWidth:(BOOL)flag
		{ [self border:MISC_COL_BORDER constrainMinTotalSize:flag]; }
- (void) constrainMaxTotalWidth:(BOOL)flag
		{ [self border:MISC_COL_BORDER constrainMaxTotalSize:flag]; }
- (void) constrainMinTotalHeight:(BOOL)flag
		{ [self border:MISC_ROW_BORDER constrainMinTotalSize:flag]; }
- (void) constrainMaxTotalHeight:(BOOL)flag
		{ [self border:MISC_ROW_BORDER constrainMaxTotalSize:flag]; }


//-----------------------------------------------------------------------------
// - setFrameSize:
//-----------------------------------------------------------------------------
- (void) setFrameSize:(NSSize)s
	{
	[super setFrameSize:s];
	[self constrainSize];
	}


//-----------------------------------------------------------------------------
// - forwardBGColor
//-----------------------------------------------------------------------------
- (void) forwardBGColor
	{
	[super setBackgroundColor:backgroundColor];
	[colInfo.clip setBackgroundColor:backgroundColor];
	[rowInfo.clip setBackgroundColor:backgroundColor];
	}


//-----------------------------------------------------------------------------
// - initBorder:type:
//-----------------------------------------------------------------------------
- (void) initBorder:(MiscBorderInfo*) p type:(MiscBorderType) type
	{
	NSZone* const z = [self zone];
	if (p->border == 0)
		p->border = new( NSZoneMalloc(z,sizeof(*(p->border))) )
						MiscTableBorder( type );
	p->border->setOwner( self );

	if (type == MISC_COL_BORDER)
		p->view = [[MiscColView allocWithZone:z] 
						initWithFrame:NSZeroRect scroll:self info:p->border];
	else
		p->view = [[MiscRowView allocWithZone:z]
						initWithFrame:NSZeroRect scroll:self info:p->border];

	p->clip = [[NSClipView allocWithZone:z] initWithFrame:NSZeroRect];
	[p->clip setDocumentView:p->view];
	if (p->isOn)
		{
		[self addSubview: p->clip];
		[[self window] invalidateCursorRectsForView:p->view];
		}
	}


//-----------------------------------------------------------------------------
// - doInit:
//
//		NOTE *1*: The tableView was initialized with an argument of 0 so it's
//				frame is invalid.  Calling constrainSize with an invalid frame
//				can result in crashers when it tries to access slots outside
//				of the appropriate area.  Calling -tile first ensures that the
//				frame is valid.
//-----------------------------------------------------------------------------
- (void) doInit:(int)ver
	{
	[self setBorderType:NSBezelBorder];
	[super setHasHorizontalScroller:YES];
	[super setHasVerticalScroller:YES];

	info[ MISC_COL_BORDER ] = &colInfo;
	info[ MISC_ROW_BORDER ] = &rowInfo;

	[self initBorder:&colInfo type:MISC_COL_BORDER];
	[self initBorder:&rowInfo type:MISC_ROW_BORDER];

	NSZone* const z = [self zone];

	tableView = [[MiscTableView allocWithZone:z] initWithFrame:NSZeroRect
				scroll:self colInfo:colInfo.border rowInfo:rowInfo.border];
	nullView = [[MiscNullView allocWithZone:z] initWithFrame:NSZeroRect];

	if (colInfo.isOn && rowInfo.isOn)
		[self addSubview:nullView];

	[self setDocumentView:tableView];
	[self tile];										// NOTE *1*
	[self constrainSize];
	[self forwardBGColor];

	[self registerServicesTypes];
	}


//-----------------------------------------------------------------------------
// - initWithFrame:
//-----------------------------------------------------------------------------
- (id) initWithFrame:(NSRect)frameRect
	{
	[super initWithFrame:frameRect];

	tag = 0;
	enabled = YES;

	colInfo.border = 0;
	colInfo.isOn = YES;
	colInfo.constrain_min = YES;
	colInfo.constrain_max = NO;

	rowInfo.border = 0;
	rowInfo.isOn = NO;
	rowInfo.constrain_min = NO;
	rowInfo.constrain_max = NO;

	id classObj = [self class];
	font = [[classObj defaultFont] retain];
	textColor = [[classObj defaultTextColor] retain];
	backgroundColor = [[classObj defaultBackgroundColor] retain];
	highlightTextColor = [[classObj defaultHighlightTextColor] retain];
	highlightBackgroundColor =
				[[classObj defaultHighlightBackgroundColor] retain];

	[self doInit:MISC_TS_VERSION];
	return self;
	}


//-----------------------------------------------------------------------------
// - setFrame:
//-----------------------------------------------------------------------------
- (void) setFrame:(NSRect)frameRect
	{
	[super setFrame:frameRect];
	[self constrainSize];
	}


//-----------------------------------------------------------------------------
// - adjustSizes
//-----------------------------------------------------------------------------
- (void) adjustSizes
	{
	if (colInfo.isOn && colInfo.view != 0)
		[colInfo.view adjustSize];
	if (rowInfo.isOn && rowInfo.view != 0)
		[rowInfo.view adjustSize];
	[tableView adjustSize];
	}


//-----------------------------------------------------------------------------
// - tile
//-----------------------------------------------------------------------------
- (void) tile
	{
	[super tile];
	[self adjustSizes];
	NSClipView* cv = [self contentView];

	if (colInfo.isOn && !rowInfo.isOn)
		{
		float colHeight = [colInfo.view frameHeight];
		NSRect colRect;
		NSRect docRect;

		docRect = [cv frame];

		NSDivideRect(docRect, &colRect, &docRect, colHeight, NSMinYEdge);

		[colInfo.clip setFrame:colRect];
		[cv setFrame:docRect];
		[[self window] invalidateCursorRectsForView:colInfo.view];
		}
	else if (!colInfo.isOn && rowInfo.isOn)
		{
		float rowWidth	= [rowInfo.view frameHeight];
		NSRect rowRect;
		NSRect docRect;

		docRect = [cv frame];

		NSDivideRect(docRect, &rowRect, &docRect, rowWidth, NSMinXEdge);

		[rowInfo.clip setFrame:rowRect];
		[cv setFrame:docRect];
		[[self window] invalidateCursorRectsForView:rowInfo.view];
		}
	else if (colInfo.isOn && rowInfo.isOn)
		{
		float colHeight = [colInfo.view frameHeight];
		float rowWidth	= [rowInfo.view frameHeight];
		NSRect nulRect;
		NSRect colRect;
		NSRect rowRect;
		NSRect docRect;

		docRect = [cv frame];

		NSDivideRect( docRect, &colRect, &docRect, colHeight, NSMinYEdge );
		NSDivideRect( colRect, &nulRect, &colRect, rowWidth , NSMinXEdge );
		NSDivideRect( docRect, &rowRect, &docRect, rowWidth , NSMinXEdge );

		[nullView setFrame:nulRect];
		[colInfo.clip setFrame:colRect];
		[rowInfo.clip setFrame:rowRect];
		[cv setFrame:docRect];
		[[self window] invalidateCursorRectsForView:colInfo.view];
		[[self window] invalidateCursorRectsForView:rowInfo.view];
		}
	}


//-----------------------------------------------------------------------------
// - reflectScroll:
//-----------------------------------------------------------------------------
- (void) reflectScrolledClipView:(NSClipView*)aView
	{
	if (aView == [self contentView]) // only reflect position of contentView
		[super reflectScrolledClipView:aView];
	}


//-----------------------------------------------------------------------------
// - scrollClip:to:
//-----------------------------------------------------------------------------
- (void) scrollClipView:(NSClipView*)aClipView toPoint:(NSPoint)aPoint
	{
	if (aClipView == [self contentView])		// contentView only.
		{
		NSRect rect;
		NSWindow* win = [self window];
		[win disableFlushWindow];

		[aClipView scrollToPoint:aPoint];				// Scroll content

		if (colInfo.isOn)
			{
			rect = [colInfo.clip bounds];		// Maybe scroll col headings.
			if (rect.origin.x != aPoint.x)
				{
				rect.origin.x = aPoint.x;
				[colInfo.clip scrollToPoint:rect.origin];
				[[self window] invalidateCursorRectsForView:colInfo.view];
				}
			}

		if (rowInfo.isOn)
			{
			rect = [rowInfo.clip bounds];		// Maybe scroll row labels.
			if (rect.origin.y != aPoint.y)
				{
				rect.origin.y = aPoint.y;
				[rowInfo.clip scrollToPoint:rect.origin];
				[[self window] invalidateCursorRectsForView:rowInfo.view];
				}
			}

		[win enableFlushWindow];
		[win flushWindow];
		}
	else if (aClipView == (id)colInfo.clip)
		{
		NSClipView* cv = [self contentView];
		NSRect rect = [cv bounds];
		rect.origin.x = aPoint.x;
		[self scrollClipView:cv toPoint:rect.origin];
		[self reflectScrolledClipView:cv];
		}
	else if (aClipView == (id)rowInfo.clip)
		{
		NSClipView* cv = [self contentView];
		NSRect rect = [cv bounds];
		rect.origin.y = aPoint.y;
		[self scrollClipView:cv toPoint:rect.origin];
		[self reflectScrolledClipView:cv];
		}
	}



//-----------------------------------------------------------------------------
// Delegate Stuff
//-----------------------------------------------------------------------------
- (id) delegate					{ return delegate; }
- (void) setDelegate:(id)obj	{ delegate = obj; }
- (id) dataDelegate				{ return dataDelegate; }
- (void) setDataDelegate:obj	{ dataDelegate = obj; }

- (NSString*) border:(MiscBorderType)b getDelegateSlotTitle:(int)slot
	{
	if (delegate != 0 &&
		[delegate respondsToSelector:
				@selector(tableScroll:border:slotTitle:)])
		return [delegate tableScroll:self border:b slotTitle:slot];

	if (dataDelegate != 0 &&
		[dataDelegate respondsToSelector:
				@selector(tableScroll:border:slotTitle:)])
		return [dataDelegate tableScroll:self border:b slotTitle:slot];

	NSString* s = @"";
	return s;
	}

- (id) border:(MiscBorderType)b getDelegateSlotPrototype:(int)s
	{
	if (delegate != 0 &&
		[delegate respondsToSelector:
				@selector(tableScroll:border:slotPrototype:)])
		return [delegate tableScroll:self border:b slotPrototype:s];

	if (dataDelegate != 0 &&
		[dataDelegate respondsToSelector:
				@selector(tableScroll:border:slotPrototype:)])
		return [dataDelegate tableScroll:self border:b slotPrototype:s];

	return 0;
	}

- (void) border:(MiscBorderType)b slotDraggedFrom:(int)fromPos to:(int)toPos
	{
	MiscBorderType const ob = otherBorder(b);
	if ([self autoSortSlots:ob])
		{
		int const slot = [self border:b slotAtPosition:toPos];
		if ([self border:b slotSortType:slot] != MISC_SORT_SKIP ||
			[self border:b slotSortFunc:slot] != 0)
			[self sortSlots:ob]; // Don't resort if it doesn't affect ordering.
		}
	if (delegate != 0 &&
		[delegate respondsToSelector:
				@selector(tableScroll:border:slotDraggedFrom:to:)])
		[delegate tableScroll:self border:b slotDraggedFrom:fromPos to:toPos];
	if (dataDelegate != 0 &&
		[dataDelegate respondsToSelector:
				@selector(tableScroll:border:slotDraggedFrom:to:)])
		[dataDelegate tableScroll:self border:b
				slotDraggedFrom:fromPos to:toPos];
	}

- (void) border:(MiscBorderType)b slotResized:(int)n
	{
	if (delegate != 0 &&
		[delegate respondsToSelector:
				@selector(tableScroll:border:slotResized:)])
		[delegate tableScroll:self border:b slotResized:n];
	if (dataDelegate != 0 &&
		[dataDelegate respondsToSelector:
				@selector(tableScroll:border:slotResized:)])
		[dataDelegate tableScroll:self border:b slotResized:n];
	}


//-----------------------------------------------------------------------------
// Target / Action
//-----------------------------------------------------------------------------
- (BOOL) isEnabled						{ return enabled; }
- (void) setEnabled:(BOOL)flag			{ enabled = flag; }

- (id) target							{ return target; }
- (void) setTarget:obj					{ target = obj; }
- (id) doubleTarget						{ return doubleTarget; }
- (void) setDoubleTarget:obj			{ doubleTarget = obj; }
- (SEL) action							{ return action; }
- (void) setAction:(SEL)new_sel			{ action = new_sel; }
- (void) setDoubleAction:(SEL)new_sel	{ doubleAction = new_sel; }
- (SEL) doubleAction					{ return doubleAction; }

- (BOOL) sendAction:(SEL)aSel to:(id)obj
		{
		if (aSel == 0)
			aSel = action;
		if (obj == 0)
			obj = target;
		return [NSApp sendAction:aSel to:obj from:self];
		}

- (SEL) getCellAction:(id)cell
		{
		if ([cell respondsToSelector:@selector(action)])
			return [cell action];
		return 0;
		}

- (id) getCellTarget:(id)cell
		{
		if ([cell respondsToSelector:@selector(target)])
			return [cell target];
		return 0;
		}

- (BOOL) sendAction
		{
		id cell = [self selectedCell];
		return [self sendAction:[self getCellAction:cell]
						to:[self getCellTarget:cell]];
		}

- (BOOL) sendDoubleAction
		{
		return [self sendAction:doubleAction to:doubleTarget];
		}

- (BOOL) sendActionIfEnabled
		{
		if ([self isEnabled]) { [self sendAction]; return YES; }
		return NO;
		}

- (BOOL) sendDoubleActionIfEnabled
		{
		if ([self isEnabled]) { [self sendDoubleAction]; return YES; }
		return NO;
		}


//-----------------------------------------------------------------------------
// FONT
//-----------------------------------------------------------------------------
+ (NSFont*) defaultFont
	{
	return [NSFont userFontOfSize:12.0];
	}

- (NSFont*) font
	{
	return font;
	}


static double get_height( NSFont* font )
	{
	return NSHeight( [font boundingRectForFont] );
	}


- (void) setFont:(NSFont*)newFont
	{
	if (![newFont isEqual:font])
		{
		NSFont* oldFont = [font autorelease];
		font = [newFont retain];
		float old_size = [self uniformSizeRows];
		if (old_size != 0)
			{
			// FIXME: Handle this better.  Different cell-types will have
			// different amounts of "fixed-size" border stuff.
			float const BORDER_THICKNESS = 1;
			old_size -= BORDER_THICKNESS;
			float const new_size =
				floor(	0.5 + (double(old_size) *
						(get_height( newFont ) / get_height( oldFont ))));
			if (new_size != old_size)
				[self setUniformSizeRows:(new_size + BORDER_THICKNESS)];
			}
		// FIXME: Set font in all existing prototype cells.
		// WARNING: Currently, just asking the border for a prototype
		// in a given slot allocates and initializes an array of prototypes.
		if (![self isLazy])				// Eager beaver sets all cells now.
			{
			int const NRows = num_rows;
			int const NCols = num_cols;
			for (int r = 0;	 r < NRows;	 r++)
				{
				for (int c = 0;	 c < NCols;	 c++)
					{
					id cell = [self cellAtRow:r column:c];
					if (cell != 0)
						{
						if ([cell respondsToSelector:@selector(setOwnerFont:)])
							[cell setOwnerFont:newFont];
						else if ([cell respondsToSelector:@selector(setFont:)])
							[cell setFont:newFont];
						}
					}
				}
			}
		if (delegate != 0 &&
			[delegate respondsToSelector:
			@selector(tableScroll:fontChangedFrom:to:)])
			[delegate tableScroll:self fontChangedFrom:oldFont to:newFont];
		if (dataDelegate != 0 &&
			[dataDelegate respondsToSelector:
			@selector(tableScroll:fontChangedFrom:to:)])
			[dataDelegate tableScroll:self fontChangedFrom:oldFont to:newFont];

		[self setNeedsDisplay:YES];
		}
	}


- (void) changeFont:(id)sender
	{
	NSFontManager* fontMgr = [NSFontManager sharedFontManager];
	NSFont* newFont = [fontMgr convertFont:[fontMgr selectedFont]];
	if (newFont != 0 && ![newFont isEqual:font])
		{
		NSFont* oldFont = [[font retain] autorelease];
		[self setFont:newFont];
		if (delegate != 0 &&
			[delegate respondsToSelector:
			@selector(tableScroll:changeFont:to:)])
			[delegate tableScroll:self changeFont:oldFont to:newFont];
		if (dataDelegate != 0 &&
			[dataDelegate respondsToSelector:
			@selector(tableScroll:changeFont:to:)])
			[dataDelegate tableScroll:self changeFont:oldFont to:newFont];
		}
	}


//-----------------------------------------------------------------------------
// COLOR
//
// The following macros expand into the implementations for these functions.
// Their names are listed so they can be found when searched for.
// setBackgroundColor:, setTextColor:,
// setHighlightBackgroundColor:, setHighlightTextColor:
//-----------------------------------------------------------------------------
+ (NSColor*) defaultBackgroundColor			{ return [NSColor lightGrayColor];}
+ (NSColor*) defaultTextColor				{ return [NSColor blackColor]; }
+ (NSColor*) defaultHighlightBackgroundColor{ return [NSColor whiteColor]; }
+ (NSColor*) defaultHighlightTextColor		{ return [NSColor blackColor]; }

- (NSColor*) backgroundColor			{ return backgroundColor; }
- (NSColor*) textColor					{ return textColor; }
- (NSColor*) highlightBackgroundColor	{ return highlightBackgroundColor; }
- (NSColor*) highlightTextColor			{ return highlightTextColor; }

- (void) setColor:(NSColor*)value		{ [self setBackgroundColor:value]; }
- (NSColor*) color						{ return [self backgroundColor]; }

- (void) setColor:(NSColor*)x			// New color value
		var:(NSColor**)v				// Instance variable for the color
		sel1:(SEL)sel1					// "setOwner...Color:" message for cell
		sel2:(SEL)sel2					// "set...Color:" message for cell
		notifySel:(SEL)notifySel		// "..ColorChangedTo:" message
	{
	if (![x isEqual:*v])
		{
		[*v autorelease];
		*v = [x retain];
		if (v == &backgroundColor) [self forwardBGColor];

		if (![self isLazy])
			{
			int const NRows = num_rows;
			int const NCols = num_cols;
			for (int r = 0; r < NRows;	r++)
				for (int c = 0;	 c < NCols;	 c++)
					{
					id cell = [self cellAtRow:r column:c];
					if (cell != 0)
						{
						if ([cell respondsToSelector:sel1])
							(*[cell methodForSelector:sel1])( cell, sel1, x );
						else if ([cell respondsToSelector:sel2])
							(*[cell methodForSelector:sel2])( cell, sel2, x );
						}
					}
			}

		if (delegate != 0 && [delegate respondsToSelector:notifySel])
			(*[delegate methodForSelector:notifySel])
						( delegate, notifySel, self, x );

		if (dataDelegate != 0 && [dataDelegate respondsToSelector:notifySel])
			(*[dataDelegate methodForSelector:notifySel])
						( dataDelegate, notifySel, self, x );

		[self setNeedsDisplay:YES];
		}
	}


#define MISC_SET_COLOR_FUNC(LNAME,CNAME)\
- (void) set##CNAME##Color: (NSColor*) value\
	{ [self setColor:value\
				var:& LNAME##Color\
				sel1:@selector(setOwner##CNAME##Color:)\
				sel2:@selector(set##CNAME##Color:)\
				notifySel:@selector(tableScroll:LNAME##ColorChangedTo:)]; }

MISC_SET_COLOR_FUNC( background, Background )
MISC_SET_COLOR_FUNC( text, Text )
MISC_SET_COLOR_FUNC( highlightBackground, HighlightBackground )
MISC_SET_COLOR_FUNC( highlightText, HighlightText )
#undef MISC_SET_COLOR_FUNC


//-----------------------------------------------------------------------------
// SAVE / RESTORE
//-----------------------------------------------------------------------------

- (NSString*) stringFromIntArray:(NSArray*)array
	{
	NSMutableString* s = [[[NSMutableString alloc] init] autorelease];
	for (int i = 0, lim = [array count]; i < lim; i++)
		{
		if (i > 0) [s appendString:@" "];
		[s appendString: [[array objectAtIndex:i] stringValue]];
		}
	return s;
	}

- (NSArray*) intArrayFromString:(NSString*)s
	{
	NSMutableArray* array = [NSMutableArray array];
	NSScanner* scanner = [NSScanner scannerWithString:s];
	int i;
	while (![scanner isAtEnd])
		if ([scanner scanInt:&i])
			[array addObject: [NSNumber numberWithInt:i]];
	return array;
	}


- (NSArray*) slotOrder:(MiscBorderType)b
	{
	NSMutableArray* array = [NSMutableArray array];
	MiscTableBorder const* const bp = info[b]->border;
	MiscCoord_P const* const vmap = bp->getVMap();
	int const lim = bp->count();
	for (int i = 0;	 i < lim;  i++)
		[array addObject: [NSNumber numberWithInt: (vmap ? vmap[i] : i)]];
	return array;
	}
- (NSArray*) colOrder { return [self slotOrder:MISC_COL_BORDER]; }
- (NSArray*) rowOrder { return [self slotOrder:MISC_ROW_BORDER]; }

- (BOOL) border:(MiscBorderType)b setSlotOrder:(NSArray*)array
	{
	BOOL ret = NO;
	if (array != 0)
		{
		MiscTableBorder* const bp = info[b]->border;
		int const lim = bp->count();
		if ([array count] == lim)
			{
			MiscCoord_V* vmap;
			vmap = (MiscCoord_V*) malloc( lim * sizeof(MiscCoord_V) );
			if (vmap != 0)
				{
				for (int i = 0; i < lim; i++)
					vmap[i] = [[array objectAtIndex:i] intValue];

				if (bp->setVMap( vmap ))
					{
					MiscBorderType const ob = otherBorder(b);
					if ([self autoSortSlots:ob])
						[self sortSlots:ob];
					ret = YES;
					}
				free( vmap );
				}
			}
		}
	return ret;
	}
- (BOOL) setColOrder:(NSArray*)list
	{ return [self border:MISC_COL_BORDER setSlotOrder:list]; }
- (BOOL) setRowOrder:(NSArray*)list
	{ return [self border:MISC_ROW_BORDER setSlotOrder:list]; }

- (NSString*) slotOrderAsString:(MiscBorderType)b
	{ return [self stringFromIntArray: [self slotOrder:b]]; }
- (NSString*) colOrderAsString;
	{ return [self slotOrderAsString:MISC_COL_BORDER]; }
- (NSString*) rowOrderAsString
	{ return [self slotOrderAsString:MISC_ROW_BORDER]; }

- (BOOL) border:(MiscBorderType)b setSlotOrderFromString:(NSString*)s
	{ return [self border:b setSlotOrder:[self intArrayFromString:s]]; }
- (BOOL) setColOrderFromString:(NSString*)s
	{ return [self border:MISC_COL_BORDER setSlotOrderFromString:s]; }
- (BOOL) setRowOrderFromString:(NSString*)s
	{ return [self border:MISC_ROW_BORDER setSlotOrderFromString:s]; }


- (NSArray*) slotSizes:(MiscBorderType)b
	{
	NSMutableArray* array = [NSMutableArray array];
	MiscTableBorder const* const bp = info[b]->border;
	int const lim = bp->count();
	for (int i = 0;	 i < lim;  i++)
		[array addObject: [NSNumber numberWithInt:(int)bp->getSize_P(i)]];
	return array;
	}
- (NSArray*) colSizes { return [self slotSizes:MISC_COL_BORDER]; }
- (NSArray*) rowSizes { return [self slotSizes:MISC_ROW_BORDER]; }

- (BOOL) border:(MiscBorderType)b setSlotSizes:(NSArray*)array
	{
	BOOL ret = NO;
	if (array != 0)
		{
		MiscTableBorder* const bp = info[b]->border;
		int const lim = bp->count();
		if ([array count] == lim)
			{
			for (int i = 0;	 i < lim;  i++)
				bp->setSize_P( i, (MiscPixels)
						[[array objectAtIndex:i] intValue] );
			[self constrainSize];
			ret = YES;
			}
		}
	return ret;
	}
- (BOOL) setColSizes:(NSArray*)list
	{ return [self border:MISC_COL_BORDER setSlotSizes:list]; }
- (BOOL) setRowSizes:(NSArray*)list
	{ return [self border:MISC_ROW_BORDER setSlotSizes:list]; }

- (NSString*) slotSizesAsString:(MiscBorderType)b
	{ return [self stringFromIntArray: [self slotSizes:b]]; }
- (NSString*) colSizesAsString
	{ return [self slotSizesAsString:MISC_COL_BORDER]; }
- (NSString*) rowSizesAsString
	{ return [self slotSizesAsString:MISC_ROW_BORDER]; }

- (BOOL) border:(MiscBorderType)b setSlotSizesFromString:(NSString*)s
	{ return [self border:b setSlotSizes: [self intArrayFromString:s]]; }
- (BOOL) setColSizesFromString:(NSString*)s
	{ return [self border:MISC_COL_BORDER setSlotSizesFromString:s]; }
- (BOOL) setRowSizesFromString:(NSString*)s
	{ return [self border:MISC_ROW_BORDER setSlotSizesFromString:s]; }



//-----------------------------------------------------------------------------
// Border Views
//-----------------------------------------------------------------------------
- (BOOL) setBorder:(MiscBorderType)type on:(BOOL)on
	{
	MiscBorderInfo& b = *(info[type]);
	if (b.isOn != on)
		{
		BOOL const other_border_is_on = info[otherBorder(type)]->isOn;
		b.isOn = on;
		if (on)
			{
			[self addSubview: b.clip];
			[[self window] invalidateCursorRectsForView:b.view];
			if (other_border_is_on)
				[self addSubview: nullView];
			}
		else
			{
			[b.clip removeFromSuperview];
			if (other_border_is_on)
				[nullView removeFromSuperview];
			}
		[self tile];
		[self constrainSize];
		[self setNeedsDisplay:YES];
		return YES;
		}
	return NO;
	}


//-----------------------------------------------------------------------------
// SLOT methods
//-----------------------------------------------------------------------------
- (MiscTableBorder*) border:(MiscBorderType)b
		{ return info[b]->border; }
- (BOOL) border:(MiscBorderType)b setSlotTitlesOn:(BOOL)on_off
		{ return [self setBorder:b on:on_off]; }
- (BOOL) slotTitlesOn:(MiscBorderType)b
		{ return info[b]->isOn; }
- (MiscTableTitleMode) slotTitleMode:(MiscBorderType)b
		{ return info[b]->border->getTitleMode(); }
- (void) border:(MiscBorderType)b setSlotTitleMode:(MiscTableTitleMode)x
		{
		MiscBorderInfo* const ip = info[b];
		if (ip->border->setTitleMode(x) && ip->isOn && ip->border->count() > 0)
			[ip->view setNeedsDisplay:YES];
		}

- (void) border:(MiscBorderType)b moveSlotFrom:(int)fromPos to:(int)toPos
		{ info[b]->border->moveFromTo(fromPos,toPos); }
- (int) border:(MiscBorderType)b slotPosition:(int)n
		{ return info[b]->border->physicalToVisual(n); }
- (int) border:(MiscBorderType)b slotAtPosition:(int)n
		{ return info[b]->border->visualToPhysical(n); }

- (NSArray*) border:(MiscBorderType)b physicalToVisual:(NSArray*)p_array
		{
		NSMutableArray* v_array = [NSMutableArray array];
		for (unsigned int i = 0, lim = [p_array count]; i < lim; i++)
			{
			MiscCoord_P const p = [[p_array objectAtIndex:i] intValue];
			MiscCoord_V const v = [self border:b slotPosition:p];
			[v_array addObject: [NSNumber numberWithInt:v]];
			}
		return v_array;
		}

- (NSArray*) border:(MiscBorderType)b visualToPhysical:(NSArray*)v_array
		{
		NSMutableArray* p_array = [NSMutableArray array];
		for (unsigned int i = 0, lim = [v_array count]; i < lim; i++)
			{
			MiscCoord_V const v = [[v_array objectAtIndex:i] intValue];
			MiscCoord_P const p = [self border:b slotAtPosition:v];
			[p_array addObject: [NSNumber numberWithInt:p]];
			}
		return p_array;
		}


- (BOOL) sizeableSlots:(MiscBorderType)b
		{ return info[b]->border->isSizeable(); }
- (BOOL) draggableSlots:(MiscBorderType)b
		{ return info[b]->border->isDraggable(); }
- (BOOL) modifierDragSlots:(MiscBorderType)b
		{ return info[b]->border->isModifierDrag(); }
- (float) uniformSizeSlots:(MiscBorderType)b
		{ return (float) info[b]->border->getUniformSize(); }

- (float) border:(MiscBorderType)b slotAdjustedSize:(int)n
		{ return (float) info[b]->border->effectiveSize_P(n); }
- (float) border:(MiscBorderType)b slotSize:(int)n
		{ return (float) info[b]->border->getSize_P(n); }
- (float) border:(MiscBorderType)b slotMinSize:(int)n
		{ return (float) info[b]->border->getMinSize_P(n); }
- (float) border:(MiscBorderType)b slotMaxSize:(int)n
		{ return (float) info[b]->border->getMaxSize_P(n); }
- (float) border:(MiscBorderType)b slotDataSize:(int)n
		{ return (float) info[b]->border->getDataSize_P(n); }
- (BOOL) border:(MiscBorderType)b slotIsSizeable:(int)n
		{ return info[b]->border->isSizeable_P(n); }
- (BOOL) border:(MiscBorderType)b slotExpandsToData:(int)n
		{ return info[b]->border->isData_P(n); }
- (BOOL) border:(MiscBorderType)b slotIsAutosize:(int)n
		{ return info[b]->border->isSpringy_P(n); }
- (NSString*) border:(MiscBorderType)b slotTitle:(int)n
		{ return info[b]->border->getTitle_P(n); }
- (int) border:(MiscBorderType)b slotTag:(int)n
		{ return info[b]->border->getTag_P(n); }
- (MiscTableCellStyle) border:(MiscBorderType)b slotCellType:(int)n
		{ return info[b]->border->getStyle_P(n); }
- (id) border:(MiscBorderType)b slotCellPrototype:(int)n
		{ return info[b]->border->getPrototype_P(n); }

- (void) border:(MiscBorderType)b setSizeableSlots:(BOOL)flag
		{ info[b]->border->setSizeable(flag); }
- (void) border:(MiscBorderType)b setDraggableSlots:(BOOL)flag
		{ info[b]->border->setDraggable(flag); }
- (void) border:(MiscBorderType)b setModifierDragSlots:(BOOL)flag
		{ info[b]->border->setModifierDrag(flag); }
- (void) border:(MiscBorderType)b setUniformSizeSlots:(float)uniform_size
		{
		MiscBorderInfo* const ip = info[b];
		if (ip->border->setUniformSize((MiscPixels)floor(uniform_size)))
			{
			[self constrainSize];
			if (b == MISC_ROW_BORDER)
				{
				[self setLineScroll:uniform_size];
				[self setPageScroll:uniform_size];
				}
			[self setNeedsDisplay:YES];
			}
		}

- (void) border:(MiscBorderType)b setSlot:(int)n size:(float)size
		{
		info[b]->border->setSize_P( n, (MiscPixels)floor(size) );
		[self constrainSize];
		}
- (void) border:(MiscBorderType)b setSlot:(int)n minSize:(float)size
		{
		info[b]->border->setMinSize_P( n, (MiscPixels)floor(size) );
		[self constrainSize];
		}
- (void) border:(MiscBorderType)b setSlot:(int)n maxSize:(float)size
		{
		info[b]->border->setMaxSize_P( n, (MiscPixels)floor(size) );
		[self constrainSize];
		}
- (void) border:(MiscBorderType)b setSlot:(int)n dataSize:(float)size
		{ info[b]->border->setDataSize_P( n, (MiscPixels)floor(size) ); }
- (void) border:(MiscBorderType)b setSlot:(int)n sizeable:(BOOL)flag
		{ info[b]->border->setSizeable_P( n, flag ); }
- (void) border:(MiscBorderType)b setSlot:(int)n expandsToData:(BOOL)flag
		{ info[b]->border->setData_P( n, flag ); }
- (void) border:(MiscBorderType)b setSlot:(int)n autosize:(BOOL)flag
		{
		info[b]->border->setSpringy_P( n, flag );
		[self constrainSize];
		}

- (void) border:(MiscBorderType)b setSlot:(int)n title:(NSString*)title
		{
		MiscBorderInfo* const ip = info[b];
		if (ip->border->setTitle_P( n, title ) && ip->isOn)
			[self border:b drawSlotTitle:n];
		}

- (void) border:(MiscBorderType)b setSlot:(int)n tag:(int)x
		{ info[b]->border->setTag_P( n, x ); }
- (void) border:(MiscBorderType)b setSlot:(int)n
				cellType:(MiscTableCellStyle)type
		{ info[b]->border->setStyle_P(n,type); }
- (void) border:(MiscBorderType)b setSlot:(int)n cellPrototype:(id)p
		{ info[b]->border->setPrototype_P(n,p); }



//-----------------------------------------------------------------------------
// COL methods
//-----------------------------------------------------------------------------
- (MiscTableBorder*) colBorder
		{ return colInfo.border; }
- (BOOL) colTitlesOn
		{ return [self slotTitlesOn:MISC_COL_BORDER]; }
- (BOOL) setColTitlesOn:(BOOL)x
		{ return [self setBorder:MISC_COL_BORDER on:x]; }
- (MiscTableTitleMode) colTitleMode
		{ return [self slotTitleMode:MISC_COL_BORDER]; }
- (void) setColTitleMode:(MiscTableTitleMode)x
		{ [self border:MISC_COL_BORDER setSlotTitleMode:x]; }

- (void) moveColFrom:(int)fromPos to:(int)toPos
		{ return [self border:MISC_COL_BORDER moveSlotFrom:fromPos to:toPos]; }
- (int) colPosition:(int)n
		{ return [self border:MISC_COL_BORDER slotPosition:n]; }
- (int) colAtPosition:(int)n
		{ return [self border:MISC_COL_BORDER slotAtPosition:n]; }

- (BOOL) sizeableCols
		{ return [self sizeableSlots:MISC_COL_BORDER]; }
- (BOOL) draggableCols
		{ return [self draggableSlots:MISC_COL_BORDER]; }
- (BOOL) modifierDragCols
		{ return [self modifierDragSlots:MISC_COL_BORDER]; }
- (float) uniformSizeCols
		{ return [self uniformSizeSlots:MISC_COL_BORDER]; }

- (float) colAdjustedSize:(int)n
		{ return [self border:MISC_COL_BORDER slotAdjustedSize:n]; }
- (float) colSize:(int)n
		{ return [self border:MISC_COL_BORDER slotSize:n]; }
- (float) colMinSize:(int)n
		{ return [self border:MISC_COL_BORDER slotMinSize:n]; }
- (float) colMaxSize:(int)n
		{ return [self border:MISC_COL_BORDER slotMaxSize:n]; }
- (float) colDataSize:(int)n
		{ return [self border:MISC_COL_BORDER slotDataSize:n]; }
- (BOOL) colIsSizeable:(int)n
		{ return [self border:MISC_COL_BORDER slotIsSizeable:n]; }
- (BOOL) colExpandsToData:(int)n
		{ return [self border:MISC_COL_BORDER slotExpandsToData:n]; }
- (BOOL) colIsAutosize:(int)n
		{ return [self border:MISC_COL_BORDER slotIsAutosize:n]; }
- (NSString*) colTitle:(int)n
		{ return [self border:MISC_COL_BORDER slotTitle:n]; }
- (int) colTag:(int)n
		{ return [self border:MISC_COL_BORDER slotTag:n]; }
- (MiscTableCellStyle) colCellType:(int)n
		{ return [self border:MISC_COL_BORDER slotCellType:n]; }
- (id) colCellPrototype:(int)n
		{ return [self border:MISC_COL_BORDER slotCellPrototype:n]; }

- (void) setSizeableCols:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSizeableSlots:flag]; }
- (void) setDraggableCols:(BOOL)flag
		{ [self border:MISC_COL_BORDER setDraggableSlots:flag]; }
- (void) setModifierDragCols:(BOOL)flag
		{ [self border:MISC_COL_BORDER setModifierDragSlots:flag]; }
- (void) setUniformSizeCols:(float)size
		{ [self border:MISC_COL_BORDER setUniformSizeSlots:size]; }

- (void) setCol:(int)n size:(float)size
		{ [self border:MISC_COL_BORDER setSlot:n size:size]; }
- (void) setCol:(int)n minSize:(float)size
		{ [self border:MISC_COL_BORDER setSlot:n minSize:size]; }
- (void) setCol:(int)n maxSize:(float)size
		{ [self border:MISC_COL_BORDER setSlot:n maxSize:size]; }
- (void) setCol:(int)n dataSize:(float)size
		{ [self border:MISC_COL_BORDER setSlot:n dataSize:size]; }
- (void) setCol:(int)n sizeable:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSlot:n sizeable:flag]; }
- (void) setCol:(int)n expandsToData:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSlot:n expandsToData:flag]; }
- (void) setCol:(int)n autosize:(BOOL)flag
		{ [self border:MISC_COL_BORDER setSlot:n autosize:flag]; }
- (void) setCol:(int)n title:(NSString*)title
		{ [self border:MISC_COL_BORDER setSlot:n title:title]; }
- (void) setCol:(int)n tag:(int)x
		{ [self border:MISC_COL_BORDER setSlot:n tag:x]; }
- (void) setCol:(int)n cellType:(MiscTableCellStyle)x
		{ [self border:MISC_COL_BORDER setSlot:n cellType:x]; }
- (void) setCol:(int)n cellPrototype:p
		{ [self border:MISC_COL_BORDER setSlot:n cellPrototype:p]; }


//-----------------------------------------------------------------------------
// ROW methods
//-----------------------------------------------------------------------------
- (MiscTableBorder*) rowBorder
		{ return rowInfo.border; }
- (BOOL) rowTitlesOn
		{ return [self slotTitlesOn:MISC_ROW_BORDER]; }
- (BOOL) setRowTitlesOn:(BOOL)x
		{ return [self setBorder:MISC_ROW_BORDER on:x]; }
- (MiscTableTitleMode) rowTitleMode
		{ return [self slotTitleMode:MISC_ROW_BORDER]; }
- (void) setRowTitleMode:(MiscTableTitleMode)x
		{ [self border:MISC_ROW_BORDER setSlotTitleMode:x]; }

- (void) moveRowFrom:(int)fromPos to:(int)toPos
		{ return [self border:MISC_ROW_BORDER moveSlotFrom:fromPos to:toPos]; }
- (int) rowPosition:(int)n
		{ return [self border:MISC_ROW_BORDER slotPosition:n]; }
- (int) rowAtPosition:(int)n
		{ return [self border:MISC_ROW_BORDER slotAtPosition:n]; }

- (BOOL) sizeableRows
		{ return [self sizeableSlots:MISC_ROW_BORDER]; }
- (BOOL) draggableRows
		{ return [self draggableSlots:MISC_ROW_BORDER]; }
- (BOOL) modifierDragRows
		{ return [self modifierDragSlots:MISC_ROW_BORDER]; }
- (float) uniformSizeRows
		{ return [self uniformSizeSlots:MISC_ROW_BORDER]; }

- (float) rowAdjustedSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotAdjustedSize:n]; }
- (float) rowSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotSize:n]; }
- (float) rowMinSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotMinSize:n]; }
- (float) rowMaxSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotMaxSize:n]; }
- (float) rowDataSize:(int)n
		{ return [self border:MISC_ROW_BORDER slotDataSize:n]; }
- (BOOL) rowIsSizeable:(int)n
		{ return [self border:MISC_ROW_BORDER slotIsSizeable:n]; }
- (BOOL) rowExpandsToData:(int)n
		{ return [self border:MISC_ROW_BORDER slotExpandsToData:n]; }
- (BOOL) rowIsAutosize:(int)n
		{ return [self border:MISC_ROW_BORDER slotIsAutosize:n]; }
- (NSString*) rowTitle:(int)n
		{ return [self border:MISC_ROW_BORDER slotTitle:n]; }
- (int) rowTag:(int)n
		{ return [self border:MISC_ROW_BORDER slotTag:n]; }
- (MiscTableCellStyle) rowCellType:(int)n
		{ return [self border:MISC_ROW_BORDER slotCellType:n]; }
- (id) rowCellPrototype:(int)n
		{ return [self border:MISC_ROW_BORDER slotCellPrototype:n]; }

- (void) setSizeableRows:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSizeableSlots:flag]; }
- (void) setDraggableRows:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setDraggableSlots:flag]; }
- (void) setModifierDragRows:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setModifierDragSlots:flag]; }
- (void) setUniformSizeRows:(float)size
		{ [self border:MISC_ROW_BORDER setUniformSizeSlots:size]; }

- (void) setRow:(int)n size:(float)size
		{ [self border:MISC_ROW_BORDER setSlot:n size:size]; }
- (void) setRow:(int)n minSize:(float)size
		{ [self border:MISC_ROW_BORDER setSlot:n minSize:size]; }
- (void) setRow:(int)n maxSize:(float)size
		{ [self border:MISC_ROW_BORDER setSlot:n maxSize:size]; }
- (void) setRow:(int)n dataSize:(float)size
		{ [self border:MISC_ROW_BORDER setSlot:n dataSize:size]; }
- (void) setRow:(int)n sizeable:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSlot:n sizeable:flag]; }
- (void) setRow:(int)n expandsToData:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSlot:n expandsToData:flag]; }
- (void) setRow:(int)n autosize:(BOOL)flag
		{ [self border:MISC_ROW_BORDER setSlot:n autosize:flag]; }
- (void) setRow:(int)n title:(NSString*)title
		{ [self border:MISC_ROW_BORDER setSlot:n title:title]; }
- (void) setRow:(int)n tag:(int)x
		{ [self border:MISC_ROW_BORDER setSlot:n tag:x]; }
- (void) setRow:(int)n cellType:(MiscTableCellStyle)x
		{ [self border:MISC_ROW_BORDER setSlot:n cellType:x]; }
- (void) setRow:(int)n cellPrototype:p
		{ [self border:MISC_ROW_BORDER setSlot:n cellPrototype:p]; }


//-----------------------------------------------------------------------------
// - readString:isCString:
//-----------------------------------------------------------------------------
- (NSString*) readString:(NSArchiver*)decoder isCString:(BOOL)isCString
	{
	NSString* str = @"";
	if (!isCString)
		str = [decoder decodeObject];
	else
		{
		char* cstr = 0;
		[decoder decodeValueOfObjCType:@encode(char*) at:&cstr];
		if (cstr != 0)
			{
			str = [NSString stringWithCString:cstr];
			NSZoneFree( [decoder objectZone], cstr );
			}
		}
	return str;
	}


//-----------------------------------------------------------------------------
// - read:selector:
//-----------------------------------------------------------------------------
- (SEL) read:(int)ver selector:(NSArchiver*)decoder
	{
	SEL cmd = 0;
	BOOL const isCString = (ver < MISC_TS_VERSION_1000);
	NSString* str = [self readString:decoder isCString:isCString];
	if ([str length] > 0)
		cmd = NSSelectorFromString( str );
	return cmd;
	}


//-----------------------------------------------------------------------------
// - read:globalInfo:
//
//		NOTE: Cannot archive sort_entry_func and sort_slot_func because they 
//				are function addresses.	 
//-----------------------------------------------------------------------------
- (void) read:(int)ver globalInfo:(NSArchiver*)decoder
	{
	[decoder decodeValueOfObjCType:@encode(int) at:&tag];
	[decoder decodeValueOfObjCType:@encode(BOOL) at:&enabled];
	[decoder decodeValueOfObjCType:@encode(BOOL) at:&lazy];
	[decoder decodeValueOfObjCType:@encode(MiscSelectionMode) at:&mode];
	font = [[decoder decodeObject] retain];

	if (ver < MISC_TS_VERSION_1000)
		{
		textColor				= [[decoder decodeNXColor] retain];
		backgroundColor			= [[decoder decodeNXColor] retain];
		highlightTextColor		= [[decoder decodeNXColor] retain];
		highlightBackgroundColor= [[decoder decodeNXColor] retain];
		id nextText				= [decoder decodeObject];
		id previousText			= [decoder decodeObject];
		[self setNextKeyView:nextText];
		[previousText setNextKeyView:self];
		}
	else
		{
		textColor				= [[decoder decodeObject] retain];
		backgroundColor			= [[decoder decodeObject] retain];
		highlightTextColor		= [[decoder decodeObject] retain];
		highlightBackgroundColor= [[decoder decodeObject] retain];
		}

	delegate	= [decoder decodeObject];
	dataDelegate= [decoder decodeObject];
	target		= [decoder decodeObject];
	doubleTarget= [decoder decodeObject];
	action		= [self read:ver selector:decoder];
	doubleAction= [self read:ver selector:decoder];
	}


//-----------------------------------------------------------------------------
// - read:borderSortVector:decoder:
//-----------------------------------------------------------------------------
- (void) read:(int)ver borderSortVector:(MiscBorderInfo*)p
	decoder:(NSArchiver*)decoder
	{
	if (p->sort_vector != 0)
		{
		[p->sort_vector release];
		p->sort_vector = 0;
		}

	if (ver >= MISC_TS_VERSION_1001)
		{
		p->sort_vector = [[decoder decodeObject] retain];
		}
	else
		{
		int n;
		[decoder decodeValueOfObjCType:@encode(int) at:&n];
		if (n > 0)
			{
			NSMutableArray* a =
						[[NSMutableArray allocWithZone:[self zone]] init];
			for (int i = 0;	 i < n;	 i++)
				{
				int j;
				[decoder decodeValueOfObjCType:@encode(int) at:&j];
				[a addObject:[NSNumber numberWithInt:j]];
				}
			p->sort_vector = a;
			}
		}
	}


//-----------------------------------------------------------------------------
// - read:border:decoder:
//-----------------------------------------------------------------------------
- (void) read:(int)ver border:(MiscBorderInfo*)p decoder:(NSArchiver*)decoder
	{
	[self read:ver borderSortVector:p decoder:decoder];
	[decoder decodeValueOfObjCType:@encode(BOOL) at:&(p->isOn)];
	[decoder decodeValueOfObjCType:@encode(BOOL) at:&(p->autoSort)];
	[decoder decodeValueOfObjCType:@encode(BOOL) at:&(p->constrain_min)];
	[decoder decodeValueOfObjCType:@encode(BOOL) at:&(p->constrain_max)];
	if (p->border == 0)
		p->border = new( NSZoneMalloc( [self zone], sizeof(*(p->border)) ) )
						MiscTableBorder( MISC_ROW_BORDER );
	p->border->initWithCoder( decoder, ver );
	}


//-----------------------------------------------------------------------------
// - read:cells:
//
// FIXME: What happens when a non-empty non-lazy TableScroll reads the data
// from a lazy, empty one.	Does the memory get freed?
//-----------------------------------------------------------------------------
- (void) read:(int)ver cells: (NSArchiver*) decoder
	{
	[decoder decodeValueOfObjCType:@encode(int) at:&num_cols];
	[decoder decodeValueOfObjCType:@encode(int) at:&num_rows];
	if (lazy)
		{
		max_rows = 0;			// No rows have been allocated.
		max_cells = 0;
		cells = 0;
		}
	else // (!lazy)
		{
		max_rows = num_rows;
		max_cells = max_rows * num_cols;
		if (max_cells == 0)
			cells = 0;
		else
			{
			int const nbytes = max_cells * sizeof(*cells);
			cells = (id*) NSZoneMalloc( [self zone], nbytes );
			id* p = cells;
			for (int r = 0; r < num_rows; r++)
				for (int c = 0; c < num_cols; c++)
					*p++ = [[decoder decodeObject] retain];
			}
		}
	}


//-----------------------------------------------------------------------------
// - initWithCoder:
//-----------------------------------------------------------------------------
- (id) initWithCoder:(NSCoder*)decoder
	{
	[super initWithCoder:decoder];
	NSString* class_name = [[MiscTableScroll class] description];
	int const ver = [decoder versionForClassName:class_name];
	if (((unsigned int)ver) > ((unsigned int)MISC_TS_VERSION))
		[NSException raise:NSGenericException format:
		@"%@: old library (version %d), can't -read: new object (version %d)",
		class_name, MISC_TS_VERSION, ver ];
	[self read:ver globalInfo:decoder];
	[self read:ver border:&colInfo decoder:decoder];
	[self read:ver border:&rowInfo decoder:decoder];
	[self read:ver cells:decoder];
	[self doInit:ver];
	return self;
	}


//-----------------------------------------------------------------------------
// - writeSelector:encoder:
//-----------------------------------------------------------------------------
- (void) writeSelector: (SEL) cmd encoder: (NSArchiver*) encoder
	{
	NSString* str = @"";
	if (cmd != 0)
		str = NSStringFromSelector( cmd );
	[encoder encodeObject:str];
	}


//-----------------------------------------------------------------------------
// - writeGlobalInfo:
//
//		NOTE: Cannot archive sort_entry_func and sort_slot_func because they 
//				are function addresses.	 
//-----------------------------------------------------------------------------
- (void) writeGlobalInfo: (NSArchiver*) encoder
	{
	[encoder encodeValueOfObjCType:@encode(int) at:&tag];
	[encoder encodeValueOfObjCType:@encode(BOOL) at:&enabled];
	[encoder encodeValueOfObjCType:@encode(BOOL) at:&lazy];
	[encoder encodeValueOfObjCType:@encode(MiscSelectionMode) at:&mode];
	[encoder encodeObject:font];
	[encoder encodeObject:textColor];
	[encoder encodeObject:backgroundColor];
	[encoder encodeObject:highlightTextColor];
	[encoder encodeObject:highlightBackgroundColor];
	[encoder encodeConditionalObject:delegate];
	[encoder encodeConditionalObject:dataDelegate];
	[encoder encodeConditionalObject:target];
	[encoder encodeConditionalObject:doubleTarget];
	[self writeSelector:action encoder:encoder];
	[self writeSelector:doubleAction encoder:encoder];
	}


//-----------------------------------------------------------------------------
// - writeBorder:encoder:
//-----------------------------------------------------------------------------
- (void) writeBorder:(MiscBorderInfo*)p encoder:(NSArchiver*)encoder
	{
	[encoder encodeObject:p->sort_vector];
	[encoder encodeValueOfObjCType:@encode(BOOL) at:&(p->isOn)];
	[encoder encodeValueOfObjCType:@encode(BOOL) at:&(p->autoSort)];
	[encoder encodeValueOfObjCType:@encode(BOOL) at:&(p->constrain_min)];
	[encoder encodeValueOfObjCType:@encode(BOOL) at:&(p->constrain_max)];
	p->border->encodeWithCoder( encoder );
	}


//-----------------------------------------------------------------------------
// - writeCells:
//-----------------------------------------------------------------------------
- (void) writeCells: (NSArchiver*) encoder
	{
	[encoder encodeValueOfObjCType:@encode(int) at:&num_cols];
	[encoder encodeValueOfObjCType:@encode(int) at:&num_rows];
	if (!lazy)
		for (int r = 0;	 r < num_rows;	r++)
			for (int c = 0;	 c < num_cols;	c++)
				[encoder encodeObject:[self cellAtRow:r column:c]];
	}


//-----------------------------------------------------------------------------
// - encodeWithCoder:
//-----------------------------------------------------------------------------
- (void) encodeWithCoder:(NSCoder*)encoder
	{
	// Ensure that subviews are NOT archived.
	if (colInfo.isOn)	[colInfo.clip removeFromSuperview];
	if (rowInfo.isOn)	[rowInfo.clip removeFromSuperview];
	if (rowInfo.isOn && colInfo.isOn) [nullView removeFromSuperview];
	[tableView removeFromSuperview];

	[super encodeWithCoder:encoder];
	[self writeGlobalInfo:encoder];
	[self writeBorder:&colInfo encoder:encoder];
	[self writeBorder:&rowInfo encoder:encoder];
	[self writeCells:encoder];

	// Restore state.
	[self setDocumentView:tableView];
	if (colInfo.isOn)	[self addSubview:colInfo.clip];
	if (rowInfo.isOn)	[self addSubview:rowInfo.clip];
	if (rowInfo.isOn && colInfo.isOn) [self addSubview:nullView];
	}


//-----------------------------------------------------------------------------
// DRAWING
//-----------------------------------------------------------------------------

- (void) drawCellAtRow:(int)row column:(int)col
	{
	[tableView drawCellAtRow:row column:col];
	[[self window] flushWindow];
	}
- (void) drawRow:(int)row
	{
	[tableView drawRow:row];
	[[self window] flushWindow]; 
	}
- (void) drawCol:(int)col
	{
	[tableView drawCol:col];
	[[self window] flushWindow]; 
	}
- (void) border:(MiscBorderType)b drawSlot:(int)n
	{ return (b == MISC_COL_BORDER) ? [self drawCol:n] : [self drawRow:n]; }

- (void) border:(MiscBorderType)b drawSlotTitle:(int)n	// physical position.
	{
	MiscBorderInfo* const ip = info[b];
	if (ip->isOn)										// visual position.
		{
		[ip->view drawSlot: [self border:b slotPosition:n]];
		[[self window] flushWindow];
		} 
	}
- (void) drawRowTitle:(int)n
	{ return [self border:MISC_ROW_BORDER drawSlotTitle:n]; }
- (void) drawColTitle:(int)n
	{ return [self border:MISC_COL_BORDER drawSlotTitle:n]; }

- (void) selectionChanged
	{
	[rowInfo.view selectionChanged];
	[colInfo.view selectionChanged];
	[tableView selectionChanged];
	}


//-----------------------------------------------------------------------------
// VISIBLE / SCROLLING
//-----------------------------------------------------------------------------

- (void) scrollCellToVisibleAtRow:(int)row column:(int)col
		{ [tableView scrollCellToVisibleAtRow:row column:col]; }
- (void) scrollRowToVisible: (int)row { [tableView scrollRowToVisible:row]; }
- (void) scrollColToVisible: (int)col { [tableView scrollColToVisible:col]; }
- (void) scrollSelToVisible
	{
	if ([self hasRowSelection])
		[self scrollRowToVisible: [self selectedRow]];
	else if ([self hasColSelection])
		[self scrollColToVisible: [self selectedColumn]];
	}


- (int) numVisibleSlots:(MiscBorderType)b
		{ return [tableView numVisibleSlots:b]; }
- (int) firstVisibleSlot:(MiscBorderType)b
		{ return [tableView firstVisibleSlot:b]; }
- (int) lastVisibleSlot:(MiscBorderType)b
		{ return [tableView lastVisibleSlot:b]; }
- (BOOL) border:(MiscBorderType)b slotIsVisible:(int)n
		{ return [tableView border:b slotIsVisible:n]; }
- (void) border:(MiscBorderType)b setFirstVisibleSlot:(int)n
		{ [tableView border:b setFirstVisibleSlot:n]; }
- (void) border:(MiscBorderType)b setLastVisibleSlot:(int)n
		{ [tableView border:b setLastVisibleSlot:n]; }

- (int) numVisibleCols
		{ return [self numVisibleSlots:MISC_COL_BORDER]; }
- (int) firstVisibleCol
		{ return [self firstVisibleSlot:MISC_COL_BORDER]; }
- (int) lastVisibleCol
		{ return [self lastVisibleSlot:MISC_COL_BORDER]; }
- (BOOL) colIsVisible:(int)n
		{ return [self border:MISC_COL_BORDER slotIsVisible:n]; }
- (void) setFirstVisibleCol:(int)n
		{ [self border:MISC_COL_BORDER setFirstVisibleSlot:n]; }
- (void) setLastVisibleCol:(int)n
		{ [self border:MISC_COL_BORDER setLastVisibleSlot:n]; }

- (int) numVisibleRows
		{ return [self numVisibleSlots:MISC_ROW_BORDER]; }
- (int) firstVisibleRow
		{ return [self firstVisibleSlot:MISC_ROW_BORDER]; }
- (int) lastVisibleRow
		{ return [self lastVisibleSlot:MISC_ROW_BORDER]; }
- (BOOL) rowIsVisible:(int)n
		{ return [self border:MISC_ROW_BORDER slotIsVisible:n]; }
- (void) setFirstVisibleRow:(int)n
		{ [self border:MISC_ROW_BORDER setFirstVisibleSlot:n]; }
- (void) setLastVisibleRow:(int)n
		{ [self border:MISC_ROW_BORDER setLastVisibleSlot:n]; }


//-----------------------------------------------------------------------------
// MOUSE & KEYBOARD TRACKING
//-----------------------------------------------------------------------------
- (void) trackBy: (MiscBorderType)b { [tableView trackBy:b]; }
- (MiscBorderType) trackingBy { return [tableView trackingBy]; }

@end
