//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableView.M
//
//		General-purpose 2-D display object that works with the
//		MiscTableScroll to provide row/column sizing and dragging.
//
//		This object is responsible for drawing, mouse and keyboard
//		events in the content portion of the display.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableView.M,v 1.15 96/05/07 09:49:33 sunshine Exp $
// $Log:		MiscTableView.M,v $
//	Revision 1.15  96/05/07	 09:49:33  sunshine
//	Ditched -reflectSelection in favor of -selectionChanged:.
//	-setCursor: ->> -setCursorSlot:	 
//	
//	Revision 1.14  96/05/07	 02:19:42  sunshine
//	Uses new MiscGeometry instead of gobs "type variation" methods.
//	Ripped out all "type variation" methods (which were also duplicated in
//	BorderView).
//	-firstVisibleSlot: now returns the first slot whose leading edge is visible
//	if possible, else the first slot that has some content showing.	 Previously
//	it just returned the first slot with some content showing.	This behavior
//	change results in more pleasant and intuitive scrolling via new keyboard
//	movement functions (page-up, down, etc.).  Ditto for -lastVisibleSlot:.
//	For OpenStep conformance, keyboard selection events (space, alt-space) are
//	now treated the same as mouse events so keybpoardSelect: now calls the
//	tracker's mouseDown:atPos: methods instead of the obsolete keyDown:atPos:.
//	TableFocus's -initOwner: renamed to -initWithOwner:.
//	
//	Revision 1.13  96/05/06	 04:35:35  sunshine
//	Moved keyDown: from MiscTableView to MiscTableScroll.
//-----------------------------------------------------------------------------
#import "MiscTableView.h"
#import "MiscGeometry.h"
#import "MiscHighlightTracker.h"
#import "MiscListTracker.h"
#import "MiscRadioTracker.h"
#import "MiscSparseSet.h"
#import "MiscTableBorder.h"
#import "MiscTableFocus.h"
#import <misckit/MiscTableScroll.h>

#import <new.h>

extern "Objective-C" {
#import <AppKit/NSApplication.h>
#import <AppKit/NSCell.h>
#import <AppKit/NSFontManager.h>
}

extern "C" {
#import <math.h>		// for floor()
}


//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation MiscTableView

//-----------------------------------------------------------------------------
// - initWithFrame:scroll:colInfo:rowInfo:
//
//		NOTE *1*: Default behavior is to take keyboard-cursor information from
//				the row-border.
//-----------------------------------------------------------------------------
- (id) initWithFrame:(NSRect)rect
	scroll:(MiscTableScroll*)i_scroll
	colInfo:(MiscTableBorder*)i_col_border
	rowInfo:(MiscTableBorder*)i_row_border
	{
	NSZone* const z = [self zone];

	rect.size.width = i_col_border->totalSize();
	rect.size.height = i_row_border->totalSize();

	[super initWithFrame:rect];

	scroll = i_scroll;
	col_border = i_col_border;
	row_border = i_row_border;
	tracker_border = MISC_ROW_BORDER;					// NOTE *1*

	oldColSel = new( NSZoneMalloc(z,sizeof(*oldColSel)) ) MiscSparseSet;
	oldRowSel = new( NSZoneMalloc(z,sizeof(*oldRowSel)) ) MiscSparseSet;
	[self setSelectionMode: [scroll selectionMode]];

	focuser = [[MiscTableFocus allocWithZone:z] initWithOwner: self];
	return self;
	}


//-----------------------------------------------------------------------------
// - dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
	[focuser release];
	[tracker release];
	NSZone* const z = [self zone];
	if (oldColSel != 0)
		{
		oldColSel->MiscSparseSet::~MiscSparseSet();
		NSZoneFree( z, oldColSel );
		}
	if (oldRowSel != 0)
		{
		oldRowSel->MiscSparseSet::~MiscSparseSet();
		NSZoneFree( z, oldRowSel );
		}
	[super dealloc];
	}


//-----------------------------------------------------------------------------
// - isFlipped
//-----------------------------------------------------------------------------
- (BOOL) isFlipped
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - isOpaque
//-----------------------------------------------------------------------------
- (BOOL) isOpaque
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// TYPE VARIATIONS
//-----------------------------------------------------------------------------

- (MiscTableBorder*) borderFor: (MiscBorderType)b
		{ return (b == MISC_ROW_BORDER ? row_border : col_border); }
- (MiscBorderType) otherBorder: (MiscBorderType)b
		{ return (b == MISC_ROW_BORDER ? MISC_COL_BORDER : MISC_ROW_BORDER); }



//-----------------------------------------------------------------------------
// drewCellAt:border:oldSel:
//
//		Keeps the old selection sets up to date whenever we draw a cell.  This
//		way -selectionChanged has valid data to work from.
//-----------------------------------------------------------------------------
- (void) drewCellAt: (MiscCoord_V)slot border: (MiscTableBorder*)b
	oldSel: (MiscSparseSet*)oldSel
	{
	BOOL isSelected = b->selectionSet().contains( slot );
	if (isSelected != oldSel->contains(slot))
		{
		if (isSelected)
			oldSel->add( slot );
		else
			oldSel->remove( slot );
		}
	}


//-----------------------------------------------------------------------------
// drewCellAtRow:column:
//-----------------------------------------------------------------------------
- (void) drewCellAtRow: (MiscCoord_V)r column: (MiscCoord_V)c
	{
	[self drewCellAt:r border:row_border oldSel: oldRowSel];
	[self drewCellAt:c border:col_border oldSel: oldColSel];
	}


//-----------------------------------------------------------------------------
// - cellFrameAtRow:column: -- Visual coords
//-----------------------------------------------------------------------------
- (NSRect) cellFrameAtRow:(int)row column:(int)col
	{
	if (row < 0 || row >= row_border->count() ||
		col < 0 || col >= col_border->count())
		return NSMakeRect( 0, 0, 0, 0 );
	else
		return NSMakeRect(	col_border->getOffset(col),
							row_border->getOffset(row),
							col_border->effectiveSize(col),
							row_border->effectiveSize(row) );
	}


//-----------------------------------------------------------------------------
// - getVisFrameAtSlot:from:
//-----------------------------------------------------------------------------
- (NSRect) getVisFrameAtSlot:(MiscCoord_V)slot from:(MiscBorderType)border
	{
	MiscTableBorder* const b = [self borderFor:border];
	MiscRect_O r( border, [self visibleRect] );
	r.setX_O( b->getOffset(slot) );
	r.setWidth_O( b->effectiveSize(slot) );
	return r;
	}


//-----------------------------------------------------------------------------
// - firstVisibleVSlot:
//-----------------------------------------------------------------------------
- (MiscCoord_V) firstVisibleVSlot: (MiscBorderType)bdr
	{
	MiscCoord_V ret = -1;
	MiscTableBorder* const b = [self borderFor:bdr];
	if (b->count() > 0)
		{
		MiscPixels vis_origin = MiscRect_O( bdr, [self visibleRect] ).getX_O();
		ret = b->visualForOffset( vis_origin );
		MiscPixels slot_origin = b->getOffset( ret );
		if (slot_origin < vis_origin)
			ret++;
		}
	return ret;
	}


//-----------------------------------------------------------------------------
// VISIBLE / SCROLLING STUFF
//-----------------------------------------------------------------------------
- (int) firstVisibleSlot:(MiscBorderType)bdr
	{
	MiscCoord_P ret = -1;
	MiscTableBorder* const b = [self borderFor:bdr];
	int const lim = b->count();
	if (lim > 0)
		{
		MiscRect_O r( bdr, [self visibleRect] );
		MiscPixels const vorg = r.getX_O();
		MiscPixels const vlim = r.getMaxX_O();
		MiscCoord_V v = b->visualForOffset( vorg );
		if (v < lim - 1 && b->getOffset(v) < vorg && b->getOffset(v+1) < vlim)
			v++;				// First slot whose leading edge is visible
		ret = b->visualToPhysical( v );
		}
	return ret;
	}


- (int) lastVisibleSlot:(MiscBorderType)bdr
	{
	MiscCoord_P ret = -1;
	MiscTableBorder* const b = [self borderFor:bdr];
	if (b->count() > 0)
		{
		MiscRect_O r( bdr, [self visibleRect] );
		MiscPixels const vorg = r.getX_O();
		MiscPixels const vlim = r.getMaxX_O();
		MiscCoord_V v = b->visualForOffset( vlim );
		if (v > 0 && b->getOffset(v) + b->effectiveSize(v) > vlim &&
					 b->getOffset(v-1) + b->effectiveSize(v-1) > vorg)
			v--;				// Last slot whose trailing edge is visible
		ret = b->visualToPhysical( v );
		}
	return ret;
	}


- (int) numVisibleSlots:(MiscBorderType)bdr
	{
	MiscTableBorder* const b = [self borderFor:bdr];
	if (b->count() > 0)
		{
		MiscRect_O r( bdr, [self visibleRect] );
		MiscPixels const vorg = r.getX_O();
		MiscPixels const vlim = r.getMaxX_O();
		MiscCoord_V const first = b->visualForOffset( vorg );
		MiscCoord_V const last	= b->visualForOffset( vlim );
		return last - first + 1;
		}
	return 0;
	}


- (BOOL) border:(MiscBorderType)bdr slotIsVisible:(int)n
	{
	MiscTableBorder* const b = [self borderFor:bdr];
	if (b->count() > 0)
		{
		MiscCoord_V const x = b->physicalToVisual(n);
		MiscPixels const xorg = b->getOffset(x);
		MiscPixels const xlim = xorg + b->effectiveSize(x);
		MiscRect_O r( bdr, [self visibleRect] );
		MiscPixels const vorg = r.getX_O();
		MiscPixels const vlim = r.getMaxX_O();
		return xorg <= vlim && xlim >= vorg;
		}
	return NO;
	}


- (void) border:(MiscBorderType)bdr setFirstVisibleSlot:(int)n
	{
	MiscTableBorder* const b = [self borderFor:bdr];
	int const num_slots = b->count();
	if (0 <= n && n < num_slots)
		{
		MiscCoord_V const x = b->physicalToVisual(n);
		MiscPixels const xorg = b->getOffset(x);
		MiscRect_O r( bdr, [[self superview] bounds] );
		r.setX_O( xorg );
		[self scrollPoint:r];
		}
	}


- (void) border:(MiscBorderType)bdr setLastVisibleSlot:(int)n
	{
	MiscTableBorder* const b = [self borderFor:bdr];
	int const num_slots = b->count();
	if (0 <= n && n < num_slots)
		{
		MiscPixels const vwidth =
						MiscRect_O( bdr, [self visibleRect] ).getWidth_O();
		MiscCoord_V const x = b->physicalToVisual(n);
		MiscPixels const vorg = b->getOffset(x) + b->effectiveSize(x) - vwidth;
		MiscRect_O r( bdr, [[self superview] bounds] );
		r.setX_O( vorg );
		[self scrollPoint:r];
		}
	}


//-----------------------------------------------------------------------------
// - acceptsFirstMouse:
//-----------------------------------------------------------------------------
- (BOOL) acceptsFirstMouse:(NSEvent*)theEvent
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - acceptsFirstResponder
//-----------------------------------------------------------------------------
- (BOOL) acceptsFirstResponder
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - setFocusFrame
//-----------------------------------------------------------------------------
- (void) setFocusFrame
	{
	NSRect r = NSZeroRect;
	if (row_border->count() > 0 && col_border->count() > 0)
		{
		MiscTableBorder* const b = [self borderFor: tracker_border];
		MiscCoord_V focusSlot = b->getCursor();
		if (focusSlot < 0 || focusSlot >= b->count())
			{
			focusSlot = (b->hasSelection() ? b->selectedSlot() :
						[self firstVisibleVSlot: tracker_border]);
			b->setCursor( focusSlot );
			}
		r = [self getVisFrameAtSlot:focusSlot from:tracker_border];
		}
	[focuser setFrame:r];
	}


//-----------------------------------------------------------------------------
// - startFocus
//-----------------------------------------------------------------------------
- (void) startFocus
	{
	[self setFocusFrame];
	[focuser startFocus];
	}


//-----------------------------------------------------------------------------
// - stopFocus
//-----------------------------------------------------------------------------
- (void) stopFocus
	{
	[focuser stopFocus];
	}


//-----------------------------------------------------------------------------
// - becomeFirstResponder
//-----------------------------------------------------------------------------
- (BOOL) becomeFirstResponder
	{
	NSWindow* win = [self window];
	if (win && [win isKeyWindow])
		{
		[self startFocus];
		[[NSFontManager sharedFontManager]
				setSelectedFont:[scroll font] isMultiple:NO];
		}
	return YES;
	}


//-----------------------------------------------------------------------------
// - resignFirstResponder
//-----------------------------------------------------------------------------
- (BOOL) resignFirstResponder
	{
	[self stopFocus];
	return YES;
	}


//-----------------------------------------------------------------------------
// - becomeKeyWindow
//-----------------------------------------------------------------------------
- (void) becomeKeyWindow
	{
	[self startFocus];
	[[NSFontManager sharedFontManager]
				setSelectedFont:[scroll font] isMultiple:NO];
	}


//-----------------------------------------------------------------------------
// - resignKeyWindow
//-----------------------------------------------------------------------------
- (void) resignKeyWindow
	{
	[self stopFocus];
	}


//-----------------------------------------------------------------------------
// - viewWillMoveToSuperview:
//-----------------------------------------------------------------------------
- (void) viewWillMoveToSuperview:(NSView*)newSuperview
	{
	[self stopFocus];
	}


//-----------------------------------------------------------------------------
// - viewWillMoveToWindow:
//-----------------------------------------------------------------------------
- (void) viewWillMoveToWindow:(NSWindow*)newWindow
	{
	[self stopFocus];
	}


//-----------------------------------------------------------------------------
// - resetCursorRects
//
//		NOTE: This catches srolling.  The blinking cursor is always placed
//				within the visibleRect.	 After scrolling the visibleRect has
//				changed so we need to readjust the cursor frame.
//-----------------------------------------------------------------------------
- (void) resetCursorRects
	{
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// - setFrameSize:
//-----------------------------------------------------------------------------
- (void) setFrameSize:(NSSize)newSize
	{
	[super setFrameSize:newSize];
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// - setFrame:
//-----------------------------------------------------------------------------
- (void) setFrame:(NSRect)r
	{
	[super setFrame:r];
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// - adjustSize
//-----------------------------------------------------------------------------
- (void) adjustSize
	{
	[self setFrameSize:
			NSMakeSize( col_border->totalSize(), row_border->totalSize() )];
	}


//-----------------------------------------------------------------------------
// - scrollCellToVisibleAtRow:column: -- Physical coords
//-----------------------------------------------------------------------------
- (void) scrollCellToVisibleAtRow:(int)row column:(int)col
	{
	NSRect r = [self cellFrameAtRow:row_border->physicalToVisual(row)
				column:col_border->physicalToVisual(col)];
	[self scrollRectToVisible:r];
	}


//-----------------------------------------------------------------------------
// - border:scrollToVisible: -- Physical coord
//-----------------------------------------------------------------------------
- (void) border: (MiscBorderType)bdr scrollToVisible: (MiscCoord_P)slot
	{
	MiscTableBorder* const b = [self borderFor:bdr];
	NSRect r = [self getVisFrameAtSlot:b->physicalToVisual(slot) from:bdr];
	[self scrollRectToVisible:r];
	}


//-----------------------------------------------------------------------------
// - scrollRowToVisible: -- Physical coord
//-----------------------------------------------------------------------------
- (void) scrollRowToVisible: (int)row
	{
	[self border:MISC_ROW_BORDER scrollToVisible:row];
	}


//-----------------------------------------------------------------------------
// - scrollColToVisible: -- Physical coord
//-----------------------------------------------------------------------------
- (void) scrollColToVisible: (int)col
	{
	[self border:MISC_COL_BORDER scrollToVisible:col];
	}


//-----------------------------------------------------------------------------
// - cMin:cMax:rMin:rMax:forRect:
//-----------------------------------------------------------------------------
- (void) cMin: (MiscCoord_V*)cmin cMax: (MiscCoord_V*)cmax
		 rMin: (MiscCoord_V*)rmin rMax: (MiscCoord_V*)rmax
		 forRect: (NSRect) nsrect
	{
	*cmin = col_border->visualForOffset( (MiscPixels) nsrect.origin.x );
	*cmax = col_border->visualForOffset( 
				(MiscPixels) (nsrect.origin.x + nsrect.size.width) - 1 );
	*rmin = row_border->visualForOffset( (MiscPixels) nsrect.origin.y );
	*rmax = row_border->visualForOffset(
				(MiscPixels) (nsrect.origin.y + nsrect.size.height) - 1 );
	}


//-----------------------------------------------------------------------------
// - drawRect:
//-----------------------------------------------------------------------------
- (void) drawRect:(NSRect)nsrect
	{
	MiscCoord_V c, cmin, cmax;
	MiscCoord_V r, rmin, rmax;
	[self cMin:&cmin cMax:&cmax rMin:&rmin rMax:&rmax forRect:nsrect];

	if (cmin >= 0 && cmax >= 0 && rmin >= 0 && rmax >= 0)
		{
		float const x0 = floor( (float) col_border->getOffset( cmin ) );
		float const y0 = floor( (float) row_border->getOffset( rmin ) );
		NSRect rect = NSMakeRect( x0, y0, 0, 0 );
	
		for (r = rmin;	r <= rmax;	r++)
			{
			rect.size.height = floor( (float) row_border->effectiveSize(r) );
			for (c = cmin;	c <= cmax;	c++)
				{
				rect.size.width = floor( (float)col_border->effectiveSize(c) );
				id cell = [scroll cellAtRow:row_border->visualToPhysical(r)
						column:col_border->visualToPhysical(c)];
				[cell drawWithFrame:rect inView:self];
				[self drewCellAtRow:r column:c];
				rect.origin.x = floor( rect.origin.x + rect.size.width );
				}
			rect.origin.x = x0;
			rect.origin.y = floor( rect.origin.y + rect.size.height );
			}
		}
	}


//-----------------------------------------------------------------------------
// - drawCellAtRow:column:updateSel: -- Physical coords
//-----------------------------------------------------------------------------
- (void) drawCellAtRow: (int)row column: (int)col updateSel: (BOOL)updateSel
	{
	if (row >= 0 && row < row_border->count() &&
		col >= 0 && col < col_border->count())
		{
		NSRect v = [self visibleRect];
		MiscCoord_V const vRow = row_border->physicalToVisual(row);
		MiscCoord_V const vCol = col_border->physicalToVisual(col);

		NSRect const r = NSMakeRect(	col_border->getOffset(vCol),
										row_border->getOffset(vRow),
										col_border->effectiveSize(vCol),
										row_border->effectiveSize(vRow) );

		if (!NSIsEmptyRect( NSIntersectionRect(r,v) ))
			[self setNeedsDisplayInRect:r];
		}
	}


//-----------------------------------------------------------------------------
// - drawCellAtRow:column: -- Physical coords
//-----------------------------------------------------------------------------
- (void) drawCellAtRow:(int)row column:(int)col
	{
	[self drawCellAtRow:row column:col updateSel:YES];
	}


//-----------------------------------------------------------------------------
// - drawRow: -- Physical coord
//-----------------------------------------------------------------------------
- (void) drawRow:(int)row
	{
	if ([self canDraw] && row >= 0 && row < row_border->count())
		{
		MiscCoord_V const vRow = row_border->physicalToVisual(row);
	
		NSRect r = [self visibleRect];
		r.origin.y = row_border->getOffset(vRow);
		r.size.height = row_border->effectiveSize(vRow);

		[self setNeedsDisplayInRect:r];
		} 
	}


//-----------------------------------------------------------------------------
// - drawCol: -- Physical coord
//-----------------------------------------------------------------------------
- (void) drawCol:(int)col
	{
	if ([self canDraw] && col >= 0 && col < col_border->count())
		{
		MiscCoord_V const vCol = col_border->physicalToVisual(col);
	
		NSRect r = [self visibleRect];
		r.origin.x = col_border->getOffset(vCol);
		r.size.width = col_border->effectiveSize(vCol);

		[self setNeedsDisplayInRect:r];
		} 
	}


//-----------------------------------------------------------------------------
// - setSelectionMode:
//-----------------------------------------------------------------------------
- (void) setSelectionMode: (MiscSelectionMode) mode
	{
	NSZone* const z = [self zone];
	if (tracker != 0)
		[tracker release];
	switch (mode)
		{
		case MISC_LIST_MODE:
			tracker = [MiscListTracker allocWithZone:z];
			break;
		case MISC_RADIO_MODE:
			tracker = [MiscRadioTracker allocWithZone:z];
			break;
		case MISC_HIGHLIGHT_MODE:
			tracker = [MiscHighlightTracker allocWithZone:z];
			break;
		}
	[tracker initBorder: [self borderFor:tracker_border]];
	}


//-----------------------------------------------------------------------------
// - selectionChanged
//-----------------------------------------------------------------------------
- (void) selectionChanged
	{
	if ([self canDraw])
		{
		MiscSparseSet const& newColSel = col_border->selectionSet();
		MiscSparseSet const& newRowSel = row_border->selectionSet();

		MiscCoord_V cmin, cmax;
		MiscCoord_V rmin, rmax;
		NSRect vis = [self visibleRect];
		[self cMin:&cmin cMax:&cmax rMin:&rmin rMax:&rmax forRect:vis];
	
		for (MiscCoord_V r = rmin; r <= rmax; r++)
			{
			BOOL rowWasOn = oldRowSel->contains(r);
			BOOL rowIsOn = newRowSel.contains(r);
			for (MiscCoord_V c = cmin; c <= cmax; c++)
				{
				BOOL wasOn = rowWasOn || oldColSel->contains(c);
				BOOL isOn = rowIsOn || newColSel.contains(c);
				if (isOn != wasOn)
					{
					[self setNeedsDisplayInRect:
						NSMakeRect(		col_border->getOffset(c),
										row_border->getOffset(r),
										col_border->effectiveSize(c),
										row_border->effectiveSize(r) )];
					}
				}
			}
		*oldColSel = newColSel;
		*oldRowSel = newRowSel;
		}
	}


//-----------------------------------------------------------------------------
// - trackBy:
//-----------------------------------------------------------------------------
- (void) trackBy: (MiscBorderType)b
	{
	tracker_border = b;
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// - trackingBy
//-----------------------------------------------------------------------------
- (MiscBorderType) trackingBy
	{
	return tracker_border;
	}


//-----------------------------------------------------------------------------
// - reflectCursor
//-----------------------------------------------------------------------------
- (void) reflectCursor
	{
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// - constrainSlot:inBorder:
//-----------------------------------------------------------------------------
- (MiscCoord_V) constrainSlot:(MiscCoord_V)s inBorder: (MiscTableBorder*)b
	{
	if (s < 0)
		s = 0;
	else if (s >= b->count())
		s = b->count() - 1;
	return s;
	}


//-----------------------------------------------------------------------------
// - displayNewSelection
//-----------------------------------------------------------------------------
- (void) displayNewSelection
	{
	[scroll selectionChanged];
	[scroll displayIfNeeded];
	}


//-----------------------------------------------------------------------------
// - mouseDown:
//-----------------------------------------------------------------------------
- (void) mouseDown:(NSEvent*)p 
	{
	BOOL const doubleClicked = ([p clickCount] > 1);
	NSPoint evpt = [self convertPoint:[p locationInWindow] fromView:0];
	MiscTableBorder* const b = [self borderFor:tracker_border];
	MiscCoord_V slot =
			b->visualForOffset( MiscPoint_O(tracker_border,evpt).getX_O() );
	int WANTED = (NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSPeriodicMask);

	[focuser stopFocus];
	[tracker mouseDown:p atPos:slot];
	[self borderFor:[self otherBorder:tracker_border]]->selectionSet().empty();
	[self displayNewSelection];

	[NSEvent startPeriodicEventsAfterDelay:0.1 withPeriod:0.1];
	NSEvent* lastEvent = [p copy];

	for (;;)
		{
		p = [[self window] nextEventMatchingMask:WANTED];
		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		else if ([p type] == NSPeriodic)
			[self autoscroll:lastEvent];
		else
			{
			[lastEvent release];
			lastEvent = [p copy];
			}

		NSPoint new_loc =
				[self convertPoint:[lastEvent locationInWindow] fromView:0];
		MiscPixels const offset =
				MiscPoint_O( tracker_border, new_loc ).getX_O();
		MiscCoord_V const new_slot =
				offset < MiscRect_O(tracker_border,[self bounds]).getMaxX_O() ?
				b->visualForOffset(offset) : b->count();
		if (new_slot != slot)
			{
			slot = new_slot;
			[tracker mouseDragged: p atPos: slot];
			[self displayNewSelection];
			}
		}

	[lastEvent release];
	[NSEvent stopPeriodicEvents];

	[tracker mouseUp: p atPos: slot];
	[self displayNewSelection];
	[scroll border:tracker_border setCursorSlot:
				b->visualToPhysical([self constrainSlot:slot inBorder:b])];
	[focuser startFocus];

	if ([scroll isEnabled])
		{
		[scroll sendAction];
		if (doubleClicked)
			[scroll sendDoubleAction];
		}
	}


//-----------------------------------------------------------------------------
// - moveFocusBy:
//-----------------------------------------------------------------------------
- (void) moveFocusBy: (int) delta
	{
	MiscTableBorder* const b = [self borderFor: tracker_border];
	int const lim = b->count();
	if ([focuser isShowingFocus] && lim > 0)
		{
		MiscCoord_V focusSlot = b->getCursor() + delta;
		if (focusSlot < 0)
			focusSlot = lim - 1;
		else if (focusSlot >= lim)
			focusSlot = 0;
		b->setCursor( focusSlot );
		[self border:tracker_border scrollToVisible:
						b->visualToPhysical(focusSlot)];
		[self setFocusFrame];
		}
	}


//-----------------------------------------------------------------------------
// - keyboardSelect:
//-----------------------------------------------------------------------------
- (void) keyboardSelect:(NSEvent*)p 
	{
	[tracker mouseDown:p atPos:[self borderFor:tracker_border]->getCursor()];
	}

@end
