//=============================================================================
//
//		Copyright (C) 1995,1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//
//=============================================================================
//-----------------------------------------------------------------------------
// DirWindow.m
//
//		Manages window which displays directory listing.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id$
// $Log$
//-----------------------------------------------------------------------------
#import "DirWindow.h"
#import "NameCache.h"
#import "Defaults.h"
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableCell.h>
#import <AppKit/NSButton.h>
#import <AppKit/NSFont.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSNibLoading.h>
#import <AppKit/NSPanel.h>
#import <AppKit/NSPasteboard.h>
#import <AppKit/NSTextField.h>
#import <AppKit/NSWorkspace.h>
#import <assert.h>
#import <errno.h>
#import <libc.h>
#import <sys/dir.h>
#import <sys/stat.h>

#ifdef _POSIX_SOURCE
# define GET_CURR_DIR(_BF_,_SZ_)		getcwd(_BF_,_SZ_)
#else
# define GET_CURR_DIR(_BF_,_SZ_)		getwd(_BF_)
#endif

#define USE_PRIVATE_ZONE		1

#define CASCADE_MAX 10
static int cascadeCounter = 0;
static float cascadeOriginX;
static float cascadeOriginY;
static float cascadeDeltaX;
static float cascadeDeltaY;

static BOOL defaultAutoSort;
static BOOL defaultShowHidden;
static BOOL defaultHighlightDirs;
static NSColor* defaultColor;
static NSSize defaultWinSize;
static NSFont* defaultFont;
static NSMutableArray* openDirs = 0;
static OwnerCache* owners = 0;
static GroupCache* groups = 0;

#define COLOR_DEF		@"DirColor"
#define SIZE_DEF		@"DirSize"
#define FONT_DEF		@"DirFont"
#define SORT_DEF		@"AutoSort"
#define HIDDEN_DEF		@"ShowHidden"
#define HLIGHT_DEF		@"HighlightDirs"
#define COL_SIZES_DEF	@"ColSizes"
#define COL_ORDER_DEF	@"ColOrder"


//-----------------------------------------------------------------------------
// normalize_path
//-----------------------------------------------------------------------------
static inline NSString* normalize_path( NSString* buff )
	{
	return [buff stringByStandardizingPath];
	}


//-----------------------------------------------------------------------------
// fsize
//-----------------------------------------------------------------------------
static void fsize( size_t n, id cell )
	{
	[cell setAlignment:NSRightTextAlignment];
	[cell setIntValue:n];
	}


//-----------------------------------------------------------------------------
// mdate
//-----------------------------------------------------------------------------
static void mdate( time_t t, id cell )
	{
	NSCalendarDate* d = [NSCalendarDate dateWithTimeIntervalSince1970:t];
	NSString* s = [d descriptionWithCalendarFormat:@"%m/%d/%y %H:%M"];
	[cell setStringValue:s];
	[cell setTag:(int)t];
	}


//-----------------------------------------------------------------------------
// perms
//-----------------------------------------------------------------------------
static void perms( unsigned int mode, id cell )
	{
	char buff[ 16 ];
	char* s;
	int ftype;
	int c;

	s = buff;

	ftype = (mode & S_IFMT);
	switch (ftype)
		{
		case S_IFDIR:	c = 'd'; break;
		case S_IFCHR:	c = 'c'; break;
		case S_IFBLK:	c = 'b'; break;
		case S_IFREG:	c = '-'; break;
		case S_IFLNK:	c = 'l'; break;
		case S_IFSOCK:	c = 's'; break;
		default:		c = '-'; break;
		}
	*s++ = c;
	*s++ = (mode & 0400) ? 'r' : '-';
	*s++ = (mode & 0200) ? 'w' : '-';
	*s++ = (mode & S_ISUID) ? 's' : (mode & 0100) ? 'x' : '-';
	*s++ = (mode & 0040) ? 'r' : '-';
	*s++ = (mode & 0020) ? 'w' : '-';
	*s++ = (mode & S_ISGID) ? 's' : (mode & 0010) ? 'x' : '-';
	*s++ = (mode & 0004) ? 'r' : '-';
	*s++ = (mode & 0002) ? 'w' : '-';
	*s++ = (mode & S_ISVTX) ? 't' : (mode & 0001) ? 'x' : '-';
	*s = '\0';

	[cell setStringValue:[NSString stringWithCString:buff]];
	[cell setTag:mode];
	}


//-----------------------------------------------------------------------------
// owner
//-----------------------------------------------------------------------------
static void owner( unsigned int uid, id cell )
	{
	[cell setStringValue:[owners lookup:uid]];
	}


//-----------------------------------------------------------------------------
// group
//-----------------------------------------------------------------------------
static void group( unsigned int gid, id cell )
	{
	[cell setStringValue:[groups lookup:gid]];
	}


//-----------------------------------------------------------------------------
// image
//
//		NOTE *1*: -iconForFile: under OpenStep 4.0 for Mach PR2 no longer 
//				returns the icon for the executable inside of application 
//				wrappers (.app).  Instead if some application associates an
//				icon with the extension ".app" then that icon is returned.
//				In PR2, MallocDebug does just this with the result that all
//				".app" directories display the exact *same* icon (that of
//				MallocDebug), so browsing a directory such as /NextApps is
//				pretty pointless (so far as the iconic display goes).  We hack
//				around the problem by explicitly checking for the ".app"
//				extension and fetching the icon of the executable inside the
//				application wrapper.
//-----------------------------------------------------------------------------
static void image( NSString* name, id cell, float w, float h )
	{
	float s = (w < h ? w : h) - 1.0;
	NSSize sz = { s, s };
	NSImage* i;
	if ([[name pathExtension] isEqualToString:@"app"])			// NOTE *1*
		name = [name stringByAppendingPathComponent:
				[[name lastPathComponent] stringByDeletingPathExtension]];
	i = [[NSWorkspace sharedWorkspace] iconForFile:name];
	[i setScalesWhenResized:YES];
	[i setSize:sz];
	[cell setImage:i];
	}



//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation DirWindow

//-----------------------------------------------------------------------------
// + initCascader
//-----------------------------------------------------------------------------
+ (void) initCascader
	{
	NSSize screenSize;
	screenSize = [[NSScreen mainScreen] frame].size;
	cascadeOriginX = screenSize.width / 4;
	cascadeOriginY = screenSize.height - 20;
	cascadeDeltaX = 20;
	cascadeDeltaY = 20;
	}


//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void) initialize
	{
	if (self == [DirWindow class])
		{
		[self initCascader];
		openDirs = [[NSMutableArray alloc] init];
		owners = [OwnerCache commonInstance];
		groups = [GroupCache commonInstance];
		defaultColor = [[Defaults getColor:COLOR_DEF
						fallback:[NSColor lightGrayColor]] retain];
		defaultAutoSort = [Defaults getBool:SORT_DEF fallback:NO];
		defaultShowHidden = [Defaults getBool:HIDDEN_DEF fallback:NO];
		defaultHighlightDirs = [Defaults getBool:HLIGHT_DEF fallback:NO];
		}
	}


//-----------------------------------------------------------------------------
// - cascade
//-----------------------------------------------------------------------------
- (void) cascade
	{
	float top,left;
	left = cascadeOriginX + (cascadeDeltaX * cascadeCounter);
	top	 = cascadeOriginY - (cascadeDeltaY * cascadeCounter);
	[window setFrameTopLeftPoint:NSMakePoint(left, top)];
	if (++cascadeCounter >= CASCADE_MAX)
		cascadeCounter = 0;
	}


//-----------------------------------------------------------------------------
// - isDir:
//-----------------------------------------------------------------------------
- (BOOL) isDir:(int)r
	{
	unsigned int const m = (unsigned int) [scroll tagAtRow:r column:PERMS_SLOT];
	return ((m & S_IFMT) == S_IFDIR);
	}


//-----------------------------------------------------------------------------
// - updateButtons
//-----------------------------------------------------------------------------
- (void) updateButtons
	{
	BOOL const enable = [scroll numSelectedRows] == 1 &&
						[self isDir:[scroll selectedRow]];
	if (enable != [cdButton isEnabled])
		[cdButton setEnabled:enable];
	}


//-----------------------------------------------------------------------------
// - setRow:useOwner:color:
//-----------------------------------------------------------------------------
- (void) setRow:(int)r useOwner:(BOOL)useOwner color:(NSColor*)color
	{
	int i;
	for (i = MAX_SLOT; i-- >= 0; )
		{
		id cell = [scroll cellAtRow:r column:i];
		if (useOwner &&
			[cell respondsToSelector:@selector(setUseOwnerBackgroundColor:)])
			{
			[cell setUseOwnerBackgroundColor:YES];
			if ([cell respondsToSelector:@selector(setOwnerBackgroundColor:)])
				[cell setOwnerBackgroundColor:color];
			}
		else
			{
			if ([cell respondsToSelector:@selector(setBackgroundColor:)])
				[cell setBackgroundColor:color];
			}
		}
	}


//-----------------------------------------------------------------------------
// - highlight:row:
//-----------------------------------------------------------------------------
- (void) highlight:(BOOL)flag row:(int)r
	{
	if (flag)
		[self setRow:r useOwner:NO color:[NSColor cyanColor]];
	else
		[self setRow:r useOwner:YES color:defaultColor];
	}


//-----------------------------------------------------------------------------
// - highlightDirs:
//-----------------------------------------------------------------------------
- (void) highlightDirs:(BOOL)flag
	{
	int const N = [scroll numberOfRows];
	int i;
	for (i = 0; i < N; i++)
		if ([self isDir:i])
			[self highlight:flag row:i];
	}


//-----------------------------------------------------------------------------
// - freeImages
//-----------------------------------------------------------------------------
- (void) freeImages
	{
	int i;
	int const n = [scroll numberOfRows];
	for (i = 0; i < n; i++)		
		[[scroll cellAtRow:i column:ICON_SLOT] setImage:0];
	}


//-----------------------------------------------------------------------------
// - tableScroll:fontChangedFrom:to:
//-----------------------------------------------------------------------------
- (void) tableScroll:(MiscTableScroll*)ts
		fontChangedFrom:(NSFont*)oldFont
		to:(NSFont*)newFont
	{
	[defaultFont autorelease];
	defaultFont = [newFont retain];
	[Defaults set:FONT_DEF font:defaultFont];
	}


//-----------------------------------------------------------------------------
// - tableScroll:border:slotResized:
//-----------------------------------------------------------------------------
- (void) tableScroll:(MiscTableScroll*)ts
		border:(MiscBorderType)b
		slotResized:(int)n
	{
	NSString* s;
	assert( b == MISC_COL_BORDER );
	s = [ts colSizesAsString];
	[Defaults set:COL_SIZES_DEF str:s];
	}


//-----------------------------------------------------------------------------
// - tableScroll:border:slotDraggedFrom:to:
//-----------------------------------------------------------------------------
- (void) tableScroll:(MiscTableScroll*)ts
		border:(MiscBorderType)b
		slotDraggedFrom:(int)fromPos
		to:(int)toPos;
	{
	NSString* s;
	assert( b == MISC_COL_BORDER );
	s = [ts colOrderAsString];
	[Defaults set:COL_ORDER_DEF str:s];
	}


//-----------------------------------------------------------------------------
// - fillScroll
//-----------------------------------------------------------------------------
- (void) fillScroll
	{
	size_t total_bytes = 0;
	DIR* dirp;
	NSString* buff;

	[self freeImages];
	[scroll empty];

	if ((dirp = opendir( [path cString] )) == 0)
		{
		NSRunAlertPanel( @"Can't Open", @"Cannot open directory, %@\n%d:%s",
				@"OK", 0, 0, path, errno, strerror(errno));
		}
	else
		{
		int rc;
		int row = -1;
		struct stat st;
		struct direct const* dp;

		while ((dp = readdir(dirp)) != 0)
			{			// Do not include the "." (self) entry.
			int const dlen = dp->d_namlen;
			char const* const dname = dp->d_name;
			BOOL const dot_name = (dname[0] == '.');
			if ((dlen > 1 || !dot_name) &&		// exclude "."
				(showHidden || !dot_name ||
					(dlen == 2 && dname[1] == '.')))
				{
				NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
				NSString* fullpath = [path stringByAppendingPathComponent:
						[NSString stringWithCString:dname]];
				char const* pFullpath = [fullpath cString];

				[scroll addRow]; row++;
				[[scroll cellAtRow:row column:NAME_SLOT]
						setStringValue:[NSString stringWithCString:dname]];

				if ((rc = lstat( pFullpath, &st )) == 0)
					{
					if ((st.st_mode & S_IFMT) == S_IFLNK)		// soft-link?
						{
						char symbuf[ FILENAME_MAX ];
						int cc;
						cc = readlink( pFullpath, symbuf, FILENAME_MAX );
						if (cc >= 0)
							{
							symbuf[ cc ] = '\0';
							[[scroll cellAtRow:row column:SOFTLINK_SLOT]
										setStringValue:
										[NSString stringWithCString:symbuf]];
							rc = stat( pFullpath, &st );
							}
						else
							rc = -1;
						}
					}
				if (rc == 0)
					{
					total_bytes += st.st_size;
					image( fullpath, [scroll cellAtRow:row column:ICON_SLOT],
								[scroll colSize:ICON_SLOT],
								[scroll uniformSizeRows] );
					fsize( st.st_size,
								[scroll cellAtRow:row column:SIZE_SLOT] );
					mdate( st.st_mtime,
								[scroll cellAtRow:row column:MODIFIED_SLOT] );
					perms( st.st_mode,
								[scroll cellAtRow:row column:PERMS_SLOT] );
					owner( st.st_uid,
								[scroll cellAtRow:row column:OWNER_SLOT] );
					group( st.st_gid,
								[scroll cellAtRow:row column:GROUP_SLOT] );
					[[scroll cellAtRow:row column:HARDLINKS_SLOT]
								setIntValue:st.st_nlink];
					[[scroll cellAtRow:row column:MESSAGE_SLOT]
								setStringValue:@""];
					if (highlightDirs && (st.st_mode & S_IFMT) == S_IFDIR)
						[self highlight:YES row:row];
					}
				else // (rc != 0)
					{
					int slot;
					[[scroll cellAtRow:row column:MESSAGE_SLOT] setStringValue:
								[NSString stringWithCString:strerror(errno)]];
					for (slot = SIZE_SLOT; slot < SOFTLINK_SLOT; slot++)
						[[scroll cellAtRow:row column:slot]
								setStringValue:@"?"];
					}

				[pool release];
				}
			}
		closedir(dirp);
		}
	if ([scroll autoSortRows])
		[scroll sortRows];
	[scroll sizeToCells];
	
	buff = [NSString stringWithFormat:@"%d files   %lu bytes",
						[scroll numberOfRows], total_bytes];
	[countField setStringValue:buff];

	[self updateButtons];
	}


//-----------------------------------------------------------------------------
// - setPath:
//-----------------------------------------------------------------------------
- (void) setPath:(NSString*)dirname
	{
	[path autorelease];
	if (dirname == 0) dirname = NSHomeDirectory();
	if ([dirname length] == 0) dirname = @"/";
	path = [dirname retain];
	[window setTitle:path];
	}


//-----------------------------------------------------------------------------
// - load:
//-----------------------------------------------------------------------------
- (void) load:(NSString*)dirname
	{
	[self setPath:dirname];
	[self fillScroll];
	}


//-----------------------------------------------------------------------------
// - open:
//-----------------------------------------------------------------------------
- (void) open:(id)sender
	{
	if ([scroll hasRowSelection])
		{
		int i;
		NSArray* list = [scroll selectedRows];
		for (i = [list count]; i-- > 0; )
			{
			int const row = [[list objectAtIndex:i] intValue];

			NSString* s = [path stringByAppendingPathComponent:
				[scroll stringValueAtRow:row column:NAME_SLOT]];

			unsigned int mode =
				(unsigned int)[scroll tagAtRow:row column:PERMS_SLOT];

			if ((mode & S_IFMT) == S_IFDIR)
				[[self class] launchDir:s];
			else
				[[NSWorkspace sharedWorkspace] openFile:s];
			}
		}
	}


//-----------------------------------------------------------------------------
// - destroy:
//-----------------------------------------------------------------------------
- (void) destroy:(id)sender
	{
	if ([scroll hasRowSelection])
		{
		int i;
		NSArray* list = [scroll selectedRows];
		NSMutableArray* file_list = [NSMutableArray array];

		for (i = [list count]; i-- > 0; )
			{
			int const row = [[list objectAtIndex:i] intValue];
			NSString* s = [scroll stringValueAtRow:row column:NAME_SLOT];
			if (s != 0)
				[file_list addObject:s];
			}

		[[NSWorkspace sharedWorkspace]
				performFileOperation:NSWorkspaceDestroyOperation
				source:path destination:@"" files:file_list tag:0];

		[self fillScroll];
		}
	}


//-----------------------------------------------------------------------------
// - updatePressed:
//-----------------------------------------------------------------------------
- (void) updatePressed:(id)sender
	{
	[self fillScroll];
	}


//-----------------------------------------------------------------------------
// - cdPressed:
//-----------------------------------------------------------------------------
- (void) cdPressed:(id)sender
	{
	if ([scroll numSelectedRows] == 1)
		{
		MiscCoord_P const row = [scroll selectedRow];
		if ([self isDir:row])
			[self load:normalize_path( [path stringByAppendingPathComponent:
						[scroll stringValueAtRow:row column:NAME_SLOT]] )];
		}
	}


//-----------------------------------------------------------------------------
// - autoSortClick:
//-----------------------------------------------------------------------------
- (void) autoSortClick:(id)sender
	{
	BOOL const switchState = [autoSortSwitch state];
	if (autoSort != switchState)
		{
		defaultAutoSort = autoSort = switchState;
		[Defaults set:SORT_DEF bool:defaultAutoSort];
		[scroll setAutoSortRows:switchState];
		if (switchState)
			[scroll sortRows];
		}
	}


//-----------------------------------------------------------------------------
// - hiddenFilesClick:
//
//		FIXME: Should not have to rescan the whole directory just to show/hide
//		hidden files.  A more careful implementation would not be this sloppy.
//-----------------------------------------------------------------------------
- (void) hiddenFilesClick:(id)sender
	{
	BOOL const switchState = [hiddenFilesSwitch state];
	if (showHidden != switchState)
		{
		defaultShowHidden = showHidden = switchState;
		[Defaults set:HIDDEN_DEF bool:defaultShowHidden];
		[self fillScroll];
		}
	}


//-----------------------------------------------------------------------------
// - highlightClick:
//-----------------------------------------------------------------------------
- (void) highlightClick:(id)sender
	{
	BOOL const switchState = [highlightSwitch state];
	if (highlightDirs != switchState)
		{
		defaultHighlightDirs = highlightDirs = switchState;
		[Defaults set:HLIGHT_DEF bool:defaultHighlightDirs];
		[self highlightDirs:highlightDirs];
		[scroll setNeedsDisplay:YES];
		}
	}


//-----------------------------------------------------------------------------
// - didClick:
//-----------------------------------------------------------------------------
- (void) didClick:(id)sender
	{
	[self updateButtons];
	}


//-----------------------------------------------------------------------------
// - didDoubleClick:
//-----------------------------------------------------------------------------
- (void) didDoubleClick:(id)sender
	{
	[self open:sender];
	}


//-----------------------------------------------------------------------------
// - makeKeyAndOrderFront:
//-----------------------------------------------------------------------------
- (void) makeKeyAndOrderFront:(id)sender
	{
	[window makeKeyAndOrderFront:sender];
	}


//-----------------------------------------------------------------------------
// - windowShouldClose:
//-----------------------------------------------------------------------------
- (BOOL) windowShouldClose:(id)sender
	{
	[openDirs removeObject:self];
	[self autorelease];
	return YES;
	}


//-----------------------------------------------------------------------------
// - windowDidResize:
//-----------------------------------------------------------------------------
- (void) windowDidResize:(NSNotification*)notification
	{
	NSWindow* theWindow = [notification object];
	NSRect r = [theWindow frame];
	if (NSWidth (r) != defaultWinSize.width ||
		NSHeight(r) != defaultWinSize.height)
		{
		defaultWinSize = r.size;
		[Defaults set:SIZE_DEF size:defaultWinSize];
		}
	}


//-----------------------------------------------------------------------------
// - setDefaultColor:
//-----------------------------------------------------------------------------
- (void) setDefaultColor: (NSColor*)c
	{
	[defaultColor autorelease];
	defaultColor = [c retain];
	[Defaults set:COLOR_DEF color:c];
	}


//-----------------------------------------------------------------------------
// - setDaColors:
//-----------------------------------------------------------------------------
- (void) setDaColors: (NSColor*)c
	{
	[window setBackgroundColor:c];
	[scroll setBackgroundColor:c];
	[window display];
	}


//-----------------------------------------------------------------------------
// - draggingEntered:
//-----------------------------------------------------------------------------
- (unsigned int) draggingEntered:(id<NSDraggingInfo>)sender
	{
	return ([sender draggingSourceOperationMask] & NSDragOperationGeneric);
	}


//-----------------------------------------------------------------------------
// - performDragOperation:
//-----------------------------------------------------------------------------
- (BOOL) performDragOperation:(id<NSDraggingInfo>)sender
	{
	[self setDefaultColor:
				[NSColor colorFromPasteboard:[sender draggingPasteboard]]];
	[self setDaColors: defaultColor];
	return YES;
	}


//-----------------------------------------------------------------------------
// - initDefaults
//-----------------------------------------------------------------------------
- (void) initDefaults
	{
	static BOOL initialized = NO;
	if (!initialized)
		{
		NSRect r = [window frame];
		defaultWinSize = r.size;
		defaultFont = 
				[[Defaults getFont:FONT_DEF fallback:[scroll font]] retain];
		initialized = YES;
		}
	}


//-----------------------------------------------------------------------------
// - loadDefaults
//-----------------------------------------------------------------------------
- (void) loadDefaults
	{
	NSString* s;

	NSRect r = [window frame];
	r.size = [Defaults getSize:SIZE_DEF fallback:defaultWinSize];
	[window setFrame:r display:NO];

	autoSort = defaultAutoSort;
	showHidden = defaultShowHidden;
	highlightDirs = defaultHighlightDirs;

	[autoSortSwitch setState:autoSort];
	[hiddenFilesSwitch setState:showHidden];
	[highlightSwitch setState:highlightDirs];

	[scroll setAutoSortRows:autoSort];
	[scroll setFont:defaultFont];
	[self setDaColors:defaultColor];

	s = [Defaults getStr:COL_SIZES_DEF fallback:0];
	if (s)
		[scroll setColSizesFromString:s];

	s = [Defaults getStr:COL_ORDER_DEF fallback:0];
	if (s)
		[scroll setColOrderFromString:s];
	}


//-----------------------------------------------------------------------------
// - initWithDir:
//-----------------------------------------------------------------------------
- (id) initWithDir:(NSString*)dirname
	{
	[super init];
	[NSBundle loadNibNamed:@"DirWindow" owner:self];
	[window registerForDraggedTypes:
				[NSArray arrayWithObject:NSColorPboardType]];
	path = [[NSString alloc] init];
	[self initDefaults];
	[self loadDefaults];
	[self load:dirname];
	[openDirs addObject:self];
	[self cascade];
	[window makeKeyAndOrderFront:self];
	return self;
	}


//-----------------------------------------------------------------------------
// - init
//-----------------------------------------------------------------------------
- (id) init
	{
	return [self initWithDir:NSHomeDirectory()];
	}


//-----------------------------------------------------------------------------
// - dealloc
//-----------------------------------------------------------------------------
- (void) dealloc
	{
#if(USE_PRIVATE_ZONE)
	NSZone* const z = [self zone];
#endif
	[window setDelegate:0];
	[window close];
	[self freeImages];
	[window release];
	[path release];
	[super dealloc];
#if(USE_PRIVATE_ZONE)
	NSRecycleZone( z );
#endif
	}


//-----------------------------------------------------------------------------
// - path
//-----------------------------------------------------------------------------
- (NSString*) path
	{
	return path;
	}


//-----------------------------------------------------------------------------
// + findDir:
//-----------------------------------------------------------------------------
+ (DirWindow*) findDir:(NSString*)normalized_path
	{
	if (normalized_path != 0)
		{
		unsigned int i;
		unsigned int const lim = [openDirs count];
		for (i = 0;	 i < lim;  i++)
			{
			DirWindow* p = (DirWindow*)[openDirs objectAtIndex:i];
			NSString* s = [p path];
			if (s != 0 && [s isEqualToString:normalized_path])
				return p;
			}
		}
	return 0;
	}


//-----------------------------------------------------------------------------
// + launchDir:
//-----------------------------------------------------------------------------
+ (void) launchDir:(NSString*)dirname
	{
	DirWindow* p = 0;
	if (dirname == 0) dirname = NSHomeDirectory();
	if (dirname == 0) dirname = @"/";
	dirname = normalize_path( dirname );
	if ((p = [self findDir:dirname]) != 0)
		[p makeKeyAndOrderFront:self];
	else
		{
#if(USE_PRIVATE_ZONE)
		NSZone* z = NSCreateZone( NSPageSize(), NSPageSize(), YES );
		p = [[self allocWithZone:z] initWithDir:dirname];
#else
		p = [[self alloc] initDir:dirname];
#endif
		}
	}


@end
