/**************************************************************************
  CLASS:			MiscDragView
  INHERITS FROM:	View
  PROGRAMMERS:		Todd Thomas			(todd@avocado.supernet.ab.ca)
  					Bill Shirley		(Bill_Shirley@pcp.ca)
					Bruce McKenzie 		(spuds@netcom.com)
					Steve Hayman 		(shayman@Objectario.com)
					Daniel Boehringer	
  BEGAN:			February 14, 1994
  LAST CHANGED:		December 18, 1995
  VERSION:			1.4
  CHANGES:			See the end of MiscDragView.m
  		
  MiscDragView is an abstract class (cannot be used as is) designed to be  
  easily extensible. Please refer to the docs for a more indepth description.
 
  To make any subclass useable for both source and destination dragging:
 
  		-override initDragTypes to register the draggingTypes that your view 	
 			will accept.
 		-override setupForSourceDrag to put the data (representing the 
			dragged icon) on the pasteboard of your choice, and choose 
			the image you would like to be dragged.
			*Don't call the [super setupForSourceDrag].
 		-override performDragOperation: to get the data from the
 			pasteboard (used for a destination drag view)
 	
  The above is the minimum needed, but you'll likely want to override
  other methods to check whether incoming drags are acceptable, etc.
 
  Finally, the delegate (if there is one) will be sent messages like: 
  didInitiateSourceDrag:, didFinishSourceDrag:, didInitiateDestinationDrag:,
  and didFinishDestinationDrag:. (The delegate methods have changed to be
  more like NeXT's will/did methods. The old ones are still sent for 
  backward compatability though). The skeleton of this code came from the 
  NW_DragLab example by Greg Burd and Randy Nelson...thanks for sharing 
  your code. 
 
  This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.

 * Copyright (C) 1995 Robert Todd Thomas
 * Use is governed by the MiscKit license
 ***************************************************************************/

#import <AppKit/NSView.h>


@interface MiscDragView : NSView
{
	id  theImage,theAcceptingImage;			// the image to display the well
	NSSize  imageSize;		// the original size just in case you shrink it
	int  border;			// border type (same as Box)

// why not making a bitfield out of this
	BOOL  allowSourceDragging;
	BOOL  allowDestinationDragging;
	BOOL  freeWhenDragDone;
	BOOL  shadowIncoming;
	BOOL  dragImageIsMyImage;
	
    id  delegate; 				
	id 	target;		// For control-style target/action stuff. ..shayman
	SEL  action;
}

//-------------------- Methods to get things set up 
+ (void)initialize;
- initWithFrame:(NSRect)frameRect;
- initDragTypes;
- (void)dealloc;

//--------------------- Image Manipulation Stuff
- (void)setImage:(NSImage *)anImage;
- setAcceptingImage: (NSImage *)anImage;
- setImageByName: (const char *)aName;
- setAcceptingImageByName: (const char *)aName;
- setImageByFilename: (const char *)aFilename;
- (NSImage *)image;
- setDragImage: (NSImage *)anImage;
- setDragImage: (NSImage *)anImage freeWhenDone: (BOOL)freeIt;
- (NSImage *)dragImage;
- (NSImage *)acceptingImage;
-(const char*) imageName;
-(const char*) acceptingImageName;
//-------------------- Dragging Options
- setAllowSourceDragging: (BOOL)aBool;
- (BOOL)allowSourceDragging;
- setAllowDestinationDragging: (BOOL)aBool;
- (BOOL)allowDestinationDragging;

- (BOOL)acceptForeignDrag;
- (BOOL)acceptLocalDrag;
- (BOOL)acceptSelfDrag;
- (BOOL)retainData;
- (BOOL)shadowIncoming;
- setShadowIncoming:(BOOL) aBool;
- (NSColor *)shadowColor;

- (BOOL)dragImageIsMyImage;
- setDragImageIsMyImage:(BOOL)aDragImageIsMyImage;

//-------------------- Methods for source dragging 
- (void)mouseDown:(NSEvent *)theEvent;
- cleanupAfterSourceDrag;
- (BOOL)setupForSourceDrag;
- (BOOL)slideback;
- draggingPasteboard;
- calculateDragPoint: (NSPoint *)dragPoint andOffset: (NSPoint *)offset;
- (unsigned int)draggingSourceOperationMaskForLocal:(BOOL)flag;
- (void)draggedImage:(NSImage *)image endedAt:(NSPoint)screenPoint deposited:(BOOL)flag;
		
//-------------------- Methods for destination dragging */
- (unsigned int)draggingEntered:sender;
- (unsigned int)draggingUpdated:sender;
- (void)draggingExited:sender;
- (BOOL)prepareForDragOperation:sender;
- (BOOL)performDragOperation:sender;
- (void)concludeDragOperation:sender;
- cleanupAfterDestinationDrag;

//-------------------- Basic useful methods 
#warning ViewConversion: 'acceptsFirstMouse:' (used to be 'acceptsFirstMouse') now takes the event as an arg
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent;
- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent;

//-------------------- Display
- (void)setBorderType:(NSBorderType)aType;
- (NSBorderType)borderType;
#warning RectConversion: drawRect:(NSRect)rects (used to be drawSelf:(const NXRect *)rects :(int)rectCount) no longer takes an array of rects
- (void)drawRect:(NSRect)rects;

//-------------------- Archiving methods
- (id)initWithCoder:(NSCoder *)aDecoder;
- (void)encodeWithCoder:(NSCoder *)aCoder;
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake;

//-------------------- Control-style Target-action stuff  (shayman)
- target;
- (void)setTarget:aTarget;
- (SEL)action;
- (void)setAction:(SEL)anAction;
- (NSString *)stringValue;
- (void)setStringValue:(NSString *)aValue;
- (void)takeStringValueFrom:(id)sender;

@end


@interface NSObject (MiscDragViewDelegate)

- didInitiateSourceDrag: sender;
- didFinishSourceDrag: (BOOL)successful;
- didInitiateDestinationDrag: sender;
- didFinishDestinationDrag: (BOOL)successful;
- initDragTypes:sender;
- setupForSourceDrag:sender;
-(NSImage*) image;
-(NSImage*) acceptingImage;
-(BOOL) takeDataFromPasteboard:sender;

@end

