/**************************************************************************
 * CLASS:		MiscDragView
 * Copyright (C) 1995 Robert Todd Thomas
 * Use is governed by the MiscKit license
 *
 *	See the header file for more information on this class.
 *
 * This object is included in the MiscKit by permission from the author
 * and its use is governed by the MiscKit license, found in the file
 * "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
 * for a list of all applicable permissions and restrictions.
 **************************************************************************/

#import <misckit/MiscBase.h>
#import "MiscDragView-Delegate.h"
#import "MiscDragView.h"

// Since you can only conceivably have one drag going on at once, 
// I made this a class var. You can access it from subclasses through
// setDragImage: and dragImage.
static id  sourceDragImage = nil;

// This image is for the "ghost" image we will display if we are the
// destination view. 
static id  destDragImage = nil;
static NSSize  ddiSize = {0.0, 0.0};

// Used for shadowing images in the view.
static id sourceDragView = nil;
static id destinationDragView = nil;

/*
 * A proper implementation of versioning. If you add ivars to the class (or
 * just want to dork with the reading/writing) check out the comments tagged
 * "Archiving: READ ME". BJM 5/24/94, adjusted 6/17/94.
 */
 
/*
 * **** Archiving: READ ME **** This is the current and defined version of the
 * class. It is used to identify what data will be written and how that will
 * happen. If the read/write stuff is modified AT ALL, this must be bumped up
 * (I always bump by one) and the other comments must be followed. Failure to
 * do so will result in palettes and nibs that cannot be read. BJM 5/24/94
 */
#define MISC_DRAG_VIEW_VERSION 3

@implementation MiscDragView

+ (void)initialize
{
    if (self == [MiscDragView class]) 
	{
		/*
		 * **** Archiving: READ ME **** After bumping the _VERSION, it is
		 * considered common practice to add a comment line indicating the new
		 * version number, date, and modifier. Optionally, the reason for the
		 * change. There is no need to modify the setVersion message. BJM
		 * 5/24/94 
		 */
		// version 0: initial.  (bjm)
		// version 1: add target/action support (Steve_Hayman@Objectario.com)
		// version 2: add accepting image (e.g. opening folder)
		[[MiscDragView class] setVersion:MISC_DRAG_VIEW_VERSION];
    }
    
    return;
}



- initWithFrame:(NSRect)frameRect
{
    [super initWithFrame:frameRect];
	[self initDragTypes];
	[self setAllowSourceDragging: YES];
	[self setAllowDestinationDragging: YES];
	freeWhenDragDone = NO;
	[self setBorderType:NSBezelBorder];	
	theImage = theAcceptingImage = nil;	
	[self setDelegate:nil];
	[self setShadowIncoming:YES];
	[self setTarget:nil];
	[self setDragImageIsMyImage:YES];
    return self;
}



- initDragTypes
{
	// Override this method in subclasses to define what drag types you 
	// want to allow.
	if ([delegate respondsToSelector:@selector(initDragTypes:)])
		[delegate initDragTypes:self];
	return self;
}



- (void)dealloc
{
	if (theImage != nil)
		[theImage release];
		
	{ [super dealloc]; return; };
}



- (void)setImage:(NSImage *)newImage
{		
	// I'm not sure if delaying the freeing of the image is necessary,
	// but better safe than sorry.
	if (theImage != nil)
		[self performSelector:@selector(_freeImage:) object:theImage afterDelay:(0.0) / 1000.0];

	// Setup the new image. If newImage is nil, the view will be cleared.	
	theImage = newImage;
		
    if (theImage != nil)
    {
		imageSize = [theImage size];
		[theImage setScalesWhenResized:YES];
		[theImage setDataRetained:YES];
    }	
    else
	{
		imageSize.width = 0.0;
		imageSize.height = 0.0;
	}

	[self displayRect:[self bounds]];
}


- setAcceptingImage: (NSImage *)newImage
{		
	// I'm not sure if delaying the freeing of the image is necessary,
	// but better safe than sorry.
	if (theImage != nil)
		[self performSelector:@selector(_freeImage:) object:theAcceptingImage afterDelay:(0.0) / 1000.0];

	// Setup the new image. If newImage is nil, the view will be cleared.	
	theAcceptingImage = newImage;
	[self setShadowIncoming:NO];
	return self;
}
- setImageByName: (const char*)newImageName
{	[self setImage:[NSImage imageNamed:[NSString stringWithCString:newImageName]]];	return self;
}
- setAcceptingImageByName: (const char*)newImageName
{	return [self setAcceptingImage:[NSImage imageNamed:[NSString stringWithCString:newImageName]]];
}
-(const char*) imageName
{	return [[theImage name] cString];
}
-(const char*) acceptingImageName
{	return [[theAcceptingImage name] cString];
}

- setImageByFilename: (const char *)aFilename
{
	// If you pass a NULL for filename it will clear the image.
	if (aFilename != NULL && strlen(aFilename) > 0)
		[self setImage:[[NSImage alloc] initByReferencingFile:[NSString stringWithCString:aFilename]]];
	else
		[self setImage:nil];

	return self;
}



- (NSImage *)image
{
	// Returns the image currently displayed in the view.
	return theImage;
}

- (NSImage *)acceptingImage
{
	// Returns the image currently displayed in the view.
	return theAcceptingImage;
}



- setDragImage: (NSImage *)anImage
{
	// Set the image that is to be dragged. You are responsible for 
	// freeing the image when you are done with it (if you allocated 
	// it yourself).
	[self setDragImage: anImage freeWhenDone: NO];
	if(anImage != nil) [self setDragImageIsMyImage:NO];
	return self;
}


 
- setDragImage: (NSImage *)anImage freeWhenDone: (BOOL)freeIt
{
	// Use if you would like MiscDragView to take care of freeing
	// the dragImage when the drag is complete.
	
	sourceDragImage = anImage;
	freeWhenDragDone = freeIt;
	return self;
}



- (NSImage *)dragImage
{
	return 	dragImageIsMyImage?theImage:sourceDragImage;
}



- (void)setBorderType:(NSBorderType)aType
{
	// Border Type can be one of NX_BEZEL, NX_GROOVE, NX_LINE, or NX_NOBORDER,
	// same as the appkit's Box.	
	border = aType;
	[self displayRect:[self bounds]];
}



- (NSBorderType)borderType
{
	return border;
}



- setAllowSourceDragging: (BOOL)aBool
{
	// Sets whether you would like your view to begin a source drag.
	allowSourceDragging = aBool;	
	return self;
}


 
- (BOOL)allowSourceDragging
{
	// If YES is returned, the view can behave as the source of dragging, 
	// NO if it cannot. 
	return allowSourceDragging;
}



- setAllowDestinationDragging: (BOOL)aBool
{
	// Change whether the view can behave as the destination of dragging.
	// Under certain circumstances you may want to not allow a drag into the 
	// view.
	allowDestinationDragging = aBool;
	return self;
}



- (BOOL)allowDestinationDragging
{
	// If YES is returned, the view can behave as the destination of
	// dragging, NO if it cannot.
	return allowDestinationDragging;
}


	
- (BOOL)acceptForeignDrag
{
	// This method defines the behavior of the receiver.  If YES is returned, 
	// the view can behave as the destination of dragging from outside of the 	
	// application, NO if it cannot. Subclasses can override this to change
	// that behavior. 
	return YES;
}



- (BOOL)acceptLocalDrag
{
	// This method defines the behavior of the receiver.  If YES is 
	// returned, the view can behave as the destination of dragging from 
	// within the application, NO if it cannot. Subclasses can override
	// this to change that behavior. 
	return YES;
}



- (BOOL)acceptSelfDrag
{
	// This method returns YES or NO, defining the behavior of the receiver.
	// If YES is returned, the view can receive data that was dragged out of
	// itself, NO if it cannot. Subclasses can override this to change that
	// behavior.  If this method returns YES, then -acceptLocalDrag should
	// return YES as well.
	if ([self allowDestinationDragging] == NO)
		return NO;
		
	return YES;
}




- (BOOL)retainData
{
	// This method returns YES or NO, defining the behavior of the receiver.  
	// If YES is returned, the data being dragged from the view stays in the
	// view when dragged, NO if it does not.  The default implementation
	// returns YES for all instances.  Subclasses can override this to 
	// change that behavior.
	return YES;
}



- (BOOL)shadowIncoming
{
	// If YES is returned, when a view is able to accept a drag, the 
	// dragged object will be shadowed within it.  The default 
	// implementation returns YES for all instances.  Subclasses can 
	// override this to change that behavior. THIS DOES NOT WORK YET.
	return shadowIncoming;
}

- setShadowIncoming:(BOOL)f
{	shadowIncoming=f;
	return self;
}

- (BOOL)dragImageIsMyImage
{	return dragImageIsMyImage;
}
- setDragImageIsMyImage:(BOOL)aDragImageIsMyImage
{	dragImageIsMyImage=aDragImageIsMyImage;
	return self;
}


- (NSColor *)shadowColor
{
	// This method returns the color used to created a dimmed appearance
	// for a destination image.  The default is to return a partially
	// transparent gray.  If you want a different appearance for shadowing,
	// you can override this method.

  	NSColor *	shadowColor;
	shadowColor = [NSColor lightGrayColor];
	shadowColor = [shadowColor colorWithAlphaComponent:0.5];
	return shadowColor;
}



/*-------------------- methods to be the source of a dragging operation */
#define STICKINESS 5.0

- (void)mouseDown:(NSEvent *)theEvent 
{
	// A source drag now will not begin until the mouse has been moved 
	// more than STICKINESS pixels. This was sort of borrowed from Mike 	
	// Ferris's MODocumentWell class in the MOKit.
	
	NSEvent *origEvent, *loopEvent;
  	NSPoint  zero = {0.0, 0.0};
  	NSPoint  origPoint, newLoc;
  	float  distance = 0.0;
  	int  oldMask;

 	if (theImage == nil || [self allowSourceDragging] == NO) {[super mouseDown:theEvent]; return;} 

	// if the overridden setupForSourceDrag returns YES continue
	if (!([self setupForSourceDrag] && [self dragImage]))
		return;

	origPoint = [theEvent locationInWindow];
	origPoint = *origPoint = [self convertPoint:*origPoint fromView:nil]; 	
	origEvent = theEvent;		
#error EventConversion: addToEventMask:NX_MOUSEDRAGGEDMASK: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
#error EventConversion: addToEventMask:NSLeftMouseDraggedMask: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	oldMask = [[self window] addToEventMask:NSLeftMouseDraggedMask];
					
	loopEvent = [[self window] nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
	while ([loopEvent type] != NSLeftMouseUp && distance < STICKINESS)
	{
		newLoc = [loopEvent locationInWindow];
		newLoc = *newLoc = [self convertPoint:*newLoc fromView:nil];

		// if we've gone at least STICKINESS units from the original
		// mousedown, start dragging our file.
		distance = sqrt(((newLoc.x - origPoint.x) * (newLoc.x - origPoint.x)) + 
					((newLoc.y - origPoint.y) * (newLoc.y - origPoint.y)));

		loopEvent = [[self window] nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];		
	 }

#error EventConversion: setEventMask:oldMask: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
#error EventConversion: setEventMask:oldMask: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	[[self window] setEventMask:oldMask];
	[self calculateDragPoint: &origPoint andOffset: &zero];
	
	// begin the drag session if the mouse moved far enough
	if (distance >= STICKINESS)
	{  		
		// send out the delegate message
		[self _sourceDragInitiated: self];		

		sourceDragView = self;
		[self display];
		[self dragImage:[self dragImage] at:*origPoint offset:NSMakeSize((NSMakeSize((&zero)->x,(&zero)->y))->x,(NSMakeSize((&zero)->x,(&zero)->y))->y) event:origEvent pasteboard:[self draggingPasteboard] source:self slideBack:[self slideback]];		
		[self cleanupAfterSourceDrag];
		[self display];
	 }
}



- cleanupAfterSourceDrag
{
	// If we are not supposed to remember the data, then free the image
	// but on a delay. (It crashes after dragging in and out of the view
	// a few times if I use theImage = [theImage free] instead of the 
	// delayed free. It seems to still need the image in the window's
	// drag methods.) Anyway, this seems to work fine this way.
	
	if (destinationDragView != self)
		// If the drag came from ourselves then don't free the image.
		if ([self retainData] == NO)
			[self setImage:nil];
				
	if (freeWhenDragDone == YES)
		[self performSelector:@selector(_freeImage:) object:[self dragImage] afterDelay:(0.0) / 1000.0];
	[self setDragImage: nil];
	
	sourceDragView = nil;
	destinationDragView = nil;
	return self;
}



- (BOOL)setupForSourceDrag
{
	// A method that should be overridden in the subclass to put the 
	// information on the desired pasteboard and select an image to drag. 
	// If done successfully return YES for the drag to continue.

	if ([delegate respondsToSelector:@selector(setupForSourceDrag:)])
	{	[delegate setupForSourceDrag:self];
		return YES;
	} else return NO;
}



- (BOOL)slideback
{
	// Override this method if you like to change when images slide back to
	// their destination view when they aren't deposited in another view.
	// The default is to slide back if the view is retaining the data that
	// is being dragged.
	return [self retainData];
}



- draggingPasteboard
{
	// Override this if you would like to use another pasteboard to do the
	// dragging.
	return [NSPasteboard pasteboardWithName:NSDragPboard];
}



- calculateDragPoint: (NSPoint *)dragPoint andOffset: (NSPoint *)offset
{
	// This method should be overridden if you want to have some control
	// on where the dragged image is first placed, and how far it is offset
	// from the original mousedown (dragPoint).
	// Making the dragPoint be the middle of the image is the default,
	// so it looks nice.
	dragPoint->x -= imageSize.width/2.0;
    dragPoint->y -= imageSize.height/2.0;
	return self;
}



- (unsigned int)draggingSourceOperationMaskForLocal:(BOOL)flag
{
    return NSDragOperationAll;
}



- (void)draggedImage:(NSImage *)image endedAt:(NSPoint)screenPoint deposited:(BOOL)flag
{
	// Sends a "didFinishSourceDrag:" to the delegate if it responds.
	[self _sourceDragFinished: flag];
}



/*-------------------- methods to be the destination of a dragging operation */

- (unsigned int)draggingEntered: sender
{
	if ([self allowDestinationDragging] == NO)
		return NSDragOperationNone;
		
 	if ([self acceptSelfDrag] == NO && [sender draggingSource] == self)
		return NSDragOperationNone;
		
#error GeneralNamingConversion: 'isDraggingSourceLocal' is obsolete. draggingSource returns nil if the dragging source is not local.
#error GeneralNamingConversion: 'isDraggingSourceLocal' is obsolete. draggingSource returns nil if the dragging source is not local.
	if ([self acceptForeignDrag] == NO && [sender isDraggingSourceLocal] == NO)
		return NSDragOperationNone;
		
#error GeneralNamingConversion: 'isDraggingSourceLocal' is obsolete. draggingSource returns nil if the dragging source is not local.
#error GeneralNamingConversion: 'isDraggingSourceLocal' is obsolete. draggingSource returns nil if the dragging source is not local.
    if ([self acceptLocalDrag] == NO && [sender isDraggingSourceLocal] == YES)
		return NSDragOperationNone;

	// Send a "didInitiateDestinationDrag:" to the delegate
	[self _destinationDragInitiated: self];

	destinationDragView = self;
	// We will free the image in our cleanupAfterDestinationDrag method.
	destDragImage = [[sender draggedImage] retain];
	[destDragImage setScalesWhenResized:YES];
	ddiSize = [destDragImage size];
	if ([self shadowIncoming] && [destDragImage lockFocus] == YES)
	{	
		// Dim the image.
		[[self shadowColor] set];
		PScompositerect(0.0, 0.0, ddiSize.width, ddiSize.height, NSCompositeSourceAtop);
	 	[destDragImage unlockFocus];
	 }
	 	
	[self display];
	 
	return NSDragOperationCopy;
}



- (unsigned int)draggingUpdated: sender
{	
	return [super draggingUpdated: sender]; 
}



- (void)draggingExited:sender
{
	// Destination drag came to an end, but was not successful, so send a NO
	// to the delegate
	[self _destinationDragFinished: NO];
	
	// Free the dragged image we made a copy of.
	[self cleanupAfterDestinationDrag];
	destinationDragView = nil;
	[self displayRect:[self bounds]];
}



- (BOOL)prepareForDragOperation:sender
{
    return YES;
}



- (BOOL)performDragOperation:sender
{	if ([delegate respondsToSelector:@selector(takeDataFromPasteboard:)])
		return [delegate takeDataFromPasteboard:self];
	
    return YES;
}



- (void)concludeDragOperation:sender
{
	// Send the delegate a "didFinishDestinationDrag" message. If you override
	// this method make sure to call this so the delegate and action messages 
	// will be sent.

	[self _destinationDragFinished: YES];
	
    // new - if we have a target/action combination, do it. ..shayman
    if ( target && action && [target respondsToSelector:action] ) 
	[target perform:action withObject:self];
	
	[self cleanupAfterDestinationDrag];
	if(![self shadowIncoming]) destinationDragView=nil;	// "closed" icon again
	[self display];
}



- cleanupAfterDestinationDrag
{
	// Perform cleanup after the drag has completed.
	if (destDragImage != nil)
		destDragImage = [destDragImage release];
	ddiSize.width = 0.0;
	ddiSize.height = 0.0;

	return self;
}

	

/*-------------------- methods to display, and other useful stuff */

#warning ViewConversion: 'acceptsFirstMouse:' (used to be 'acceptsFirstMouse') now takes the event as an arg
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
    return YES;
}



- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent 
{
    return YES;
}

 
 
#warning RectConversion: drawRect:(NSRect)rects (used to be drawSelf:(const NXRect *)rects :(int)rectCount) no longer takes an array of rects
- (void)drawRect:(NSRect)rects
{
	NSImage  *dispImage = nil;
	NSSize  origSize;
	NSSize  currSize;
	NSPoint  offset;
	NSRect  imageRect;

	switch (border)
	{
	    case NSGrooveBorder:
	  	  	NSDrawGroove([self bounds] , rects);
		break;
	    
		case NSBezelBorder:
	   	  	NSDrawGrayBezel([self bounds] , rects);
		break;
	    
		case NSLineBorder:
			[[NSColor lightGrayColor] set];
			NSRectFill([self bounds]);
	  	  	[[NSColor blackColor] set];
		  	NSFrameRect([self bounds]);
		break;
	  	
		case NSNoBorder:
		default:
		  	[[NSColor lightGrayColor] set];
		  	NSRectFill([self bounds]);
	  	break;
	  }

	// If there is no image, or we are supposed to give the illusion
	// that the data (image) was not retained then no reason to continue.
	// There seems to be alot of confusing choices where the view 
	// should not show any images...
	if ((sourceDragView != nil && destinationDragView == sourceDragView && 
					[self shadowIncoming] == NO && [self retainData] == NO) 	
					|| (destinationDragView != self && ([self image] == nil || 
					([self retainData] == NO && self == sourceDragView))))
		return;
	 
	// Set so the image will not wipe the border (if any).
    imageRect.origin.x = NSMinX([self bounds]) + 2;
   	imageRect.origin.y = NSMinY([self bounds]) + 2;
    imageRect.size.width = NSWidth([self bounds]) - 4;
    imageRect.size.height = NSHeight([self bounds]) - 4;
	
	// Choose which image should be displayed in the view. If we
	// are the destinationDragView and we are supposed to shadow
	// the image, then use the destDragImage.
	if ([self shadowIncoming] && destinationDragView == self && 
					destDragImage != nil)
	{
		dispImage = destDragImage;
		origSize.width = ddiSize.width;
		origSize.height = ddiSize.height;
	 }
	else
	{	// Otherwise display our image.
		if(destinationDragView == self && theAcceptingImage)
			 dispImage = theAcceptingImage;
		else dispImage = theImage;
		origSize.width = imageSize.width;
		origSize.height = imageSize.height;
	 }
	currSize = [dispImage size];
	
	// Use different offsets depending on if image scaled or not.
	if (currSize.width > NSWidth(imageRect) ||
			currSize.height > NSHeight(imageRect))
	{  
	  	NSSize  scaledSize;
	  	double xscale, yscale, scale;

		// scale the image and keep aspect ratio
	 	xscale = NSWidth(imageRect) / origSize.width;
		yscale = NSHeight(imageRect) / origSize.height;
		scale = ((xscale < yscale) ? xscale : yscale);
		scaledSize.width = (origSize.width * scale); // - 2.0; 
		scaledSize.height = (origSize.height * scale); // - 2.0;
		
		offset = imageRect.origin;
		offset.x += (NSWidth(imageRect) - scaledSize.width) / 2;
		offset.y += (NSHeight(imageRect) - scaledSize.height) / 2;
		[dispImage setSize:*scaledSize];
		
		if (dispImage == destDragImage && [self shadowIncoming])
			// It seems like scaling the image makes it lose it's
			// dimmed look (that we applied in draggingEntered) so
			// do it again.
			if ([dispImage lockFocus] == YES)
			{
				[[self shadowColor] set];
				PScompositerect(0.0, 0.0, scaledSize.width,
								scaledSize.height, NSCompositeSourceAtop);
			 	[dispImage unlockFocus];
			 }	
			
	 }
	else
	{ 
		// no scaling needed
		offset = imageRect.origin;
		offset.x += (NSWidth(imageRect) - currSize.width) / 2.0;
		offset.y += (NSHeight(imageRect) - currSize.height) / 2.0;
	 }

	[dispImage compositeToPoint:offset operation:NSCompositeSourceOver];
}



// Archiving methods

- (id)initWithCoder:(NSCoder *)aDecoder
{
    int version;
	BOOL acceptForeignDrag, acceptLocalDrag;
	BOOL acceptSelfDrag, retainData;

    [super initWithCoder:aDecoder];
    
	version = [aDecoder versionForClassName:@"MiscDragView"];
    
	/*
	 * **** Archiving: READ ME **** This code (and its analogue in write:) is
	 * critical. When you bump MISC_DRAG_VIEW_VERSION, copy the whole _current_
	 * case ("case MISC_DRAG_VIEW_VERSION: ... break;"), and duplicate it.
	 * Change the "MISC_DRAG_VIEW_VERSION" in the old case to the OLD
	 * (pre-bump) version number. Now, dork the "new current" version into
	 * whatever you want. See how this code can now read EITHER version out of
	 * the typed stream?
	 *
	 * If you are adding yet another version, leave the previous versions in
	 * place. That way you can still read archived objects that are more than
	 * one version old. Don't forget to take whatever actions are necessary to
	 * harmonize those old values. For example, if you converted an "int" ivar
	 * to "double", you'd need to (in the old version) to read the int version
	 * into a temp variable, and convert it in to the double var. BJM 5/24/94
	 */
    switch (version)
    {
    case 0:
		theImage = [[aDecoder decodeObject] retain];
		imageSize = [aDecoder decodeSize];
		[aDecoder decodeValuesOfObjCTypes:"i", &border];	
		
		delegate = [[aDecoder decodeObject] retain];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aDecoder decodeValuesOfObjCTypes:"cc", &acceptForeignDrag, &acceptLocalDrag];
		[aDecoder decodeValuesOfObjCTypes:"cc", &acceptSelfDrag, &retainData];
		break;
		
    case 1:
		theImage = [[aDecoder decodeObject] retain];
		imageSize = [aDecoder decodeSize];
		[aDecoder decodeValuesOfObjCTypes:"i", &border];	
		
		delegate = [[aDecoder decodeObject] retain];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aDecoder decodeValuesOfObjCTypes:"cc", &acceptForeignDrag, &acceptLocalDrag];
		[aDecoder decodeValuesOfObjCTypes:"cc", &acceptSelfDrag, &retainData];
		// Version 1 adds target/action stuff -- shayman
		[aDecoder decodeValuesOfObjCTypes:"@:", &target, &action];
		break;

    case 2:
		theImage = [[aDecoder decodeObject] retain];
		imageSize = [aDecoder decodeSize];
		[aDecoder decodeValuesOfObjCTypes:"i", &border];	
		
		delegate = [[aDecoder decodeObject] retain];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aDecoder decodeValuesOfObjCTypes:"@:", &target, &action];
		break;

    case MISC_DRAG_VIEW_VERSION:
		theImage = [[aDecoder decodeObject] retain];
		imageSize = [aDecoder decodeSize];
		[aDecoder decodeValuesOfObjCTypes:"i", &border];	
		
		delegate = [[aDecoder decodeObject] retain];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aDecoder decodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aDecoder decodeValuesOfObjCTypes:"@:", &target, &action];

		[aDecoder decodeValuesOfObjCTypes:"cc", &shadowIncoming,&dragImageIsMyImage];	// new
		theAcceptingImage = [[aDecoder decodeObject] retain];
		break;

    default:
#warning StringConversion: NSStringFromSelector return value is an NSString; use [ cString] if needed, otherwise change the target type to an NSString.
		NSLog(@"[%s %s] - unknown version of %s in typed stream", [[[self class] name] cString], NSStringFromSelector(_cmd), 
			[[[MiscDragView class] name] cString]);
		break;
    }

    return self;
}



- (void)encodeWithCoder:(NSCoder *)aCoder
{
	BOOL acceptForeignDrag, acceptLocalDrag;
	BOOL acceptSelfDrag, retainData;

    [super encodeWithCoder:aCoder];
    
	/*
	 * **** Archiving: READ ME **** Home stretch. Now (just like read:)
	 * duplicate the current case ("case MISC_DRAG_VIEW_VERSION: ... break;").
	 * Once again, change the "MISC_DRAG_VIEW_VERSION" of the first one to the
	 * OLD version number. Now adjust the new/current MISC_DRAG_VIEW_VERSION
	 * (remember, you've bumped it) to the new way of writing vars. See how
	 * this code (because the constant is in the switch statement) always
	 * writes out ONLY the current version, but leaves the old version(s)
	 * around to posterity. DO NOT DELETE THE OLD VERSIONS. Leave them to
	 * clutter up the world. BJM 5/24/94  
	 */
    switch (MISC_DRAG_VIEW_VERSION)
    {
    case 0:
		[aCoder encodeObject:theImage];
		[aCoder encodeSize:imageSize];	
		[aCoder encodeValuesOfObjCTypes:"i", &border];
	
		[aCoder encodeConditionalObject:delegate];
		[aCoder encodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aCoder encodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aCoder encodeValuesOfObjCTypes:"cc", &acceptForeignDrag, &acceptLocalDrag];
		[aCoder encodeValuesOfObjCTypes:"cc", &acceptSelfDrag, &retainData];
		break;

    case 1:
		[aCoder encodeObject:theImage];
		[aCoder encodeSize:imageSize];	
		[aCoder encodeValuesOfObjCTypes:"i", &border];
	
		[aCoder encodeConditionalObject:delegate];
		[aCoder encodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aCoder encodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aCoder encodeValuesOfObjCTypes:"cc", &acceptForeignDrag, &acceptLocalDrag];
		[aCoder encodeValuesOfObjCTypes:"cc", &acceptSelfDrag, &retainData];
		// Version 1 adds support for target/action -- shayman
		[aCoder encodeValuesOfObjCTypes:"@:", &target, &action];
		break;

    case 2:
		[aCoder encodeObject:theImage];
		[aCoder encodeSize:imageSize];	
		[aCoder encodeValuesOfObjCTypes:"i", &border];
	
		[aCoder encodeConditionalObject:delegate];
		[aCoder encodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aCoder encodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aCoder encodeValuesOfObjCTypes:"@:", &target, &action];
		break;

    case MISC_DRAG_VIEW_VERSION:
		[aCoder encodeObject:theImage];
		[aCoder encodeSize:imageSize];	
		[aCoder encodeValuesOfObjCTypes:"i", &border];
	
		[aCoder encodeConditionalObject:delegate];
		[aCoder encodeValuesOfObjCTypes:"c", &allowSourceDragging];
		[aCoder encodeValuesOfObjCTypes:"c", &allowDestinationDragging];
		[aCoder encodeValuesOfObjCTypes:"@:", &target, &action];

		[aCoder encodeValuesOfObjCTypes:"cc", &shadowIncoming,&dragImageIsMyImage];	// new
		[aCoder encodeObject:theAcceptingImage];
		break;

    default:
#warning StringConversion: NSStringFromSelector return value is an NSString; use [ cString] if needed, otherwise change the target type to an NSString.
		NSLog(@"[%s %s] - unknown version of %s in typed stream", [[[self class] name] cString], NSStringFromSelector(_cmd), [[[self class] name] cString]);
		break;
    }
}



#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake
{
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
	[super awake];
	[self initDragTypes];
	if(theAcceptingImage) [self setShadowIncoming:NO];

	return self;
}

/*
 * new for version 1 - some basic target/action handling, and the idea
 * of a "stringValue" so that we can use this class just like various
 * Control objects.
 * steve hayman, aug 2 1994
 */
- target { return target; }
- (void)setTarget:aTarget
{
    target = aTarget;
}
- (SEL)action { return action; }
- (void)setAction:(SEL)anAction
{
    action = anAction;
}

/*
 * these should be overridden in the appropriate way by subclasses  -- shayman
 */
- (NSString *)stringValue { return nil; }
- (void)setStringValue:(NSString *)aValue {}
- (void)takeStringValueFrom:(id)sender 
{    
	if ([sender respondsToSelector:@selector(stringValue)]) {[self setStringValue:[sender stringValue]]; return;}
    return nil;
}

@end



@interface MiscDragView (PrivateMethods)

- _freeImage: image;

@end

@implementation MiscDragView (PrivateMethods)

// Tries to fix the problem of freeing the image too early if the data
// is not retained in the view. This method is called after a short
// delay. Why isn't there a message perform:afterDelay:cancelPrevious?
// Then you could just say [theImage perform:@selector(free)afterDelay..].

- _freeImage: image
{
	// Leak leak leak.... Uncomment at your own risk (probably OK if you
	// are not using any of the methods that use NXImage's +findImageNamed:
	//image = [image free];
	return self;
}


@end


/***************************************************************************
  CHANGES:
  
  Bruce McKenzie (spuds@netcom.com) (aka BJM) on 5/24/94:
     1) Added proper versioning/archiving
     2) Reset variables so that class properly recognizes freed objects
     3) Changed "- setImageByFilename: (char *)aFilename" to 
        "- setImageByFilename: (const char *)aFilename
  Steve Hayman (shayman@Objectario.com)  
	 4) Added Control's target/action paradigm. Added target and action
	    ivars and target/setTarget:, action/setAction:, 
		stringValue/setStringValue and takeStringValueFrom.
  Bill Shirley (Bill_Shirley@WilTel.com)
  	 5) Made dragImage a class ivar since only one drag can take place 
	 	at a time. Added dragImage/setDragImage: to set the var.
	 6) Removed the ivars acceptForeignDrag/localDrag/selfDrag/retainData
	    and the methods to set them. (They have been put in the Backward
		compatability category) Each subclass of MiscDragView can
		make each choice for itself and you cannot change it at runtime.
	 7) Added initDragTypes method so you don't have to override both
	   	initFrame: and awake: in order to register your pasteboard types.
	 8) Added shadowIncoming/shadowOutgoing (note: they are not functional
	 	yet.)
	 9) Added slideback method to determine if an unsuccessfully dragged
	    image should slideback to it's destination view.
  Todd Thomas (todd@avocado.supernet.ab.ca) on Sept. 20, 1994
     10) Put all the delegate methods in a category and renamed all the
	 	methods sent to the delegate to be more inline with NeXT's.
	 11) Changed mouseDown so you have to move the mouse a little
	 	before a drag will start.
  Version 1.4: (Significant bug fixes and small additions) (Todd)
  	 12) Fixed all (ok, so I'm being optimistic) the bugs related to
	 	non-retained views.
	 13) Added -cleanupAfterDestinationDrag method which must be called
	 	from your prepareForDragOperation: or performForDragOperation:
		if you decide to stop the drag operation (by returning NO). This
		frees a copy of the dragged image we made in draggingEntered:.
	 14) Changed method -cleanupImages to -cleanupAfterSourceDrag to
	 	be more in line with the newly created method in #13.
	 15) Added the ability of the destination drag views to show a 
	 	"dimmed" image if it will accept the drag.
	 16) Just some general cleanup in drawSelf and some other methods.
My personal changes:
	1)	Added accepting image outlet and supporting methods
	2)	dragImageIsMyImage flag introduced so that draggingImage
		can default to theImage
	3)	changed drawSelf to support an accepting image
	4)	setUpForDraggingOperation queries the delegate first
	5)	un-hardwiring shadowIncoming
	6)	Extended the Inspector so that icons can be set in IB
	7)	Providing a means for settin images by name
	8)	calculateDragPoint modified so that dragging looks nicer
		(as in MiscIconWell). Thought thas such a generic behaviour sooks nice
		in a generic class.
	9)	extended setDelegate in that it asks the delegate to provide an image
		(via - image method)
	10)	initDragTypes modified in that it first asks the delegate to load the types
toDo	1)	make freeing old theImage in setImage optional
		2)	supporting a Cell like Label
		3)	providing a delegate protocol
	 
 ****************************************************************************/
 	 
	 
