// Copyright (C) 1995 Jon Kutemeier
// Use is governed by the MiscKit license

#import <objc/List.h>
#import <AppKit/NSApplication.h>
#import <apps/InterfaceBuilder.h>
#import <AppKit/NSText.h>
#import <AppKit/NSTextField.h>
#import <AppKit/NSFont.h>
#import <AppKit/NSMatrix.h>
#import <AppKit/NSBox.h>
#import <AppKit/NSColor.h>
#import <AppKit/NSDragging.h>

#import "misckit/MiscCalendarViewConstants.h"
#import "misckit/DateDelegateProtocol.h"
#import "misckit/MiscCalendarView.h"

#import "DateSelectionCell.h"
#import "MiscCalendarMatrix.h"

typedef struct {
    int	xdelta;
    int width;
    const char *value;
} calHeaders;


static calHeaders headerSpecs[] = {
{0, 17, "S"},
{16, 19, "M"},
{17, 17, "T"},
{16, 20, "W"},
{18, 17, "T"},
{18, 17, "F"},
{17, 17, "S"},
{-1, -1, ""},
};

@interface MiscCalendarView (TextDelegate)

- (BOOL)textShouldBeginEditing:(NSText *)sender;
#warning NotificationConversion: 'textDidBeginEditing:' used to be 'textDidChange:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
- (void)textDidBeginEditing:(NSNotification *)notification;
- (BOOL)textShouldEndEditing:(NSText *)sender;
#warning NotificationConversion: 'textDidEndEditing:' used to be 'textDidEnd:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
- (void)textDidEndEditing:(NSNotification *)notification;
#warning NotificationConversion: 'textDidChange:' used to be 'textDidGetKeys:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
- (void)textDidChange:(NSNotification *)notification;

@end

@interface MiscCalendarView (Dragging)

- (unsigned int)draggingEntered:(id <NSDraggingInfo>)sender;
- (void)draggingExited:(id <NSDraggingInfo>)sender;
- (unsigned int)draggingUpdated:(id <NSDraggingInfo>)sender;
- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender;
- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender;
- (void)concludeDragOperation:(id <NSDraggingInfo>)sender;

@end

@interface MiscCalendarView (IBView)

- depositIBColor:(NSColor *)aColor at:(NSPoint *)aPoint;
- acceptsIBColorAt:(NSPoint *)aPoint;

@end

@interface MiscCalendarView (Private)

- _createTextFieldWithFrame:(NSRect *)textFrame 
  stringValue:(const char *)aString;
- _acceptColor:(NSColor *)aColor at:(NSPoint *)aPoint;

@end

/* This helps us get rid of unwanted compiler warnings and serves */
/* no useful purpose */

@protocol FAKEOUTCC

- getInspectorFor:object owner:owner;
- cellColorMatrix;

@end


@implementation MiscCalendarView

- awakeAfterUsingCoder:(NSCoder *)aDecoder
{
    /* This is called by IB In test interface mode. */
    /* Register a timed-entry to be called right after */
    /* this event-loop to do our final initialization. */

   /* only for IB */

    const char *pbTypes[] = {[NSColorPboardType cString]};

    [self registerForDraggedTypes:[NSArray arrayWithObject:[NSString stringWithCString:*pbTypes]]];
    [[self window] registerForDraggedTypes:[NSArray arrayWithObject:[NSString stringWithCString:*pbTypes]]];

   if([NSApp conformsToProtocol:@protocol(IB)])
   {
       [self performSelector:@selector(awakeFromNib) object:nil afterDelay:(0.01) / 1000.0];
   }

   return self;
}

- (void)awakeFromNib
{
    const char *pbTypes[] = {[NSColorPboardType cString]};

    [self registerForDraggedTypes:[NSArray arrayWithObject:[NSString stringWithCString:*pbTypes]]];
    [[self window] registerForDraggedTypes:[NSArray arrayWithObject:[NSString stringWithCString:*pbTypes]]];

    helvetica_11 = [NSFont fontWithName:@"Helvetica" size:11];
    system_11_bold = [NSFont boldSystemFontOfSize:11];

    helvetica_12 = [NSFont fontWithName:@"Helvetica" size:12];
    system_12_bold = [NSFont boldSystemFontOfSize:12];
    
    /* Setup standard calendar action method */

    [calendarMatrix setAction:@selector(updateDate:)];
    [calendarMatrix setTarget:self];

    /* If there is a date delegate, set it */

    if (dateDelegate)
      [calendarMatrix setDateDelegate:dateDelegate];
    else
      dateDelegate = [calendarMatrix dateDelegate];

    /* If there is an external month field, set its value to */
    /* the month value */

    if (monthField && [monthField respondsToSelector:@selector(setStringValue:)])
      [monthField setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				  monthStringValue]]];

    /* If we are displaying the standard month header, set its value */
    /* to the month */

    if (displayMonthHeader)
      [monthSubview setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				    monthStringValue]]];

    /* If there is an external year field, set its value to */
    /* the year value */

    if (yearField && [yearField respondsToSelector:@selector(setIntValue:)])
      [yearField setIntValue:[[calendarMatrix dateDelegate] year]];

    /* If we are displaying the standard year header, set its value */
    /* to the year */

    if (displayYearHeader)
      [yearSubview setIntValue:[[calendarMatrix dateDelegate]
				year]];

    if (displayMonthAndYearHeader)
    {
	[inlineMonthSubview setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
					    monthStringValue]]];

	[inlineYearSubview setIntValue:[[calendarMatrix dateDelegate]
					year]];
    }

    /* If there is a date field, set its value to the curent date */

    if (dateField && [dateField respondsToSelector:@selector(setStringValue:)])
      [dateField setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				 dateStringValue]]];
}

- initWithFrame:(NSRect)frameRect
{
    NSRect 	subViewFrame;
    id		newSubview;
    calHeaders	*headers;
    const char *pbTypes[] = {[NSColorPboardType cString]};

    [super initWithFrame:frameRect];

    helvetica_11 = [NSFont fontWithName:@"Helvetica" size:11];
    system_11_bold = [NSFont boldSystemFontOfSize:11];

    helvetica_12 = [NSFont fontWithName:@"Helvetica" size:12];
    system_12_bold = [NSFont boldSystemFontOfSize:12];

    [self registerForDraggedTypes:[NSArray arrayWithObject:[NSString stringWithCString:*pbTypes]]];
    [[self window] registerForDraggedTypes:[NSArray arrayWithObject:[NSString stringWithCString:*pbTypes]]];

    colorDragAndDrop = NO;

    /* The list of textfield cells that holds the day of the week */
    /* headings */

    dowSubviewList = [[List alloc] init];

    displayDOWHeader = YES;
    displayMonthHeader = NO;
    displayYearHeader = NO;
    displayMonthAndYearHeader = NO;

    headers = headerSpecs;

    /* static view information for day-of-the-week textfield headers */

    subViewFrame = NSMakeRect(DEFAULT_MATRIX_XPOS+1, DEFAULT_HEADER_YPOS, DEFAULT_MATRIX_WIDTH, DEFAULT_HEADER_HEIGHT);
		 
    /* Create the day of the week headers (S, M, T, W...) */

    while(headers->xdelta > -1)
    {
      (&subViewFrame)->origin.x += headers->xdelta;
      (&subViewFrame)->size.width = headers->width;

      newSubview = [self _createTextFieldWithFrame:&subViewFrame
		    stringValue:headers->value];

      [dowSubviewList addObject:newSubview];

      [self addSubview:newSubview];

      headers++;
    }

    /* Create the month and year textfield headers */

    subViewFrame = NSMakeRect(DEFAULT_MATRIX_XPOS+1, DEFAULT_HEADER_YPOS, DEFAULT_MATRIX_WIDTH, DEFAULT_HEADER_HEIGHT);

    monthSubview = [self _createTextFieldWithFrame:&subViewFrame
		    stringValue:"Month"];

    yearSubview = [self _createTextFieldWithFrame:&subViewFrame
		   stringValue:"Year"];

    [NSFont setUserFont:helvetica_12];

    [monthSubview setFont:system_12_bold];

    [yearSubview setFont:system_12_bold];

    /* Create the month and year textfield headers, where the month */
    /* and year are on the same line */

    subViewFrame = NSMakeRect(DEFAULT_MATRIX_XPOS+1, DEFAULT_HEADER_YPOS, (int)(DEFAULT_MATRIX_WIDTH/2), DEFAULT_HEADER_HEIGHT);

    inlineMonthSubview = [self _createTextFieldWithFrame:&subViewFrame
			  stringValue:"Month"];

    [inlineMonthSubview setAlignment:NSLeftTextAlignment];

    [inlineMonthSubview setFont:system_12_bold];

    subViewFrame = NSMakeRect(((DEFAULT_MATRIX_WIDTH - (int)(DEFAULT_MATRIX_WIDTH/2) + 
		DEFAULT_MATRIX_XPOS+1)), DEFAULT_HEADER_YPOS, (int)(DEFAULT_MATRIX_WIDTH/2), DEFAULT_HEADER_HEIGHT);

    inlineYearSubview = [self _createTextFieldWithFrame:&subViewFrame
			 stringValue:"Year"];

    [inlineYearSubview setAlignment:NSRightTextAlignment];

    [inlineYearSubview setFont:system_12_bold];

    /* Create the actual calendar matrix view with the given */
    /* dimensions */

    subViewFrame = NSMakeRect(DEFAULT_MATRIX_XPOS, DEFAULT_MATRIX_YPOS, DEFAULT_MATRIX_WIDTH, DEFAULT_MATRIX_HEIGHT);

    calendarMatrix = [[MiscCalendarMatrix alloc] initWithFrame:subViewFrame];

    [calendarMatrix setAutoresizingMask:NSViewMinYMargin | 
     NSViewMaxYMargin | NSViewWidthSizable | NSViewHeightSizable];

    [calendarMatrix setAutosizesCells:YES];

    [calendarMatrix setAction:@selector(updateDate:)];
    [calendarMatrix setTarget:self];

    if (dateDelegate)
      [calendarMatrix setDateDelegate:dateDelegate];
    else
      dateDelegate = [calendarMatrix dateDelegate];

    previousDayCell = [calendarMatrix selectedCell];

    /* Added the calendar matrix to our view */

    [self addSubview:calendarMatrix];

    if (monthField && [monthField respondsToSelector:@selector(setStringValue:)])
      [monthField setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				  monthStringValue]]];

    if (yearField && [yearField respondsToSelector:@selector(setIntValue:)])
      [yearField setIntValue:[[calendarMatrix dateDelegate] year]];

    if (dateField && [dateField respondsToSelector:@selector(setStringValue:)])
      [dateField setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				 dateStringValue]]];

    (&defaultFrame)->size.width = frameRect.size.width;
    (&defaultFrame)->size.height = frameRect.size.height;

    return self;
}

- getMinSize:(NSSize *)minSize maxSize:(NSSize *)maxSize from:(int)where
{
    /* The view is not sizable at this time */

    minSize->width = NSWidth(defaultFrame);
    minSize->height = NSHeight(defaultFrame);

    maxSize->width = NSWidth(defaultFrame);
    maxSize->height = NSHeight(defaultFrame);

    return self;
}

- calendarMatrix
{
    return calendarMatrix;
}

- preserveCellColors:(BOOL)yn
{
    [calendarMatrix overwriteCellColors:yn];

    return self;
}

- (BOOL)cellColorsPreserved
{
    return [calendarMatrix cellColorsOverwritten];
}

- setColorOf:(int)element to:(NSColor *)aColor forCellAt:(int)xPos :(int)yPos
{
    switch (element)
    {
      case MISC_CV_CELLBACKGROUND:
	[calendarMatrix setBackgroundColor:aColor forCellAt:xPos :yPos];
	[calendarMatrix display];
	break;

      case MISC_CV_CELLTEXT:
	[calendarMatrix setTextColor:aColor forCellAt:xPos :yPos];
	[calendarMatrix display];
	break;

      case MISC_CV_CELLHIGHLIGHT:
	[calendarMatrix setHighlightColor:aColor forCellAt:xPos :yPos];
	[calendarMatrix display];
	break;

      case MISC_CV_CELLTEXTHIGHLIGHT:
	[calendarMatrix setHighlightTextColor:aColor forCellAt:xPos :yPos];
	[calendarMatrix display];
	break;

      default:
	break;
    }
    
    return self;
}

- (NSColor *)colorOf:(int)element forCellAt:(int)xPos :(int)yPos
{
    NSColor * color;

    color = [NSColor blackColor];

    switch (element)
    {
      case MISC_CV_CELLBACKGROUND:
	color = [calendarMatrix backgroundColorForCellAt:xPos :yPos];
	break;

      case MISC_CV_CELLTEXT:
	color = [calendarMatrix textColorForCellAt:xPos :yPos];
	break;

      case MISC_CV_CELLHIGHLIGHT:
	color = [calendarMatrix highlightColorForCellAt:xPos :yPos];
	break;

      case MISC_CV_CELLTEXTHIGHLIGHT:
	color =[calendarMatrix highlightTextColorForCellAt:xPos :yPos];
	break;

      default:
	break;
    }
    
    return color;
}

- setColorOf:(int)element to:(NSColor *)aColor
{
    switch (element)
    {
      case MISC_CV_CELLBACKGROUND:
	[calendarMatrix setCellBackgroundColor:aColor];
	[calendarMatrix display];
	break;

      case MISC_CV_CELLTEXT:
	[calendarMatrix setCellTextColor:aColor];
	[calendarMatrix display];
	break;

      case MISC_CV_CELLHIGHLIGHT:
	[calendarMatrix setCellHighlightColor:aColor];
	[calendarMatrix display];
	break;

      case MISC_CV_CELLTEXTHIGHLIGHT:
	[calendarMatrix setCellHighlightTextColor:aColor];
	[calendarMatrix display];
	break;

      case MISC_CV_MONTHHEADER:
	[monthSubview setTextColor:aColor];
	[monthSubview display];
	break;

      case MISC_CV_YEARHEADER:
	[yearSubview setTextColor:aColor];
	[yearSubview display];
	break;

      case MISC_CV_INLINEMONTHHEADER:
	[inlineMonthSubview setTextColor:aColor];
	[inlineMonthSubview display];
	break;

      case MISC_CV_INLINEYEARHEADER:
	[inlineYearSubview setTextColor:aColor];
	[inlineYearSubview display];
	break;

      default:
	break;
    }
    
    return self;
}

- (NSColor *)colorOf:(int)element
{
    NSColor * color;

    switch (element)
    {
      case MISC_CV_CELLBACKGROUND:
	color = [calendarMatrix cellBackgroundColor];
	break;

      case MISC_CV_CELLTEXT:
	color = [calendarMatrix cellTextColor];
	break;

      case MISC_CV_CELLHIGHLIGHT:
	color = [calendarMatrix cellHighlightColor];
	break;

      case MISC_CV_CELLTEXTHIGHLIGHT:
	color = [calendarMatrix cellHighlightTextColor];
	break;

      case MISC_CV_MONTHHEADER:
	color = [monthSubview textColor];
	break;

      case MISC_CV_YEARHEADER:
	color = [yearSubview textColor];
	break;

      case MISC_CV_INLINEMONTHHEADER:
	color = [inlineMonthSubview textColor];
	break;

      case MISC_CV_INLINEYEARHEADER:
	color = [inlineYearSubview textColor];
	break;

      default:
	color = [NSColor blackColor];
	break;
    }
    
    return color;
}

- cellAt:(int)xPos :(int)yPos lock:(BOOL)yn
{
    [[calendarMatrix cellAtRow:xPos column:yPos] setEnabled:yn];

    return self;
}

- forRow:(int)row lock:(BOOL)yn
{
    int i;

    for (i = 0; i < NUM_MATRIX_COLS; i++)
      [self cellAt:row :i lock:yn];

    return self;
}

- forColumn:(int)col lock:(BOOL)yn
{
    int i;

    for (i = 0; i < NUM_MATRIX_ROWS; i++)
      [self cellAt:i :col lock:yn];

    return self;
}

- setHighlightMode:(int)mode
{
    [calendarMatrix setMode:mode];
    return calendarMatrix;
}

- (int)highlightMode
{
    return [calendarMatrix mode];
}

- allowColorDragAndDrop:(BOOL)yn
{
    colorDragAndDrop = yn;

    return self;
}

- (BOOL)canDragAndDropColor
{
    return colorDragAndDrop;
}

- forHeader:(int)aHeader display:(BOOL)doDisplay
{
    int		yOffset,
		i;
    
    (&defaultFrame)->size.height += doDisplay ? DEFAULT_HEADER_HEIGHT : 
      -DEFAULT_HEADER_HEIGHT;

    [self sizeBy:0 :(doDisplay ? DEFAULT_HEADER_HEIGHT : 
		     -DEFAULT_HEADER_HEIGHT)];
    
    yOffset = doDisplay ? DEFAULT_HEADER_HEIGHT :
      -DEFAULT_HEADER_HEIGHT;
	  
    switch (aHeader)
    {
      case MISC_CV_DOWHEADER:
	displayDOWHeader = doDisplay;

	if (displayMonthAndYearHeader)
	{
	    [inlineMonthSubview moveBy:0 :yOffset];
	    [inlineYearSubview moveBy:0 :yOffset];
	}

	if (displayMonthHeader)
	  [monthSubview moveBy:0 :yOffset];

	if (displayYearHeader)
	  [yearSubview moveBy:0 :yOffset];
	
	if (displayDOWHeader)
	  for (i = 0; i < [dowSubviewList count]; i++)
	    [self addSubview:[dowSubviewList objectAt:i]];
	else
	    for (i = 0; i < [dowSubviewList count]; i++)
	      [[dowSubviewList objectAt:i] removeFromSuperview];

	break;

      case MISC_CV_MONTHANDYEARHEADER:
	displayMonthAndYearHeader = doDisplay;
    
	if (displayYearHeader)
	  [yearSubview moveBy:0 :yOffset];

	if (displayMonthHeader)
	  [monthSubview moveBy:0 :yOffset];

	if (displayDOWHeader)
	{
	    [inlineMonthSubview moveBy:0 :yOffset];
	    [inlineYearSubview moveBy:0 :yOffset];
	}

	if (displayMonthAndYearHeader)
	{
	    [self addSubview:inlineMonthSubview];
	    [self addSubview:inlineYearSubview];
	}
	else
	{
	    [inlineMonthSubview removeFromSuperview];
	    [inlineYearSubview removeFromSuperview];
	}

	break;

      case MISC_CV_MONTHHEADER:
	displayMonthHeader = doDisplay;

	if (displayYearHeader)
	  [yearSubview moveBy:0 :yOffset];

	if (displayDOWHeader)
	  [monthSubview moveBy:0 :yOffset];

	if (displayMonthAndYearHeader)
	  [monthSubview moveBy:0 :yOffset];

	if (displayMonthHeader)
	  [self addSubview:monthSubview];
	else
	  [monthSubview removeFromSuperview];

	break;

      case MISC_CV_YEARHEADER:
	displayYearHeader = doDisplay;

	if (displayDOWHeader)
	  [yearSubview moveBy:0 :yOffset];

	if (displayMonthAndYearHeader)
	  [yearSubview moveBy:0 :yOffset];

	if (displayMonthHeader)
	  [yearSubview moveBy:0 :yOffset];

	if (displayYearHeader)
	  [self addSubview:yearSubview];
	else
	  [yearSubview removeFromSuperview];

	break;
	
      default:
	(&defaultFrame)->size.height += !doDisplay ? DEFAULT_HEADER_HEIGHT : 
	  -DEFAULT_HEADER_HEIGHT;

	[self sizeBy:0 :(!doDisplay ? DEFAULT_HEADER_HEIGHT : 
			 -DEFAULT_HEADER_HEIGHT)];
    
	break;
    }

    return self;
}

- (BOOL)isHeaderDisplayed:(int)aHeader;
{
    switch (aHeader)
    {
      case MISC_CV_DOWHEADER:
	return displayDOWHeader;
	break;
	
      case MISC_CV_MONTHHEADER:
	return displayMonthHeader;
	break;

      case MISC_CV_YEARHEADER:
	return displayYearHeader;
	break;

      case MISC_CV_MONTHANDYEARHEADER:
	return displayMonthAndYearHeader;
	break;

      default:
	return NO;
    }
}

- setColorTo:(NSColor *)aColor forDay:(MiscCVDays)aDay
{
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
    {
    id newVar = [dowSubviewList objectAt:aDay];
    [newVar setTextColor:aColor];
    [newVar display];
}
    [[dowSubviewList objectAt:aDay] setTextColor:aColor];
    return [dowSubviewList objectAt:aDay];
}

- (NSColor *)colorForDay:(MiscCVDays)aDay
{
    return [[dowSubviewList objectAt:aDay] textColor];
}

- incrementMonth:sender
{
    id date;

    date = [calendarMatrix dateDelegate];

    [date incrementMonth];

    [self updateDate:self];

    previousDayCell = [calendarMatrix selectedCell];

    return self;
}

- decrementMonth:sender
{
    id date;

    date = [calendarMatrix dateDelegate];

    [date decrementMonth];

    [self updateDate:self];

    previousDayCell = [calendarMatrix selectedCell];

    return self;
}

- incrementYear:sender
{
    id date;

    date = [calendarMatrix dateDelegate];

    [date incrementYear];

    [self updateDate:self];

    previousDayCell = [calendarMatrix selectedCell];

    return self;
}

- decrementYear:sender
{
    id date;

    date = [calendarMatrix dateDelegate];

    [date decrementYear];

    [self updateDate:self];

    previousDayCell = [calendarMatrix selectedCell];

    return self;
}

- updateDate:sender
{
    DateSelectionCell	*dayCell;
    id			date;
    int			day,
			month,
			year;

    if (!(dayCell = [calendarMatrix selectedCell]))
    {
	if (previousDayCell)
	  [calendarMatrix selectCell:previousDayCell];

	return nil;
    }

    dayCell = [calendarMatrix selectedCell];

    previousDayCell = dayCell;

    date = [calendarMatrix dateDelegate];

    day = [dayCell intValue];
    month = [date month];
    year = [date year];

    [date setYear:year month:month day:day];

    [self displayDate:self];

    return self;
}

- displayDate:sender
{
    if (monthField && [monthField respondsToSelector:@selector(setStringValue:)])
      [monthField setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				  monthStringValue]]];

    if (displayMonthHeader)
      [monthSubview setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				    monthStringValue]]];

    if (yearField && [yearField respondsToSelector:@selector(setIntValue:)])
      [yearField setIntValue:[[calendarMatrix dateDelegate] year]];

    if (displayYearHeader)
      [yearSubview setIntValue:[[calendarMatrix dateDelegate]
				year]];

    if (dateField && [dateField respondsToSelector:@selector(setStringValue:)])
      [dateField setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
				 dateStringValue]]];

    if (displayMonthAndYearHeader)
    {
	[inlineMonthSubview setStringValue:[NSString stringWithCString:[[calendarMatrix dateDelegate]
					    monthStringValue]]];

	[inlineYearSubview setIntValue:[[calendarMatrix dateDelegate]
					year]];
    }

    [calendarMatrix displayDate:sender];

    return self;
}

- setDateDelegate:(id <DateDelegate>)anObject
{
    dateDelegate = anObject;
    [calendarMatrix setDateDelegate:dateDelegate];
    
    return self;
}

- (id <DateDelegate>)dateDelegate
{
    return dateDelegate;
}

- (const char *)getInspectorClassName
{
    return "MiscCalendarViewInspector";
}

- (const char *)getEditorClassName
{
    return "MiscCalendarViewEditor";
}

- (id)initWithCoder:(NSCoder *)aDecoder
{
    float	width,
		height;

    [super initWithCoder:aDecoder];

    [aDecoder decodeValuesOfObjCTypes:"fficcccc@@@@@@@@@@@", &width, &height,
		&titleDisplayMode, &displayDOWHeader, &displayMonthHeader, 
		&displayYearHeader, &displayMonthAndYearHeader,
		&colorDragAndDrop,
		&calendarMatrix, &monthField, &previousDayCell, &monthSubview,
		&dateDelegate, &yearField, &dateField, &dowSubviewList, 
		&yearSubview, &inlineMonthSubview, &inlineYearSubview];

    (&defaultFrame)->size.width = width;
    (&defaultFrame)->size.height = height;

    return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
{
    [super encodeWithCoder:aCoder];

    [aCoder encodeValuesOfObjCTypes:"fficcccc@@@@@@@@@@@", &(NSWidth(defaultFrame)),
		 &(NSHeight(defaultFrame)), &titleDisplayMode, 
		 &displayDOWHeader, &displayMonthHeader, &displayYearHeader,
		 &displayMonthAndYearHeader, &colorDragAndDrop, 
		 &calendarMatrix, &monthField,
		 &previousDayCell, &monthSubview, &dateDelegate, &yearField,
		 &dateField, &dowSubviewList, &yearSubview, 
		 &inlineMonthSubview, &inlineYearSubview];
}

@end

@implementation MiscCalendarView (TextDelegate)

- (BOOL)textShouldBeginEditing:(NSText *)sender
{
    if (textDelegate && [textDelegate respondsToSelector:@selector(textShouldBeginEditing:)])
      return !((![textDelegate textShouldBeginEditing:sender]));
    else
      return YES;
}

#warning NotificationConversion: 'textDidBeginEditing:' used to be 'textDidChange:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
- (void)textDidBeginEditing:(NSNotification *)notification
{
#warning NotificationConversion: 'textDidBeginEditing:' used to be 'textDidChange:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
#warning NotificationConversion: if this notification was not posted by NSText (eg. was forwarded by NSMatrix or NSTextField from the field editor to their textDelegate), then the text object is found by [[notification userInfo] objectForKey:@"NSFieldEditor"] rather than [notification object]
    NSText *theText = [notification object];
    if (textDelegate && [textDelegate respondsToSelector:@selector(textDidBeginEditing:)])
#warning NotificationConversion: 'textDidBeginEditing:' used to be 'textDidChange:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
      return [[NSNotificationCenter defaultCenter] postNotificationName:NSTextDidBeginEditingNotification object:theText];
    else
}

- (BOOL)textShouldEndEditing:(NSText *)sender
{
    if (textDelegate && [textDelegate respondsToSelector:@selector(textShouldEndEditing:)])
      return !((![textDelegate textShouldEndEditing:sender]));
    else
      return YES;
}

#warning NotificationConversion: 'textDidEndEditing:' used to be 'textDidEnd:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
- (void)textDidEndEditing:(NSNotification *)notification
{
#warning NotificationConversion: if this notification was not posted by NSText (eg. was forwarded by NSMatrix or NSTextField from the field editor to their textDelegate), then the text object is found by [[notification userInfo] objectForKey:@"NSFieldEditor"] rather than [notification object]
    NSText *theText = [notification object];
    int whyEnd = [[[notification userInfo] objectForKey:@"NSTextMovement"] intValue];
    id retCode = self;

#warning NotificationConversion: 'textDidEndEditing:' used to be 'textDidEnd:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
    if (textDelegate && [textDelegate respondsToSelector:@selector(textDidEndEditing:)])
    {
#warning NotificationConversion: 'textDidEndEditing:' used to be 'textDidEnd:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
	retCode = [[NSNotificationCenter defaultCenter] postNotificationName:NSTextDidEndEditingNotification object:theText userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:whyEnd], @"NSTextMovement", nil]];

	[self updateDate:self];
    }

    return retCode;
}

#warning NotificationConversion: 'textDidChange:' used to be 'textDidGetKeys:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
- (void)textDidChange:(NSNotification *)notification
{
#warning NotificationConversion: 'textDidChange:' used to be 'textDidGetKeys:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
#warning NotificationConversion: if this notification was not posted by NSText (eg. was forwarded by NSMatrix or NSTextField from the field editor to their textDelegate), then the text object is found by [[notification userInfo] objectForKey:@"NSFieldEditor"] rather than [notification object]
    NSText *theText = [notification object];
    BOOL flag = [[theText text] isEqualToString:@""];
    if (textDelegate && [textDelegate respondsToSelector:@selector(textDidChange:)])
#warning NotificationConversion: 'textDidChange:' used to be 'textDidGetKeys:'.  This conversion assumes this method is implemented or sent to a delegate of NSText.  If this method was implemented by a NSMatrix or NSTextField textDelegate, use the text notifications in NSControl.h.
      return [[NSNotificationCenter defaultCenter] postNotificationName:NSTextDidChangeNotification object:_theText];
    else
}

@end

@implementation MiscCalendarView (Dragging)

- (unsigned int)draggingEntered:(id <NSDraggingInfo>)sender
{
    NSRect	aRect;
    NSPoint	mouseLocation;

    if (([sender draggingSourceOperationMask] & NSDragOperationCopy) &&
	colorDragAndDrop)
    {
	aRect = NSMakeRect(DEFAULT_MATRIX_XPOS+1, DEFAULT_HEADER_YPOS, DEFAULT_MATRIX_WIDTH, 0);

	if (displayDOWHeader)
	  (&aRect)->size.height += DEFAULT_HEADER_HEIGHT;

	if (displayMonthAndYearHeader)
	  (&aRect)->size.height += DEFAULT_HEADER_HEIGHT;

	if (displayMonthHeader)
	  (&aRect)->size.height += DEFAULT_HEADER_HEIGHT;

	if (displayYearHeader)
	  (&aRect)->size.height += DEFAULT_HEADER_HEIGHT;

	mouseLocation = [sender draggingLocation];

	mouseLocation = [self convertPoint:mouseLocation toView:self];

	if (NSMouseInRect(mouseLocation , aRect , NO))
	  return NSDragOperationCopy;
    }

    return NSDragOperationNone;
}

- (void)draggingExited:(id <NSDraggingInfo>)sender
{
    
}

- (unsigned int)draggingUpdated:(id <NSDraggingInfo>)sender
{
    return [self draggingEntered:sender];
}

- (BOOL)prepareForDragOperation:(id <NSDraggingInfo>)sender
{
    if ([self draggingEntered:sender] == NSDragOperationCopy)
      return YES;

    return NO;
}

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
    NSColor *	color;
    NSPoint	mouseLocation;

    color = [NSColor colorFromPasteboard:[sender draggingPasteboard]];
    
    mouseLocation = [sender draggingLocation];
    
    if ([self _acceptColor:color at:&mouseLocation])
      return YES;

    return NO;
}

- (void)concludeDragOperation:(id <NSDraggingInfo>)sender
{
    
}

@end

/* Satan made me do it! Okay, well, maybe not, but jeez NeXT, if you */
/* would just publish some of these APIs so us lowly end-user programmer */
/* could do all the neat things you get to do, ugly rule-breaking hacks */
/* like this would not happen. sigh.... Anyway, these two methods allow */
/* my view to accept drag-and-drop color in IB, just like textfield does. */
/* Thanks to the person who wrote ClassDump and Robert Nicholson for */
/* helping me find these... */

@implementation MiscCalendarView (IBView)

- depositIBColor:(NSColor *)aColor at:(NSPoint *)aPoint
{
    NSPoint	mouseLocation;
    int		cellColorSelection;

    mouseLocation.x = aPoint->x;
    mouseLocation.y = aPoint->y;

    if (![self _acceptColor:aColor at:&mouseLocation])
    {
	cellColorSelection = [[[[NSApp getInspectorFor:self owner:self] 
			      cellColorMatrix] selectedCell] tag];

	[self setColorOf:cellColorSelection to:aColor];
    }

    return self;
}

- acceptsIBColorAt:(NSPoint *)aPoint;
{
    return self;
}

@end

@implementation MiscCalendarView (Private)

- _createTextFieldWithFrame:(NSRect *)textFrame 
  stringValue:(const char *)aString
{
    NSTextField	*textField;

    textField = [[NSTextField alloc] initWithFrame:*textFrame];
    [textField setEditable:NO];
    [textField setBezeled:NO];
    [textField setBordered:NO];
    [textField setSelectable:NO];
    [textField setAlignment:NSCenterTextAlignment];
#warning ColorConversion: [textField setDrawsBackground:NO] was [textField setBackgroundGray:-1]
    [textField setDrawsBackground:NO];
    
    [textField setAutoresizingMask:NSViewMinYMargin | 
     NSViewMaxYMargin | NSViewWidthSizable | NSViewHeightSizable];

    [NSFont setUserFont:helvetica_11];

    [textField setFont:system_11_bold];

    [textField setStringValue:[NSString stringWithCString:aString]];

    return textField;
}

- _acceptColor:(NSColor *)aColor at:(NSPoint *)aPoint
{
    NSTextField	*textField;
    NSPoint	mouseLocation;
    NSRect	aRect;
    int 	i;

    mouseLocation.x = aPoint->x;
    mouseLocation.y = aPoint->y;

    textField = nil;

    mouseLocation = [self convertPoint:mouseLocation fromView:[self superview]];

    if (displayDOWHeader)
      for (i = 0; i < [dowSubviewList count]; i++)
      {
	  aRect = [[dowSubviewList objectAt:i] frame];

	  if (NSMouseInRect(mouseLocation , aRect , NO))
	  {
	      textField = [dowSubviewList objectAt:i];
	      i = [dowSubviewList count];
	  }
      }

    if (!textField && displayMonthAndYearHeader)
    {
	aRect = [inlineMonthSubview frame];

	  if (NSMouseInRect(mouseLocation , aRect , NO))
	    textField = inlineMonthSubview;
    }

    if (!textField && displayMonthAndYearHeader)
    {
	aRect = [inlineYearSubview frame];

	  if (NSMouseInRect(mouseLocation , aRect , NO))
	    textField = inlineYearSubview;
    }

    if (!textField && displayMonthHeader)
    {
	aRect = [monthSubview frame];

	  if (NSMouseInRect(mouseLocation , aRect , NO))
	    textField = monthSubview;
    }

    if (!textField && displayYearHeader)
    {
	aRect = [yearSubview frame];

	  if (NSMouseInRect(mouseLocation , aRect , NO))
	    textField = yearSubview;
    }

    if (textField)
    {
	[textField setTextColor:aColor];
	return self;
    }
    
    return nil;
}

@end

