//
//	MiscCircularSliderCell.m -- a SliderCell for circular sliders
//		Written by Vince DeMarco and Don Yacktman
//		Copyright (c) 1994 by Vince DeMarco and Don Yacktman.
//				Version 1.0  All rights reserved.
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

// This was created by modifying Vince's CircularSliderCell class and
// merging it with Don's RotationSliderCell class.  Here's the info
// from the original files.
/*
 *    Filename:	CircularSliderCell.m 
 *    Created :	Sun Oct 18 16:56:38 1992 
 *    Author  :	Vince DeMarco
 *		<vince@whatnxt.cuc.ab.ca>
 *    LastEditDate was "Wed Feb  3 17:30:28 1993"
 */
//	RotationSliderCell.m -- written by Don Yacktman
//	Copyright 1993 by Don Yacktman.  All rights reserved.

//	Modified by Laurent Daudelin on Wed Sep 7 12:09:55 1994
//	to add support for lower and upper stop values, and
//	to support for knob width adjustment for the MISC_SHUTTLEWHEEL_STYLE slider.

//  Modified by Carl Lindberg   Wed Apr 19 02:01:15 EDT 1995
//  Added support for border types
//  Added support for background color
//  Fixed boundary-check bug, fixed bug when view is taller than it is wide
//  Reorganized the code to something approaching sanity
//  Redid drawing of knobImage, no more glitchiness on larger sliders/knob widths.
//    (Unless you get really, really big :-) )

//  Modified more by Carl Lindberg Mon Jun 12 00:36:31 EDT 1995
//  Added support for allowing which angle minValue is shown at
//       (they don't have to "start" on the right side anymore)
//  Added support for specifying which direction (clockwise or 
//       clockwise) that values increase in.  Before, this was
//       counterclockwise.
//  Changed the pswrap to the pie-style one to accomodate above fixes
//  The pie-style one previously gave strange values; doing the
//       above modifications fixed this

#import "MiscCircularSliderCell.h"
#import "MiscCSWraps.h"
#import <math.h>
#import <AppKit/AppKit.h>

@interface MiscCircularSliderCell(private)
- _cacheImages;
@end

// I changed how the knobImage gets drawn for slider-style a bit, so 
// naturally I have a new set of fudges... :-)   -Carl

#define _KNOBSIZE_FUDGE			(- 0.1)
#define _KNOBPOS_RADIUS_FUDGE	(-  .15)
#define _KNOBPOS_X_FUDGE		(  1.3)
#define _KNOBPOS_Y_FUDGE		(- 0.5)

// These are still used, though
#define _KNOB_WIDTH_FUDGE (5.0 / 8.0)
#define _KNOB_Y_FUDGE (- 1.0)
#define _KNOB_X_FUDGE (  1.0)
#define _BEZELS_FUDGE (  4.0)
#define MISC_SLIDERCELL_CURRENT_VERSION 4

#define MISC_CSDEFAULTCOLOR [NSColor colorWithCalibratedRed:0.0 green:0.0 blue:0.0 alpha:0.0]  //transparent

@implementation MiscCircularSliderCell

// returns angle in 0-360 range
inline static float in360(float ang)
{	
	while (ang < 0.0) ang += 360.0;
	while (ang >= 360.0) ang -= 360.0;
	return ang;
}

// Given a "raw" angle (the angle if not clockwise and
// startangle at 0), calculates the correct angle based
// on the current values of isClockwise and startAngle.
//  Used with currentangle() to get actual angle of
// given value.
float convertAngle(float ang, BOOL isClockwise, float startAngle)
{
	ang = in360(ang);
	if (isClockwise) ang = startAngle-ang;
	else ang = startAngle+ang;
	return in360(ang);
}

// Inverse of convertAngle().  Given the actual angle, uses
// values of startAngle and isClockwise and startAngle to
// convert to a "raw" angle (0-360, with start at 0), which then can be
// mapped against minValue and maxValue to compute the value
// of the current angle.  Used by position().
float convertAngle2(float ang, BOOL isClockwise, float startAngle)
{
	ang = in360(ang);
	ang -= startAngle;
	if (isClockwise) ang = (360-ang);
	return in360(ang);
}

inline static float angle(float x, float y)
{ // this is a little cleaner than Vince's original routine.  :-)
	double result = atan2(y, x) * (180 / M_PI);
	return in360(result); // keep in 0-360 degree range
}

inline static void xycoord(float angle, float radius,float *x, float *y)
{ // the inverse of the above; given angle and radius, determines x & y
    *x = radius * (float)cos((M_PI / 180) * angle);
    *y = radius * (float)sin((M_PI / 180) * angle);
}

// Convert between a 0-360 degree angle and the slider's position.  The
// angle is used as the parameter to pass to the pswrap that does the drawing.
// These two #defines are inverses of each other.
#define currentangle(pos, min, max) \
	(convertAngle((360.0 * ((pos) - (min)) / ((max) - (min))),isClockwise,startAngle))
#define position(angle, min, max) \
	((convertAngle2(angle, isClockwise,startAngle)) * ((max) - (min)) / 360.0 + (min))

static float distBetween(NSPoint a, NSPoint b)
{ return (float)hypot((double)(a.x-b.x),(double)(a.y-b.y)); }

+ (void)initialize 
{
    if (self == [MiscCircularSliderCell class]) {
	      [MiscCircularSliderCell setVersion:MISC_SLIDERCELL_CURRENT_VERSION];
    }
    return;
}

- init
{
    self = [super init];
    [self minValue] = 0.0;
	[self doubleValue] = 45.0;
    [self maxValue] = 360.0;
	radius = 50.0; sliderPathWidth = 10.0;
	lowerStopValue = [self minValue];
	upperStopValue = [self maxValue];
	center.x = radius;
	center.y = radius;
	style = MISC_SLIDER_STYLE;
	jumpToMousePoint = YES;
    hidden = NO;
	backgroundColor = MISC_CSDEFAULTCOLOR;
	borderType = MISC_CSNONE;
	startAngle = 0;
	isClockwise = NO;
	[self _cacheImages];
    return self;
}

#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake
{
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
	id ret = [super awake];
	[self _cacheImages];
	return ret;
}

#define _SETUPIMAGE PSsetgray(0.0); PSsetalpha(0.0); \
		NSRectFill(frame); PSsetalpha(1.0);

- _cacheImages
{	// builds the cached slider, etc., if necessary.  (ie, size changed
	// or no cache created yet... *****
	NSSize knobSize, bezelSize = { radius * 2, radius * 2 };
	NSRect frame = { { 0.0, 0.0 }, bezelSize };

    // don't cache for other styles -Carl
  	if (style != MISC_SLIDER_STYLE) return self;
		
	if (disabledBezelImage) [disabledBezelImage release];
	disabledBezelImage = [[NSImage alloc] initWithSize:bezelSize];
	[disabledBezelImage lockFocus];
	PSgsave(); _SETUPIMAGE;
	MiscCSDrawBezel(0, 0, radius, sliderPathWidth, 0.666);
	PSgrestore();
	[disabledBezelImage unlockFocus];

	if (enabledBezelImage) [enabledBezelImage release];
	enabledBezelImage = [[NSImage alloc] initWithSize:bezelSize];
	[enabledBezelImage lockFocus];
	PSgsave(); _SETUPIMAGE;
	MiscCSDrawBezel(0, 0, radius, sliderPathWidth, 0.5);
	PSgrestore();
	[enabledBezelImage unlockFocus];

	knobSize.width  = sliderPathWidth + _KNOBSIZE_FUDGE;
	knobSize.height = sliderPathWidth + _KNOBSIZE_FUDGE;
	if (knobImage) [knobImage release];
	knobImage = [[NSImage alloc] initWithSize:knobSize];
	[knobImage lockFocus];
	PSgsave(); _SETUPIMAGE;
	MiscCSDrawKnob(0,0, sliderPathWidth + _KNOBSIZE_FUDGE);
	PSgrestore();
	[knobImage unlockFocus];
	return self;
}

- setSliderPathWidth:(float)val
{
	if (val == sliderPathWidth) return self;
	sliderPathWidth = val;
	[self _cacheImages];
	return self;
}

- (float)sliderPathWidth
{
    return sliderPathWidth;
}

- setSliderStyle:(MiscCircularSliderStyle)aBOOL
{
	if (aBOOL == style) return self;
	style = aBOOL;
	[self _cacheImages];
    return self;
}

- (MiscCircularSliderStyle)sliderStyle
{
    return style;
}

- setJumpToMousePoint:(BOOL)aBOOL
{
    jumpToMousePoint = aBOOL;
    return self;
}

- (BOOL)jumpToMousePoint
{
    return jumpToMousePoint;
}

- setHidden:(BOOL)flag
{
    hidden = flag;
    return self;
}

- (BOOL)hidden
{
    return hidden;
}

- setUpperStopValue:(float)aValue
{
	upperStopValue = aValue;
	return self;
}

- (float)upperStopValue
{
	return upperStopValue;
}

- setLowerStopValue:(float)aValue
{
	lowerStopValue = aValue;
	return self;
}

- (float)lowerStopValue
{
	return lowerStopValue;
}

- setUseBoundaries:(BOOL)flag
{
	useBoundaries = flag;
	return self;
}

- (BOOL)useBoundaries
{
	return useBoundaries;
}

- (id)initWithCoder:(NSCoder *)aDecoder
{
//  Looks like version 1 added upper/lower stop values, version 2 added an
//  option to turn that feature off.
//  Version 3 adds border type and a background color.
//  Version 4 adds startAngle  -Carl
    [super initWithCoder:aDecoder];
	switch ([aDecoder versionForClassName:@"MiscCircularSliderCell"]) {
 		case MISC_SLIDERCELL_CURRENT_VERSION:
			[aDecoder decodeValuesOfObjCTypes:"iccffffcifi",
				&style,
				&jumpToMousePoint,
				&hidden,
				&radius,
				&sliderPathWidth,
				&lowerStopValue,
				&upperStopValue,
				&useBoundaries,
				&borderType,
				&startAngle,
				&isClockwise];
			center = [aDecoder decodePoint];
#warning ColorConversion: [[aDecoder decodeNXColor] retain] used to be NXReadColor(aDecoder).  Use 'decodeNXColor' to read old style colors, but use 'decodeObject' to read OpenStep colors.
			backgroundColor = [[aDecoder decodeNXColor] retain];
			break;
 		case 3:
			[aDecoder decodeValuesOfObjCTypes:"iccffffci",
				&style,
				&jumpToMousePoint,
				&hidden,
				&radius,
				&sliderPathWidth,
				&lowerStopValue,
				&upperStopValue,
				&useBoundaries,
				&borderType];
			center = [aDecoder decodePoint];
#warning ColorConversion: [[aDecoder decodeNXColor] retain] used to be NXReadColor(aDecoder).  Use 'decodeNXColor' to read old style colors, but use 'decodeObject' to read OpenStep colors.
			backgroundColor = [[aDecoder decodeNXColor] retain];
			startAngle = (style == MISC_PIECHART_STYLE) ? 90 : 0;
			isClockwise = (style == MISC_PIECHART_STYLE) ? YES : NO;
			break;
		case 2:
			[aDecoder decodeValuesOfObjCTypes:"iccffffc",
				&style,
				&jumpToMousePoint,
				&hidden,
				&radius,
				&sliderPathWidth,
				&lowerStopValue,
				&upperStopValue,
				&useBoundaries];
			center = [aDecoder decodePoint];
			borderType = MISC_CSNONE;
			backgroundColor = MISC_CSDEFAULTCOLOR;
			startAngle = (style == MISC_PIECHART_STYLE) ? 90 : 0;
			isClockwise = (style == MISC_PIECHART_STYLE) ? YES : NO;
			break;
		case 1:
			[aDecoder decodeValuesOfObjCTypes:"iccffff",
				&style,
				&jumpToMousePoint,
				&hidden,
				&radius,
				&sliderPathWidth,
				&lowerStopValue,
				&upperStopValue];
    		center = [aDecoder decodePoint];
			useBoundaries = NO;
			borderType = MISC_CSNONE;
			backgroundColor = MISC_CSDEFAULTCOLOR;
			startAngle = (style == MISC_PIECHART_STYLE) ? 90 : 0;
			isClockwise = (style == MISC_PIECHART_STYLE) ? YES : NO;
			break;
		case 0:        /* read code for old version */
			[aDecoder decodeValuesOfObjCTypes:"iccff",
				&style, &jumpToMousePoint, &hidden, &radius, &sliderPathWidth];
		    center = [aDecoder decodePoint];
			upperStopValue = [self maxValue];
			lowerStopValue = [self minValue];
			useBoundaries = NO;
			borderType = MISC_CSNONE;
			backgroundColor = MISC_CSDEFAULTCOLOR;
			startAngle = (style == MISC_PIECHART_STYLE) ? 90 : 0;
			isClockwise = (style == MISC_PIECHART_STYLE) ? YES : NO;
			break;
 	}
    return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
{
    [super encodeWithCoder:aCoder];
    [aCoder encodeValuesOfObjCTypes:"iccffffcifi",
		&style,
		&jumpToMousePoint,
		&hidden,
		&radius,
		&sliderPathWidth,
		&lowerStopValue,
		&upperStopValue,
		&useBoundaries,
		&borderType,                        /*NEW for version 3*/
		&startAngle,                        /*NEW for version 4*/
		&isClockwise];                      /*NEW for version 4*/
    [aCoder encodePoint:center];
	[aCoder encodeObject:backgroundColor];
}

// overridden methods here:
- (NSSize)cellSizeForBounds:(NSRect)aRect
{
	
    NSSize theSize;
    theSize.width = NSWidth(aRect);
	theSize.height = NSHeight(aRect);

	if (style != MISC_SLIDER_STYLE) return theSize;
	// the "slider looking" style should always be a perfect circle.
	if (theSize.height < theSize.width) theSize.width = theSize.height;
	else theSize.height = theSize.width;
	return theSize;
}

#define _FINDXYFROMPOINT(z) \
        if (imFlipped) \
          y = NSHeight(_lastViewFrame) - (z)->y - center.y; \
        else y = (z)->y - center.y; \
        x = (z)->x - center.x;

- (BOOL)startTrackingAt:(NSPoint)startPoint inView:(NSView *)controlView
{ // make sure we're within the slider's "hot" area, a circle (or ring)
	float x, y, rad, ir;

	_FINDXYFROMPOINT(&&startPoint);
	_offsetAng = angle(x, y) - currentangle([self doubleValue], [self minValue], [self maxValue]);
	rad = x * x + y * y;
	if (rad > radius * radius) return NO; // outside
	ir = radius - sliderPathWidth - 3.0;
	// If we are the slider style, don't respond to a click on the inner
	// circle.  The second condition checks if the knob width is larger
	// than the radius, meaning there is no inner circle and no reason
	// to disallow the click.  -Carl
	if ((style==MISC_SLIDER_STYLE) && (radius > sliderPathWidth) && (rad < ir*ir))
		return NO;
	return [super startTrackingAt:startPoint inView:controlView];  //return YES?
}

- (BOOL)continueTracking:(NSPoint)lastPoint at:(NSPoint)currentPoint inView:(NSView *)controlView
{
	float x, y, oldx, oldy, curAng, oldAng;

	if (hidden) return NO;

	// get the current angle of the mouse pointer
	_FINDXYFROMPOINT(&&currentPoint);
	curAng = angle(x, y);
	if (!jumpToMousePoint) {
		curAng -= _offsetAng;
		if (curAng < 0) curAng += 360.0;
	}
	// get the last angle of the mouse pointer
	oldAng = currentangle([self doubleValue], [self minValue], [self maxValue]);

	[self doubleValue] = position(curAng, [self minValue], [self maxValue]);
	
	// Do the boundary check; if we are outside those bounds, set
	// the current value to be the bound the mouse is closest to. -Carl
	if (useBoundaries && (([self doubleValue] > upperStopValue) || ([self doubleValue] < lowerStopValue))) {
		NSPoint upoint, lpoint, vpoint;
		xycoord(currentangle(lowerStopValue,[self minValue],[self maxValue]),radius,&lpoint.x,&lpoint.y);
		xycoord(currentangle(upperStopValue,[self minValue],[self maxValue]),radius,&upoint.x,&upoint.y);
		xycoord(curAng,radius,&vpoint.x,&vpoint.y);
		
		if (distBetween(vpoint,lpoint) < distBetween(vpoint,upoint)) {
			[self doubleValue] = lowerStopValue;
			curAng = currentangle(lowerStopValue, [self minValue], [self maxValue]);
		}
		else {
			[self doubleValue] = upperStopValue;
			curAng = currentangle(upperStopValue, [self minValue], [self maxValue]);
		}
	}
	
	[[controlView window] disableFlushWindow];
	PSgsave();
	if (imFlipped) {
		PSscale(1.0, -1.0);
		PStranslate(0.0, -NSHeight(_lastViewFrame));
	}
	if (style == MISC_SLIDER_STYLE) {
		// Need to erase the old knob, which is done be re-compositing just that
		// section.  The other styles do their own erasing in -drawKnob.
		// Have to do this here, because it is only here we have the info
		// on where the last knob position was, which the other styles don't need.
		NSRect eraseRect; NSPoint imgPoint;
		id bezel = ([self isEnabled] ? disabledBezelImage : enabledBezelImage);
		xycoord(oldAng, radius - sliderPathWidth / 2 - 2, &oldx, &oldy);
		oldx += center.x - sliderPathWidth * _KNOB_WIDTH_FUDGE + _KNOB_X_FUDGE;
		oldy += center.y - sliderPathWidth * _KNOB_WIDTH_FUDGE + _KNOB_Y_FUDGE;
		eraseRect = NSMakeRect(oldx, oldy, sliderPathWidth * 1.5 + 2, sliderPathWidth * 1.5 + 2);
		eraseRect = NSIntegralRect(eraseRect);
		eraseRect = NSInsetRect(eraseRect , -4.0 , -4.0);
		imgPoint.x = NSMinX(eraseRect) + NSMinX(_lastFrame);
		imgPoint.y = NSMinY(eraseRect) + NSMinY(_lastFrame);
		[bezel compositeToPoint:(imgPoint) fromRect:eraseRect operation:NSCompositeSourceOver];
	}
	PSgrestore();

	[self drawKnob:*_lastFrame];  // -drawKnob ignores its given rect; use as dummy.
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
	{
    id newVar = [controlView window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}
	PSWait();
	return YES;
}


// The basic drawing methods...

#define _COPYRECT(x, y) *(x) = NSMakeRect(\
		NSMinX(*y), NSMinY(*y), NSWidth(*y), NSHeight(*y));

// This should draw the whole slider.  In here, we draw the
// border, then the background color second (so to wipe out
// any gray drawn by the border functions).  We then inset
// the given rect a bit to account for the border, and call
// -drawBarInside, which will draw the rest (via -drawKnob).
// Hidden will just draw background color and return.
// The reason I draw to a temporary NXImage here is so that
// a background color with alpha in it will draw correctly.
// When I drew directly to the view, the background would
// start out as black, then NX_SOVER-like draw the background
// color.  Fully transparent would end up black, and semi-
// transparent would be much darker than you would expect.
// This was the way I found to make it work.  A bit slower,
// perhaps, but this method shouldn't get called all _that_
// often. :-)  -Carl

- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	NSRect ourRect;
	NSPoint pt = cellFrame.origin;	
	NSImage *img = [[NSImage alloc] initWithSize:(cellFrame.size)];

	_COPYRECT(&ourRect,&cellFrame);
	_COPYRECT(&_lastViewFrame,&cellFrame);
	if ([controlView isFlipped]) imFlipped = YES;

	[img lockFocus];
	
	if (!hidden) switch (borderType) {
		case MISC_CSGROOVE:
			NSDrawGroove(ourRect , ourRect);
			ourRect = NSInsetRect(ourRect , 2.0 , 2.0);
			break;
		case MISC_CSBEZEL:
			NSDrawGrayBezel(ourRect , ourRect);
			ourRect = NSInsetRect(ourRect , 2.0 , 2.0);
			break;
		case MISC_CSLINE:
			PSsetgray(0.0);
			NSFrameRect(ourRect);
			ourRect = NSInsetRect(ourRect , 1.0 , 1.0);
			break;
		case MISC_CSBUTTON:
			NSDrawButton(ourRect , ourRect);
			ourRect = NSInsetRect(ourRect , 2.0 , 2.0);
			break;
		case MISC_CSNONE:
		default:break;
	}
	[backgroundColor set]; NSRectFill(ourRect);
	[img unlockFocus];
	if (imFlipped) pt.y += NSHeight(cellFrame);
	[img compositeToPoint:pt operation:NSCompositeSourceOver];
	[img release];

	// I want to put this code inside of -drawBarInside, but
	// when I do, it thinks enabledBezelImage is a Window class
	// object!!  [I get a "Window does not respond to
	// -composite:fromPoint:" runtime error in the switch
	// statement in -drawBarInside.]  I am *clueless* as to
	// why that's true.  It works if I put this code here, though.
	if (!NSEqualRects(_lastFrame , ourRect)) {
		_COPYRECT(&_lastFrame, &ourRect);
		radius = MIN(NSWidth(ourRect), NSHeight(ourRect)) / 2;
		center.x = NSMinX(ourRect) + radius;
		center.y = NSMinY(ourRect) + radius;
		[self _cacheImages];
	}

	[self drawInteriorWithFrame:ourRect inView:controlView];
}


// This method should basically draw the slider, sans border and
// background color, inside cellFrame.  It resets some instance
// vars, draws whatever slider background is necessary, then calls
// -drawKnob to do the rest.
- (void)drawBarInside:(NSRect)cellFrame flipped:(BOOL)flipped
{
	NSImage *bezel = ([self isEnabled] ? disabledBezelImage : enabledBezelImage);
	float r4 = radius - _BEZELS_FUDGE;

	// reset some instance vars....

/*  //WHY DOESN'T THIS WORK HERE!?!?!?!?
	if (!NXEqualRect(&_lastFrame, cellFrame)) {
		_COPYRECT(&_lastFrame, cellFrame);
		radius = MIN(NX_WIDTH(cellFrame), NX_HEIGHT(cellFrame)) / 2;
		center.x = NX_X(cellFrame) + radius;
		center.y = NX_Y(cellFrame) + radius;
		[self _cacheImages];
	}
 */
	if (hidden) return;
	
	PSgsave();
	if (flipped) {
		PSscale(1.0, -1.0);
		PStranslate(0.0, -NSHeight(_lastViewFrame));
		imFlipped = YES;
	}
	switch(style) {
		case MISC_SLIDER_STYLE:
			[bezel compositeToPoint:(&cellFrame.origin) operation:NSCompositeSourceOver];
			break;
		case MISC_PIECHART_STYLE:
			if ([self isEnabled]) break;	//otherwise do the background below
		case MISC_DIAL_STYLE:
		case MISC_SHUTTLEWHEEL_STYLE:
			MiscCSDrawBackground(center.x, center.y, r4, radius, radius);
			break;
	}

	PSgrestore();
	[self drawKnob:cellFrame];
}


// This method will draw the filled piepiece/line/knob of the slider.
// It actually ignores the knobRect parameter; it is only there because
// this is the correct method to override.
- (void)drawKnob:(NSRect)knobRect
{
	NSPoint pos;
	float ang, r4 = radius - _BEZELS_FUDGE;
	NSSize knobSize;

	if (hidden) return;

	PSgsave();
	if (imFlipped) {
		PSscale(1.0, -1.0);
		PStranslate(0.0, -NSHeight(_lastViewFrame));
	}
	ang = currentangle([self doubleValue], [self minValue], [self maxValue]);
	switch (style) {
		case MISC_SLIDER_STYLE:
			// Composite the knobImage to the right spot. 
			knobSize = [knobImage size];
			xycoord(ang, (radius-2) - (sliderPathWidth / 2) + _KNOBPOS_RADIUS_FUDGE, &pos.x, &pos.y);
			pos.x += center.x - knobSize.width / 2.0 + _KNOBPOS_X_FUDGE;
			pos.y += center.y - knobSize.height / 2.0 + _KNOBPOS_Y_FUDGE;
			[knobImage compositeToPoint:pos operation:NSCompositeSourceOver];
			break;
		case MISC_PIECHART_STYLE: {
			// With this style, I want the gray area to start with startAngle and go to
			// the current angle.  Which one to start with depends on whether it's clockwise.
			// So, just set the two params based on if it's clockwise.
			float startang = (isClockwise)? ang:startAngle;
			float finishang= (isClockwise)? startAngle:ang;

			if ([self isEnabled]) {
				MiscCSPieChartDisabled(center.x, center.y, r4, radius, radius, startang, finishang);
			} else {
				MiscCSPieChart(center.x, center.y, r4, radius, radius, startang, finishang);
			}
			break;
		}
		case MISC_DIAL_STYLE:
			MiscCSControlDial(center.x, center.y, r4, radius, radius,
					ang, ([self isEnabled] ? 0.333 : 0.0));
			break;
		case MISC_SHUTTLEWHEEL_STYLE:
			if ([self isEnabled]) {
				MiscCSControlKnobDisabled(center.x, center.y, r4, radius, radius);
			} else {
				// Removed the following function call to allow adjustment of the
				// control know width by the following one. -Laurent Daudelin.
				//MiscCSControlKnob(center.x, center.y, r4, radius, radius, ang);
				MiscCSDrawShuttleKnob(center.x, center.y, radius, sliderPathWidth, ang);
			}
			break;
	}
	PSgrestore();
}


// I think this is what the superclass does anyways, but let's make sure
- (void)drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
	[self drawBarInside:*cellFrame flipped:imFlipped];
}

- (NSRect)knobRectFlipped:(BOOL)flipped
{
	
    NSRect knobRect;
    knobRect = [super knobRectFlipped:flipped];
	return knobRect;
}

- (void)setBorderType:(int)type
{
	borderType = type;
}

- (NSBorderType)borderType
{
	return borderType;
}

- (void)setBackgroundColor:(NSColor *)aColor
{
	backgroundColor = aColor;
}

- (NSColor *)backgroundColor
{
	return backgroundColor;
}

- (float)startAngle
{
	return startAngle;
}

- setStartAngle:(float)ang
{
	startAngle = in360(ang);
	return self;
}

- (BOOL)isClockwise
{
	return isClockwise;
}

- setClockwise:(BOOL)clockwise
{
	isClockwise = clockwise;
	return self;
}

@end
