//
//	MiscStringArray.m -- a generic class to store lists of strings
//		Originally written by Drew Davidson
//		Copyright (c) 1994 by Drew Davidson.
//		Modified and extended by Don Yacktman for inclusion into the MiscKit.
//				Version 1.4  All rights reserved.
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#define MISCSTRINGARRAY_VERSION 2

/*----------------------------------------------------------------------------
	$Source$

	SYNOPSIS
		Implements an array of char * values built from MiscString objects.

	REVISIONS
	$Log$
----------------------------------------------------------------------------*/
# import <misckit/misckit.h>

@interface MiscStringArray(private)

- _invalidateArray;
- _buildArray;

@end

@implementation MiscStringArray

/*----------------------------< PRIVATE METHODS >----------------------------*/
- _invalidateArray
{
	free(stringArray);
	stringArray = NULL;
	return(self);
}

- _buildArray
{
	if (stringArray)
		[self _invalidateArray];
	NX_MALLOC(stringArray,const char *,[strings count]+1);
	MISCforAllListObjects(strings,string)
	{	stringArray[MISCCurrentSlot] = ((uniqued) ? [string uniqueStringValue] : [[string stringValue] cString]);
		// stringArray[MISCCurrentSlot+1] = NULL; // Carl says unneeded
	} MISCendFor
	stringArray[[self count]] = NULL; // Carl added
	return(self);
}

/*---------------------------< INIT/FREE METHODS >---------------------------*/
+ (void)initialize
{
	if ( self == [MiscStringArray class] )
		[MiscStringArray setVersion:MISCSTRINGARRAY_VERSION];

	return;
}

- init
{
	[super init];
	strings = [[List allocWithZone:[self zone]] initCount:0];
	uniqued = NO;
	stringArray = NULL;
	return(self);
}

- (void)dealloc
{
	[[strings freeObjects] release];
	[self _invalidateArray];
	[super dealloc];
}

/*--------------------------< OVERRIDDEN METHODS >---------------------------*/
- copyWithZone:(NSZone *)aZone
{
	self = [super copyWithZone:aZone];
	strings = [strings deepCopyFromZone:aZone];
	stringArray = NULL;
	[self _buildArray];
	return(self);
}

/*-----------------------------< OTHER METHODS >-----------------------------*/
- (BOOL)uniqued
{
	return(uniqued);
}

- setUniqued:(BOOL)yn
{
	if (yn != uniqued)
	{	uniqued = yn;
		[self _invalidateArray];
	}
	return(self);
}

- strings
{
	[self _invalidateArray];
	return(strings);
}

- addStringObject:(MiscString *) aStringObj // -bbum
{
	[self _invalidateArray];
	[[self strings] addObject:aStringObj];
	return(self);
}

- addString:(const char *)aString
{
	[self _invalidateArray];
	[[self strings] addObject:[[MiscString allocWithZone:[self zone]] initString:aString]];
	return(self);
}

- removeString:(const char *)aString
{	// This method will remove aString from the MiscStringArray. If the string
	// doesn't exist, it will  return nil, otherwise it will return self.
	// contributed by Juergen Zeller <zet@cip.e-technik.uni-erlangen.de>
	unsigned int index = MISC_NOT_IN_ARRAY;

	index = [self indexOfString:aString];
	if (index == MISC_NOT_IN_ARRAY) return (nil);
	[self _invalidateArray];
	[[[self strings] removeObjectAt:index] release];
	return (self);
}


- insertString:(const char *)aString at:(unsigned int)index
{
	[self _invalidateArray];
	[[self strings] insertObject:[[MiscString allocWithZone:[self zone]] initString:aString] at:index];
	return(self);
}

- (unsigned int)stringCount
{
	return([strings count]);
}

- (unsigned int)count
{
	return([strings count]);
}

- empty // -bbum
{
    // invalidate stringArray
	[self _invalidateArray];

    // empty and free all strings contained in strings List
	[[strings freeObjects] empty];

	return self;
}
- sort { return [strings miscStringSort]; } // bbum

- (const char **)stringArray
{
	if (!stringArray)
		[self _buildArray];
	return(stringArray);
}

- (const char *)stringAt:(unsigned int)index
{	const char		**array = [self stringArray];

	return(array[index]);
}

- (unsigned int)indexOfString:(const char *)aString
{	unsigned int	index = MISC_NOT_IN_ARRAY;
	
	MISCforAllListObjects(strings,string)
	{	if ([string cmp:aString] == 0)
		{	index = MISCCurrentSlot;
			break;
		}
	} MISCendFor
	return(index);
}

- (void)browser:(NSBrowser *)sender createRowsForColumn:(int)column inMatrix:(NSMatrix *)matrix
{
	int i;
	NSArray *cellList, theCell;
 
	// Set matrix to have the right number of cells.
	[matrix renewRows:[self count] columns:1];

	// Get list of cells from the matrix.
	cellList = [matrix cells];

	// For each cell set its value, set whether it is a leaf or not and
	//   mark it loaded.
	for(i=0; i<[cellList count]; i++) {
		theCell = [cellList objectAtIndex:i];
		[theCell setStringValue:[NSString stringWithCString:[self stringAt:i]]];
		[theCell setLeaf:YES];
		[theCell setLoaded:YES];
	}

	// Return the number of rows.
#error BrowserConversion: Do not return count.  The matrix row count must be correct.
	return [self count];
}

/*
 * Basic read/write methods.
 * Note that we aren't writing the const char **stringArray at the moment,
 * because (a) I'm not sure how to handle that, and (b) I think we can
 * get away without it.
 * shayman Aug 12 1994
 */
- (id)initWithCoder:(NSCoder *)aDecoder
{
    int version;

    [super initWithCoder:aDecoder];
    version = [aDecoder versionForClassName:[[MiscStringArray name] cString]];
    
    switch (version) {
		case MISCSTRINGARRAY_VERSION: {
			[aDecoder decodeValuesOfObjCTypes:"@c", &strings, &uniqued];
			break;
		}
		case 0: { int un;
		// we were writing a BOOL as an int--that's dangerous, so, rather
		// than bump the version, I changed it to convert between int/BOOL
		// on the write and the read.
			[aDecoder decodeValuesOfObjCTypes:"@i", &strings, &un];
			uniqued = (un ? YES : NO);
			break;
		}
		default: {
#warning StringConversion: NSStringFromSelector return value is an NSString; use [ cString] if needed, otherwise change the target type to an NSString.
			NSLog(@"[%s %s] - unknown version of %s in typed stream", [[[self class] name] cString], NSStringFromSelector(_cmd), 
					[[[self class] name] cString]);
			break;
		}
	}
	return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
{
    [super encodeWithCoder:aCoder];
	// we were writing a BOOL as an int--that's dangerous, so now we
	// use a "c" for it.  This is safe.
	[aCoder encodeValuesOfObjCTypes:"@c", &strings, &uniqued];
}

@end
