//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscBorderView.M
//
//		View class for the row/column borders on a MiscTableView.
//		Supports resizing, dragging.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscBorderView.M,v 1.9 96/02/12 00:04:25 sunshine Exp $
// $Log:		MiscBorderView.M,v $
//	Revision 1.9  96/02/12	00:04:25  sunshine
//	no comment.
//	
//	Revision 1.8  96/01/17	05:17:09  sunshine
//	Fixed crasher.	Was allocating drag-cache window (an alloc once free never
//	object) from [self zone].  When the zone was destroyed the window pointer
//	(in a static variable) ceased to point to valid memory.
//	
//	Revision 1.7  96/01/16	19:51:25  zarnuk
//	Eliminated a lot of unnecessary drawing while dragging.
//-----------------------------------------------------------------------------
#import "MiscBorderView.h"
#import "MiscBorderCell.h"
#import "MiscHighlightTracker.h"
#import "MiscListTracker.h"
#import "MiscRadioTracker.h"
#import "MiscTableBorder.h"
#import "MiscTableView.h"
#import "MiscTableWell.h"
#import <misckit/MiscTableScroll.h>
#import <misckit/MiscTableTypes.h>

#import <new.h>

extern "Objective-C" {
#import <AppKit/NSApplication.h>
#import <AppKit/NSClipView.h>
#import <AppKit/NSFont.h>
#import <AppKit/NSCursor.h>
#import <AppKit/NSImage.h>
#import <AppKit/NSText.h>
#import <AppKit/NSWindow.h>
#import <AppKit/NSGraphics.h>
/*#import <AppKit/timer.h>*/
#import <Foundation/NSBundle.h>
#import <AppKit/dpsOpenStep.h>
#import <AppKit/NSDPSContext.h>
#import <AppKit/psopsOpenStep.h>
}

extern "C" {
#import <assert.h>
#import <float.h>
#import <limits.h>
#import <math.h>
#import <string.h>		// memset().
#import <stdio.h>		// FILENAME_MAX
}

#define MISC_RESIZE_EPSILON		5
#define MISC_FRAME_HEIGHT		18
#define MISC_FRAME_WIDTH		46

static NSCursor* horzCursor = 0;
static NSCursor* vertCursor = 0;

static inline double dmin( double a, double b ) { return (a < b ? a : b); }
static inline double dmax( double a, double b ) { return (a > b ? a : b); }


//----------------------------------------------------------------------------
// startTimer
//----------------------------------------------------------------------------
static inline void startTimer( NXTrackingTimer timer )
	{
	if (timer == 0)
		[NSEvent startPeriodicEventsAfterDelay:0.1 withPeriod:0.1];
	}


//----------------------------------------------------------------------------
// stopTimer
//----------------------------------------------------------------------------
static inline void stopTimer( NXTrackingTimer timer )
	{
	if (timer)
		{
		[NSEvent stopPeriodicEvents];
		timer = 0;
		}
	}



//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation MiscBorderView

//-----------------------------------------------------------------------------
// + cursorFromImageName:
//		NOTE: Cannot use [self zone] in a class (factory) method
//		like this one, hence no allocFromZone:.
//-----------------------------------------------------------------------------
+ (NSCursor*) cursorFromImageName:(char const*)name
	{
	NSImage* image = [NSImage imageNamed:[NSString stringWithCString:name]];
	if (image == 0)
		{
		char path[ FILENAME_MAX + 1 ];
#error StringConversion: This call to -[NXBundle getPath:forResource:ofType:] has been converted to the similar NSBundle method.  The conversion has been made assuming that the variable called path will be changed into an (NSString *).  You must change the type of the variable called path by hand.
		path = [[NSBundle bundleForClass:self] pathForResource:[NSString stringWithCString:name] ofType:@"tiff"];
		image = [[NSImage alloc] initByReferencingFile:[NSString stringWithCString:path]];
		[image setName:[NSString stringWithCString:name]];
		}
	NSPoint const hot = { 8, 8 };
	NSCursor* cursor = [[NSCursor alloc] initWithImage:image];
	[cursor setHotSpot:hot];
	return cursor;
	}


//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void)initialize
	{
	if (horzCursor == 0)
		{
		horzCursor = [self cursorFromImageName:"MiscHorzResizeCursor"];
		vertCursor = [self cursorFromImageName:"MiscVertResizeCursor"];
		}
	return;
	}


//-----------------------------------------------------------------------------
// TYPE VARIATIONS
//-----------------------------------------------------------------------------

- (float) frameHeight
		{ return (isHorz ? MISC_FRAME_HEIGHT : MISC_FRAME_WIDTH); }
- (NSCursor*) cursor
		{ return isHorz ? horzCursor : vertCursor; }
- (MiscBorderType) borderType
		{ return isHorz ? MISC_COL_BORDER : MISC_ROW_BORDER; }
- (MiscBorderType) otherBorderType
		{ return isHorz ? MISC_ROW_BORDER : MISC_COL_BORDER; }
- (MiscTableBorder*) otherBorder
		{ return [scroll border: [self otherBorderType]]; }
- (void) clearOtherBorder { [self otherBorder]->selectionSet().empty(); }
- cellAtRow:(MiscCoord_P)row column:(MiscCoord_P)col
	{ return isHorz ? [scroll cellAtRow:row column:col] : [scroll cellAtRow:col column:row]; }

- (float) sizeWidth: (NSSize const*) size
		{ return isHorz ? size->width : size->height; }
- (float) sizeHeight: (NSSize const*) size
		{ return isHorz ? size->height : size->width; }

- (float) pointX: (NSPoint const*) point
		{ return isHorz ? point->x : point->y; }
- (float) pointY: (NSPoint const*) point
		{ return isHorz ? point->y : point->x; }
- (void) point: (NSPoint*) point setX: (float) x
		{ if (isHorz) point->x = x; else point->y = x; };
- (void) point: (NSPoint*) point setY: (float) y
		{ if (isHorz) point->y = y; else point->x = y; };

- (float)rectMinX:(NSRect const*)frameRect
		{ return isHorz ? NSMinX(*frameRect) : NSMinY(*frameRect); }
- (float)rectMaxX:(NSRect const*)frameRect
		{ return isHorz ? NSMaxX(*frameRect) : NSMaxY(*frameRect); }
- (float)rectMinY:(NSRect const*)frameRect
		{ return isHorz ? NSMinY(*frameRect) : NSMinX(*frameRect); }
- (float)rectMaxY:(NSRect const*)frameRect
		{ return isHorz ? NSMaxY(*frameRect) : NSMaxX(*frameRect); }
- (float)rectWidth:(NSRect const*)frameRect
		{ return isHorz ? NSWidth(*frameRect) : NSHeight(*frameRect); }
- (float)rectHeight:(NSRect const*)frameRect
		{ return isHorz ? NSHeight(*frameRect) : NSWidth(*frameRect); }

- (void) setRect:(NSRect*)rect minX:(float)x
		{ if (isHorz) rect->origin.x = x; else rect->origin.y = x; }
- (void) setRect:(NSRect*)rect minY:(float)y
		{ if (isHorz) rect->origin.y = y; else rect->origin.x = y; }
- (void) setRect:(NSRect*)rect width:(float)width
		{ if (isHorz) rect->size.width = width; else rect->size.height = width; }
- (void) setRect:(NSRect*)rect height:(float)height
		{ if (isHorz) rect->size.height = height; else rect->size.width = height; }

- (void) changeFrameIfNeeded
	{
	MiscPixels const my_width = (MiscPixels) floor([self rectWidth:&[self frame]]);
	MiscPixels const i_width = info->totalSize();
	if (my_width != i_width)
		{
		NSRect rect = [self frame];
		[[self window] invalidateCursorRectsForView:self];
		[self setRect:&rect width:i_width];
		[self setFrameSize:NSMakeSize(rect.size.width, rect.size.height)];
		[[scroll documentView] adjustSize];
		}
	}

- (void) setPos:(MiscCoord_V)pos width:(MiscPixels)size
	{
	MiscCoord_P pPos = info->visualToPhysical(pos);
	if (isHorz)
		[scroll setCol:pPos size:(float)size];
	else
		[scroll setRow:pPos size:(float)size];
	[self changeFrameIfNeeded];
	[[self window] invalidateCursorRectsForView:self];
#warning ViewConversion:  The View 'update' method is obsolete; if the receiver of this method is an NSView convert to [<theView> setNeedsDisplay:YES]
	[scroll update];
	}

- (void) moveFrom:(MiscCoord_V)from to:(MiscCoord_V)to
	{
	if (isHorz)
		[scroll moveColFrom:from to:to];
	else
		[scroll moveRowFrom:from to:to];
	[[self window] invalidateCursorRectsForView:self];
#warning ViewConversion:  The View 'update' method is obsolete; if the receiver of this method is an NSView convert to [<theView> setNeedsDisplay:YES]
	[scroll update];
	}



//-----------------------------------------------------------------------------
// - initFrame:scroll:info:type:
//-----------------------------------------------------------------------------
- initFrame: (NSRect const*) frameRect
	 scroll: (MiscTableScroll*) i_scroll
	   info: (MiscTableBorder*) i_info
	   type: (MiscBorderViewType) type
	{
	NSZone* const z = [self zone];

	[super initWithFrame:*0];
#error ViewConversion: '[NSView setFlipped:]' is obsolete; you must override 'isFlipped' instead of setting externally. However, [NSImage setFlipped:] is not obsolete. If that is what you are using here, no change is needed.
	[self setFlipped:YES];
#error ViewConversion: View's 'setOpaque:' is obsolete; you must override 'isOpaque' instead of setting externally (if sent to text, 'setDrawsBackground:' can be used to eliminate background gray; if sent to NSImageRep, it's OK)
	[self setOpaque:YES];
#error ViewConversion: 'setClipping:' is obsolete. Views always clip to their bounds. Use PSinitclip instead.
	[self setClipping:NO];

	isHorz = (type == MISC_COL_BORDER_VIEW);
	scroll = i_scroll;
	info = i_info;
	theCell = [[MiscBorderCell allocWithZone:z] initTextCell:@"Kilroy"];

	oldSel = new( NSZoneMalloc(z,sizeof(*oldSel)) ) MiscSparseSet;
	[self setSelectionMode: [scroll selectionMode]];

	NSRect myFrame = {{0,0},{0,0}};
	if (frameRect != 0)
		myFrame.origin = frameRect->origin;
	[self setRect: &myFrame width: i_info->totalSize()];
	[self setRect: &myFrame height: [self frameHeight]];
	[self setFrame:myFrame];
	return self;
	}


//-----------------------------------------------------------------------------
// - free
//-----------------------------------------------------------------------------
- (void)dealloc
	{
	[theCell release];
	[tracker release];
	if (oldSel != 0)
		{
		oldSel->MiscSparseSet::~MiscSparseSet();
		NSZoneFree( [self zone], oldSel );
		}
	{ [super dealloc]; return; };
	}



//=============================================================================
// CONVERSIONS
//=============================================================================
//-----------------------------------------------------------------------------
// - range::fromRect:
//
//		Returns by reference the range of visual slots contained in rect.
//		rMin is inclusive, rMax is exclusive.
//-----------------------------------------------------------------------------
- (void) range: (MiscCoord_V*)rMin : (MiscCoord_V*)rMax
	fromRect: (NSRect const*)rect
	{
	*rMin = info->visualForOffset(MiscPixels(floor([self rectMinX:rect])));
	*rMax = info->visualForOffset(MiscPixels(floor([self rectMaxX:rect]))) + 1;
	}


//-----------------------------------------------------------------------------
// - rect:forPos:
//-----------------------------------------------------------------------------
- (void) rect:(NSRect*)nxrect forPos:(MiscCoord_V)pos
	{
	*nxrect = [self bounds];
	[self setRect: nxrect minX: info->getOffset(pos)];
	[self setRect: nxrect width: info->effectiveSize(pos)];
	}



//=============================================================================
// DRAWING
//=============================================================================
//-----------------------------------------------------------------------------
// - getVisibleRange::
//
//		Returns by reference the range of visual slots contained in the visible
//		rectangle.	rMin is inclusive, rMax is exclusive.
//-----------------------------------------------------------------------------
- (void) getVisibleRange: (MiscCoord_V*)rMin : (MiscCoord_V*)rMax
	{
	NSRect visRect;
	visRect = [[self superview] documentVisibleRect];
	[self range:rMin:rMax fromRect:&visRect];
	}


//-----------------------------------------------------------------------------
// drewPos:
//
//		Keeps the old selection sets up to date whenever we draw a cell.  This
//		way -reflectSelection has valid data to work from.
//-----------------------------------------------------------------------------
- (void) drewPos: (MiscCoord_V)slot
	{
	BOOL isSelected = info->selectionSet().contains( slot );
	if (isSelected != oldSel->contains(slot))
		{
		if (isSelected)
			oldSel->add( slot );
		else
			oldSel->remove( slot );
		}
	}


//-----------------------------------------------------------------------------
// - drawPos:inRect:controlView:
//-----------------------------------------------------------------------------
- (void) drawPos: (MiscCoord_V) pos inRect:(NSRect const*) rect
	controlView: (NSView *)controlView
	{
	[theCell setStringValue:[NSString stringWithCString:info->getTitle(pos)]];
	[theCell setState:info->selectionSet().contains(pos)];
	[theCell drawWithFrame:*rect inView:controlView];
	}


//-----------------------------------------------------------------------------
// - drawPos:updateSel:
//-----------------------------------------------------------------------------
- (void) drawPos: (MiscCoord_V) pos updateSel: (BOOL) updateSel
	{
	if (0 <= pos && pos < info->count())
		{
		NSRect rect;
		[self rect:&rect forPos:pos];
		[self drawPos:pos inRect:&rect controlView: self];
		if (updateSel)
			[self drewPos:pos];
		}
	}


//-----------------------------------------------------------------------------
// - drawPos:
//-----------------------------------------------------------------------------
- (void) drawPos: (MiscCoord_V) pos
	{
	[self drawPos:pos updateSel:YES];
	}


//-----------------------------------------------------------------------------
// - drawRect:
//-----------------------------------------------------------------------------
- (void)drawRect:(NSRect const*)rect
	{
	MiscCoord_V pos_min;
	MiscCoord_V pos_max;
	[self range:&pos_min:&pos_max fromRect:rect];
	for (MiscCoord_V pos = pos_min;	 pos < pos_max;	 pos++)
		[self drawPos:pos];
	}


//-----------------------------------------------------------------------------
// - drawSelf::
//-----------------------------------------------------------------------------
#warning RectConversion: drawRect:(NSRect)rects (used to be drawSelf:(NXRect const*)rects :(int)nrects) no longer takes an array of rects
- (void)drawRect:(NSRect)rects
	{
	if (nrects == 1)
		[self drawRect:&rects];
	else if (nrects == 3)
		{
		[self drawRect:++&rects];
		[self drawRect:++&rects];
		}
}



//-----------------------------------------------------------------------------
// - drawSlot:	[public]
//-----------------------------------------------------------------------------
- (void) drawSlot: (MiscCoord_V) n
	{
	MiscCoord_V rMin, rMax;
	[self getVisibleRange:&rMin:&rMax];
	if (rMin <= n && n < rMax)
		{
		if (NO /* was [<view> isAutoDisplay] */ && [self canDraw])
			{
			[self lockFocus];
			[self drawPos:n];
			[self unlockFocus];
			}
		else
			{
			[self setNeedsDisplay:YES];
			}
		}
	}



//=============================================================================
// SELECTION
//=============================================================================
//-----------------------------------------------------------------------------
// - setSelectionMode:
//-----------------------------------------------------------------------------
- (void) setSelectionMode: (MiscSelectionMode) mode
	{
	NSZone* const z = [self zone];
	if (tracker)
		[tracker release];
	switch (mode)
		{
		case MISC_LIST_MODE:
			tracker = [MiscListTracker allocWithZone:z];
			break;
		case MISC_RADIO_MODE:
			tracker = [MiscRadioTracker allocWithZone:z];
			break;
		case MISC_HIGHLIGHT_MODE:
			tracker = [MiscHighlightTracker allocWithZone:z];
			break;
		}
	[tracker initBorder: info];
	}


//-----------------------------------------------------------------------------
// - reflectSelection
//-----------------------------------------------------------------------------
- (void) reflectSelection
	{
	if ([self canDraw])
		{
		MiscSparseSet const& newSel = info->selectionSet();

		MiscCoord_V rMin, rMax;
		[self getVisibleRange:&rMin:&rMax];
	
		[[self window] disableFlushWindow];
		int locked = 0;
		for (MiscCoord_V i = rMin;	i < rMax;  i++)
			{
			if (oldSel->contains(i) != newSel.contains(i))
				{
				if (!locked)
					{
					locked = 1;
					[self lockFocus];
					}
				[self drawPos:i updateSel:NO];
				}
			}
		if (locked)
			[self unlockFocus];
	
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
		{
    id newVar = [self window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}
		*oldSel = newSel;
		}
	}


//-----------------------------------------------------------------------------
// - selectPos:
//-----------------------------------------------------------------------------
- (void) selectPos: (MiscCoord_V) pos
	{
	MiscSparseSet& s = info->selectionSet();
	s.empty();
	s.add( pos );
	[scroll selectionChanged];
	}



//=============================================================================
// CURSOR MANAGEMENT
//=============================================================================
//-----------------------------------------------------------------------------
// - cursorRect:forPos:
//-----------------------------------------------------------------------------
- (BOOL) cursorRect:(NSRect*)nxrect forPos:(MiscCoord_V)pos;
	{
	MiscPixels const pos_width = info->effectiveSize(pos);
	if (pos_width > 0)
		{
		MiscPixels const pos_offset = info->getOffset(pos);
		MiscPixels const min_x = pos_offset + pos_width - MISC_RESIZE_EPSILON;
		*nxrect = [self bounds];
		[self setRect:nxrect minX:min_x];
		[self setRect:nxrect width:MISC_RESIZE_EPSILON];
		return YES;
		}
	return NO;
	}


//-----------------------------------------------------------------------------
// - resetCursorRects
//-----------------------------------------------------------------------------
- (void)resetCursorRects
	{
	if (info->isSizeable())
		{
		NSCursor* const cursor = [self cursor];
		MiscCoord_V min_pos, max_pos;
		[self getVisibleRange:&min_pos:&max_pos];
		int const count = info->count();
		if (0 <= min_pos && min_pos < count)
			{
			NSRect rect;
			int const lim = (max_pos <= count) ? max_pos : count;
			for (MiscCoord_V pos = min_pos;	 pos < lim;	 pos++)
				if (info->isSizeable(pos) && [self cursorRect:&rect forPos:pos])
					[self addCursorRect:rect cursor:cursor];
			}
		}
}


//-----------------------------------------------------------------------------
// - refresh
//-----------------------------------------------------------------------------
- (void) refresh
	{
	[self changeFrameIfNeeded];
	if ([self canDraw])
		{
		[[self window] disableFlushWindow];
		NSRect rect;
		rect = [[self superview] documentVisibleRect];
		[self lockFocus];
		[self drawRect:rect];
		[self unlockFocus];
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
		{
    id newVar = [self window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}
		}
	}



//=============================================================================
// MOUSE-TRACKING
//=============================================================================
//-----------------------------------------------------------------------------
// - startTracking
//-----------------------------------------------------------------------------
- (void) startTracking
	{
	[self clearOtherBorder];
	if ([[self window] firstResponder] == [scroll documentView])
		[[self window] makeFirstResponder:[self window]];
	}


//-----------------------------------------------------------------------------
// - endTracking:
//-----------------------------------------------------------------------------
- (void) endTracking: (MiscCoord_V)pos
	{
	if (pos < 0)
		pos = 0;
	else if (pos >= info->count())
		pos = info->count() - 1;
	[scroll border:[self borderType] setCursor:info->visualToPhysical(pos)];
	[scroll selectText:self];
	}


//-----------------------------------------------------------------------------
// - posForMousePt:
//-----------------------------------------------------------------------------
- (MiscCoord_V) posForMousePt: (NSPoint const*) p
	{
	MiscPixels pix = MiscPixels( floor([self pointX: p]) );
	return (pix < [self rectMaxX: &[self bounds]] ?
				info->visualForOffset( pix ) : info->count());
	}



//=============================================================================
// RESIZING
//=============================================================================
//-----------------------------------------------------------------------------
// - resizeEvent:x:deltaX:minX:maxX:
//-----------------------------------------------------------------------------
- (int) resizeEvent:(NSEvent *)p x:(MiscPixels)x
		deltaX:(MiscPixels)deltaX
		minX:(MiscPixels)minX
		maxX:(MiscPixels)maxX
	{
	NSRect docFrame;	[scroll getDocClipFrame: &docFrame];
	NSRect clipFrame;	clipFrame = [[self superview] frame];
	float const minDrawX = [self rectMinX: &clipFrame]; // scroll coords
	float const maxDrawX = [self rectMaxX: &clipFrame]; // scroll coords

	x += deltaX;

	float draw_x;				// scroll coords
	NSRect line;
	[self setRect:&line height:[self rectHeight:&clipFrame] +
								[self rectHeight:&docFrame] - 1 ];
	[self setRect:&line width: 2 ];
	[self setRect:&line minX: x - 1 ];
	[self setRect:&line minY: 0 ];
	(line.origin) = [scroll convertPoint:(line.origin) fromView:self];
	draw_x = [self rectMinX:&line];

	BOOL did_scroll = NO;
	BOOL in_bounds = (minDrawX <= draw_x && draw_x <= maxDrawX);

	[scroll lockFocus];
	PSsetinstance(YES);
	PSsetgray(NSBlack);

	if (in_bounds)
		PSrectfill( line.origin.x, line.origin.y,
						line.size.width, line.size.height );

#error EventConversion: NXTrackingTimer is obsolete; use NSEvent 'startPeriodicEventsAfterDelay:withPeriod:' and 'stopPeriodicEvents'.  You can probably delete this variable and its uses.
	NXTrackingTimer * timer = 0;
	startTimer( timer );
	NSEvent *lastEvent = p;

	while (1)
		{
		p = [[self window] nextEventMatchingMask:NSLeftMouseUpMask |
								NSLeftMouseDraggedMask |
								NSPeriodicMask];
		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		if ([p type] == NSPeriodic)
			{
			NSPoint mousePt = [lastEvent locationInWindow];
			mousePt = [scroll convertPoint:mousePt fromView:0];
			if (mousePt.x < minDrawX || maxDrawX < mousePt.x)
				{
				PSsetinstance(NO);
				[self autoscroll:lastEvent];
				PSsetinstance(YES);
				did_scroll = YES;
				}
			}
		else
			lastEvent = p;

		NSPoint new_loc = [lastEvent locationInWindow];
		new_loc = [self convertPoint:new_loc fromView:0];
		MiscPixels mouseX = (MiscPixels) floor([self pointX:&new_loc]);
		MiscPixels new_x = (mouseX + deltaX);
		if (new_x < minX)
			new_x = minX;
		else if (new_x >= maxX)
			new_x = maxX - 1;

		BOOL const did_move = (new_x != x);
		if (did_move || did_scroll)
			{
			if (in_bounds && !did_scroll)
				PShideinstance( line.origin.x, line.origin.y,
								line.size.width, line.size.height );
			if (did_move)
				{
				x = new_x;
				[self setRect:&line minX: x - 1 ];
				[self setRect:&line minY:0];
				(line.origin) = [scroll convertPoint:(line.origin) fromView:self];
				draw_x = [self rectMinX:&line];
				in_bounds = (minDrawX <= draw_x && draw_x <= maxDrawX);
				}

			if (in_bounds)
				PSrectfill( line.origin.x, line.origin.y,
								line.size.width, line.size.height );
			did_scroll = NO;
			}
		}

	stopTimer( timer );

	if (in_bounds)
		PShideinstance( line.origin.x, line.origin.y,
						line.size.width, line.size.height );

	PSsetinstance(NO);
	[scroll unlockFocus];

	x -= deltaX;
	return x;
	}


//-----------------------------------------------------------------------------
// - resizeEvent:inPos:atX:deltaX:finalWidth:
//-----------------------------------------------------------------------------
- (BOOL) resizeEvent:(NSEvent *)p inPos: (MiscCoord_V) pos
		 atX: (MiscPixels) x_origin
		 deltaX: (MiscPixels) delta_x
		 finalWidth: (MiscPixels*) finalWidth
	{
	MiscPixels const org_x = info->getOffset(pos);
	MiscPixels const min_x = org_x + info->effectiveMinSize(pos);
	MiscPixels const max_x = org_x + info->effectiveMaxSize(pos) + 1;
	MiscPixels const curr_x = [self resizeEvent:p x:x_origin deltaX:delta_x
								minX:min_x maxX:max_x];
	MiscPixels final_delta = curr_x - x_origin;
	if (final_delta != 0)
		{
		*finalWidth = info->effectiveSize(pos) + final_delta;
		return YES;
		}

	return NO;
	}


//-----------------------------------------------------------------------------
// - inHotZone:forPos:atX:deltaX:
//-----------------------------------------------------------------------------
- (BOOL) inHotZone: (NSPoint const*) pt
		forPos: (MiscCoord_V*) pos
		atX: (MiscPixels*) pos_x
		deltaX: (MiscPixels*) delta_x
	{
	if (info->isSizeable())
		{
		MiscCoord_V const plim = info->count();
		MiscPixels x = (MiscPixels) floor([self pointX:pt]);
		MiscCoord_V p = info->visualForOffset(x);
		if (0 <= p && p < plim)
			{
			MiscPixels max_x = info->getOffset(p) + info->effectiveSize(p);
			MiscPixels delta = max_x - x;
		
			if (0 <= delta && delta <= MISC_RESIZE_EPSILON)
				{
				do { p++; } while (p < plim && info->effectiveSize(p) <= 0);
				p--;
				if (0 <= p && p < plim && info->isSizeable(p))
					{
					*pos = p;
					*pos_x = x;
					*delta_x = delta;
					return YES;
					}
				}
			}
		}

	return NO;
	}


//-----------------------------------------------------------------------------
// - adjustSize
//-----------------------------------------------------------------------------
- (void) adjustSize
	{
	[self changeFrameIfNeeded];
	}



//=============================================================================
// DRAGGING
//=============================================================================
//-----------------------------------------------------------------------------
// - drawDocCells:at:controlView
//-----------------------------------------------------------------------------
- (void) drawDocCells: (MiscCoord_V) pos at: (NSPoint const*) pt
	controlView: (NSView *)controlView
	{
	NSRect rDoc;
	rDoc = [[scroll documentView] visibleRect];
	MiscTableBorder* b = [self otherBorder];
	MiscCoord_V sMin = b->visualForOffset(MiscPixels([self rectMinY: &rDoc]));
	MiscCoord_V sMax = b->visualForOffset(MiscPixels([self rectMaxY: &rDoc]));
	if (sMin >= 0 && sMax >= 0)
		{
		MiscCoord_P p_pos = info->visualToPhysical( pos );
		float delta = [self rectMinY: &rDoc] - b->getOffset( sMin );
	
		NSRect r = {{0,0},{0,0}};
		[self setRect: &r minY: [self pointY: pt] - delta];
		[self setRect: &r width: info->effectiveSize( pos )];
	
		for (int i = sMin; i <= sMax; i++)
			{
			id cell = [self cellAtRow:b->visualToPhysical(i) column:p_pos];
			[self setRect: &r height: b->effectiveSize(i)];
			[cell drawWithFrame:r inView:controlView];
			[self setRect:&r minY:[self rectMinY:&r] + [self rectHeight:&r]];
			}
		}
	}


//-----------------------------------------------------------------------------
// - setDragCache:forPos:
//
//		NOTE *1*: We create a special window just for the sake of getting the
//				text on the drag-cache to display properly. 1) By creating a
//				window, the cells have an object to which they can send their
//				getFieldEditor:for: message.  2) By providing a view which is
//				at least as large as the drag-cache we avoid clipping problems.
//				If it wasn't large enough the text would get clipped at the
//				wrong places.
//		NOTE *2*: The weird combination of non-deferred window plus
//				-reenableDisplay was necessary to get the text to actually
//				display.  Without both of them text wouldn't draw at all.
//				Let's have a look at that AppKit source!
//-----------------------------------------------------------------------------
- (void) setDragCache: (NSImage*) cache forPos: (MiscCoord_V) pos
	{
	NSWindow* w = [[NSWindow allocWithZone:[self zone]] initWithContentRect:*0 styleMask:NSBorderlessWindowMask backing:NSBackingStoreNonretained defer:NO];
#error WindowConversion: 'reenableDisplay' is obsolete.  You can probably remove this call.  Typically drawing should happen as part of the update mechanism after every event.  Display is now optimized using the View setNeedsDisplay: method.  See the conversion doc for more info.
	[w reenableDisplay];										// NOTE *2*
#error ViewConversion: '[NSView setFlipped:]' is obsolete; you must override 'isFlipped' instead of setting externally. However, [NSImage setFlipped:] is not obsolete. If that is what you are using here, no change is needed.
	[[w contentView] setFlipped:YES];

	NSRect rDoc;		rDoc = [[scroll documentView] visibleRect];
	NSRect rClip;		rClip = [[self superview] frame];
	[scroll convertRect:rDoc fromView:[scroll documentView]];

	NSRect r = {{0,0},{0,0}};
	[self setRect: &r width: info->effectiveSize( pos )];
	[self setRect: &r height:
				([self rectHeight: &rClip] + [self rectHeight: &rDoc])];

	NSRect rBorder = r;
	[self setRect: &rBorder height: [self rectHeight: &rClip]];

	NSRect rContent = r;
	[self setRect: &rContent minY: [self rectHeight: &rClip]];

	[w setContentSize:NSMakeSize(NSWidth(r), NSHeight(r))];
#error ViewConversion: '[NSView setFlipped:]' is obsolete; you must override 'isFlipped' instead of setting externally. However, [NSImage setFlipped:] is not obsolete. If that is what you are using here, no change is needed.
	[cache setFlipped:YES];
	[cache setSize:r.size];
	[cache lockFocus];
	[self drawDocCells: pos at: &rContent.origin controlView: [w contentView]];
	[self drawPos: pos inRect: &rBorder controlView: [w contentView]];
	[cache unlockFocus];

	[w release];
	}


//-----------------------------------------------------------------------------
// - setVisibleCache:min:max:
//-----------------------------------------------------------------------------
- (void) setVisibleCache:(NSImage*)cache min:(float*)pMin max:(float*)pMax
	{
	NSRect rDoc;		rDoc = [[scroll documentView] visibleRect];
	NSRect rVis;		rVis = [self visibleRect];
	*pMin = [self rectMinX: &rVis];
	*pMax = [self rectMaxX: &rVis];
	[scroll convertRect:rDoc fromView:[scroll documentView]];
	[scroll convertRect:rVis fromView:self];

	NSRect r;
	[self setRect: &r minX: [self rectMinX: &rVis]];
	[self setRect: &r minY: [self rectMinY: &rVis]];
	[self setRect: &r width: [self rectWidth: &rVis]];
	[self setRect: &r height:
				([self rectHeight: &rVis] + [self rectHeight: &rDoc])];

	[scroll convertRect:r toView:0];
	[cache setSize:r.size];
	[cache lockFocus];
	PScomposite( NSMinX(r), NSMinY(r), NSWidth(r), NSHeight(r),
				[[self window] gState], 0.0, 0.0, NSCompositeCopy );
	[cache unlockFocus];
	}


//-----------------------------------------------------------------------------
// - setWells::forPos:
//-----------------------------------------------------------------------------
- (void) setWells: (MiscTableWell**) w1 : (MiscTableWell**) w2
	forPos: (MiscCoord_V) pos
	{
	NSZone* const z = [self zone];
	MiscTableView* doc = [scroll documentView];
	NSRect rDoc;		rDoc = [doc visibleRect];
	NSRect rClip;		rClip = [[self superview] frame];

	NSRect r = {{0,0},{0,0}};
	[self setRect: &r minX: info->getOffset( pos )];
	[self setRect: &r width: info->effectiveSize( pos )];
	[self setRect: &r height: [self rectHeight: &rClip]];

	*w1 = [[MiscTableWell allocWithZone:z] initWithFrame:r];
	[self addSubview: *w1];

	[self setRect: &r minY: [self rectMinY: &rDoc]];
	[self setRect: &r height: [self rectHeight: &rDoc]];
	*w2 = [[MiscTableWell allocWithZone:z] initWithFrame:r];
	[doc addSubview: *w2];

	[*w1 display];
	[*w2 display];
	}


//-----------------------------------------------------------------------------
// - clearWells::
//-----------------------------------------------------------------------------
- (void) clearWells: (MiscTableWell**) w1 : (MiscTableWell**) w2
	{
	[*w1 removeFromSuperview];
	[*w2 removeFromSuperview];
	[*w1 release];
	[*w2 release];
	*w1 = 0;
	*w2 = 0;
	// NOTE: Does not need display here.  Everything will get displayed later.
	}


//-----------------------------------------------------------------------------
// - offsetFromEvent:
//-----------------------------------------------------------------------------
- (float) offsetFromEvent:(NSEvent *)ev 
	{
	NSPoint mLoc = [ev locationInWindow];
	mLoc = [self convertPoint:mLoc fromView:0];
	return floor( [self pointX: &mLoc] );
	}


//-----------------------------------------------------------------------------
// - calcDrop::
//-----------------------------------------------------------------------------
- (MiscCoord_V) calcDrop: (MiscCoord_V) fromPos
						: (NSPoint const*) mouseDownPt
						: (NSPoint const*) mouseUpPt
	{
	MiscCoord_V toPos = fromPos;

	MiscPixels const start_pos = (MiscPixels) floor([self pointX:mouseDownPt]);
	MiscPixels const end_pos   = (MiscPixels) floor([self pointX:mouseUpPt]);
	MiscPixels const delta_pos = (end_pos - start_pos);

	MiscPixels const SLOP = 4;
	if (delta_pos < -SLOP || SLOP < delta_pos)
		{
		MiscPixels const start_ofs = info->getOffset( fromPos );
		MiscPixels drop_pos = start_ofs + delta_pos;
		if (delta_pos < 0)
			drop_pos += SLOP;
		else
			drop_pos += info->effectiveSize(fromPos) - SLOP;
		toPos = info->visualForOffset( drop_pos );
		if (toPos < 0)
			toPos = 0;
		}

	return toPos;
	}


//-----------------------------------------------------------------------------
// - dragEvent:inPos:
//-----------------------------------------------------------------------------
- (MiscCoord_V) dragEvent:(NSEvent *)event inPos: (MiscCoord_V) pos
	{
	int const WANTED = (NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSPeriodicMask);
	NSPoint mouseDownPt = [event locationInWindow];

	mouseDownPt = [self convertPoint:mouseDownPt fromView:0];
	NSPoint mouseUpPt = mouseDownPt;

	[[self window] disableFlushWindow];

	MiscTableWell *w1, *w2;
	[self setWells: &w1 : &w2 forPos: pos];

	NSZone* const z = [self zone];
	float pMin,pMax;
	NSImage* visCache = [[NSImage allocWithZone:z] init];
	NSImage* dragCache = [[NSImage allocWithZone:z] init];
	[self setVisibleCache: visCache min: &pMin max: &pMax];
	[self setDragCache: dragCache forPos: pos];

	NSSize size;
	size = [dragCache size];
	float pLoc = info->getOffset( pos );
	float delta = [self offsetFromEvent: event] - pLoc;

#error EventConversion: NXTrackingTimer is obsolete; use NSEvent 'startPeriodicEventsAfterDelay:withPeriod:' and 'stopPeriodicEvents'.  You can probably delete this variable and its uses.
	NXTrackingTimer * timer = 0;
	NSEvent *lastEvent = event;
	[scroll lockFocus];
	while (1)
		{
		NSPoint pt;
		NSRect rDrag = {{0,0},{0,0}};
		float const w = [self sizeWidth: &size];
		float const dw =
			dmin( dmin( dmin(w, pMax - pMin), pMax - pLoc), pLoc + w - pMin );
		BOOL const shouldDraw = (dw > 0.0);
		if (shouldDraw)
			{
			if (isHorz)
				{
				pt.x = dmax( pLoc, pMin );
				pt.y = size.height;
				if (pLoc < pMin)
					rDrag.origin.x = (pMin - pLoc);
				}
			else
				{
				pt.x = 0.0;
				pt.y = dmax( pLoc, pMin ) + dw;
				if (pLoc + w >= pMax)
					rDrag.origin.y = pLoc + w - pMax;
				}
			pt = [scroll convertPoint:pt fromView:self];
			[self setRect: &rDrag width: dw];
			[self setRect: &rDrag height: [self sizeHeight: &size]];
			[dragCache compositeToPoint:pt fromRect:rDrag operation:NSCompositeCopy];
			}
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
		{
    id newVar = [self window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}

		event = [[self window] nextEventMatchingMask:WANTED];

		[[self window] disableFlushWindow];
		if (shouldDraw)
			{
			NSRect rVis = {{0,0},{0,0}};
			NSSize s; s = [visCache size];
			float xTarg;
			if (isHorz)
				{
				xTarg = (pLoc < pMin ? 0.0 : pLoc - pMin);
				}
			else
				{
				if (pLoc < pMin)
					xTarg = s.height - dw;
				else if (pLoc < pMax)
					xTarg = pMax - pLoc - dw;
				else
					xTarg = 0.0;
				}
			[self setRect: &rVis minX: xTarg];
			[self setRect: &rVis width: [self rectWidth: &rDrag]];
			[self setRect: &rVis height: [self rectHeight: &rDrag]];
			[visCache compositeToPoint:pt fromRect:rVis operation:NSCompositeCopy];
			}

		if (event == 0)
			break;
		if ([event type] == NSLeftMouseUp)
			{
			mouseUpPt = [event locationInWindow];
			mouseUpPt = [self convertPoint:mouseUpPt fromView:0];
			break;
			}
		if ([event type] != NSPeriodic)
			lastEvent = event;

		float mLoc = [self offsetFromEvent: lastEvent];
		if ((mLoc < pMin && pMin > 0.0) ||
			(mLoc > pMax && pMax < [self rectMaxX: &[self bounds]]))
			{
			[self autoscroll:lastEvent];
			[self setVisibleCache: visCache min: &pMin max: &pMax];
			mLoc = [self offsetFromEvent: lastEvent];
			startTimer( timer );
			}
		else
			{
			stopTimer( timer );
			}

		pLoc = mLoc - delta;
		if (pLoc < pMin - [self sizeWidth: &size])
			pLoc = pMin - [self sizeWidth: &size];
		else if (pLoc > pMax)
			pLoc = pMax;
		}

	stopTimer( timer );
	[scroll unlockFocus];
	[self clearWells: &w1 : &w2];
	[visCache release];
	[dragCache release];

	MiscCoord_V const toPos = [self calcDrop:pos :&mouseDownPt :&mouseUpPt];

	if (toPos != pos)
		{
		BOOL const was_auto = NO /* was [<view> isAutoDisplay] */;
#error ViewConversion: 'setAutodisplay:' is obsolete
		[scroll setAutodisplay:NO];
		[self moveFrom: pos to: toPos];
		[self selectPos: toPos];
		[scroll border:[self borderType] slotDraggedFrom:pos to:toPos];
#error ViewConversion: 'setAutodisplay:' is obsolete
		[scroll setAutodisplay:was_auto];		// Will display if needed.
		}
	else
		{
		// Need to redisplay the slot that the wells were covering.
		MiscBorderType const b = [self borderType];
		int const phys_pos = [scroll border:b slotAtPosition:toPos];
		[scroll border:b drawSlotTitle:phys_pos];
		[scroll border:b drawSlot:phys_pos];
		}

#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
	{
    id newVar = [self window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}
	return toPos;
	}


//-----------------------------------------------------------------------------
// - awaitDragEvent:inPos:
//-----------------------------------------------------------------------------
- (MiscCoord_V) awaitDragEvent:(NSEvent *)event inPos: (MiscCoord_V) pos
	{
	MiscCoord_V toPos = pos;
	int const WANTED = (NSLeftMouseUpMask | NSLeftMouseDraggedMask);
	float const FOREVER = FLT_MAX;
	float const SLOP = 2.0;
	NSEvent *mouseDown = event;
	for (; ;)
		{
		NSEvent *peek;
		NSEvent *p = (peek = [[self window] nextEventMatchingMask:WANTED untilDate:[NSDate dateWithTimeIntervalSinceNow:FOREVER] inMode:NSEventTrackingRunLoopMode dequeue:NO]);
		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		else if ([p locationInWindow].x > [mouseDown locationInWindow].x + SLOP ||
				 [p locationInWindow].y > [mouseDown locationInWindow].y + SLOP ||
				 [p locationInWindow].x < [mouseDown locationInWindow].x - SLOP ||
				 [p locationInWindow].y < [mouseDown locationInWindow].y - SLOP)
			{
			toPos = [self dragEvent: mouseDown inPos: pos];
			break;
			}
		else
			[[self window] nextEventMatchingMask:NSLeftMouseDraggedMask];
		}
	return toPos;
	}
	


//=============================================================================
// SELECTION
//=============================================================================
//-----------------------------------------------------------------------------
// - selectionEvent:fromPos:
//-----------------------------------------------------------------------------
- (void) selectionEvent:(NSEvent *)p fromPos:(MiscCoord_V)pos
	{
	BOOL doubleClicked = ([p clickCount] > 1);

	[tracker mouseDown: p atPos: pos];
	[scroll reflectSelection];

#error EventConversion: NXTrackingTimer is obsolete; use NSEvent 'startPeriodicEventsAfterDelay:withPeriod:' and 'stopPeriodicEvents'.  You can probably delete this variable and its uses.
	NXTrackingTimer * timer = 0;
	startTimer( timer );
	NSEvent *lastEvent = p;

	while (1)
		{
		p = [[self window] nextEventMatchingMask:NSLeftMouseUpMask |
								NSLeftMouseDraggedMask |
								NSPeriodicMask];

		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		else if ([p type] == NSPeriodic)
			[self autoscroll:lastEvent];
		else
			lastEvent = p;

		NSPoint new_loc = [lastEvent locationInWindow];
		new_loc = [self convertPoint:new_loc fromView:0];
		MiscCoord_V new_pos = [self posForMousePt: &new_loc];
		if (new_pos != pos)
			{
			pos = new_pos;
			[tracker mouseDragged: p atPos: pos];
			[scroll reflectSelection];
			}
		}

	stopTimer( timer );

	[tracker mouseUp: p atPos: pos];
	[scroll reflectSelection];
	[self endTracking: pos];

	if (doubleClicked)
		[scroll sendDoubleActionIfEnabled];
	else
		[scroll sendActionIfEnabled];
	}



//=============================================================================
// MOUSE-EVENTS
//=============================================================================
//-----------------------------------------------------------------------------
// - acceptsFirstMouse
//-----------------------------------------------------------------------------
#warning ViewConversion: 'acceptsFirstMouse:' (used to be 'acceptsFirstMouse') now takes the event as an arg
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - mouseDown:
//-----------------------------------------------------------------------------
- (void)mouseDown:(NSEvent *)p 
	{
	NSPoint evpt = [p locationInWindow];
	evpt = [self convertPoint:evpt fromView:0];
	MiscCoord_V pos = [self posForMousePt: &evpt];
#error EventConversion: addToEventMask:NX_MOUSEDRAGGEDMASK: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	int old_mask = [[self window] addToEventMask:NSLeftMouseDraggedMask];
	MiscPixels x, delta_x;

	if ([self inHotZone:&evpt forPos:&pos atX:&x deltaX:&delta_x])
		{
		[self startTracking];
		[self selectPos:pos];
		MiscPixels finalWidth;
		BOOL doit;
		doit = [self resizeEvent:p inPos:pos atX:x deltaX:delta_x
				finalWidth:&finalWidth];
		if (doit)
			{
			[[self window] disableFlushWindow];
			[self setPos:pos width:finalWidth];
			[scroll border:[self borderType] slotResized:pos];
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
			{
    id newVar = [self window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}
			}
		[self endTracking:pos];
		[scroll sendActionIfEnabled];
		}
	else if (info->isDraggable() &&
		(info->isModifierDrag() == (([p modifierFlags] & NSCommandKeyMask) != 0)))
		{
		[self startTracking];
		[self selectPos:pos];
		[[self window] flushWindow];
		PSFlush();
		MiscCoord_V toPos = [self awaitDragEvent:p inPos:pos];
		[self endTracking:toPos];
		[scroll sendActionIfEnabled];
		}
	else
		{
		assert( tracker != 0 );
		[self startTracking];
		[self selectionEvent: p fromPos: pos];
		}

#error EventConversion: setEventMask:old_mask: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	[[self window] setEventMask:old_mask];
}

@end
