//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableCell.M
//
//		Default cell class used by MiscTableScroll to display text.
//
// NOTE *SET-OWNER-VALUE*
//		We must implement these methods so that MiscTableScroll will call 
//		these methods instead of the explicit instance -setXxx methods.	 
//		However, we don't need to do any work.	We will ask the owner for 
//		it's value whenever we need it.	 
//
// FIXME: optional-allocation stuff does not address alignment requirements.
// FIXME: automate most of the optional-allocation stuff.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableCell.M,v 1.7 96/01/13 23:41:21 zarnuk Exp $
// $Log:		MiscTableCell.M,v $
// Revision 1.7	 96/01/13  23:41:21	 zarnuk
// Tag is no longer optional.
// 
// Revision 1.6	 95/10/14  16:48:53	 zarnuk
// Now calls -setUseOwnerFont:YES in -initTextCell:
// 
// Revision 1.5	 95/10/03  05:01:54	 sunshine
// Implemented -copyFromZone: instead of -copy.	 Now all allocations are made
// from [self zone] instead of default-malloc-zone.
// 
// Revision 1.4	 95/10/01  15:07:07	 sunshine
// Added highlightTextColor and highlightBackgroundColor and corresponding
// manipulation methods and 'gray' methods.	 Fixed a bunch of bugs.	 For
// instance the -init methods were unconditionally setting the font and color
// (and allocated memory for them) and clearing the useOwner... stuff.
// Simplified -*pos methods somewhat by removing copy/paste code.  Now each
// one calls its predecessor.  Made default background color LTGRAY once again
// because Paul hatez it and it's much more consistent with the rest of the
// AppKit (every scrolling list in practically every application has a LTGRAY
// background).
// 
// Revision 1.3	 95/09/29  16:01:02	 zarnuk
// lazy-allocation for tags, colors.
// useOwner options for font, colors.
//-----------------------------------------------------------------------------
#import <misckit/MiscTableCell.h>

extern "Objective-C" {
#import <AppKit/NSFont.h>
#import <AppKit/NSText.h>
#import <AppKit/psopsOpenStep.h>
}

extern "C" {
#import <string.h>
}

#define MISC_TC_VERSION_0		0
#define MISC_TC_VERSION_1		1
#define MISC_TC_VERSION			MISC_TC_VERSION_1

@implementation MiscTableCell

//-----------------------------------------------------------------------------
// + initialize
//-----------------------------------------------------------------------------
+ (void)initialize
	{
	if (self == [MiscTableCell class])
		{
		[self setVersion: MISC_TC_VERSION];
		}
	return;
	}


//-----------------------------------------------------------------------------
// + defaultFont
//-----------------------------------------------------------------------------
+ (NSFont *)defaultFont
	{
	return [NSFont userFontOfSize:12.0];
	}


//-----------------------------------------------------------------------------
// initTextCell:
//		NOTE *1*  -[Cell initTextCell:] invokes -setStringValue: which
//		invokes -setFont:.	This means that this cell will get an explicit
//		-setFont call before we finish initialization, and we need to 
//		restore the useOwnerFont setting.
//-----------------------------------------------------------------------------
- initTextCell:(NSString *)s
	{
	[super initTextCell:s];								// NOTE *1*
	[self setBordered:NO];
	[self setWrap:NO];
	[self setAlignment:NSLeftTextAlignment];
	[self setUseOwnerFont:YES];							// NOTE *1*
	[self setOwnerFont: [[self class] defaultFont]];
	return self;
	}


//-----------------------------------------------------------------------------
// initIconCell:
//-----------------------------------------------------------------------------
- initImageCell:(NSImage *)s
	{
	[super initImageCell:s];
	[self setBordered:NO];
	return self;
	}


//-----------------------------------------------------------------------------
// tag
//-----------------------------------------------------------------------------
- (int) tag
	{
	return tag;
	}


//-----------------------------------------------------------------------------
// setTag:
//-----------------------------------------------------------------------------
- (void)setTag:(int)x
	{
	tag = x;
}


//-----------------------------------------------------------------------------
// image
//-----------------------------------------------------------------------------
- image
	{
	return ([self type] == NSIconCellType ? support : 0);
	}


//-----------------------------------------------------------------------------
// setImage:
//-----------------------------------------------------------------------------
- (void)setImage:(NSImage *)image
	{
	if ([self type] == NSIconCellType)
		{
		char const* name = [[image name] cString];
		if (name)
			{
			[self setImage:[NSImage imageNamed:[NSString stringWithCString:name]]];
			}
		else
			{
			support = image;
			if ([self stringValue] != 0 && cFlags1.freeText)
				NSZoneFree( [self zone], [self stringValue] );
			[self stringValue] = 0;
			cFlags1.freeText = NO;
			}
		}
}


//-----------------------------------------------------------------------------
// owner
//-----------------------------------------------------------------------------
- owner
	{
	return owner;
	}


//-----------------------------------------------------------------------------
// setOwner:
//-----------------------------------------------------------------------------
- setOwner:obj
	{
	if (obj != owner)
		{
		owner = obj;
		if (owner != 0)
			{
			if ([owner respondsToSelector:@selector(font)])
				[self setOwnerFont:[owner font]];
			if ([owner respondsToSelector:@selector(textColor)])
				[self setOwnerTextColor:[owner textColor]];
			if ([owner respondsToSelector:@selector(backgroundColor)])
				[self setOwnerBackgroundColor:[owner backgroundColor]];
			if ([owner respondsToSelector:@selector(highlightTextColor)])
				[self setOwnerTextColor:[owner highlightTextColor]];
			if ([owner respondsToSelector:@selector(highlightBackgroundColor)])
				[self setOwnerBackgroundColor:[owner highlightBackgroundColor]];
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// font
//-----------------------------------------------------------------------------
- font
	{
	return [super font];
	}


//-----------------------------------------------------------------------------
// setFont:
//-----------------------------------------------------------------------------
- (void)setFont:(NSFont *)fontObj
	{
	[self setUseOwnerFont:NO];
	[super setFont:fontObj];
}


//-----------------------------------------------------------------------------
// tc1Flags
//-----------------------------------------------------------------------------
- (unsigned int) tc1Flags
	{
	return tc1_flags;
	}


//-----------------------------------------------------------------------------
// tc1DataSize
//-----------------------------------------------------------------------------
- (unsigned int) tc1DataSize
	{
	unsigned int size = 0;
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR)
		size += [self tc1TextColorLen];
	if (tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR)
		size += [self tc1BackgroundColorLen];
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR_H)
		size += [self tc1HighlightTextColorLen];
	if (tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR_H)
		size += [self tc1HighlightBackgroundColorLen];
	return size;
	}


//-----------------------------------------------------------------------------
// tc1TextColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1TextColorPos
	{
	return 0;
	}


//-----------------------------------------------------------------------------
// tc1BackgroundColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1BackgroundColorPos
	{
	unsigned int pos = [self tc1TextColorPos];
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR)
		pos += [self tc1TextColorLen];
	return pos;
	}


//-----------------------------------------------------------------------------
// tc1HighlightTextColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightTextColorPos
	{
	unsigned int pos = [self tc1BackgroundColorPos];
	if (tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR)
		pos += [self tc1BackgroundColorLen];
	return pos;
	}


//-----------------------------------------------------------------------------
// tc1HighlightBackgroundColorPos
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightBackgroundColorPos
	{
	unsigned int pos = [self tc1HighlightTextColorPos];
	if (tc1_flags & MISC_TC1_SELF_TEXT_COLOR_H)
		pos += [self tc1HighlightTextColorLen];
	return pos;
	}


//-----------------------------------------------------------------------------
// tc1TextColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1TextColorLen
	{
	return sizeof(NSColor *);
	}


//-----------------------------------------------------------------------------
// tc1BackgroundColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1BackgroundColorLen
	{
	return sizeof(NSColor *);
	}


//-----------------------------------------------------------------------------
// tc1HighlightTextColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightTextColorLen
	{
	return sizeof(NSColor *);
	}


//-----------------------------------------------------------------------------
// tc1HighlightBackgroundColorLen
//-----------------------------------------------------------------------------
- (unsigned int) tc1HighlightBackgroundColorLen
	{
	return sizeof(NSColor *);
	}


//-----------------------------------------------------------------------------
// tc1TextColorPtr
//-----------------------------------------------------------------------------
- (NSColor *) tc1TextColorPtr
	{
	return (NSColor *) ((char*)tc1_data + [self tc1TextColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1BackgroundColorPtr
//-----------------------------------------------------------------------------
- (NSColor *) tc1BackgroundColorPtr
	{
	return (NSColor *) ((char*)tc1_data + [self tc1BackgroundColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1HighlightTextColorPtr
//-----------------------------------------------------------------------------
- (NSColor *) tc1HighlightTextColorPtr
	{
	return (NSColor *) ((char*)tc1_data + [self tc1HighlightTextColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1HighlightBackgroundColorPtr
//-----------------------------------------------------------------------------
- (NSColor *) tc1HighlightBackgroundColorPtr
	{
	return (NSColor *) ((char*)tc1_data + [self tc1HighlightBackgroundColorPos]);
	}


//-----------------------------------------------------------------------------
// tc1InsertData:len:pos:
//-----------------------------------------------------------------------------
- (void*) tc1InsertData:(void const*)data
		pos:(unsigned int)pos
		len:(unsigned int)len
	{
	if (len > 0)
		{
		unsigned int const old_size = [self tc1DataSize];
		unsigned int const new_size = old_size + len;
		if (old_size == 0)
			tc1_data = NSZoneMalloc( [self zone], new_size );
		else
			{
			tc1_data = NSZoneRealloc( [self zone], tc1_data, new_size );
			if (pos < old_size)
				memmove( (char*) tc1_data + pos + len,			// Destination
						(char*) tc1_data + pos,					// Source
						old_size - pos );						// # bytes.
			}
		if (data != 0)
			memcpy( (char*) tc1_data + pos, data, len );
		else
			memset( (char*) tc1_data + pos, 0, len );
		return (void*)((char*) tc1_data + pos );
		}
	return 0;
	}


//-----------------------------------------------------------------------------
// tc1DeleteDataPos:len:
//-----------------------------------------------------------------------------
- (void) tc1DeleteDataPos:(unsigned int)pos len:(unsigned int)len
	{
	if (len > 0)
		{
		unsigned int const old_size = [self tc1DataSize];
		unsigned int const new_size = old_size - len;
		if (new_size == 0)
			{
			NSZoneFree( [self zone], tc1_data );
			tc1_data = 0;
			}
		else
			{
			if (pos < new_size)
				memmove( (char*) tc1_data + pos,				// Destination
						(char*) tc1_data + pos + len,			// Source
						new_size - pos );						// # bytes
			tc1_data = NSZoneRealloc( [self zone], tc1_data, new_size );
			}
		}
	}


//-----------------------------------------------------------------------------
// useOwnerFont
//-----------------------------------------------------------------------------
- (BOOL) useOwnerFont
	{
	return ((tc1_flags & MISC_TC1_SELF_FONT) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerTextColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerTextColor
	{
	return ((tc1_flags & MISC_TC1_SELF_TEXT_COLOR) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerBackgroundColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerBackgroundColor
	{
	return ((tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerHighlightTextColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerHighlightTextColor
	{
	return ((tc1_flags & MISC_TC1_SELF_TEXT_COLOR_H) == 0);
	}


//-----------------------------------------------------------------------------
// useOwnerHighlightBackgroundColor
//-----------------------------------------------------------------------------
- (BOOL) useOwnerHighlightBackgroundColor
	{
	return ((tc1_flags & MISC_TC1_SELF_BACKGROUND_COLOR_H) == 0);
	}


//-----------------------------------------------------------------------------
// setUseOwnerFont:
//-----------------------------------------------------------------------------
- setUseOwnerFont: (BOOL) flag
	{
	if ([self useOwnerFont] != flag)
		{
		if (flag)
			{
			tc1_flags &= ~(MISC_TC1_SELF_FONT);
			if (owner != 0 && [owner respondsToSelector:@selector(font)])
				[self setOwnerFont:[owner font]];
			}
		else // (!flag)
			{
			tc1_flags |= MISC_TC1_SELF_FONT;
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// setUseOwnerTextColor:
//-----------------------------------------------------------------------------
- setUseOwnerTextColor: (BOOL) flag
	{
	if ([self useOwnerTextColor] != flag)
		{
		unsigned int const pos = [self tc1TextColorPos];
		unsigned int const len = [self tc1TextColorLen];
		if (flag)
			{
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_TEXT_COLOR);
			}
		else // (!flag)
			{
			NSColor * color = [[self class] defaultTextColor];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_TEXT_COLOR;
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// setUseOwnerBackgroundColor:
//-----------------------------------------------------------------------------
- setUseOwnerBackgroundColor: (BOOL) flag
	{
	if ([self useOwnerBackgroundColor] != flag)
		{
		unsigned int const pos = [self tc1BackgroundColorPos];
		unsigned int const len = [self tc1BackgroundColorLen];
		if (flag)
			{
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_BACKGROUND_COLOR);
			}
		else // (!flag)
			{
			NSColor * color = [[self class] defaultBackgroundColor];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_BACKGROUND_COLOR;
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// setUseOwnerHighlightTextColor:
//-----------------------------------------------------------------------------
- setUseOwnerHighlightTextColor: (BOOL) flag
	{
	if ([self useOwnerHighlightTextColor] != flag)
		{
		unsigned int const pos = [self tc1HighlightTextColorPos];
		unsigned int const len = [self tc1HighlightTextColorLen];
		if (flag)
			{
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_TEXT_COLOR_H);
			}
		else // (!flag)
			{
			NSColor * color = [[self class] defaultHighlightTextColor];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_TEXT_COLOR_H;
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// setUseOwnerHighlightBackgroundColor:
//-----------------------------------------------------------------------------
- setUseOwnerHighlightBackgroundColor: (BOOL) flag
	{
	if ([self useOwnerHighlightBackgroundColor] != flag)
		{
		unsigned int const pos = [self tc1HighlightBackgroundColorPos];
		unsigned int const len = [self tc1HighlightBackgroundColorLen];
		if (flag)
			{
			[self tc1DeleteDataPos:pos len:len];
			tc1_flags &= ~(MISC_TC1_SELF_BACKGROUND_COLOR_H);
			}
		else // (!flag)
			{
			NSColor * color = [[self class] defaultHighlightBackgroundColor];
			[self tc1InsertData:&color pos:pos len:len];
			tc1_flags |= MISC_TC1_SELF_BACKGROUND_COLOR_H;
			}
		}
	return self;
	}


//-----------------------------------------------------------------------------
// setOwnerFont:
//-----------------------------------------------------------------------------
- setOwnerFont:fontObj
	{
	if ([self useOwnerFont])
		[super setFont:fontObj];
	return self;
	}


//-----------------------------------------------------------------------------
// setOwner values...	NOTE *SET-OWNER-VALUE*
//-----------------------------------------------------------------------------

- setOwnerTextColor: (NSColor *) pcolor					{ return self; }
- setOwnerBackgroundColor: (NSColor *) pcolor				{ return self; }
- setOwnerHighlightTextColor: (NSColor *) pcolor			{ return self; }
- setOwnerHighlightBackgroundColor: (NSColor *) pcolor	{ return self; }


//-----------------------------------------------------------------------------
// textGray
//-----------------------------------------------------------------------------
- (float)textGray
	{
	float gray;
	[[[self textColor] colorUsingColorSpaceName:NSCalibratedWhiteColorSpace] getWhite:&gray alpha:NULL];
	return gray;
	}


//-----------------------------------------------------------------------------
// setTextGray:
//-----------------------------------------------------------------------------
- setTextGray:(float)value
	{
	[self setTextColor:[NSColor colorWithCalibratedWhite:value alpha:1.0]];
	return self;
	}


//-----------------------------------------------------------------------------
// backgroundGray
//-----------------------------------------------------------------------------
- (float)backgroundGray
	{
	float gray;
	[[[self backgroundColor] colorUsingColorSpaceName:NSCalibratedWhiteColorSpace] getWhite:&gray alpha:NULL];
	return gray;
	}


//-----------------------------------------------------------------------------
// setBackgroundGray:
//-----------------------------------------------------------------------------
- setBackgroundGray:(float)value
	{
	[self setBackgroundColor:[NSColor colorWithCalibratedWhite:value alpha:1.0]];
	return self;
	}


//-----------------------------------------------------------------------------
// highlightTextGray
//-----------------------------------------------------------------------------
- (float)highlightTextGray
	{
	float gray;
	[[[self highlightTextColor] colorUsingColorSpaceName:NSCalibratedWhiteColorSpace] getWhite:&gray alpha:NULL];
	return gray;
	}


//-----------------------------------------------------------------------------
// setHighlightTextGray:
//-----------------------------------------------------------------------------
- setHighlightTextGray:(float)value
	{
	return [self setHighlightTextColor: [NSColor colorWithCalibratedWhite:value alpha:1.0]];
	}


//-----------------------------------------------------------------------------
// highlightBackgroundGray
//-----------------------------------------------------------------------------
- (float)highlightBackgroundGray
	{
	float gray;
	[[[self highlightBackgroundColor] colorUsingColorSpaceName:NSCalibratedWhiteColorSpace] getWhite:&gray alpha:NULL];
	return gray;
	}


//-----------------------------------------------------------------------------
// setHighlightBackgroundGray:
//-----------------------------------------------------------------------------
- setHighlightBackgroundGray:(float)value
	{
	return [self setHighlightBackgroundColor: [NSColor colorWithCalibratedWhite:value alpha:1.0] ];
	}


//-----------------------------------------------------------------------------
// +defaultBackgroundColor
//-----------------------------------------------------------------------------
+ (NSColor *) defaultBackgroundColor
	{
	return [NSColor lightGrayColor];
	}


//-----------------------------------------------------------------------------
// backgroundColor
//-----------------------------------------------------------------------------
- (NSColor *) backgroundColor
	{
	if ([self useOwnerBackgroundColor])
		{
		if (owner != 0 && [owner respondsToSelector:@selector(backgroundColor)])
			return [owner backgroundColor];
		return [[self class] defaultBackgroundColor];
		}
	NSColor * const* const p = [self tc1BackgroundColorPtr];
	return *p;
	}


//-----------------------------------------------------------------------------
// setBackgroundColor
//-----------------------------------------------------------------------------
- (void)setBackgroundColor:(NSColor *)c
	{
	NSColor *p;
	[self setUseOwnerBackgroundColor:NO];
	p = [self tc1BackgroundColorPtr];
	*p = c;
}


//-----------------------------------------------------------------------------
// +defaultHighlightBackgroundColor
//-----------------------------------------------------------------------------
+ (NSColor *) defaultHighlightBackgroundColor
	{
	return [NSColor whiteColor];
	}


//-----------------------------------------------------------------------------
// highlightBackgroundColor
//-----------------------------------------------------------------------------
- (NSColor *) highlightBackgroundColor
	{
	if ([self useOwnerHighlightBackgroundColor])
		{
		if (owner && [owner respondsToSelector:@selector(highlightBackgroundColor)])
			return [owner highlightBackgroundColor];
		return [[self class] defaultHighlightBackgroundColor];
		}
	NSColor * const* const p = [self tc1HighlightBackgroundColorPtr];
	return *p;
	}


//-----------------------------------------------------------------------------
// setHighlightBackgroundColor
//-----------------------------------------------------------------------------
- setHighlightBackgroundColor: (NSColor *) c
	{
	NSColor *p;
	[self setUseOwnerHighlightBackgroundColor:NO];
	p = [self tc1HighlightBackgroundColorPtr];
	*p = c;
	return self;
	}


//-----------------------------------------------------------------------------
// +defaultTextColor
//-----------------------------------------------------------------------------
+ (NSColor *) defaultTextColor
	{
	return [NSColor blackColor];
	}


//-----------------------------------------------------------------------------
// textColor
//-----------------------------------------------------------------------------
- (NSColor *) textColor
	{
	if ([self useOwnerTextColor])
		{
		if (owner != 0 && [owner respondsToSelector:@selector(textColor)])
			return [owner textColor];
		return [[self class] defaultTextColor];
		}
	NSColor * const* const p = [self tc1TextColorPtr];
	return *p;
	}


//-----------------------------------------------------------------------------
// setTextColor
//-----------------------------------------------------------------------------
- (void)setTextColor:(NSColor *)c
	{
	NSColor *p;
	[self setUseOwnerTextColor:NO];
	p = [self tc1TextColorPtr];
	*p = c;
}


//-----------------------------------------------------------------------------
// +defaultHighlightTextColor
//-----------------------------------------------------------------------------
+ (NSColor *) defaultHighlightTextColor
	{
	return [NSColor blackColor];
	}


//-----------------------------------------------------------------------------
// highlightTextColor
//-----------------------------------------------------------------------------
- (NSColor *) highlightTextColor
	{
	if ([self useOwnerHighlightTextColor])
		{
		if (owner != 0 && [owner respondsToSelector:@selector(highlightTextColor)])
			return [owner highlightTextColor];
		return [[self class] defaultHighlightTextColor];
		}
	NSColor * const* const p = [self tc1HighlightTextColorPtr];
	return *p;
	}


//-----------------------------------------------------------------------------
// setHighlightTextColor
//-----------------------------------------------------------------------------
- setHighlightTextColor: (NSColor *) c
	{
	NSColor *p;
	[self setUseOwnerHighlightTextColor:NO];
	p = [self tc1HighlightTextColorPtr];
	*p = c;
	return self;
	}


//-----------------------------------------------------------------------------
// bgColor
//-----------------------------------------------------------------------------
- (NSColor *) bgColor
	{
	// FIXME: Use [self isHighlighted] when we figure out how to support
	// this for lazy mode.
	return [self state] ?
				highlightBackgroundColor: [self backgroundColor];
	}


//-----------------------------------------------------------------------------
// fgColor
//-----------------------------------------------------------------------------
- (NSColor *) fgColor
	{
	// FIXME: Use [self isHighlighted] when we figure out how to support
	// this for lazy mode.
	return [self state] ? highlightTextColor: [self textColor];
	}


//-----------------------------------------------------------------------------
// calcCellSize:inRect:
//-----------------------------------------------------------------------------
- (NSSize)cellSizeForBounds:(NSRect)aRect
	{
	
    NSSize theSize;
    theSize = [super cellSizeForBounds:aRect];
	theSize.width += 1;
	theSize.height += 1;
	return theSize;
	}


//-----------------------------------------------------------------------------
// setTextAttributes:
//-----------------------------------------------------------------------------
- setUpFieldEditorAttributes:textObject
	{
	[super setUpFieldEditorAttributes:textObject];
	[textObject setTextColor:[self fgColor]];
	[textObject setBackgroundColor:[self bgColor]];
	return textObject;
	}


//-----------------------------------------------------------------------------
// drawInside:inView:
//-----------------------------------------------------------------------------
- (void)drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
	{
	NSRect rect;

	[[self bgColor] set];
	rect = NSMakeRect(NSMinX(cellFrame), NSMinY(cellFrame), NSWidth(cellFrame) - 1, NSHeight(cellFrame) - 1);
	NSRectFill(rect);
	[super drawInteriorWithFrame:rect inView:controlView];
}


//-----------------------------------------------------------------------------
// drawSelf:inView:
//-----------------------------------------------------------------------------
- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)aView
	{
	float const MEDGRAY = 0.5;
	NSRect r[2];

	PSsetgray( MEDGRAY );
	(r[0]) = NSMakeRect(NSMaxX(cellFrame) - 1, NSMinY(cellFrame), 1, NSHeight(cellFrame));
	(r[1]) = NSMakeRect(NSMinX(cellFrame), NSMaxY(cellFrame) - 1, NSWidth(cellFrame), 1);
	NSRectFillList(r, 2);

	[self drawInteriorWithFrame:cellFrame inView:aView];
}


//-----------------------------------------------------------------------------
// tc1DestroyData
//-----------------------------------------------------------------------------
- (void) tc1DestroyData
	{
	tc1_flags = 0;
	}


//-----------------------------------------------------------------------------
// tc1FreeData
//-----------------------------------------------------------------------------
- (void) tc1FreeData
	{
	if (tc1_data != 0)
		{
		NSZoneFree( [self zone], tc1_data );
		tc1_data = 0;
		}
	}


//-----------------------------------------------------------------------------
// free
//-----------------------------------------------------------------------------
- (void)dealloc
	{
	[self tc1DestroyData];
	[self tc1FreeData];
	{ [super dealloc]; return; };
	}


//-----------------------------------------------------------------------------
// copyFromZone:
//-----------------------------------------------------------------------------
- copyWithZone:(NSZone*)zone;
	{
	MiscTableCell* the_clone = [super copyWithZone:zone];
	if (tc1_data != 0)
		{
		unsigned int const size = [self tc1DataSize];
		void* p = NSZoneMalloc( [self zone], size );
		memcpy( p, tc1_data, size );
		the_clone->tc1_data = p;
		}
	return the_clone;
	}


//-----------------------------------------------------------------------------
// read:
//-----------------------------------------------------------------------------
- (id)initWithCoder:(NSCoder *)aDecoder
	{
	[super initWithCoder:aDecoder];

	[self tc1DestroyData];
	[self tc1FreeData];

	unsigned int x;
	int ver = [aDecoder versionForClassName:[[MiscTableCell class] name]];

	owner = [[aDecoder decodeObject] retain];

	if (ver == MISC_TC_VERSION_0)
		{
		[aDecoder decodeValueOfObjCType:@encode(unsigned int) at:&x];
		if (x & MISC_TC1_HAS_TAG)
			[aDecoder decodeValueOfObjCType:@encode(int) at:&tag];
		else
			tag = 0;
		}
	else if (ver == MISC_TC_VERSION_1)
		{
		[aDecoder decodeValueOfObjCType:@encode(int) at:&tag];
		[aDecoder decodeValueOfObjCType:@encode(unsigned int) at:&x];
		}
	else
		{
		[self error:"Cannot read: unknown version."];
		}

	if (x & MISC_TC1_SELF_TEXT_COLOR)
		{
#warning ColorConversion: [[aDecoder decodeNXColor] retain] used to be NXReadColor(aDecoder).  Use 'decodeNXColor' to read old style colors, but use 'decodeObject' to read OpenStep colors.
		NSColor * c = [[aDecoder decodeNXColor] retain];
		[self setTextColor:c];
		}
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR)
		{
#warning ColorConversion: [[aDecoder decodeNXColor] retain] used to be NXReadColor(aDecoder).  Use 'decodeNXColor' to read old style colors, but use 'decodeObject' to read OpenStep colors.
		NSColor * c = [[aDecoder decodeNXColor] retain];
		[self setBackgroundColor:c];
		}
	if (x & MISC_TC1_SELF_TEXT_COLOR_H)
		{
#warning ColorConversion: [[aDecoder decodeNXColor] retain] used to be NXReadColor(aDecoder).  Use 'decodeNXColor' to read old style colors, but use 'decodeObject' to read OpenStep colors.
		NSColor * c = [[aDecoder decodeNXColor] retain];
		[self setHighlightTextColor:c];
		}
	if (x & MISC_TC1_SELF_BACKGROUND_COLOR_H)
		{
#warning ColorConversion: [[aDecoder decodeNXColor] retain] used to be NXReadColor(aDecoder).  Use 'decodeNXColor' to read old style colors, but use 'decodeObject' to read OpenStep colors.
		NSColor * c = [[aDecoder decodeNXColor] retain];
		[self setHighlightBackgroundColor:c];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// write:
//-----------------------------------------------------------------------------
- (void)encodeWithCoder:(NSCoder *)aCoder
	{
	[super encodeWithCoder:aCoder];
	[aCoder encodeConditionalObject:owner];
	[aCoder encodeValueOfObjCType:@encode(int) at:&tag];
	[aCoder encodeValueOfObjCType:@encode(unsigned int) at:&tc1_flags];
	if (![self useOwnerTextColor])
		[aCoder encodeObject:[self textColor]];
	if (![self useOwnerBackgroundColor])
		[aCoder encodeObject:[self backgroundColor]];
	if (![self useOwnerHighlightTextColor])
		[aCoder encodeObject:[self highlightTextColor]];
	if (![self useOwnerHighlightBackgroundColor])
		[aCoder encodeObject:[self highlightBackgroundColor]];
}

@end
