//=============================================================================
//
//		Copyright (C) 1995, 1996 by Paul S. McCarthy and Eric Sunshine.
//				Written by Paul S. McCarthy and Eric Sunshine.
//							All Rights Reserved.
//
//		This notice may not be removed from this source code.
//
//		This object is included in the MiscKit by permission from the authors
//		and its use is governed by the MiscKit license, found in the file
//		"License.rtf" in the MiscKit distribution.	Please refer to that file
//		for a list of all applicable permissions and restrictions.
//		
//=============================================================================
//-----------------------------------------------------------------------------
// MiscTableView.M
//
//		General-purpose 2-D display object that works with the
//		MiscTableScroll to provide row/column sizing and dragging.
//
//		This object is responsible for drawing, mouse and keyboard
//		events in the content portion of the display.
//
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// $Id: MiscTableView.M,v 1.8 96/01/13 23:46:17 zarnuk Exp $
// $Log:		MiscTableView.M,v $
//	Revision 1.8  96/01/13	23:46:17  zarnuk
//	All allocations are now done from [self zone].
//	
//	Revision 1.7  95/10/20	00:13:39  sunshine
//	Was including MiscTableScroll.h with "" instead of <>.
//	
//	Revision 1.6  95/10/19	05:42:00  sunshine
//	Now sends -setSelFont: to FontManager at appropriate times so that font
//	selection in FontManager is up-to-date.
//	
//	Revision 1.5  95/10/08	16:12:30  zarnuk
//	Cell* --> id.
//-----------------------------------------------------------------------------
#import "MiscTableView.h"
#import "MiscHighlightTracker.h"
#import "MiscListTracker.h"
#import "MiscRadioTracker.h"
#import "MiscSparseSet.h"
#import "MiscTableBorder.h"
#import "MiscTableFocus.h"
#import <misckit/MiscTableScroll.h>

#import <new.h>

extern "Objective-C" {
#import <AppKit/NSApplication.h>
#import <AppKit/NSCell.h>
#import <AppKit/NSFontManager.h>
/*#import <AppKit/timer.h>*/
}

extern "C" {
#import <math.h>		// for floor()
}


//----------------------------------------------------------------------------
// startTimer
//----------------------------------------------------------------------------
static inline void startTimer( NXTrackingTimer timer )
	{
	if (timer == 0)
		[NSEvent startPeriodicEventsAfterDelay:0.1 withPeriod:0.1];
	}


//----------------------------------------------------------------------------
// stopTimer
//----------------------------------------------------------------------------
static inline void stopTimer( NXTrackingTimer timer )
	{
	if (timer)
		{
		[NSEvent stopPeriodicEvents];
		timer = 0;
		}
	}



//=============================================================================
// IMPLEMENTATION
//=============================================================================
@implementation MiscTableView

//-----------------------------------------------------------------------------
// - initFrame:
//
//		NOTE *1*: Default behavior is to take keyboard-cursor information from
//				the row-border.
//-----------------------------------------------------------------------------
- initFrame:(NSRect const*)frameRect
		scroll:(MiscTableScroll*)i_scroll
		colInfo:(MiscTableBorder*)i_col_border
		rowInfo:(MiscTableBorder*)i_row_border
	{
	NSZone* const z = [self zone];
	NSRect rect;
	if (frameRect != 0)
		rect.origin = frameRect->origin;
	else
		{
		rect.origin.x = 0;
		rect.origin.y = 0;
		}

	rect.size.width = i_col_border->totalSize();
	rect.size.height = i_row_border->totalSize();

	[super initWithFrame:rect];
#error ViewConversion: '[NSView setFlipped:]' is obsolete; you must override 'isFlipped' instead of setting externally. However, [NSImage setFlipped:] is not obsolete. If that is what you are using here, no change is needed.
	[self setFlipped:YES];
#error ViewConversion: 'setClipping:' is obsolete. Views always clip to their bounds. Use PSinitclip instead.
	[self setClipping:NO];
#error ViewConversion: View's 'setOpaque:' is obsolete; you must override 'isOpaque' instead of setting externally (if sent to text, 'setDrawsBackground:' can be used to eliminate background gray; if sent to NSImageRep, it's OK)
	[self setOpaque:YES];

	scroll = i_scroll;
	col_border = i_col_border;
	row_border = i_row_border;
	tracker_border = MISC_ROW_BORDER;					// NOTE *1*

	oldColSel = new( NSZoneMalloc(z,sizeof(*oldColSel)) ) MiscSparseSet;
	oldRowSel = new( NSZoneMalloc(z,sizeof(*oldRowSel)) ) MiscSparseSet;
	[self setSelectionMode: [scroll selectionMode]];

	focuser = [[MiscTableFocus allocWithZone:z] initOwner: self];
	return self;
	}


//-----------------------------------------------------------------------------
// - free
//-----------------------------------------------------------------------------
- (void)dealloc
	{
	[focuser release];
	[tracker release];
	NSZone* const z = [self zone];
	if (oldColSel != 0)
		{
		oldColSel->MiscSparseSet::~MiscSparseSet();
		NSZoneFree( z, oldColSel );
		}
	if (oldRowSel != 0)
		{
		oldRowSel->MiscSparseSet::~MiscSparseSet();
		NSZoneFree( z, oldRowSel );
		}
	{ [super dealloc]; return; };
	}


//-----------------------------------------------------------------------------
// TYPE VARIATIONS
//-----------------------------------------------------------------------------

- (MiscTableBorder*) borderFor: (MiscBorderType)b
		{ return (b == MISC_ROW_BORDER ? row_border : col_border); }
- (MiscBorderType) otherBorder: (MiscBorderType)b
		{ return (b == MISC_ROW_BORDER ? MISC_COL_BORDER : MISC_ROW_BORDER); }
- (BOOL) isHorz: (MiscBorderType)b { return (b == MISC_COL_BORDER); }

- (float) border:(MiscBorderType)b pointX: (NSPoint const*) point
		{ return [self isHorz:b] ? point->x : point->y; }
- (float) border:(MiscBorderType)b pointY: (NSPoint const*) point
		{ return [self isHorz:b] ? point->y : point->x; }

- (float) border:(MiscBorderType)b rectMinX:(NSRect const*)frameRect
		{ return [self isHorz:b] ? NSMinX(*frameRect) : NSMinY(*frameRect); }
- (float) border:(MiscBorderType)b rectMaxX:(NSRect const*)frameRect
		{ return [self isHorz:b] ? NSMaxX(*frameRect) : NSMaxY(*frameRect); }
- (float) border:(MiscBorderType)b rectMinY:(NSRect const*)frameRect
		{ return [self isHorz:b] ? NSMinY(*frameRect) : NSMinX(*frameRect); }
- (float) border:(MiscBorderType)b rectMaxY:(NSRect const*)frameRect
		{ return [self isHorz:b] ? NSMaxY(*frameRect) : NSMaxX(*frameRect); }
- (float) border:(MiscBorderType)b rectWidth:(NSRect const*)frameRect
		{ return [self isHorz:b] ? NSWidth(*frameRect) : NSHeight(*frameRect); }
- (float) border:(MiscBorderType)b rectHeight:(NSRect const*)frameRect
		{ return [self isHorz:b] ? NSHeight(*frameRect) : NSWidth(*frameRect); }

- (void) border:(MiscBorderType)b setRect:(NSRect*)rect minX:(float)x
		{ if ([self isHorz:b]) rect->origin.x = x; else rect->origin.y = x; }
- (void) border:(MiscBorderType)b setRect:(NSRect*)rect minY:(float)y
		{ if ([self isHorz:b]) rect->origin.y = y; else rect->origin.x = y; }
- (void) border:(MiscBorderType)b setRect:(NSRect*)rect width:(float)width
		{ if ([self isHorz:b]) rect->size.width = width;
				else rect->size.height = width; }
- (void) border:(MiscBorderType)b setRect:(NSRect*)rect height:(float)height
		{ if ([self isHorz:b]) rect->size.height = height;
				else rect->size.width = height; }


//-----------------------------------------------------------------------------
// drewCellAt:border:oldSel:
//
//		Keeps the old selection sets up to date whenever we draw a cell.  This
//		way -reflectSelection has valid data to work from.
//-----------------------------------------------------------------------------
- (void) drewCellAt: (MiscCoord_V)slot border: (MiscTableBorder*)b
	oldSel: (MiscSparseSet*)oldSel
	{
	BOOL isSelected = b->selectionSet().contains( slot );
	if (isSelected != oldSel->contains(slot))
		{
		if (isSelected)
			oldSel->add( slot );
		else
			oldSel->remove( slot );
		}
	}


//-----------------------------------------------------------------------------
// drewCellAt::
//-----------------------------------------------------------------------------
- (void) drewCellAt: (MiscCoord_V)r : (MiscCoord_V)c
	{
	[self drewCellAt:r border:row_border oldSel: oldRowSel];
	[self drewCellAt:c border:col_border oldSel: oldColSel];
	}


//-----------------------------------------------------------------------------
// - getCellFrame:at:: -- Visual coords
//-----------------------------------------------------------------------------
- (NSRect)cellFrameAtRow:(int)row column:(int)col
	{
	
    NSRect r;
    if (row < 0 || row >= row_border->count() ||
		col < 0 || col >= col_border->count())
		r = NSMakeRect(0, 0, 0, 0);
	else
		r = NSMakeRect(col_border->getOffset(col), row_border->getOffset(row), col_border->effectiveSize(col), row_border->effectiveSize(row));
	return r;
	}


//-----------------------------------------------------------------------------
// - getVisFrame:at:from:
//-----------------------------------------------------------------------------
- (void) getVisFrame:(NSRect*)r at:(MiscCoord_V)slot from:(MiscBorderType)bdr
	{
	MiscTableBorder* const b = [self borderFor:bdr];
	*r = [self visibleRect];
	[self border:bdr setRect:r minX:b->getOffset(slot)];
	[self border:bdr setRect:r width:b->effectiveSize(slot)];
	}


//-----------------------------------------------------------------------------
// - firstVisibleSlot:
//-----------------------------------------------------------------------------
- (MiscCoord_V) firstVisibleSlot: (MiscBorderType)bdr
	{
	MiscCoord_V ret = -1;
	MiscTableBorder* const b = [self borderFor:bdr];
	if (b->count() > 0)
		{
		NSRect r; r = [self visibleRect];
		MiscPixels vis_origin = MiscPixels( [self border:bdr rectMinX:&r] );
		ret = b->visualForOffset( vis_origin );
		MiscPixels slot_origin = b->getOffset( ret );
		if (slot_origin < vis_origin)
			ret++;
		}
	return ret;
	}


//-----------------------------------------------------------------------------
// - acceptsFirstMouse
//-----------------------------------------------------------------------------
#warning ViewConversion: 'acceptsFirstMouse:' (used to be 'acceptsFirstMouse') now takes the event as an arg
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - acceptsFirstResponder
//-----------------------------------------------------------------------------
- (BOOL) acceptsFirstResponder
	{
	return YES;
	}


//-----------------------------------------------------------------------------
// - setFocusFrame
//-----------------------------------------------------------------------------
- (void) setFocusFrame
	{
	NSRect r = {{ 0,0 }, { 0,0 }};
	if (row_border->count() > 0 && col_border->count() > 0)
		{
		MiscTableBorder* const b = [self borderFor: tracker_border];
		MiscCoord_V focusSlot = b->getCursor();
		if (focusSlot < 0 || focusSlot >= b->count())
			focusSlot = (b->hasSelection() ? b->selectedSlot() :
						[self firstVisibleSlot: tracker_border]);
		b->setCursor( focusSlot );
		[self getVisFrame:&r at:focusSlot from:tracker_border];
		}
	[focuser setFrame:r];
	}


//-----------------------------------------------------------------------------
// - startFocus
//-----------------------------------------------------------------------------
- (void) startFocus
	{
	[self setFocusFrame];
	[focuser startFocus];
	}


//-----------------------------------------------------------------------------
// - stopFocus
//-----------------------------------------------------------------------------
- (void) stopFocus
	{
	[focuser stopFocus];
	}


//-----------------------------------------------------------------------------
// - becomeFirstResponder
//-----------------------------------------------------------------------------
- (BOOL)becomeFirstResponder
	{
	NSWindow* win = [self window];
	if (win && [win isKeyWindow])
		{
		[self startFocus];
		[[NSFontManager new] setSelFont:[scroll font] isMultiple:NO];
		}
	return YES;
	}


//-----------------------------------------------------------------------------
// - resignFirstResponder
//-----------------------------------------------------------------------------
- (BOOL)resignFirstResponder
	{
	[self stopFocus];
	return YES;
	}


//-----------------------------------------------------------------------------
// - becomeKeyWindow
//-----------------------------------------------------------------------------
- (void)becomeKeyWindow
	{
	[self startFocus];
	[[NSFontManager new] setSelFont:[scroll font] isMultiple:NO];
}


//-----------------------------------------------------------------------------
// - resignKeyWindow
//-----------------------------------------------------------------------------
- (void)resignKeyWindow
	{
	[self stopFocus];
}


//-----------------------------------------------------------------------------
// - windowChanged:
//-----------------------------------------------------------------------------
- (void)viewWillMoveToWindow:(NSWindow *)newWindow
	{
	[self stopFocus];
}


//-----------------------------------------------------------------------------
// - resetCursorRects
//
//		NOTE: This catches srolling.  The blinking cursor is always placed
//				within the visibleRect.	 After scrolling the visibleRect has
//				changed so we need to readjust the cursor frame.
//-----------------------------------------------------------------------------
- (void)resetCursorRects
	{
	[self setFocusFrame];
}


//-----------------------------------------------------------------------------
// - sizeTo::
//-----------------------------------------------------------------------------
- (void)setFrameSize:(NSSize)_newSize
	{
	[super setFrameSize:_newSize];
	[self setFocusFrame];
}


//-----------------------------------------------------------------------------
// - setFrame:
//-----------------------------------------------------------------------------
- (void)setFrame:(NSRect)r
	{
	[super setFrame:r];
	[self setFocusFrame];
}


//-----------------------------------------------------------------------------
// - adjustSize
//-----------------------------------------------------------------------------
- (void) adjustSize
	{
	[self setFrameSize:NSMakeSize(col_border->totalSize(), row_border->totalSize())];
	}


//-----------------------------------------------------------------------------
// - scrollCellToVisible:: -- Physical coords
//-----------------------------------------------------------------------------
- (void)scrollCellToVisibleAtRow:(int)row column:(int)col
	{
	NSRect r;
	r = [self cellFrameAtRow:row_border->physicalToVisual(row) column:col_border->physicalToVisual(col)];
	[self scrollRectToVisible:r];
	}


//-----------------------------------------------------------------------------
// - border:scrollToVisible: -- Physical coord
//-----------------------------------------------------------------------------
- (void) border: (MiscBorderType)bdr scrollToVisible: (MiscCoord_P)slot
	{
	NSRect r;
	MiscTableBorder* const b = [self borderFor:bdr];
	[self getVisFrame:&r at:b->physicalToVisual(slot) from:bdr];
	[self scrollRectToVisible:r];
	}


//-----------------------------------------------------------------------------
// - scrollRowToVisible: -- Physical coord
//-----------------------------------------------------------------------------
- (void) scrollRowToVisible: (int)row
	{
	[self border:MISC_ROW_BORDER scrollToVisible:row];
	}


//-----------------------------------------------------------------------------
// - scrollColToVisible: -- Physical coord
//-----------------------------------------------------------------------------
- (void) scrollColToVisible: (int)col
	{
	[self border:MISC_COL_BORDER scrollToVisible:col];
	}


//-----------------------------------------------------------------------------
// - cMin:cMax:rMin:rMax:forRect:
//-----------------------------------------------------------------------------
- (void) cMin: (MiscCoord_V*)cmin cMax: (MiscCoord_V*)cmax
		 rMin: (MiscCoord_V*)rmin rMax: (MiscCoord_V*)rmax
		 forRect: (NSRect const*) nxrect
	{
	*cmin = col_border->visualForOffset( (MiscPixels) nxrect->origin.x );
	*cmax = col_border->visualForOffset( 
				(MiscPixels) (nxrect->origin.x + nxrect->size.width) - 1 );
	*rmin = row_border->visualForOffset( (MiscPixels) nxrect->origin.y );
	*rmax = row_border->visualForOffset(
				(MiscPixels) (nxrect->origin.y + nxrect->size.height) - 1 );
	}


//-----------------------------------------------------------------------------
// - drawRect:
//-----------------------------------------------------------------------------
- (void)drawRect:(NSRect const*)nxrect
	{
	MiscCoord_V c, cmin, cmax;
	MiscCoord_V r, rmin, rmax;
	[self cMin:&cmin cMax:&cmax rMin:&rmin rMax:&rmax forRect:nxrect];

	if (cmin >= 0 && cmax >= 0 && rmin >= 0 && rmax >= 0)
		{
		float const x0 = floor( (float) col_border->getOffset( cmin ) );
		float const y0 = floor( (float) row_border->getOffset( rmin ) );
		NSRect rect = { {x0,y0}, {0,0} };
	
		for (r = rmin;	r <= rmax;	r++)
			{
			rect.size.height = floor( (float) row_border->effectiveSize(r) );
			for (c = cmin;	c <= cmax;	c++)
				{
				rect.size.width = floor( (float)col_border->effectiveSize(c) );
				id cell = [scroll cellAtRow:row_border->visualToPhysical(r) column:col_border->visualToPhysical(c)];
				[cell drawWithFrame:rect inView:self];
				[self drewCellAt:r:c];
				rect.origin.x = floor( rect.origin.x + rect.size.width );
				}
			rect.origin.x = x0;
			rect.origin.y = floor( rect.origin.y + rect.size.height );
			}
		}
	}


//-----------------------------------------------------------------------------
// - drawSelf::
//-----------------------------------------------------------------------------
#warning RectConversion: drawRect:(NSRect)rects (used to be drawSelf:(NXRect const*)rects :(int)nrects) no longer takes an array of rects
- (void)drawRect:(NSRect)rects
	{
	if (nrects == 1)
		[self drawRect:&rects];
	else if (nrects == 3)
		{
		[self drawRect:++&rects];
		[self drawRect:++&rects];
		}
}


//-----------------------------------------------------------------------------
// - drawCellAt::updateSel: -- Physical coords
//-----------------------------------------------------------------------------
- (void) drawCellAt: (int)row : (int)col updateSel: (BOOL)updateSel
	{
	if ([self canDraw] &&
		row >= 0 && row < row_border->count() &&
		col >= 0 && col < col_border->count())
		{
		NSRect v;		v = [self visibleRect];
		MiscCoord_V const vRow = row_border->physicalToVisual(row);
		MiscCoord_V const vCol = col_border->physicalToVisual(col);

		NSRect const r = {
		{ col_border->getOffset(vCol), row_border->getOffset(vRow) },
		{ col_border->effectiveSize(vCol), row_border->effectiveSize(vRow) }};

		if (!NSIsEmptyRect(NSIntersectionRect(r , v)))
			{
			id const cell = [scroll cellAtRow:row column:col];
#error ViewConversion: '-focusView' in NSApplication has been replaced by '+focusView' in NSView
			BOOL const didLock = !([NSView focusView] == self);
			if (didLock) [self lockFocus];
			[cell drawWithFrame:r inView:self];
			if (updateSel)
				[self drewCellAt: vRow : vCol];
			if (didLock) [self unlockFocus];
			}
		}
	}


//-----------------------------------------------------------------------------
// - drawCellAt:: -- Physical coords
//-----------------------------------------------------------------------------
- (void)drawCellAtRow:(int)row column:(int)col
	{
	[self drawCellAt:row:col updateSel:YES];
}


//-----------------------------------------------------------------------------
// - drawRow: -- Physical coord
//-----------------------------------------------------------------------------
- drawRow: (int)row
	{
	if ([self canDraw] && row >= 0 && row < row_border->count())
		{
		MiscCoord_V const vRow = row_border->physicalToVisual(row);
	
		NSRect r;
		r = [self visibleRect];
		r.origin.y = row_border->getOffset(vRow);
		r.size.height = row_border->effectiveSize(vRow);

#error ViewConversion: '-focusView' in NSApplication has been replaced by '+focusView' in NSView
		BOOL const didLock = !([NSView focusView] == self);
		if (didLock) [self lockFocus];
		[self drawRect:&r];
		if (didLock) [self unlockFocus];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - drawCol: -- Physical coord
//-----------------------------------------------------------------------------
- drawCol: (int)col
	{
	if ([self canDraw] && col >= 0 && col < col_border->count())
		{
		MiscCoord_V const vCol = col_border->physicalToVisual(col);
	
		NSRect r;
		r = [self visibleRect];
		r.origin.x = col_border->getOffset(vCol);
		r.size.width = col_border->effectiveSize(vCol);

#error ViewConversion: '-focusView' in NSApplication has been replaced by '+focusView' in NSView
		BOOL const didLock = !([NSView focusView] == self);
		if (didLock) [self lockFocus];
		[self drawRect:&r];
		if (didLock) [self unlockFocus];
		}
	return self;
	}


//-----------------------------------------------------------------------------
// - setSelectionMode:
//-----------------------------------------------------------------------------
- (void) setSelectionMode: (MiscSelectionMode) mode
	{
	NSZone* const z = [self zone];
	if (tracker != 0)
		[tracker release];
	switch (mode)
		{
		case MISC_LIST_MODE:
			tracker = [MiscListTracker allocWithZone:z];
			break;
		case MISC_RADIO_MODE:
			tracker = [MiscRadioTracker allocWithZone:z];
			break;
		case MISC_HIGHLIGHT_MODE:
			tracker = [MiscHighlightTracker allocWithZone:z];
			break;
		}
	[tracker initBorder: [self borderFor:tracker_border]];
	}


//-----------------------------------------------------------------------------
// - reflectSelection
//-----------------------------------------------------------------------------
- (void) reflectSelection
	{
	if ([self canDraw])
		{
		MiscSparseSet const& newColSel = col_border->selectionSet();
		MiscSparseSet const& newRowSel = row_border->selectionSet();

		MiscCoord_V cmin, cmax;
		MiscCoord_V rmin, rmax;
		NSRect vis;		vis = [self visibleRect];
		[self cMin:&cmin cMax:&cmax rMin:&rmin rMax:&rmax forRect:&vis];
	
		[[self window] disableFlushWindow];
		BOOL locked = NO;
		for (MiscCoord_V r = rmin; r <= rmax; r++)
			{
			BOOL rowWasOn = oldRowSel->contains(r);
			BOOL rowIsOn = newRowSel.contains(r);
			for (MiscCoord_V c = cmin; c <= cmax; c++)
				{
				BOOL wasOn = rowWasOn || oldColSel->contains(c);
				BOOL isOn = rowIsOn || newColSel.contains(c);
				if (isOn != wasOn)
					{
					if (!locked)
						{
						locked = YES;
						[self lockFocus];
						}
					MiscCoord_P const pCol = col_border->visualToPhysical(c);
					MiscCoord_P const pRow = row_border->visualToPhysical(r);
					[self drawCellAt: pRow : pCol updateSel: NO];
					}
				}
			}
		if (locked)
			[self unlockFocus];
#warning VoidConversion:  creating new variable to deal with methods returning void or bool (previously self)
		{
    id newVar = [self window];
    [newVar enableFlushWindow];
    [newVar flushWindow];
}

		*oldColSel = newColSel;
		*oldRowSel = newRowSel;
		}
	}


//-----------------------------------------------------------------------------
// - trackBy:
//-----------------------------------------------------------------------------
- (void) trackBy: (MiscBorderType)b
	{
	tracker_border = b;
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// - trackingBy
//-----------------------------------------------------------------------------
- (MiscBorderType) trackingBy
	{
	return tracker_border;
	}


//-----------------------------------------------------------------------------
// - reflectCursor
//-----------------------------------------------------------------------------
- (void) reflectCursor
	{
	[self setFocusFrame];
	}


//-----------------------------------------------------------------------------
// constrainSlot:inBorder:
//-----------------------------------------------------------------------------
- (MiscCoord_V) constrainSlot:(MiscCoord_V)s inBorder: (MiscTableBorder*)b
	{
	if (s < 0)
		s = 0;
	else if (s >= b->count())
		s = b->count() - 1;
	return s;
	}


//-----------------------------------------------------------------------------
// - mouseDown:
//-----------------------------------------------------------------------------
- (void)mouseDown:(NSEvent *)p 
	{
	BOOL doubleClicked = ([p clickCount] > 1);
	NSPoint evpt = [p locationInWindow];
	evpt = [self convertPoint:evpt fromView:0];
	MiscTableBorder* const b = [self borderFor:tracker_border];
	MiscCoord_V slot = b->visualForOffset(
				MiscPixels([self border:tracker_border pointX:&evpt]) );
	int const WANTED = (NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSPeriodicMask);

	[focuser stopFocus];
	[tracker mouseDown: p atPos: slot];
	[self borderFor: [self otherBorder:tracker_border]]->selectionSet().empty();
	[scroll reflectSelection];

#error EventConversion: addToEventMask:NX_MOUSEDRAGGEDMASK: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	int old_mask = [[self window] addToEventMask:NSLeftMouseDraggedMask];
#error EventConversion: NXTrackingTimer is obsolete; use NSEvent 'startPeriodicEventsAfterDelay:withPeriod:' and 'stopPeriodicEvents'.  You can probably delete this variable and its uses.
	NXTrackingTimer * timer = 0;
	startTimer( timer );
	NSEvent *lastEvent = p;

	for (; ;)
		{
		p = [[self window] nextEventMatchingMask:WANTED];
		if (p == 0 || [p type] == NSLeftMouseUp)
			break;
		else if ([p type] == NSPeriodic)
			[self autoscroll:lastEvent];
		else
			lastEvent = p;

		NSPoint new_loc = [lastEvent locationInWindow];
		new_loc = [self convertPoint:new_loc fromView:0];
		MiscPixels const offset =
				MiscPixels( [self border:tracker_border pointX:&new_loc] );
		MiscCoord_V const new_slot =
				(offset < [self border:tracker_border rectMaxX:&[self bounds]] ?
				b->visualForOffset(offset) : b->count());
		if (new_slot != slot)
			{
			slot = new_slot;
			[tracker mouseDragged: p atPos: slot];
			[scroll reflectSelection];
			}
		}

	stopTimer( timer );
#error EventConversion: setEventMask:old_mask: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	[[self window] setEventMask:old_mask];

	[tracker mouseUp: p atPos: slot];
	[scroll reflectSelection];
	[scroll border: tracker_border setCursor:
				b->visualToPhysical([self constrainSlot:slot inBorder:b])];
	[scroll selectText:self];
	[focuser startFocus];

	if ([scroll isEnabled])
		{
		if (doubleClicked)
			[scroll sendDoubleAction];
		else
			[scroll sendAction];
		}
}


//-----------------------------------------------------------------------------
// - moveFocusBy:
//-----------------------------------------------------------------------------
- (void) moveFocusBy: (int) delta
	{
	MiscTableBorder* const b = [self borderFor: tracker_border];
	int const lim = b->count();
	if ([focuser isShowingFocus] && lim > 0)
		{
		MiscCoord_V focusSlot = b->getCursor() + delta;
		if (focusSlot < 0)
			focusSlot = lim - 1;
		else if (focusSlot >= lim)
			focusSlot = 0;
		b->setCursor( focusSlot );
		[self border:tracker_border scrollToVisible:
						b->visualToPhysical(focusSlot)];
		[self setFocusFrame];
		}
	}


//-----------------------------------------------------------------------------
// - gotoNextText
//-----------------------------------------------------------------------------
- (void) gotoNextText
	{
	id targ = [scroll nextText];
	if (targ && [targ respondsToSelector:@selector(selectText:)])
		[targ selectText:scroll];
	}


//-----------------------------------------------------------------------------
// - gotoPreviousText
//-----------------------------------------------------------------------------
- (void) gotoPreviousText
	{
	id targ = [scroll previousText];
	if (targ && [targ respondsToSelector:@selector(selectText:)])
		[targ selectText:scroll];
	}


//-----------------------------------------------------------------------------
// - keyboardSelect:
//-----------------------------------------------------------------------------
- (void) keyboardSelect:(NSEvent *)p 
	{
	[tracker keyDown: p atPos: [self borderFor:tracker_border]->getCursor()];
	[scroll borderClearSelection:[self otherBorder: tracker_border]];
	[scroll reflectSelection];
	if ([scroll isEnabled])
		[scroll sendAction];
	}


//-----------------------------------------------------------------------------
// - keyboardPerform
//-----------------------------------------------------------------------------
- (void) keyboardPerform
	{
	if ([scroll isEnabled])
		[scroll sendDoubleAction];
	}


//-----------------------------------------------------------------------------
// - keyDown:
//-----------------------------------------------------------------------------
- (void)keyDown:(NSEvent *)p 
	{
	enum
		{
		K_TAB	= '\t',
		K_BTAB	= 0x19,
		K_RETURN= '\r',
		K_SPACE = ' ' ,
		K_LEFT	= 0xac,
		K_UP	= 0xad,
		K_RIGHT = 0xae,
		K_DOWN	= 0xaf,
		};

	if (([p modifierFlags] & NSCommandKeyMask) == 0)
		{
#error EventConversion: the 'characters' method of NSEvent replaces the '.data.key.charCode' field of NXEvent. Use 'charactersIgnoringModifiers' to get the chars that would have been generated regardless of modifier keys (except shift)
		switch ([p characters])
			{
			case K_UP			: [self moveFocusBy: -1			]; break;
			case K_DOWN			: [self moveFocusBy:  1			]; break;
			case K_LEFT			: [self moveFocusBy: -1			]; break;
			case K_RIGHT		: [self moveFocusBy:  1			]; break;
			case K_TAB			: [self gotoNextText			]; break;
			case K_BTAB			: [self gotoPreviousText		]; break;
			case K_SPACE		: [self keyboardSelect: p		]; break;
			case K_RETURN		: [self keyboardPerform			]; break;
			}
		}
}

@end
