/*

File MiscAnnouncer.m

Copyright (C) 1994 by H. Scott Roy

		H. Scott Roy
		2573 Stowe Ct.
		Northbrook, IL  60062-8103
		iconkit@cs.stanford.edu

For your editing convenience, this file is best viewed using an editor that automatically wraps long lines, in a fixed point font at 80 columns, with tabs every 4 spaces.

*/


/* ========================================================================== */


/*

The MiscAnnouncer class is a public version of the MiscAnnouncer class found in IconKit.  It provides a simple way to conform to the MiscDependency protocol.  A MiscAnnouncer lets one send announcements to notify users and listeners of changes to an object.

There are two ways to incorporate a MiscAnnouncer in a new class: one can either inherit from it directly, or else keep a MiscAnnouncer as an instance variable and forward all the MiscDependency messages.

*/

#import <AppKit/AppKit.h>
#import <misckit/MiscAnnouncer.h>

@implementation MiscAnnouncer


/* ========================================================================== */


/*

The add and remove messages can be set to print debugging information.  Such information has proved to be of great value in tracking down bugs in the dependency structure among objects.  Debugging information is printed to stderr if the announceDebug variable is YES.  The simplest way to turn this variable on is to specify the flag -MiscAnnounceDebug on the command line.

*/


static BOOL
	announceDebug = NO;


static void
dprintf (const char * format, ...)
{
#error DefaultsConversion: the NXDefaultsVector type is obsolete. Construct a dictionary of default registrations and use the NSUserDefaults 'registerDefaults:' method
	static NXDefaultsVector defaults = {{ "MiscAnnounceDebug", NULL }, { NULL }};
	static BOOL
		firstTime = YES;
	
	if (firstTime)
	{
		const char
			* app = [[[NSProcessInfo processInfo] processName] cString];
		
#error DefaultsConversion: NXRegisterDefaults() is obsolete. Construct a dictionary of default registrations and use the NSUserDefaults 'registerDefaults:' method
    	NXRegisterDefaults(app, defaults);
		announceDebug = ([[[NSUserDefaults standardUserDefaults] objectForKey:@"MiscAnnounceDebug"] cString] != NULL);
		firstTime = NO;
	}
	
	if (announceDebug)
	{
		va_list
			args;
		
		va_start (args, format);
		vfprintf (stderr, format, args);
		va_end (args);
	}
}


/* ========================================================================== */


/*

Here is the initialization method.  The owner of a MiscAnnouncer will be given as the first argument in any announce messages.

*/


- initOwner: theOwner
{
	if ((self = [super  init]) != nil)
	{
		owner = theOwner;
		usersAndListeners = [[List  alloc]  init];
		numUsers = 0;
		sendAnnouncements = YES;
	}
	
	return self;
}


- (void)dealloc
{
	[usersAndListeners release];
	return [super dealloc]; self = self;
}


- usersAndListeners
{
	return usersAndListeners;
}


- (int) numUsers
{
	return numUsers;
}


- (BOOL) sendAnnouncements
{
	return sendAnnouncements;
}


- setSendAnnouncements: (BOOL) flag;
{
	sendAnnouncements = flag;
	return self;
}


/* ========================================================================== */


/*

Here are the methods to add and remove users and listeners.  They all return the MiscAnnouncer owner.  It's important to permit objects to add themselves multiple times, so that all the add and remove calls correctly balance.

*/


- addUser: who
{
	dprintf ("adding user:         %-17s (%p)  for: %s (%p)\n", [[who name] cString],
		who, [[owner name] cString], owner);
	
	[usersAndListeners  addObject: who];
	numUsers++;
	return owner;
}


- addListener: who
{
	dprintf ("adding listener:     %-17s (%p)  for: %s (%p)\n", [[who name] cString],
		who, [[owner name] cString], owner);
	
	[usersAndListeners  addObject: who];
	return owner;
}


- removeUser: who
{
	dprintf ("removing user:       %-17s (%p)  for: %s (%p)\n", [[who name] cString],
		who, [[owner name] cString], owner);
	
	[usersAndListeners  removeObject: who];
	numUsers--;
	return owner;
}


- removeListener: who
{
	dprintf ("removing listener:   %-17s (%p)  for: %s (%p)\n", [[who name] cString],
		who, [[owner name] cString], owner);
	
	[usersAndListeners  removeObject: who];
	return owner;
}


/* ========================================================================== */


/*

Here are the methods to send out notification messages to all users and listeners.  The methods try to be safe in case objects try to add or delete themselves from the user list during an announcement, but keep in mind that it's probably a bad idea to muck too much with an object during an announcement.

There's a fair amount of overhead to these announcement messages.  A more performance minded implementation is certainly possible.

Users and listeners only receive the messages to which they can actually respond.  Notification messages should all take forms like,

		didChange: sender
		didChange: sender  info: arg

*/

- announce: (SEL) message
{
	id
		recipients,
		who;
	int
		i = [usersAndListeners  count];
	
	if (i > 0  &&  sendAnnouncements)
	{
		recipients = [usersAndListeners  copy];
		while (i--) if ([who = [recipients  objectAt: i] respondsToSelector:message])
				[who perform:message withObject:owner];
		
		[recipients release];
	}
	
	return owner;
}


- announce: (SEL) message  with: arg
{
	id
		recipients,
		who;
	int
		i = [usersAndListeners  count];
	
	if (i > 0  &&  sendAnnouncements)
	{
		recipients = [usersAndListeners  copy];
		while (i--) if ([who = [recipients  objectAt: i] respondsToSelector:message])
				[who perform:message withObject:owner withObject:arg];
		
		[recipients release];
	}
	
	return owner;
}


/* ========================================================================== */


/*

Here are the methods to poll the users of an object for approval before performing an operation.  The result is a logical AND of all the responses.  Both methods return as soon as any user responds NO.  Some users might therefore never see a polling message.

*/

- (BOOL) poll: (SEL) message
{
	id
		recipients,
		who;
	int
		i = [usersAndListeners  count];
	
	if (i > 0)
	{
		recipients = [usersAndListeners  copy];
		while (i--) if (
			[who = [recipients  objectAt: i] respondsToSelector:message]	&&
			[who perform:message withObject:owner] == NO)
		
					 		break;
		
		[recipients release];
		return (i < 0) ? YES : NO;
	}
	
	else return YES;
}


- (BOOL) poll: (SEL) message  with: arg
{
	id
		recipients,
		who;
	int
		i = [usersAndListeners  count];
	
	if (i > 0)
	{
		recipients = [usersAndListeners  copy];
		while (i--) if (
			[who = [recipients  objectAt: i] respondsToSelector:message]	&&
			[who perform:message withObject:owner withObject:arg] == NO)
		
					 		break;
		
		[recipients release];
		return (i < 0) ? YES : NO;
	}
	
	else return YES;
}


@end
