//
//	MiscAppIconAnimator.m -- This is an Object to animate the icon
//				for an application.  This Object will allow
//				similiar activity to NeXTMail when new mail arrives.
//		Written by Craig Laurent Copyright 1995 by Craig Laurent.
//				Version 1.0.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	
//#import "MiscAppIconAnimator.h"
#import <misckit/MiscAppIconAnimator.h> 

@implementation MiscAppIconAnimator

/*  This is the callback function for the DPS Timed Entry. 
It in turn calls our animateIcon method.  DPS Timed Entries require functions not methods, as callbacks. */
void MiscAppIconAnimate (NSTimer *timedEntry, double  now, void *who)
{
	[(id)who animateIcon];
}


- init
{
	return [self initWithImageArray:[[List alloc] init] animationPattern:[[List alloc] init] andPatternLoops:0];
}

//** designated initializer
- initWithImageArray:(List*)anArray animationPattern:(List*)patternArray andPatternLoops:(int)loops
{
	if (self = [super init]) {
		//** set default value for instance variables
		appTile = nil;
		appIconContentView = nil;
		imageArray = nil;
		animationPattern = nil;
		patternLoops = 0;
		timeInterval = 0.2;
		indexOfStopIcon = 0;
		indexOfEndIcon = 1;
		timedEntry = 0;
		patternIterator = 0;
		loopCount = 0;
		patternEnd = 0;
		offset = 0;

		//** set value for instance variables
		[self setImageArray:anArray];
		[self setAnimationPattern:patternArray];
		[self setPatternLoops:loops];
		
		return self;
	}
	return nil;
}


- (void)dealloc
{
    [imageArray release];
    [animationPattern release];
	{ [super dealloc]; return; };
}


//** instance methods
- (List*)imageArray
{ return imageArray; }
- (void)setImageArray:(List*)anArray
{
	[imageArray release];
	imageArray = anArray;
}

- (List*)animationPattern
{ return animationPattern; }
- (void)setAnimationPattern:(List*)anArray
{
	[animationPattern release];
	animationPattern = anArray;

	patternEnd = [animationPattern count];
}

- (int)patternLoops
{	return patternLoops; }
- (void)setPatternLoops:(int)numLoops
{	patternLoops = numLoops; }

- (float)timeInterval
{	return timeInterval; }
- (void)setTimeInterval:(float)interval
{	timeInterval = interval; }

- (int)indexOfEndIcon
{	return indexOfEndIcon; }
- (void)setIndexOfEndIcon:(int)endIcon
{	indexOfEndIcon = endIcon; }

- (int)indexOfStopIcon
{	return indexOfStopIcon; }
- (void)setIndexOfStopIcon:(int)stopIcon
{	indexOfEndIcon = stopIcon; }


//** methods to change Animation
/* startAnimation - starts running the animation sequence */
- (void)startAnimation:sender
{
//  if none exists, add a timed entry to begin the animation sequence...
	
	if (!timedEntry) {
#error DPSConversion: timedEntry should be converted from a DPSTimedEntry to an NSTimer object
	    
#error DPSConversion: 'timerWithTimeInterval:target:selector:userInfo:repeats:' used to be DPSAddTimedEntry([self timeInterval], MiscAppIconAnimate, (void*)self, NSModalResponseThreshold).  MiscAppIconAnimate should be converted to a method with one argument (the timer), and <target> should be converted to the implementor of that method ([aTarget aSelector:timer]).  The argument to userInfo: needs to be converted to an object and can be retrieved with [timer userInfo].
	{
	timedEntry = [[NSTimer timerWithTimeInterval:[self timeInterval] target:<target> selector:MiscAppIconAnimate userInfo:(void*)self repeats:YES] retain];
	[[NSRunLoop currentRunLoop] addTimer:timedEntry forMode:NSDefaultRunLoopMode];
	[[NSRunLoop currentRunLoop] addTimer:timedEntry forMode:NSModalPanelRunLoopMode];
	[[NSRunLoop currentRunLoop] addTimer:timedEntry forMode:NSEventTrackingRunLoopMode];
	}
	}
}

/* stopAnimation - stops the animation sequence and sets the application icon to a standard image */
- (void)stopAnimation
{
//  Indicate Animation stopped by setting image to image one.  Do necessary housekeeping.
	
	[self removeTimedEntry];
	if ([[self imageArray] count]) {
		[self displayImage: (NSImage*)[[self imageArray] objectAt: [self indexOfStopIcon]]];
	}
	patternIterator = 0;
	loopCount = 0;
}


//** Internal methods
/* removeTimedEntry - removes the timedEntry if it is running */
- (void)removeTimedEntry
{
	if (timedEntry) {
		[timedEntry invalidate]; [timedEntry release];;
		timedEntry = 0;
	}
}

/* cornerPointForImage: -  calculates and returns the point that should be used as the corner for this image.  This is used to make sure the image is centered on the application icon. */
- (NSPoint)cornerPointForImage:(NSImage*)anImage
{
//*  You must center the composited image within
//*   the contentView of the appIcon window.
	NSSize	appTileSize;
	NSSize	imageSize;
	NSPoint	cornerPoint = { 0.0, 0.0 };

	appTileSize = [appTile size];
	imageSize = [anImage size];
	
	if ( imageSize.width < appTileSize.width ) 
		cornerPoint.x = (appTileSize.width - imageSize.width ) / 2.0;
	
	if ( imageSize.height < appTileSize.height ) 
		cornerPoint.y = (appTileSize.height - imageSize.height ) / 2.0;
		
	return cornerPoint;
}

/* displayImage: - displays the given image into the application icon. */
- (void)displayImage:(NSImage*)anImage
{
//*  NXAppTile is composited first at 0,0 of the icon window's content view 
//*  (this is required in order to maintain the NeXT icon look).
//*  'anImage' is then composited at center (centering is also a requirement).

	NSPoint	contentViewOrigin = { 0.0, 0.0 };
	NSPoint	cornerPoint;

	//** get application image every time
#error IconConversion: The image - NXAppTile - is no longer available in the API. Try using [NSApp setApplicationIconImage:].
	appTile = [NSImage imageNamed:@"NXAppTile"];
	appIconContentView = [[NSApp applicationIconImage] contentView];
	cornerPoint = [self cornerPointForImage:anImage];

	[appIconContentView lockFocus];
	[appTile compositeToPoint:contentViewOrigin operation:NSCompositeSourceOver];
	[anImage compositeToPoint:cornerPoint operation:NSCompositeSourceOver];
	[appIconContentView unlockFocus];
	[appIconContentView display];
}


/* displayImageAtIndex: - displays the image from the array, at the specified index into the application icon. */
- (void)displayImageAtIndex:(int)imageNum;
{
	if (imageNum < [[self imageArray] count])
		[self displayImage:[[self imageArray] objectAt:imageNum]];
}

/* animateIcon - called every time a timedEntry executes.  This contains logic for the icon animation (image sequencing).  The images contained in imageArray are used.  The animationPattern provides an index into that list, giving us the right image at the right time. */
- (void)animateIcon
{
	[self displayImage: (NSImage*)[[self imageArray] objectAt:(int)[[self animationPattern] objectAt:patternIterator++]]];

	if (patternIterator == (patternEnd - offset)) {
		[self removeTimedEntry];
		patternIterator = 0;
		loopCount++;
		if (loopCount < [self patternLoops]) {
			offset = 1;
//** the following only works for Object, not NSObject
			[NSObject cancelPreviousPerformRequestsWithTarget:self selector:@selector(startAnimation:) object:nil], [self performSelector:@selector(startAnimation:) object:nil afterDelay:(500) / 1000.0];
//			[self perform: @selector(startAnimation:) withObject:nil];
		} else {
			loopCount = offset = 0;
			[self displayImage:(NSImage*)[imageArray objectAt: [self indexOfEndIcon]]];
		}			
	}
}



//*********************
//** archiving methods
- (void)encodeWithCoder:(NSCoder *)aCoder
{
	int	count;
	int	entries;
	int temp;
	id	tempObj;

	[super encodeWithCoder:aCoder];

	// write animationPattern
	entries = [animationPattern count];
	[aCoder encodeValueOfObjCType:"i" at:&entries];
	for (count=0; count<entries; count++) {
		temp = (int)[animationPattern objectAt:count];
		[aCoder encodeValueOfObjCType:"i" at:&temp];
	}

	// write imageArray
	entries = [imageArray count];
	[aCoder encodeValueOfObjCType:"i" at:&entries];
	for (count=0; count<entries; count++) {
		tempObj = [imageArray objectAt:count];
		[aCoder encodeValueOfObjCType:"@" at:&tempObj];
	}

	[aCoder encodeValueOfObjCType:"i" at:&patternLoops];
	[aCoder encodeValueOfObjCType:"f" at:&timeInterval];
	[aCoder encodeValueOfObjCType:"i" at:&indexOfEndIcon];
	[aCoder encodeValueOfObjCType:"i" at:&indexOfStopIcon];
}

- (id)initWithCoder:(NSCoder *)aDecoder
{
	int	count;
	int	entries;
	int temp;
	id	tempObj;

	[super initWithCoder:aDecoder];

	// read animationPattern
	animationPattern = [[List alloc] init];
	[aDecoder decodeValueOfObjCType:"i" at:&entries];
	for (count=0; count<entries; count++) {
		[aDecoder decodeValueOfObjCType:"i" at:&temp];
		[animationPattern addObject:(id)temp];
	}

	// read imageArray
	imageArray = [[List alloc] init];
	[aDecoder decodeValueOfObjCType:"i" at:&entries];
	for (count=0; count<entries; count++) {
		[aDecoder decodeValueOfObjCType:"@" at:&tempObj];
		[imageArray addObject:tempObj];
	}

	[aDecoder decodeValueOfObjCType:"i" at:&patternLoops];
	[aDecoder decodeValueOfObjCType:"f" at:&timeInterval];
	[aDecoder decodeValueOfObjCType:"i" at:&indexOfEndIcon];
	[aDecoder decodeValueOfObjCType:"i" at:&indexOfStopIcon];
	return self;
}


//****************************************
//** Distributed Object encoding methods
/* encodeRemotelyFor:.. -  Distributed Object method to indicate what should be passed, The object or a copy.  NSObject won't be passed as a copy. */
#warning DO Conversion: if freeAfterEncoding was true use autorelease instead.
- replacementObjectForPortCoder:(NSPortCoder *)aPortCoder
{
	if ([aPortCoder isBycopy])
		return self;	//** return copy of object, not a proxy
//	return nil;	//** NSObject doesn't encode for DO.
#warning DO Conversion: if freeAfterEncoding was true use autorelease instead.
	return [super replacementObjectForPortCoder:[aPortCoder connection]];
}

/* encodeUsing: -  Distributed Object method to encode the data for transfer */
- (void)encodeWithCoder:(NSPortCoder *)aPortCoder
{
	[aPortCoder encodeValueOfObjCType:"@" at:&animationPattern];
	[aPortCoder encodeValueOfObjCType:"@" at:&imageArray];
	[aPortCoder encodeValueOfObjCType:"i" at:&patternLoops];
	[aPortCoder encodeValueOfObjCType:"f" at:&timeInterval];
	[aPortCoder encodeValueOfObjCType:"i" at:&indexOfEndIcon];
	[aPortCoder encodeValueOfObjCType:"i" at:&indexOfStopIcon];
}

/* decodeUsing: -  Distributed Object method to decode the data after transfer */
- (id)initWithCoder:(NSPortCoder *)aPortCoder
{
	self = [self init];

	//** decode data
	[aPortCoder decodeValueOfObjCType:"@" at:&animationPattern];
	patternEnd = [animationPattern count];	// done in set method
	[aPortCoder decodeValueOfObjCType:"@" at:&imageArray];
	[aPortCoder decodeValueOfObjCType:"i" at:&patternLoops];
	[aPortCoder decodeValueOfObjCType:"f" at:&timeInterval];
	[aPortCoder decodeValueOfObjCType:"i" at:&indexOfEndIcon];
	[aPortCoder decodeValueOfObjCType:"i" at:&indexOfStopIcon];

	return self;
}


@end
