/*============================= MiscBackWindow.m ===========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Class that impliments a full screen window with no title bars or other
   ornamentation visible. It can be set to a user selectable tier and will
   notify a delegate when it changes it's tier. It's content view is a
   MiscFramingView.

HISTORY
12-Jul-93  Dale Amon at GPL
	   Added basic archive methods

7-Feb-93   Dale Amon at GPL
	   Created. Based loosely on Backspace demo code.

NOTES:		There is a kludge to handle what appears to be a problem
		with the setting of the Window delegate.

		The awake method may be incomplete. I have not tested
		archiving at all.

		Assumes all screens are the same size as the main screen,
		which I don't like, so we'll just have to see if it still
		works if and when we get a bunch of screens of different sizes.
		ie, what happens when we go into the PC world?

*/

#import <misckit/misckit.h>
#import "MiscBWPS.h"

#define BACKGROUND_COLOR	NSBlack

@implementation MiscBackWindow

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/

#define ILLEGAL_METHOD [NSLocalizedStringFromTable(@"  %s objects should not be sent '%s' messages\n", @"MiscKit", "Programmer error, not allowed to use the named method") cString]

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/
/* Initialize the class */

+ (void)initialize {[MiscBackWindow setVersion:MISC_BACKWIN_VERSION_ID]; return;}


/*===========================================================================*/

+ (NSRect)frameRectForContentRect:(NSRect)cRect styleMask:(int)aStyle
 {
	
    NSRect fRect;
    fRect.origin.x=fRect.origin.y=0;
	(fRect.size) = [[NSScreen mainScreen] frame].size;
	return fRect;
 }

/*---------------------------------------------------------------------------*/

+ (NSRect)contentRectForFrameRect:(NSRect)fRect styleMask:(int)aStyle
 {
	
    NSRect cRect;
    cRect.origin.x=cRect.origin.y=0;
	(cRect.size) = [[NSScreen mainScreen] frame].size;
	return cRect;
 }

/*---------------------------------------------------------------------------*/

+ (float)minFrameWidthWithTitle:(NSString *)aTitle styleMask:(unsigned int)aStyle
 {
	NSSize s;
	<style>;
	return <style>;
 }


/*===========================================================================*/
/* Init, alloc and free methods */
/*===========================================================================*/
/* Block the designated initializer of our parent class */

- initWithContentRect:(NSRect)contentRect styleMask:(unsigned int)aStyle backing:(NSBackingStoreType)backingType defer:(BOOL)flag
#warning StringConversion: NSStringFromSelector return value is an NSString; use [ cString] if needed, otherwise change the target type to an NSString.
 {	[self error: ILLEGAL_METHOD, [[[self class] name] cString], NSStringFromSelector(_cmd)];
	return self;
 }


/*---------------------------------------------------------------------------*/
/* Initialize a window that is the size of the screen, has no borders and is
   in the background tier.
*/

- init
  {	NSRect	screenRect;
	id	content;

	screenRect.origin.x = screenRect.origin.y = 0;
	(screenRect.size) = [[NSScreen mainScreen] frame].size;

	/* I selected buffered simply because that was what I saw
	   in BackSpace Controller.m */
	[super initWithContentRect:screenRect styleMask:NX_TOKENSTYLE backing:NSBackingStoreBuffered defer:NO];

	[self setBackgroundColor:[NSColor colorWithCalibratedWhite:BACKGROUND_COLOR alpha:1.0]];

	/* Replace current content with a MiscFramingView and free the old one */
	content    = [[MiscFramingView allocWithZone:[self zone]] init];
	[content setOutline:MISC_FV_PLAIN];
	[content setBackgroundColor:[NSColor colorWithCalibratedWhite:BACKGROUND_COLOR alpha:1.0]];
	[[self setContentView:content] release];

	currentTier = MISC_BACKGROUNDTIER;
	_MiscTweakWindow([self windowNumber], MISC_BACKGROUNDTIER);

	[self display];
	return self;
  }

/*===========================================================================*/
/* KLUDGE ALERT!
   This is a hack until I figure out what is going on. After either
   [self setDelegate: anObject]] or a [super setDelegate: anObject]
   delegate is still nil, ie 0. So I have overridden the Window method
   here to force the issue. I haven't a CLUE why it doesn't work the way
   it is supposed to...
*/

- (void)setDelegate:(id)anObject {delegate = anObject;
}


/*===========================================================================*/
/* Set the window tier, if it has changed, and notify the Window delegate */

- setWindowTier: (int) tier
 {
	if (tier != currentTier) {
		currentTier = tier;
		_MiscSetWindowLevel([self windowNumber], tier);
#ifdef DEBUG
		NSLog(@"Changing tier to %d", tier);
#endif
		if ([delegate respondsToSelector:@selector(windowDidChangeTierTo:)])
			[delegate windowDidChangeTierTo: tier];
		[self windowDidChangeTierTo: tier];
	}
	return self;
  }

- windowDidChangeTierTo:(int)aTier { return self; }

/*---------------------------------------------------------------------------*/
/* Put the full screen image in front of everything. */

- makeFrontdrop:sender
 {	[self setWindowTier: MISC_SAVERTIER];
#error EventConversion: addToEventMask:NX_MOUSEMOVEDMASK: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	[self addToEventMask:NSMouseMovedMask];
	return self;
  }


/*---------------------------------------------------------------------------*/
/* Put the full screen image in back of everything. */

- makeBackdrop:sender
 {	[self setWindowTier: MISC_BACKGROUNDTIER];
#error EventConversion: removeFromEventMask:NX_MOUSEMOVEDMASK: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
	[self removeFromEventMask:NSMouseMovedMask];
	return self;
 }


// Debug scaffolding
//- (BOOL)acceptsFirstResponder	   {NXLogError("Window Set first responder");
//				    return YES;}
//
//- keyDown:    (NXEvent *) eventPtr {NXLogError("Window Key down");
//				    return self;}
//- mouseDown:  (NXEvent *) eventPtr {NXLogError("Window Mouse Down");
//				    return self;}
//- mouseMoved: (NXEvent *) eventPtr {NXLogError("Window Mouse Moved");
//				    return self;}

/*===========================================================================*/
/* Archive methods */
/*===========================================================================*/

- (void)encodeWithCoder:(NSCoder *)aCoder
  {	[super encodeWithCoder:aCoder];
	[aCoder encodeValueOfObjCType:"i" at:&currentTier];
}


- (id)initWithCoder:(NSCoder *)aDecoder
  {	[super initWithCoder:aDecoder];
	[aDecoder decodeValueOfObjCType:"i" at:&currentTier];
	return self;
  }

/*---------------------------------------------------------------------------*/
/* play games with the tier number so that we can correctly reset the window
   status in the window manager. It might seems strange, but it gets by
   the check that ignores window tier values unless they have changed from
   the previous setting.

   NOTE: Will I have to handle the setting of the event mask? Will the
    	 delegate problem still occur?
*/

#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake
  {	int	tmp;

#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
	[super awake];
	tmp = currentTier++;
	[self setWindowTier: tmp];
	return self;
  }

@end
