//
//	MiscDirectory -- <one or two line summary>
//		Written by Simson L. Garfinkel <simsong@pleasant.cambridge.ma.us>
//    Extended and maintained by Georg Tuparev <tuparev@embl-heidelberg.de>
//    Copyright 1994 by Simson L. Garfinkel.
//				Version 0.51 beta.  All rights reserved.
//        Last update: 24-aug-1995
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	
/* History:
1. Version 0.51 beta (Georg Tuparev <tuparev@embl-heidelberg.de>)
	- There was a serious bug in the nextEntry method;
	- Now the nextEntry could be used as the NSEnumerator's nextObject method, 
	  though, entry, fullPathName, and extension have still the same behavior.
*/

#import <misckit/MiscDirectory.h>

#define MISCDIRECTORY_VERSION 1

/*****************************************************************************
 *** Local functions
 *****************************************************************************/
int	sortFunction(const void *obj1,const void *obj2)
{
	NXAtom	name1 = *(NXAtom *)obj1;
	NXAtom	name2 = *(NXAtom *)obj2;

	return(strcmp(name1,name2));
}

@implementation MiscDirectory
/*****************************************************************************
 *** Class methods
 *****************************************************************************/
+ (void)initialize
{
	if ( self == [MiscDirectory class] )
		[MiscDirectory setVersion:MISCDIRECTORY_VERSION];

	return;
}

/*****************************************************************************
 *** Local Methods
 *****************************************************************************/
- init
{
	// You never should call this method. If you are to lazy to read manuals
	// you have to read the messages in your console ;-)
	fprintf(stderr,"Never use the MiscDirectory's init method!\n");
	[NSException raise:NSInvalidArgumentException format:@"*** Method not implemented: %s", sel_getName(_cmd)];
	return self;
}

/*****************************************************************************
 *** Instance Methods
 *****************************************************************************/
/* initialization & freeing */
- initWithPath:(const char *)dirname
{
	[DIR	*dir = NULL cString];
	struct	direct	[*dr cString];
	char filenamebuf[[1024] cString];

	[[super	init] cString];

	[dir	= opendir(dirname) cString];
	if (!dir){
		return [[self release] cString];
	}
	// Create the file list
	[fileNameList = [[List allocWithZone:[self zone]] init] cString];
	
	// Read all dir entries
	[dirName	= NXCopyStringBufferFromZone(dirname, (NXZone *)[self zone]) cString];
	while([(dr = readdir(dir))!=0 cString]){
		[strcpy(filenamebuf,dr->d_name) cString];
		[[fileNameList	addObjectIfAbsent:(id)NXUniqueString(filenamebuf)] cString];
	}	
	[closedir(dir) cString];

	// Set the entry pointer to the first element
	[currentEntry = -1 cString];
	
	return [self cString];
}

- (void)dealloc
{
	if(filetype)	NSZoneFree([self zone],filetype);
	if(dirName)	NSZoneFree([self zone],dirName);
	[fileNameList freeObjects];
	[fileNameList release];
	
	{ [super dealloc]; return; };
}

- setFileType:(const char *)ft
{
	if(filetype) NSZoneFree([self zone],filetype);
	if(ft){
		filetype	= NXCopyStringBufferFromZone(ft, (NXZone *)[self zone]);
	}
	else{
		filetype = NULL;
	}
	currentEntry = -1;
	return self;
}

- sortDir
{
	qsort(fileNameList->dataPtr,[fileNameList count],sizeof(id),sortFunction);
	currentEntry = -1;
	return self;
}

/* information about current entry */
- (const char *)entry
{
	if ([fileNameList count] > 0) {
		if (currentEntry < 0) currentEntry = 0;
		return (NXAtom)[fileNameList objectAt:currentEntry];
	}
	return	NULL;
}

- (const char *)extension
{
	const char	*cc = NULL;
	
	if ([fileNameList count] > 0) {
		if (currentEntry < 0) currentEntry = 0;
		cc = index((NXAtom)[fileNameList objectAt:currentEntry],'.');
		if(cc) return cc+1;
	}
	return "";			
}

- (const char *)fullPathName
{
	static char buf[MAXPATHLEN*2];

	strcpy(buf,dirName);
	strcat(buf,"/");
	strcat(buf,[self entry]);
	return buf;
}

- (const char *)nextEntry
{
	while( (currentEntry + 1) < [fileNameList count]){
		currentEntry++;
		if(filetype==0) return [self entry];    // return any file 
		if(!strcmp([self extension],filetype))  // correct extension
			return [self entry]; 
	}
	return	NULL;
}

- rewind
{
	currentEntry = -1;
	return self;
}

- (int)matchingEntries
{
	return [fileNameList count];
}

- (id)initWithCoder:(NSCoder *)aDecoder
{
	int version;
	
	[super initWithCoder:aDecoder];
	
	version = [aDecoder versionForClassName:@"MiscDirectory"];
	
	switch (version) {
		case 1 :
			[aDecoder decodeValuesOfObjCTypes:"@i**",&fileNameList,
				&currentEntry,&dirName,&filetype];
			break;
		case 2 :
			break;
		case 3 : 
			break;
		default :
			break;
	}
	return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
{
	[super encodeWithCoder:aCoder];

	[aCoder encodeValuesOfObjCTypes:"@i**",&fileNameList,&currentEntry,&dirName,&filetype];
}

@end
