//
//	MiscDocument.m -- document controller
//		Written by Mike Ferris Copyright (c) 1993, 1995 by Mike Ferris.
//		Modified for use in the MiscKit by Don Yacktman from MODocController.
//				Version 1.0.  All rights reserved.
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import <misckit/MiscDocument.h>
#import <misckit/MiscDocType.h>
#import <misckit/MiscDocManager.h>
#import <misckit/MiscString.h>
#import <misckit/MiscClassVariable.h>
#import <objc/objc-runtime.h>

#define CLASS_VERSION					0
#define CLASS_NAME						"MiscDocument"

#define SAVEACCESSORY_NIBNAME			"MiscDocSaveAccessory.nib"

#define UNTITLED_STRING				[[MiscKitBundle localizedStringForKey:@"UntitledString" value:@"Untitled" table:@"MiscDocMenu"] cString]

#define CLOSEALERTPANEL_TITLE			[[MiscKitBundle localizedStringForKey:@"CloseTitle" value:@"Close" table:@"MiscDocAlertPanels"] cString]

#define CLOSEALERTPANEL_TEXT			[[MiscKitBundle localizedStringForKey:@"CloseText" value:@"Document %s has unsaved changes.  Do you wish to save, close anyway, or cancel?" table:@"MiscDocAlertPanels"] cString]

#define CLOSEALERTPANEL_SAVE			[[MiscKitBundle localizedStringForKey:@"CloseSaveButton" value:@"Save" table:@"MiscDocAlertPanels"] cString]

#define CLOSEALERTPANEL_CLOSEANYWAY		[[MiscKitBundle localizedStringForKey:@"CloseCloseButton" value:@"Close Anyway" table:@"MiscDocAlertPanels"] cString]

#define CLOSEALERTPANEL_CANCEL			[[MiscKitBundle localizedStringForKey:@"CloseCancelButton" value:@"Cancel" table:@"MiscDocAlertPanels"] cString]

#define ERRORALERTPANEL_TITLE			[[MiscKitBundle localizedStringForKey:@"ErrorTitle" value:@"Error" table:@"MiscDocAlertPanels"] cString]

#define ERRORALERTPANEL_TEXT			[[MiscKitBundle localizedStringForKey:@"ErrorText" value:@"Failed to revert document %s." table:@"MiscDocAlertPanels"] cString]

#define ERRORALERTPANEL_OK			[[MiscKitBundle localizedStringForKey:@"ErrorOKButton" value:@"OK" table:@"MiscDocAlertPanels"] cString]

@implementation MiscDocument

// True class variables via hashtables
static MiscClassVariable *Misc_classControllerName;
static MiscClassVariable *Misc_classDocTypes;

static id MiscKitBundle = nil;

+ (void)initialize
// Set the version.  Load classes.  Load the MiscKit bundle.  
// Init class variables.
{
#warning NSNameConversion:  Class names have changed. Check that the class name passed in is real.
	if (self == objc_lookUpClass(CLASS_NAME))  {
		[self setVersion:CLASS_VERSION];		
		
		// Set up class variables
		Misc_classControllerName = [[MiscClassVariable allocWithZone:[self zone]] 
					initDoesFreeValues:YES];
		Misc_classDocTypes = [[MiscClassVariable allocWithZone:[self zone]] 
					initDoesFreeValues:YES];
	}
	return;
}

+ setControllerName:(const char *)typeName
// Sets the name of this document class type.  This is used in New 
// and Open submenus that MiscDocManager creates if there is more 
// than one document type being managed.
{
	id newVal;
	
	if (typeName)  {
		newVal = [[MiscString allocWithZone:[self zone]] 
				initString:typeName];
	}  else  {
		newVal = nil;
	}
	[Misc_classControllerName setObject:newVal forClass:self];
	return self;
}

+ (MiscString *)controllerName
// Returns the controller name.
{
	return [Misc_classControllerName getObjectForClass:self];
}

+ addDocType:(int)tag name:(const char *)name extension:(const char *)ext 
			openSelector:(SEL)openSel saveSelector:(SEL)saveSel
// Adds a new document type that controllers of this class support.  
// The name is the full name of the doc type.  Extension is the file 
// extension for docs of that type.  Tag is for your own use.  open 
// and save selectors take a single MiscDocType as their argument.  
// Either can be NULL if this is a read-only or write-only file type.
{
	List *typeList = [self docTypes];
	
	if (!name || !*name)  {
		NSLog([NSString stringWithCString:"[%s addDocType]:  Attempted to add a document type "
					"with no name.  Document types must have names."], [[self name] cString]);
		return nil;
	}
	if (!typeList)  {
		typeList = [[List allocWithZone:[self zone]] init];
		[Misc_classDocTypes setObject:typeList forClass:self];
	}

	[typeList addObject:[[MiscDocType allocWithZone:[self zone]] 
				initForControllerClass:self tag:tag name:name extension:ext 
				openSelector:openSel saveSelector:saveSel]];
	return self;
}

+ (List *)docTypes
// Returns the list of document types supported by this class.
{
	return (List *)[Misc_classDocTypes getObjectForClass:self];
}

+ (MiscDocType *)docTypeForTag:(int)tag
// Returns the doc type whose tag is tag.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if ([type typeTag] == tag)  {
			return type;
		}
	}
	return nil;
}

+ (MiscDocType *)docTypeForName:(const char *)name
// Returns the doc type whose name is name.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if (([type typeName]) && (name) && 
					(strcmp([type typeName], name) == 0))  {
			return type;
		}
	}
	return nil;
}

+ (MiscDocType *)docTypeForExtension:(const char *)extension
// Returns the doc type whose extension is extension.
// If we find a match, return it.  If we don't find a match, and we have
// an extension-less doc type, return that.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	const char *curExt = NULL;
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		curExt = [type typeExtension];
		if ((!extension || !*extension) && (!curExt || !*curExt))  {
			return type;
		}
		if ((curExt) && (extension) && 
					(strcmp(curExt, extension) == 0))  {
			return type;
		}
	}  
	if (extension && *extension)  {
		return [self docTypeForExtension:NULL];
	}
	return nil;
}

+ (List *)getOpenTypesList:(List *)list
// Fills list with the types which support reading from our docTypes list.
// list is returned.  If list is nil, a List object is allocated for it.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	if (!list)  {
		list = [[List allocWithZone:[self zone]] init];
	}
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if ([type canOpenType])  {
			[list addObject:type];
		}
	}
	
	return list;
}

+ (List *)getSaveTypesList:(List *)list
// Fills list with the types which support saving from our docTypes list.
// list is returned.  If list is nil, a List object is allocated for it.
{
	List *docTypes = [self docTypes];
	int i, c = [docTypes count];
	MiscDocType *type = nil;
	
	if (!list)  {
		list = [[List allocWithZone:[self zone]] init];
	}
	
	for (i=0; i<c; i++)  {
		type = [docTypes objectAt:i];
		if ([type canSaveType])  {
			[list addObject:type];
		}
	}
	
	return list;
}

+ startUnloading
// We have to deallocate the class variables.
{
	[Misc_classControllerName release];
	[Misc_classDocTypes release];
	
	return self;
}

- initFromFile:(const char *)path manager:aManager
// This is the initializer that will be used normally.  It just calls the DI.
{
	return [self initWithFrameName:NULL fromFile:path manager:aManager];
}

- initWithFrameName:(const char *)theFrameName
// Just call the DI.
{
	return [self initWithFrameName:theFrameName fromFile:NULL manager:nil];
}

- initWithFrameName:(const char *)theFrameName 
			fromFile:(const char *)path manager:aManager
// Designated Initializer.  A null file means a new untitled doc.
{
	[super initWithFrameName:theFrameName];
	
	[self setManager:aManager];
	docName = [[MiscString allocWithZone:[self zone]] init];
	untitledString = [[MiscString allocWithZone:[self zone]] init];
	docType = nil;
	
	isDirty = NO;
	saveAccessoryPanel = nil;
	saveAccessoryBox = nil;
	saveAccessoryPopupButton = nil;
	
	[self setFile:path];
	
	return self;
}

- (void)dealloc
// free our ivars
{
	[docName release];
	[untitledString release];
	if (manager)  [manager removeDocument:self];
	
	{ [super dealloc]; return; };
}

- nibDidLoad
// Set the window title, move the window according to the frameCycle.
{
	[super nibDidLoad];
	
	if (window)  {
		[self resetWindowTitle];
		if (manager)  {
			NSPoint pt;
			[manager getNextWindowLocation:&pt];
			[window setFrameTopLeftPoint:pt];
		}
	}
	return self;
}

- (BOOL)doClear
// override this to totally clear this doc into a blank one.
{
	return YES;
}

- (BOOL)doPrint
// override this to print the document (including running print panel, etc...).
{
	return YES;
}

- saveAccessoryPopupAction:sender
// The action of the save accessory document type popup.  It sets 
// the Save panels required file extension appropriately.
{
#warning FactoryMethods: [SavePanel savePanel] used to be [SavePanel new].  Save panels are no longer shared.  'savePanel' returns a new, autoreleased save panel in the default configuration.  To maintain state, retain and reuse one save panel (or manually re-set the state each time.)
	id savePanel = [NSSavePanel savePanel];
	List *saveTypes = [[self class] getSaveTypesList:nil];
	
	[savePanel setRequiredFileType:[NSString stringWithCString:[[saveTypes objectAt:
				[sender selectedRow]] typeExtension]]];
	
	[saveTypes release];
	return self;
}

- getSaveAccessoryForSaveTypes:(List *)saveTypes 
			currentType:(MiscDocType *)currentType
// This will prepare and return the saveAccessory view for the given sender 
// document.  If necessary it will load the accessory nib first.  It returns 
// nil if there is only a single save type since no accessory is needed then.
// Subclasses may override this to use a different nib file.  If a subclass
// does override, it is important that the new nib file NOT use the existing
// saveAccessory... outlets for MiscDocument.  Declare all new outlets for
// your save accessory nib in your subclass and use those.  The saveAccessory
// outlets in this class are not used outside this method, but other subclasses
// may not override, and they should get the old nib.
{
	id popup = nil;
	int i, c;
	MiscString *scratch = nil;
	MiscDocType *type = nil;
	
	c = [saveTypes count];
	if (c <= 1)  {
		return nil;
	}
	
	if (!saveAccessoryBox)  {
		if (MiscKitBundle)  {
			char pathBuff[MAXPATHLEN+1];
#error StringConversion: This call to -[NXBundle getPath:forResource:ofType:] has been converted to the similar NSBundle method.  The conversion has been made assuming that the variable called pathBuff will be changed into an (NSString *).  You must change the type of the variable called pathBuff by hand.
			if (((pathBuff = [MiscKitBundle pathForResource:[NSString stringWithCString:SAVEACCESSORY_NIBNAME] ofType:@""]) == nil))  {
				NSLog(@"%s: failed to find nib file %s in MiscKit bundle.", [[[self class] name] cString], SAVEACCESSORY_NIBNAME);
				return nil;
			}
			if (![NSBundle loadNibFile:[NSString stringWithCString:pathBuff] externalNameTable:[NSDictionary dictionaryWithObjectsAndKeys:self, @"NSOwner", nil] withZone:[self zone]])  {
				NSLog(@"%s: failed to load nib file %s.", [[[self class] name] cString], pathBuff);
				return nil;
			}
		}  else  {
			if (![NSBundle loadNibNamed:[NSString stringWithCString:SAVEACCESSORY_NIBNAME] owner:self])  {
				NSLog(@"%s: failed to load nib file %s.", [[[self class] name] cString], SAVEACCESSORY_NIBNAME);
				return nil;
			}
		}
		[saveAccessoryBox removeFromSuperview];
		[saveAccessoryPanel release];
		saveAccessoryPanel = nil;
	}
	
	popup = [saveAccessoryPopupButton target];
	
	// remove all but the first item in the list
	for (i=[popup count]-1; i>0; i--)  {
		[[popup removeItemAtIndex:i] release];
	}
	
	// set the first item's title to ""
#warning PopUpConversion: This message should be sent to an NSPopUpButton, but is probably being sent to an NSPopUpList
#warning PopUpConversion: Consider NSPopUpButton methods instead of using itemMatrix to access items in a pop-up list.
	[[[popup itemMatrix] cellAtRow:0 column:0] setTitle:@""];
	
	scratch = [[MiscString allocWithZone:[self zone]] init];

	// add the new items
	for (i=0; i<c; i++)  {
		type = [saveTypes objectAt:i];
		[scratch setFromFormat:"%s (%s)", 
					[type typeName], 
					([type typeExtension]?[type typeExtension]:"<none>")];
		[popup addItem:[scratch stringValue]];
		// If this is the current type...
		if (type == currentType)  {
			[saveAccessoryPopupButton setTitle:[scratch stringValue]];
		}
	}
	
	// If there is no current type, select the first one.
	if (currentType == nil)  {
		type = [saveTypes objectAt:0];
		[scratch setFromFormat:"%s (%s)", 
					[type typeName], 
					([type typeExtension]?[type typeExtension]:"<none>")];
		[saveAccessoryPopupButton setTitle:[scratch stringValue]];
	}
	
	// remove the first item (which was still left from the old.
	[[popup removeItemAtIndex:0] release];
	
	// set the target/action
	[popup setTarget:self];
	[popup setAction:@selector(saveAccessoryPopupAction:)];
	
	[popup sizeToFit];
			
	return saveAccessoryBox;
}

- save:sender
// Called from various places when we are supposed to save.
{
	BOOL ret = YES;

	if ([docName emptyString])  {
		return [self saveAs:self];
	}
	ret = ([self perform:[docType saveSelector] withObject:docType] != nil);
	if (ret)  {
		[self setDirty:NO];
	}
	return ret?self:nil;
}

- saveAs:sender
// Called by DocManager in response to Save As command.
{
#warning FactoryMethods: [SavePanel savePanel] used to be [SavePanel new].  Save panels are no longer shared.  'savePanel' returns a new, autoreleased save panel in the default configuration.  To maintain state, retain and reuse one save panel (or manually re-set the state each time.)
	id savePanel = [NSSavePanel savePanel];
	MiscString *dir, *file, *extension;
	
	List *saveTypes = [[self class] getSaveTypesList:nil];
	id saveAccessory = nil;
	
	if ([saveTypes count] == 0)  {
		[saveTypes release];
		return nil;
	}
	
	saveAccessory = [self getSaveAccessoryForSaveTypes:saveTypes 
			currentType:docType];
	
	[savePanel setAccessoryView:saveAccessory];
	[savePanel setDelegate:self];
	[savePanel setTreatsFilePackagesAsDirectories:NO];
	if (!docType)  {
		[savePanel setRequiredFileType:[NSString stringWithCString:[[saveTypes objectAt:0] typeExtension]]];
	}  else  {
		[savePanel setRequiredFileType:[NSString stringWithCString:[docType typeExtension]]];
	}
	if ([docName emptyString])  {
		dir = [[MiscString allocWithZone:[self zone]] 
					initString:[NSHomeDirectory() cString]];
		file = [untitledString copy];
	}  else  {
		file = [docName fileName];
		dir = [docName pathName];
	}
	if (![savePanel runModalForDirectory:@"" file:@""])  {
		[file release];
		[dir release];
		[saveTypes release];
		return nil;
	}
	
	[docName setStringValue:[savePanel filename]];
	extension = [docName fileExtension];
	docType = [[self class] docTypeForExtension:[[extension stringValue] cString]];
	[extension release];
	[self resetWindowTitle];
	[self save:self];
	
	[file release];
	[dir release];
	[saveTypes release];
	return self;
}

#warning ColorConversion: 'writeToFile:' was 'saveTo:' ; conversion should only apply to NSColorList ; 'writeToFile:' returns BOOL rather than raising
- (BOOL)writeToFile:(id)sender
// Called by DocManager in response to Save To command.  The bit about 
// the flushWindow is so the title bar doesn't flash the save to path and 
// then reset to the old path visibly.
{
	id copyString = [docName copy];
	BOOL copyDirty = [self isDirty];
	
	if (window)  {
		[window disableFlushWindow];
	}
	[self saveAs:self];
	
	[self setDirty:copyDirty];
	[docName release];
	docName = copyString;
	[self resetWindowTitle];
	
	if (window)  {
		[window enableFlushWindow];
		[window flushWindow];
	}
	return YES;
}

- revert:sender
// Called when a file is opened or reverted.
{
	BOOL ret = YES;
	
	if ([docName emptyString])  {
		ret = [self doClear];
	}  else  {
		ret = ([self perform:[docType openSelector] withObject:docType] != nil);
	}
	if (ret)  {
		[self resetWindowTitle];
		[self setDirty:NO];
	}
	return ret?self:nil;
}

- close:sender
// Closes the document's window.
{
	[window performClose:self];
	return self;
}

#warning PrintingConversion:  printPSCode: has been renamed to print:.  Rename this method?
- (void)print:(id)sender
// Prints the doc.
{
	BOOL ret = [self doPrint];
	return ret?self:nil;
}

- saveIfNeeded
// Save ourself only if we are dirty.
{
	if ([self isDirty])  [self save:self];
	return self;
}

- setManager:aManager
// Set the docManager that manages us.
{
	if (manager)  [manager removeDocument:self];
	manager = aManager;
	if (manager)  [manager addDocument:self];
	return self;
}

- manager
// Return our manager.
{
	return manager;
}

- setFile:(const char *)fName
// Name our file.  If fName is NULL, set up an untitled name.  The file
// should be of a type we can read.  Also, if the filename is not of a 
// type we can save, make it untitled after loading.
{
	id extension = nil;
	
	[docName setStringValue:[NSString stringWithCString:fName]];
	if ([docName emptyString])  {
		docType = nil;
		if (manager)  {
			[untitledString setFromFormat:"%s-%d", UNTITLED_STRING, 
						[manager nextUntitledNum]];
		}  else  {
			[untitledString setFromFormat:"%s", UNTITLED_STRING];
		}
	}
	
	[self window:YES];
	if (window)  {
		[window disableFlushWindow];
	}
	if (![docName emptyString])  {
		extension = [docName fileExtension];
		docType = [[self class] docTypeForExtension:[[extension stringValue] cString]];
	}
	if (![self revert:self])  {
		NSRunAlertPanel([NSString stringWithCString:ERRORALERTPANEL_TITLE], [NSString stringWithCString:ERRORALERTPANEL_TEXT], [NSString stringWithCString:ERRORALERTPANEL_OK], nil, nil, [[docName stringValue] cString]);
	}
	
	// If we can't save files of whatever type this is, set us to untitled.
	// This must be done after the revert.
	if (![docName emptyString])  {
		if (!docType || (![docType canSaveType]))  {
			[docName freeString];
			docType = nil;
			if (manager)  {
				[untitledString setFromFormat:"%s-%d", UNTITLED_STRING, 
							[manager nextUntitledNum]];
			}  else  {
				[untitledString setFromFormat:"%s", UNTITLED_STRING];
			}
			[self resetWindowTitle];
		}
	}
	if (extension)  [extension release];
	if (window)  {
		[window enableFlushWindow];
		[window flushWindow];
	}
	[self showWindow:self];

	return self;
}

- (BOOL)isUntitled
// Returns YES if this doc has no file.
{
	return [docName emptyString];
}

- (const char *)file
// Return our filename (or the untitled name if we have no file).
{
	if (![docName emptyString])  {
		return [[docName stringValue] cString];
	}  else  {
		return [[untitledString stringValue] cString];
	}
}

- docName
// Return the actual MiscString used to store our doc name.  This method
// can be useful if you want to use the MiscString features.  Otherwise, 
// use the -file method.
{
	return docName;
}

- resetWindowTitle
// Make sure that our window title is correct.
{
	if (window)  [window setTitleWithRepresentedFilename:[NSString stringWithCString:[self file]]];
	return self;
}

- setDirty:(BOOL)flag
// Mark ourselves as edited.
{
	isDirty = flag;
	if (window)  [window setDocumentEdited:flag];
	return self;
}

- (BOOL)isDirty
// Returns whether or not the doc is dirty.
{
	return isDirty;
}

#warning NotificationConversion: windowDidBecomeMain:(NSNotification *)notification is an NSWindow notification method (used to be a delegate method); delegates of NSWindow are automatically set to observe this notification; subclasses of NSWindow do not automatically receive this notification
- (void)windowDidBecomeMain:(NSNotification *)notification
// Inform the manager that we should be the current document.
{
	NSWindow *theWindow = [notification object];
    if (manager)  [manager setCurrentDocument:self];
}

#warning NotificationConversion: windowDidResignMain:(NSNotification *)notification is an NSWindow notification method (used to be a delegate method); delegates of NSWindow are automatically set to observe this notification; subclasses of NSWindow do not automatically receive this notification
- (void)windowDidResignMain:(NSNotification *)notification
// Inform the manager that we should no longer be the current document.
{
	NSWindow *theWindow = [notification object];
    if (manager)  [manager setCurrentDocument:nil];
}

- (BOOL)windowShouldClose:(id)sender
// If we are dirty, ask if we should save before closing and give the chance
// to cancel.
{
	int ret;
	
	if ([self isDirty])  {
		ret = NSRunAlertPanel([NSString stringWithCString:CLOSEALERTPANEL_TITLE], [NSString stringWithCString:CLOSEALERTPANEL_TEXT], [NSString stringWithCString:CLOSEALERTPANEL_SAVE], [NSString stringWithCString:CLOSEALERTPANEL_CLOSEANYWAY], [NSString stringWithCString:CLOSEALERTPANEL_CANCEL], [self file]);
		if (ret==NSAlertDefaultReturn)  {
			// save
			if (![self save:self])  {
				return NO;
			}
		}  else if (ret==NSAlertAlternateReturn)  {
			// close anyway
			// do nothing, just let it close
		
			// close anyway
			// do nothing, just let it close
		}  else  {
			// cancel
			return NO;
		}
	}
	// if we get to here, we are closing
	[self autorelease];
	return YES;
}

#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake
// Initialize stuff we don't archive.
{
	saveAccessoryPanel = nil;
	saveAccessoryBox = nil;
	saveAccessoryPopupButton = nil;
	return self;
}

- (id)initWithCoder:(NSCoder *)aDecoder
// This method is probably not useful.  But here it is anyway.
{
	int classVersion;

	[super initWithCoder:aDecoder];
	
	classVersion = [aDecoder versionForClassName:[NSString stringWithCString:CLASS_NAME]];
	
	switch (classVersion)  {
		case 0:		// First version.
			docName = [[aDecoder decodeObject] retain];
			untitledString = [[aDecoder decodeObject] retain];
			manager = [[aDecoder decodeObject] retain]; // ***** I'm not 100% convinced we want to do this; perhaps we want to write an object reference...
			[aDecoder decodeValueOfObjCType:"C" at:&isDirty];
			docType = nil;
			[self setFile:[[docName stringValue] cString]];
			break;
		default:
			NSLog([NSString stringWithCString:"[%s read:] class version %d cannot read "
						"instances archived with version %d"], CLASS_NAME, CLASS_VERSION, classVersion);
			docName = [[MiscString allocWithZone:[self zone]] init];
			untitledString = [[MiscString allocWithZone:[self zone]] init];
			manager = nil;
			isDirty = NO;
			docType = nil;
			[self setFile:NULL];
			break;
	}
	return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
// This method is probably not useful.  But here it is anyway.
{
	[super encodeWithCoder:aCoder];
	[aCoder encodeObject:docName];
	[aCoder encodeObject:untitledString];
	[aCoder encodeConditionalObject:manager];
	[aCoder encodeValueOfObjCType:"C" at:&isDirty];
}

@end

