/************************************************************************
  CATEGORY:			MiscFile (Creation)

  See the header for more information.

  This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.

  Copyright (c) 1995 by author.  All rights reserved.

 *************************************************************************/

/*#import <AppKit/nextstd.h>*/		// for NX_MALLOC/NX_FREE
#import <misckit/MiscFile.h>


@implementation MiscFile (Creation)

- (MiscFile *)createCopyNamed: (const char *)newPath
{
	// Same as method below except this one always overwrites if newPath
	// already exists. You are responsible for freeing the MiscFile instance
	// returned.
	return [self createCopyNamed: newPath overwrite: YES];
}


- (MiscFile *)createCopyNamed: (const char *)newPath overwrite: (BOOL)kill
{
	// Actually copies the contents of the receiver into newFile and
	// returns the MiscFile if successful. Otherwise nil is returned.
	// You are responsible for freeing the MiscFile returned.

	char  *buf;
	int  sFd, dFd;
	int  flags = O_WRONLY | O_CREAT;
	int  readSize = [self size];
	int  bytesRead, totalBytes = 0;
	
	if ([self isDirectory] == YES || strcmp (newPath, [self fullPath]) == 0)
		// We cannot make a copy of a directory yet.
		return nil;
		
	if (kill == NO)
		flags |= O_EXCL;		// Generate an error if file exists.
	else
		flags |= O_TRUNC;		// If exists, erase what is there.

	// Create with the same permissions as the receiver.	
	dFd = open (newPath, flags, [self permissions]);
	
	if (dFd == -1)
	{
		[self _setErrorCode: dFd];	
		return nil;
	 }
	 
	sFd = open ([self fullPath], O_RDONLY);
	if (sFd == -1)
	{
		close (dFd);
		[self _setErrorCode: sFd];
		return nil;
	 }
	
	buf = malloc((readSize) * sizeof(char));
	do
	{
		bytesRead = read (sFd, buf, readSize);
		if (bytesRead > 0)
			write (dFd, buf, bytesRead);
		totalBytes += bytesRead;
	 }
	while (totalBytes < readSize);
	
	free (buf);
	close (sFd);
	close (dFd);
	return [ [ [self class] alloc] initWithPath: newPath];
}


- (MiscFile *)createSymbolicLinkNamed: (const char *)newLink
{
	// Creates a symbolic link to the receiver and returns it. If there
	// was an error, nil is returned.
	// You are responsible for freeing the MiscFile returned.
	if (newLink == NULL)
		return nil;
		
	if ([self symlink: newLink] != MISCFILE_NOERROR)
		return nil;

	// The link must have been successful, so create an
	// instance and return it.
	return [ [MiscFile alloc] initWithPath: newLink];

}


- (MiscFile *)createHardLinkNamed: (const char *)newLink
{
	// Creates a hard link to the receiver and returns it. If there was
	// an error nil is returned.
	// You are responsible for freeing the MiscFile returned.
	if (newLink == NULL)
		return nil;
		
	if ([self link: newLink] != MISCFILE_NOERROR)
		return nil;

	// The link must have been successful, so create an
	// instance and return it.
	return [ [MiscFile alloc] initWithPath: newLink];
}


- (MiscFile *)symbolicLinkSource
{
	// If we are a symbolic link, this returns a MiscFile representing the 
	// source of the link. Nil is returned otherwise. You are responsible for
	// freeing the returned MiscFile.
	char  buf[MAXPATHLEN+1];
	int  succ;
	
	if ([self isSymbolicLink] == NO)
		return nil;
		
	succ = [self readlink: buf pathLen: MAXPATHLEN+1];
	if (succ == MISCFILE_ERROR)
		return nil;
		
	return [ [ [self class] alloc] initWithPath: buf];
}

@end


/***************************************************************************
  CHANGES:
  Version 0.2, January 2, 1994 (Todd)
   1. Added -createCopyNamed:overwrite: to give the choice to overwrite
    if the destination already exits.
  January 8, 1994 (Todd)
   2. Added -symblolicLinkSource method.
   
 ***************************************************************************/
 