/************************************************************************
  CATEGORY:			MiscFile

  See the header for more information.
 
   This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.

  Copyright (c) 1995 by author.  All rights reserved.

 *************************************************************************/

#import <sys/param.h>
#import <AppKit/NSApplication.h>	// for NXApp
#import <misckit/MiscTime.h>
#import <misckit/MiscUser.h>
#import <misckit/MiscUserGroup.h>
#import <misckit/MiscFile.h>


@implementation MiscFile (Info)

-(BOOL)hasPermissions: (int)perms for: (int)who
{
	// This method allows inquiries into whether the owner,group or others
	// have permissions (read,write,execute) on the MiscFile. Currently it
	// doesn't check for MISCFILE_STICKY, MISCFILE_SETUID, or MISCFILE_SETGID.
	
	short  realPerms = [self permissions];
	// Copy top 7 bits and zero out other 9 (where rwxrwxrwx info is).
	short  mode = realPerms & 0xfe00;
	
	// Fill up with the permissions we want to know about.
	if (who & MISCFILE_OWNER)
		mode |= perms << 6;
	if (who & MISCFILE_GROUP)
		mode |= perms << 3;
	if (who & MISCFILE_OTHER)
		mode |= perms;
		
	if ((realPerms & mode) == mode)
		return YES;
		
	return NO;
}


- (BOOL)isSymbolicLink
{
  	struct stat  *s;
   	s = [self lstat];

	if (s != NULL && (s->st_mode & S_IFMT) == S_IFLNK)
		return YES;
	return NO;
}


- (BOOL)isDirectory
{
	// You could probably use MiscGraphNode's -isLeaf method but that loads
	// all the children just to find out if it is a leaf. This way we just use
	// stat.
  	struct stat  *s;
	s= [self stat];
	
	if (s != NULL && (s->st_mode & S_IFMT) == S_IFDIR) 
		return YES;
	return NO;
}


- (dev_t)device
{
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_dev;
}


- (ino_t)inode
{
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_ino;
}

	
- (u_short)permissions
{
	// This doesn't use whichStat because we always want the permissions
	// of the real file (not the symbolic link).
	struct stat  *s;
	s = [self stat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_mode;

}


- (short)hardLinks
{
	struct stat  *s;
	s = [self stat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_nlink;
}


- (MiscUser *)owner
{
	// Returns a MiscUser representing the uid of the MiscFile. It is your
	// responsibility to free the returned object.
	struct stat  *s;
	s= [self whichStat];
	if (s == NULL)
		return nil;
	return [ [MiscUser alloc] initWithUserId: s->st_uid];
}


- (MiscUserGroup *)group
{
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return nil;
	return [ [MiscUserGroup alloc] initWithGroupId: s->st_gid];
}


- (dev_t)deviceType
{
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_rdev;
}

	
- (off_t)size
{
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_size;
}


- (long)blocksAllocated
{
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_blocks;
}


- (long)optimalBlockSize
{
	struct stat *s;
	s = [self whichStat];
	if (s == NULL)
		return MISCFILE_ERROR;
	return s->st_blksize;
}


- (MiscTime *)lastAccessTime
{
	// You are responsible for freeing the MiscTime instance returned. If 
	// something went wrong with stating the file, nil is returned.
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return nil;
	return [ [MiscTime alloc] initWithTime_t: s->st_atime];
}


- (MiscTime *)lastModifyTime
{
	// You are responsible for freeing the MiscTime instance returned. If 
	// something went wrong with stating the file, nil is returned.
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return nil;
	return [ [MiscTime alloc] initWithTime_t: s->st_mtime];
}


- (MiscTime *)lastStatusChangeTime
{
	// You are responsible for freeing the MiscTime instance returned. If 
	// something went wrong with stating the file, nil is returned.
	struct stat  *s;
	s = [self whichStat];
	if (s == NULL)
		return nil;
	return [ [MiscTime alloc] initWithTime_t: s->st_ctime];
}


- (BOOL)readAccess
{
	if ([self access: MISCFILE_READ] == MISCFILE_SUCCESS)
		return YES;
	return NO;
}

	
- (BOOL)writeAccess
{
	if ([self access: MISCFILE_WRITE] == MISCFILE_SUCCESS)
		return YES;
	return NO;
}


- (BOOL)executeAccess
{
	if ([self access: MISCFILE_EXECUTE] == MISCFILE_SUCCESS)
		return YES;
	return NO;
}


- (BOOL)doesExist
{
	if ([self access: MISCFILE_EXISTS] == MISCFILE_SUCCESS)
		return YES;
	return NO;
}


- (NSImage *)iconImage
{
	return [[[NSWorkspace sharedWorkspace] iconForFile:[NSString stringWithCString:[self fullPath]]] retain];
}

@end
