/************************************************************************
  CATEGORY:			MiscFile (Searching)

  See the header for more information on this category.
 
  This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.

  Copyright (c) 1995 by author.  All rights reserved.

 *************************************************************************/

#import <AppKit/AppKit.h>
#import <misckit/MiscFile.h>


@interface MiscFile (PrivateSearching)

- (BOOL)_addFile: (MiscFile *)aFile;

@end


@implementation MiscFile (Searching)

- (MiscFile *)searchFor: (const char *)filename 
{
	return [self searchFor: filename recursive: NO followLinks: NO];
}


- (MiscFile *)searchFor: (const char *)filename recursive: (BOOL)recursive
{
	return [self searchFor: filename recursive: recursive followLinks: NO];	
}


- (MiscFile *)searchFor: (const char *)filename recursive: (BOOL)recursive 
				followLinks: (BOOL)followLinks
{
	// The method that does the actual searching. First it checks itself, then
	// if recursion is YES, passes the message onto it's children until it is
	// either done, or one of them matches. If rercursion is NO this method
	// simply checks it's children itself and returns the child if any of them
	// match.

  	id  found = nil;
  	int  i = 0;
  
	if (filename == NULL || mfFlags.visited == YES)
		return nil;

	if ([self isSymbolicLink] && followLinks == NO)
		return nil;
		
	if ([[self filename] isEqualToString:[NSString stringWithCString:filename]])
	{
		mfFlags.visited = NO;
		return self;
	 }
	 
	// If we don't match, then check our children
	mfFlags.visited = YES;
	for (i=0; i<[ [self children] count]; i++)
	{
		if (recursive)
		{	
			// just pass the message along to all our children
			found = [ [ [self children] objectAt: i] searchFor: filename
							recursive: YES followLinks: followLinks];
			if (found != nil)
			{
				mfFlags.visited = NO;
				return found;
			 }		
		 }	
		else
		{
			// just compare our children's filenames to what we are searching
			if ([[[ [self children] objectAt: i] filename] isEqualToString:[NSString stringWithCString:filename]])
			{
				mfFlags.visited = NO;
				return [ [self children] objectAt: i];
			 }
		 }
	 }

	// If we've gotten this far, neither us nor our children match filename.
	mfFlags.visited = NO;
	return nil;
}


- (List *)searchFiles
{
	return [self searchFilesAndRecurse: NO followLinks: NO];
}


- (List *)searchFilesAndRecurse: (BOOL)recursive
{
	return [self searchFilesAndRecurse: recursive followLinks: NO];
}


- (List *)searchFilesAndRecurse: (BOOL)recursive 
				followLinks: (BOOL)followLinks
{
  id  matchList = nil;
  id  theirList;
  int  i = 0;
  
	if (mfFlags.visited == YES)
		return nil;

	if ([self isSymbolicLink] && followLinks == NO)
		return nil;
		
	if ([self _addFile: self] == YES)
	{
		matchList = [ [List alloc] init];
		[matchList addObject: self];
	 }
	 
	// Ask our children.
	mfFlags.visited = YES;
	for (i=0; i<[ [self children] count]; i++)
	{
		if (recursive)
		{	
			// just pass the message along to all our children
			theirList = [ [ [self children] objectAt: i] 
							searchFilesAndRecurse:YES followLinks: followLinks];
			if (theirList != nil)
			{
				if (matchList == nil)
					matchList = [ [List alloc] init];
				[matchList appendList: theirList];
				theirList = [theirList release];
			 }		
		 }	
		else
		{
			// just compare our children's filenames to what we are searching
			if ([self _addFile: [ [self children] objectAt: i] ] == YES)
			{
				if (matchList == nil)
					matchList = [ [List alloc] init];
				[matchList addObject: [ [self children] objectAt: i] ];
			 }
		 }
	 }

	mfFlags.visited = NO;
	return matchList;
}
	
@end


@implementation MiscFile (PrivateSearching)

- (BOOL)_addFile: (MiscFile *)aFile
{
	// Consults the class delegate (if there is one) and asks if
	// aFile should be added to the list of MiscFiles being searched.
	id  classDelegate = [ [self class] classDelegate];
	
	if (classDelegate != nil && 
					[classDelegate respondsToSelector:@selector(addFile:)])
		return [classDelegate addFile: aFile];

	return NO;
}

@end
