/************************************************************************
  CLASS:			MiscFile

	See the header for more information on this class. 

  This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.

  Copyright (c) 1995 by author.  All rights reserved.

 *************************************************************************/

/*#import <AppKit/nextstd.h>*/
#import <AppKit/AppKit.h>
#import <objc/HashTable.h>
#import <misckit/MiscAnnouncer.h>
#import <misckit/_MiscDirectoryStream.h>
#import <misckit/MiscFile.h>

// For archiving an instance of MiscFile. If you change the instance vars
// you would like archived with this class, bump up the verison number
// and make the changes to the read: method so all versions will be 
// unarchivable.

#define MISCFILE_VERSION 1
#define MISCFILE_CLASSNAME "MiscFile"

// The defaultClass contains the class that the children will be created as.
// Therefore you could have the root be an instance of MiscFile and it's
// children be instances of MiscBundleFile (a hypothetical subclass of 
// MiscFile) See class method +setDefaultFileClass:.

static id  defaultClass = nil;

// This is used to extend the defaultClass to operate for some classes but
// not others. If there is no entry for the subclass, then the defaultClass
// is used for backwards compatibilty (AAT 20/11/95)
static HashTable *defaultClassTable = nil;

// Used so children can have a different classes depending upon their 
// filename extension. Therefore you could create a host of children that are 
// instances of different classes. (MiscTIFFFile, MiscBundleFile, 
// MiscAppWrapperFile, etc). Key = extension, Value = class

static HashTable  *extensionTable = nil;

// Here in case you would like all instances of MiscFile to either let
// a delegate know about something or else consult with the "global delegate"
// before taking some action. Here mostly for future expansion.

static id  classDelegate = nil;

// A hashtable that will keep all the allocated paths and their object's
// addresses so that you can only have one instance for a given path
// in the filesystem. (Only used if pathMustBeValid is NO)
// Key = path, Value = instance address

static HashTable  *instanceTable = nil;


@interface MiscFile (ReferenceCounting)
// Private private private...
- _incrementRefCount;

@end


@implementation MiscFile 

+ (void)initialize
{
	// set the version for archiving purposes
	if (self == [MiscFile class])
	{
		[self setVersion: MISCFILE_VERSION];
		
		// init to keep track of instances (keys are strings (paths),
		//	values are ids (instance addresses))
		instanceTable = [ [HashTable alloc] initKeyDesc: "*"];
	 }
	return;
}


+ setDefaultFileClass: classId
{
	// Sets the default file class for any children that are created. If you
	// would like to use the instance's class, just setDefaultFileClass: nil
	// (which is the default).
 	defaultClass = classId;
	[self setDefaultFileClass:classId forClass:self];
	return self;
}


+ setDefaultFileClass:classId forClass:aClass
	{
	// Sets the default class for the specified class
	if (defaultClassTable == nil)
		{
		defaultClassTable = [[HashTable alloc] init];
		}
	[defaultClassTable insertKey:aClass value:classId];
	return self;
	}

+ registerFileClass: classId forExtension: (const char *)extension
{
	// Associates a classId with an extension, so when enabled, will allow
	// children of different classes to be created as they are read from
	// the filesystem. You could create MiscAppFile classes for every file
	// with the .app extension, and MiscTIFFFile classes for every .tiff.
	// Don't really know how useful this will turn out to be, but it seemed
	// interesting.

	// Use a hashtable to keep track of extensions that could
	// be used when creating children. (key = extension, value = classId)
	if (extensionTable == nil)
		extensionTable = [ [HashTable alloc] initKeyDesc: "*" valueDesc: "@"];	
	[extensionTable insertKey: NXUniqueString(extension) value: classId];
	return self;
}


+ lookupClassForFilename: (const char *)filename
{
	// Looks up the extension of filename in the extensionTable and returns
	// the class name if a match is found. Otherwise nil is returned.
	const char  *extension;
  
  	if (filename == NULL)
		return NULL;
		
	extension = strrchr (filename, '.');
	if (extension)	
		extension += 1;
			
	// will return nil if extension is not a key in the table
	return [extensionTable valueForKey: NXUniqueString(extension)];
}

	
+ (BOOL)isPathValid: (const char *)fullPath
{
	// Checks whether fullPath exists in the filesystem.
  	struct stat  s;
  	
	if (stat (fullPath, &s) == MISCFILE_NOERROR)
		return YES;
	return NO;
}



+ classDelegate
{
	// These methods were added primarily for the searching category and the 
	// need for a delegate, but they could be useful for different situations. 	// Anyone else used a classDelegate for anything useful? Or this not 
	// generally a good idea?

	return classDelegate;
}


+ setClassDelegate: aDelegate
{
	classDelegate = aDelegate;
	return self;
}


- initWithPath: (const char *)fullPath
{
	// The designated initializer. 
	// Note: If you subclass this class and override this method, make sure to
	// check whether [super initWithPath:] returns nil (the pathname was 
	// invalid) before continuing with further initialization.

  	id  instanceExists = nil;
  
	if (!(self = [super init]))
		return nil;
		
	// if the path already exists, free this instance and return
	// the old one.
	instanceExists = [instanceTable valueForKey: NXUniqueString(fullPath)];
	if (instanceExists)
	{
		[super release];
		// Increase reference count.
		[instanceExists _incrementRefCount];
		return instanceExists;
	 }

	// if fullPath has to refer to a valid file and does not, then
	// free the receiver and return nil
	if ([ [self class] isPathValid: fullPath] == NO)
	{
		[super release];
		return nil;
	 }
	
	// Must have a valid filename or not care, so
	// add this instance and the path to the hashtable.	
	[instanceTable insertKey: NXUniqueString(fullPath) value: self];
	// We are the first instance of this pathname.
	refCount = 1;
			 			
	mfFlags.childrenLoaded = NO;
	[self setCreateChildrenUsingExtensions: NO];
	[self setDisplayAsLeaf: NO];
	[self setChildrenInherit: YES];
	mfFlags.visited = NO;
	st = NULL;
	errorCode = MISCFILE_SUCCESS;
	[self setEnableCaching: YES];
	
	if (path)
		free (path);
	path = malloc((strlen(fullPath)+1) * sizeof(char));
	strcpy (path, fullPath);
	
	return self;
}

	
- initWithName: (const char *)filename inDirectory: (const char *)directory
{
  	char  fullPath[MAXPATHLEN+1];
  	
	if (directory == NULL || filename == NULL)
		return nil;
		
	// check to see if directory already has a trailing '/'
	if (*(directory+strlen(directory)-1) == '/') 
		sprintf (fullPath, "%s%s", directory, filename);		else
		sprintf (fullPath, "%s/%s", directory, filename);

	return [self initWithPath: fullPath];
}


- (void)dealloc
{
	// Decrement the number of references to this instance. If it is
	// zero then we free ourselves.
	refCount--;
	if (refCount > 0)
		return nil;

	// Since this instance is going away, remove the entry from the
	// hashtable.
	[instanceTable removeKey: NXUniqueString (path)];
	
	if (path != NULL)
		free (path);
	{ [super dealloc]; return; };
}


- (void)setLeaf:(BOOL)isALeaf
{
	// Overridden from MiscGraphNode since you cannot take an instance of 	// MiscFile and change it from a leaf to non-leaf. The file is either a 	// directory or it isn't. If you want to change the way it is displayed in 
	// a browser, see displayAsLeaf.

	
}



- (BOOL)displayAsLeaf
{
	// Methods for allowing files that are really non-leaf nodes to be 	
	// displayed as leaf nodes in browsers.

	// Much speedier than [self isLeaf] which actually loads the
	// children in order to tell you whether or not the node has any.
	
	if ([self isDirectory] == NO)
		return YES;
	
	// If it is a directory, then give it the choice.	
	return mfFlags.displayAsLeaf;
}


- setDisplayAsLeaf: (BOOL)asLeaf
{
	mfFlags.displayAsLeaf = asLeaf;
	return self;
}


- (const char *)fullPath
{
	return path;
}


- (NSString *)filename
{
	// Returns the filename of the MiscFile. For example, given 
	// "/abc/def/ghi.jkl" this method would return "ghi.jkl". 
  	const char  *fname;
  
  	if (path == NULL)
		return nil;
	fname = strrchr (path, '/');

	if (fname)
		return [NSString stringWithCString:fname+1];
	return nil;
}		


- (const char *)extension
{
	// Returns the extension of the MiscFile. For example, given 	
	// /abc/def/ghi.jkl this method would return "jkl". If there is no 	
	// extension (no period is found) then NULL is returned.

  	const char  *extension;
  	const char  *slash;
  
  	if (path == NULL)
		return NULL;

	slash = strrchr (path, '/');
	extension = strrchr (path, '.');

	// make sure that the '.' is within the filename and not another
	// part of the path
	if (slash > extension)	
		return NULL;
			
	if (extension)
		return extension+1;
	return NULL;
}


- children
{
	// Overridden method from MiscGraphNode. Needed so we can create our 	// children when asked. There is much code here to make sure that if the 	// filesystem is reread that everything will still be correct. We also make 	// sure that annoucements go out if there has been any changes since the 	// last time it was read.
  	id  oldList = nil;
  	int  i;
  	int  diff = 0;
  
	if ([self childrenLoaded])
		return children;
	mfFlags.childrenLoaded = YES;
			
	if (children && [children count] > 0)
	{
		oldList = [children copyWithZone:[self zone]];
		[children empty];
	 }	
			
	// Load the children. If we are updating the list we will do our own
	// announcing.
	if (oldList != nil)
		[announcer setSendAnnouncements: NO];
	[self loadChildrenFromFilesystem];

	if (oldList != nil)
		[announcer setSendAnnouncements: YES];
	else
		return children;
		
	// Send some announcements depending upon list sizes
	diff = [children count] - [oldList count];
	
	// This of course fails if you add one file and remove one file...
	// You should get a didAddChild and a didRemoveChild but you
	// won't. I will fix this in a later release.
	if (diff > 0)
		if (diff == 1)
			[self announce: @selector (didAddChild:)];
		else
			[self announce: @selector (didAddChildren:)];
	else
	if (diff < 0)
		if (diff == -1)
			[self announce: @selector (didRemoveChild:)];
		else
			[self announce: @selector (didRemoveChildren:)];
				
	// Free all old objects that are no longer in the newly read list
	for (i=0; i<[oldList count]; i++)
		if ([children indexOf: [oldList objectAt: i] ] == NSNotFound)
			[[oldList removeObjectAt: i] release];
			
	[oldList release];
	return children;
}	

	
- (List *)updateChildren
{
	// Just in case you have reason to believe that the filesystem has changed,
	// this will reread the filesystem and return the list of children.

	mfFlags.childrenLoaded = NO;
	return [self children];
}

	
- loadChildrenFromFilesystem
{
	// Loads the children from the filesystem. This is here just in case you 
	// would like to read the files more efficiently than I do here. Make sure
	// though if you override this class to instantiate the children from
	// the right class.
	// I will be replacing this code with something faster in a later release.
	// Perhaps I will steal code from MiscDirectory so all the directory 
	// entries can be sorted, etc.
  	id  newList = nil;
  	id  childsClass = nil;
  	id  dirStream = nil;
  	id  newChild = nil;
  	char  childsPath[MAXPATHLEN+1];

  	dirStream = [ [_MiscDirectoryStream alloc]
			initWithDirectory: [self fullPath] ];
		
	while ([dirStream readDir])
	{
		if (newList == nil)
			newList = [ [List alloc] init];
			
		if ([@"." compare:[dirStream filename]] != 0 && 
						[@".." compare:[dirStream filename]] != 0)
		{	
			// Need a special case when the MiscFile is "/" (the root).
			if (strcmp ("/", [self fullPath]) != 0)
				sprintf (childsPath, "%s/%s", [self fullPath], 
								[[dirStream filename] cString]);	
			else
				sprintf (childsPath, "/%s", [[dirStream filename] cString]);
				
			// See what class the new child should be instantiated from.
			childsClass = [self _childrensClassGivenFilename: 
							[[dirStream filename] cString] ];

			newChild = [ [childsClass alloc] initWithPath: childsPath];
			[newChild passOnTraits: self];
			[newList addObject: newChild];
		 }
	 }
	
	[self addChildren: newList];	
	
	dirStream = [dirStream release];
	if (newList != nil)
		[newList release];
	return self;
}


- (BOOL)childrenLoaded
{
	// If the children have already been loaded from the filesystem, then
	// YES will be returned.

	return mfFlags.childrenLoaded;
}


- (BOOL)createChildrenUsingExtensions
{
	// Determines whether you should take into account a file's extension when
	// choosing it's class. This is not a class method because each instance 
	// can then choose their children's class. This should be set to YES if you 
	// will be using registerFileClass:forExtension:.
	return mfFlags.useExtensions;
}
	


- setCreateChildrenUsingExtensions: (BOOL)useExts
{
	mfFlags.useExtensions = useExts;
	return self;
}


- (BOOL)childrenInherit
{
	// Returns YES if your children will inherit your attributes. See 
	// the passOnTraits method for what attributes are inherited.
	
	return mfFlags.childrenInherit;
}


- setChildrenInherit: (BOOL)inherit
{
	mfFlags.childrenInherit = inherit;
	return self;
}



- passOnTraits: parent
{
	// If children should inherit their parent's attributes, then the parent
	// should call their child's passOnTraits method like in this class's
	// loadChildrenFromFilesystem. I am not sure you should
	// do stuff like this. Maybe createChildrenUsingExtensions stuff should
	// be class ivars and methods instead. Anyone with an opinion?

	if ([self childrenInherit] == NO)
		return self;
		
	[self setCreateChildrenUsingExtensions: 
					[parent createChildrenUsingExtensions] ];
	return self;
}



- _childrensClassGivenFilename: (const char *)filenameOrPath
{
	// Internally used method that returns the class that a new child will be
	// instantiated from. If we should be using extensions and there is a 	// match, then that class is returned. Next, the defaultClass is returned 	// if it  not nil. If none of the above yeild a class, [self class] is 	// returned.  
  	id  theClass = nil;
  
	if ([self createChildrenUsingExtensions])
		theClass = [ [self class] lookupClassForFilename: filenameOrPath];
	
	if (theClass)
		return theClass;
		

	// Check the defaultClassTable to see if there is an entry for
	// our class, if there is, return that class, otherwise
	// return defaultClass
	if (defaultClassTable)
		{
		id theClass = [defaultClassTable valueForKey:[self class]];
		if (theClass)
			return theClass;
		}

	if (defaultClass)
		return defaultClass;
	
	// If we have got this far, then may as well return our class, which
	// will be used to create the children.
	return [self class];
}


#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake
{
#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
	[super awake];
	mfFlags.visited = NO;
	st = NULL;
	errorCode = MISCFILE_NOERROR;
	return self;
}


- (id)initWithCoder:(NSCoder *)aDecoder
{
  	int  version;
  	BOOL  childrenLoaded, useExtensions, displayAsLeaf, childrenInherit;
  	BOOL  enableCaching;

	[super initWithCoder:aDecoder];
	version = [aDecoder versionForClassName:[NSString stringWithCString:MISCFILE_CLASSNAME]];
	
	switch (version)
	{
		case 0:
			[aDecoder decodeValuesOfObjCTypes:"*cc", &path, &childrenLoaded, &useExtensions];
			[aDecoder decodeValuesOfObjCTypes:"cc", &displayAsLeaf, &childrenInherit];
			mfFlags.enableCaching = YES;
			refCount = 1;
			break;
		
		case 1:
			[aDecoder decodeValuesOfObjCTypes:"*cc", &path, &childrenLoaded, &useExtensions];
			[aDecoder decodeValuesOfObjCTypes:"cc", &displayAsLeaf, &childrenInherit];
			[aDecoder decodeValueOfObjCType:"c" at:&enableCaching];
			mfFlags.enableCaching = enableCaching;
			[aDecoder decodeValueOfObjCType:"s" at:&refCount];
		default:
			break;
	 }

	mfFlags.childrenLoaded = childrenLoaded;
	mfFlags.useExtensions = useExtensions;
	mfFlags.displayAsLeaf = displayAsLeaf;
	mfFlags.childrenInherit = childrenInherit;
	return self;
}


- (void)encodeWithCoder:(NSCoder *)aCoder
{
  	BOOL  childrenLoaded, useExtensions, displayAsLeaf, childrenInherit;
  	BOOL  enableCaching;

	[super encodeWithCoder:aCoder];

	childrenLoaded = mfFlags.childrenLoaded;
	useExtensions = mfFlags.useExtensions;
	displayAsLeaf = mfFlags.displayAsLeaf;
	childrenInherit = mfFlags.childrenInherit;
	enableCaching = mfFlags.enableCaching;
		
	[aCoder encodeValuesOfObjCTypes:"*cc", &path, &childrenLoaded, &useExtensions];
	[aCoder encodeValuesOfObjCTypes:"cc", &displayAsLeaf, &childrenInherit];
	[aCoder encodeValueOfObjCType:"c" at:&enableCaching];
	[aCoder encodeValueOfObjCType:"s" at:&refCount];
}

@end


@implementation MiscFile (ReferenceCounting)
// Private private private...
- _incrementRefCount
{
	refCount++;
	return self;
}

@end

/*************************************************************************
  CHANGES:
  Sept. 12, 1994 (Todd)
  	1. Allows for only one instance for any given pathname, which made 
	it much easier when rereading the filesystem when updating the 
	children. 
	2. Added -updateChildren, +initialize.
  Version 0.3, December 5, 1994 (Todd)
	3. Added ivar visited to stop circular searches. 
  Version 0.5, December 31, 1994 (Todd)
  	4. Added stat struct ivar for caching stat results.
	5. Added enableCaching ivar so you could have the choice to use the
	cached stat structure or stat the file everytime.
	6. Changed all the BOOL ivars into a bitfield struct to save space.
	7. Removed class methods pathMustBeValid and setPathMustBeValid.
  Version 0.6, January 25, 1995
	8. Found that having a single instance without any reference counting
	can be rather dangerous, so I added the refCount ivar. Therefore if
	you did (previously)
		root1 = [ [MiscFile alloc] initWithPath: "/"],   
		root2 = [ [MiscFile alloc] initWithPath: "/"], 
		[root1 free],
		[root2 whatever].
	it would crash since root1 and root2 pointed to the same instance. This
	will be removed for NS4.0 since it will have it's own reference counting.
	
   Dec 6, 1995 (Alastair Thomson <alastair@csarc.otago.ac.nz>
	9. Extended the defaultFileClass handling so that it can be different 
	   for differnt classes. The old behaviour still exists.
	
 *************************************************************************/
  
