/*============================= MiscFileFinder.h ===========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Class for objects which define a type of file and how to where to look 
   for them. If objects are named, they are kept in a hash list and no other
   object of the same name may be created.

HISTORY
18-Sep-93  Dale Amon at GPL
	   Created. Split some portions off RemoteApplication class to create
	   a general means of finding files with a particular type, regardless
	   of location.

*/

#import <misckit/misckit.h>
#import <ctype.h>

@implementation MiscFileFinder

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/

#define SEARCH_ENVIRONMENT [NSLocalizedString(@"Searching FileType environment %s: %s", "Diagnostic, searching environment variable.") cString]

#define SEARCH_DEFAULT [NSLocalizedString(@"Searching FileType default path: %s", "Diagnostic, searching default path.") cString]

#define LOOKUP_FAILED [NSLocalizedString(@"Lookup for %s failed", "Diagnostic message, file was not found.") cString]

#define LOOKUP_WORKED [NSLocalizedString(@"Found file at %s", "Diagnostic message, file was found.") cString]

/*===========================================================================*/
/* Class Variables */
/*===========================================================================*/

static BOOL		_MiscFF_debug = NO;
static HashTable	*types;

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/

+ (void)initialize
  {	if ( self == [MiscFileFinder class] )
	   {[MiscFileFinder setVersion: MISC_FILEFINDER_VERSION_ID];
	    _MiscFF_debug = NO;
	    types = [[HashTable alloc] initKeyDesc:"*" valueDesc: "@"];
	   }
	return;
  }

/*---------------------------------------------------------------------------*/
/* Turn on the class debug trace. */

+ (BOOL) debug			{			return _MiscFF_debug;}
+ setDebug:	    (BOOL) flg	{_MiscFF_debug = flg;		return self;}

/*---------------------------------------------------------------------------*/
/* If there is already a file type by this name, get it */

+ findFileTypeNamed: (const char*) aKey
  {return [types valueForKey: (const void *)aKey];}


/*===========================================================================*/
/* Initalize and free methods */
/*===========================================================================*/
/* DESIGNATED INITIALIZER: for the MiscFileFinder class. If the named file type
   "name" already exists, nil is returned and the object is freed unless the
   found object is self, ie someone is reiniting an existing object.

   Once it has been set, the name of a file type is unchangeable.
*/

- initName:	   (const char  *) name
      defaultPath: (const char	*) path
     pathVariable: (const char	*) var
	     mode: (int)	   mode
  {	id	aType;

	[super init];

	aType = [MiscFileFinder findFileTypeNamed: name];
	if (aType && (aType != self)) {[super release]; return nil;}

	[self setTypeName:	name];
	[self setDefaultPath:	path];
	[self setPathVariable:	var];
	[self setFileMode:	mode];
	pathTmp	 = NULL;

	[types insertKey: (const void*) typeName value: self];
	return self;
   }


/*---------------------------------------------------------------------------*/
/* A unnamed object for a readable file to be found on PATH. */

- init
  {	[self  initName: NULL
	    defaultPath: NULL pathVariable: "PATH" mode: R_OK];
	return self;
  }


/*---------------------------------------------------------------------------*/
/* free whatever we created. */

- (void)dealloc
  {	if (typeName)	  {[types removeKey: (const void*) typeName];
			   free((char*)typeName);
			  }

	if (defaultPath)  free((char*)defaultPath);
	if (pathVariable) free((char*)pathVariable);

	{ [super dealloc]; return; };
  }


/*===========================================================================*/
/* Searching for files */
/*===========================================================================*/
/* INTERNAL: file has been found, get a full path for it. The string is
   allocated from the default zone since it will become the callers
   property
*/

- (void) _setapppath: (const char*) new
  {pathTmp= (new) ? NXCopyStringBuffer(new) : NULL;}
	

/*---------------------------------------------------------------------------*/
/* INTERNAL: local fn for checking a file found in a path search */

#import	<sys/file.h>
#import	<stdio.h>

#define	SKIPFILE	1
#define	FOUNDFILE	0
#define	GIVEUP		-1

static int findit (char *filespec, void *sender)
 {	int	accessible; // This is unused:  -don          , result = 1;

	accessible = access(filespec, [(id)sender fileMode]);
	if (accessible == 0)
	   {	[(id)sender _setapppath: (const char*)filespec];
		return FOUNDFILE;
	   }

	return SKIPFILE;
 }

/*---------------------------------------------------------------------------*/
/* Return a full pathname if we successfully find a file with the specified
   name. The following places are searched, in order:

	1) The paths in your environment variable
	2) The default paths in this class as set by you.
	   If you have never set it, it will be the same as your PATH
	   environment variable

   The first path found is the one returned. A typical situation: you have a
   user copy in a standard place, ie one of the App directories; and you are
   debugging a related set of apps in a set of working directories.

   The returned malloc'd pathname string is allocated from the default zone
   and becomes the callers responsibility.

*/

- (char *) fullPathForFile: (const char*) filename
  {	const char	*envpath;
	int		flag, i;

	pathTmp = NULL;
	if (!filename || strlen(filename) == 0) return NULL;

	/* Check for absolute pathnames and skip the rest if it is. */
	for (i=0; i < strlen(filename); i++)
	 {	if (isspace(filename[i])) continue;
		if (filename[i] == '/') {[self _setapppath: filename];
					 return pathTmp;}
		else			break;
	 }

	if (_MiscFF_debug)
	  {if (envpath = (getenv(pathVariable)))
		NSLog([NSString stringWithCString:SEARCH_ENVIRONMENT], pathVariable, envpath);
	   else
		NSLog([NSString stringWithCString:SEARCH_DEFAULT], defaultPath);
	  }

	flag = NXFilePathSearch (pathVariable, defaultPath, YES,
				 filename, findit, (void *) self);

	if (_MiscFF_debug)
	  {if (!pathTmp) NSLog([NSString stringWithCString:LOOKUP_FAILED], filename);
	   else		 NSLog([NSString stringWithCString:LOOKUP_WORKED], pathTmp);
	  }
	return pathTmp;
  }



/*===========================================================================*/
/* Ivar methods */
/*===========================================================================*/
/* Set the default search path. This path will be used if a file is not found
   on the path defined by the environment variable. */

- setDefaultPath: (const char *) path
  {	if (defaultPath) free((char*)defaultPath);
	defaultPath = (path) ?
		NXCopyStringBufferFromZone(path, (NXZone *)[self zone]) : NULL;

	return self;
  }


/*---------------------------------------------------------------------------*/
/* Set the default environment path variable. */

- setPathVariable: (const char *) var
  {	if (pathVariable) free((char*)pathVariable);
	pathVariable = (var) ?
		NXCopyStringBufferFromZone(var, (NXZone *)[self zone]) : NULL;

	return self;
  }

/*---------------------------------------------------------------------------*/
/* Set the default file mode. */

- setFileMode: (int) fmode	{fileMode = fmode; return self;}

/*---------------------------------------------------------------------------*/
/* If the object doesn't already have a name, and the suggested name is
   non NULL, and non empty and not in use, give the object a name and return
   YES.

   Once named, it can never be changed.
*/

- (BOOL) setTypeName: (const char*) name
  {	if (typeName || !name || (strlen(name) == 0) ||
	   [MiscFileFinder findFileTypeNamed: name]) return NO;

	typeName = NXCopyStringBufferFromZone(name, (NXZone *)[self zone]);

	return YES;
  }

/*---------------------------------------------------------------------------*/
/* read ivars */

- (const char*) typeName	{return typeName;}
- (const char*) pathVariable	{return pathVariable;}
- (const char*) defaultPath	{return defaultPath;}
- (int)		fileMode	{return fileMode;}


/*===========================================================================*/
/* Archive methods */
/*===========================================================================*/

- (void)encodeWithCoder:(NSCoder *)aCoder
  {	[super encodeWithCoder:aCoder];
	[aCoder encodeValuesOfObjCTypes:"**i", &defaultPath, &pathVariable, &fileMode];
}


- (id)initWithCoder:(NSCoder *)aDecoder
  {	[super initWithCoder:aDecoder];
	[aDecoder decodeValuesOfObjCTypes:"**i", &defaultPath, &pathVariable, &fileMode];
	return self;
  }

@end
