//
//	MiscListExtensions.m -- various extensions to List
//		Originally written by Drew Davidson and Don Yacktman
//		Copyright (c) 1994 by Drew Davidson.
//		Modified and extended by Don Yacktman for inclusion into the MiscKit.
//				Version 1.0.  All rights reserved.
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

// check for respondsTo: added by Don for safety's sake.  We really need
// to have the Object class have a deepCopy category...

#import <misckit/misckit.h>

@implementation List(Extensions)

- deepCopy
{
	int i;

	self = [self copy]; // none of the contained objects are copied
	for (i = 0; i < numElements; i++) {
		id theObject = [self objectAt:i];
		if ([theObject respondsToSelector:@selector(deepCopy)])
			[self replaceObject:theObject with:[theObject deepCopy]];
		else
			[self replaceObject:theObject with:[theObject copy]];
	}
	return(self);
}

- deepCopyFromZone:(NSZone *)zone
{
	int i;

	// none of the contained objects are copied by this:
	self = [self copyWithZone:zone];
	for (i=0; i<numElements; i++) {
		id theObject = [self objectAt:i];
		if ([theObject respondsToSelector:@selector(deepCopyFromZone:)])
			[self replaceObject:theObject
					with:[theObject deepCopyFromZone:zone]];
		else
			[self replaceObject:theObject with:[theObject copyWithZone:zone]];
	}
	return(self);
}

// This goes in reverse order like the ones in the AppKit.
// Use the ...PerformInOrder: methods to go the other way.
- makeObjectsPerform:(SEL)aSelector with:object1 with:object2
{	int i;

	// This is fast but it isn't a good idea to play with instance variables
	// in this way... so if NeXT changes List and this breaks, we do have the
	// "slow" code in here, too..
	for(i = numElements-1; i >= 0; i--)
		[dataPtr[i] perform:aSelector withObject:object1 withObject:object2];
	// Slow way:
	// for(i = [self count]-1; i >= 0; i--)
	//	[[self objectAt:i] perform:aSelector with:object1 with:object2];
	return self;
}

- makeObjectsPerformInOrder:(SEL)aSelector
{
	int i;

	// Fast way:
	for(i=0; i<numElements; i++) [dataPtr[i] perform:aSelector];
	// Slow way:
	// for(i=0; i<[self count]; i++) [[self objectAt:i] perform:aSelector];
	return self;
}

- makeObjectsPerformInOrder:(SEL)aSelector with:anObject
{
	int i;

	// Fast way:
	for(i=0; i<numElements; i++) [dataPtr[i] perform:aSelector withObject:anObject];
	// Slow way:
	// for(i=0; i < [self count]; i++)
	//	[[self objectAt:i] perform:aSelector with:anObject];
	return self;
}

- makeObjectsPerformInOrder:(SEL)aSelector with:object1 with:object2
{
	int i;

	// Fast way:
	for(i=0; i < numElements; i++)
		[dataPtr[i] perform:aSelector withObject:object1 withObject:object2];
	// Slow way:
	// for(i=0; i < [self count]; i++)
	//	[[self objectAt:i] perform:aSelector with:object1 with:object2];
	return self;
}

- addLists:firstList, ...
{
	va_list theLists;
	id tempList;

	va_start(theLists, firstList);
	tempList = firstList;
	while (tempList) {
		int i;
		for (i=0; i<[tempList count]; i++) {
			[self addObject:[tempList objectAt:i]];
		}
		tempList = va_arg(theLists, id);
	}
	va_end(theLists);
	return self;
}

- mergeLists:firstList, ...
{
	va_list theLists;
	id tempList;

	va_start(theLists, firstList);
	tempList = firstList;
	while (tempList) {
		int i;
		for (i=0; i<[tempList count]; i++) {
			[self addObjectIfAbsent:[tempList objectAt:i]];
		}
		tempList = va_arg(theLists, id);
	}
	va_end(theLists);
	return self;
}

@end


@implementation NSObject(MiscListForeach)

- foreach:aList perform:(SEL)aSelector
{
	int i;
	
	if (![aList isKindOfClass:[List class]]) return nil; // sanity check
	for (i=0; i<[aList count]; i++) {
		[self perform:aSelector withObject:[aList objectAt:i]];
	}
	return self;
}

- foreach:aList perform:(SEL)aSelector with:anObject
{
	int i;
	
	if (![aList isKindOfClass:[List class]]) return nil; // sanity check
	for (i=0; i<[aList count]; i++) {
		[self perform:aSelector withObject:[aList objectAt:i] withObject:anObject];
	}
	return self;
}

@end

@implementation List (MiscStringSort)

// This category is written by Bill Bumgarner and (C) Bill Bumgarner.

static int stringValueCompare(const void *arg1, const void *arg2)
{
	MiscString *str1 = *(id *)arg1;
	MiscString *str2 = *(id *)arg2;
	return [str1 compareTo:str2 caseSensitive:YES];
}

static int noCaseStringValueCompare(const void *arg1, const void *arg2)
{
	MiscString *str1 = *(id *)arg1;
	MiscString *str2 = *(id *)arg2;
	return [str1 compareTo:str2 caseSensitive:NO];
}

- miscStringSort
{
	[self miscStringSortCaseSensitive:YES]; // we use only one core method...
	return self;
}

- miscStringSortCaseSensitive:(BOOL)aFlag // added by DAY
{	// Danger:  we don't first check to assure that all objects are
	// MiscStrings, but we'll crash and burn if they aren't, so beware!
	// Warning:  This is a scary way to sort since changing the way
	// a List does its instance variables will require the code to change;
	// this code makes assuptions about how things are laid out in memory.
	// If only qsort would allow us to pass one other variable!  We could
	// use a single subroutine, with a shared static variable, to decide
	// the sensitivity of the compare, but then we would not be thread
	// safe.  Thus we choose to use two subroutines.  -DAY
	if (aFlag) // original behavior is case sensitive sorting
		qsort(NX_ADDRESS(self), [self count], sizeof(id), stringValueCompare);
	else
		qsort(NX_ADDRESS(self), [self count], sizeof(id),
				noCaseStringValueCompare);
	return self;
}

@end



// This next category is from R. Todd Thomas and was implemented to
// help support the MiscUser* classes.

@implementation List (StringFinding)

- (BOOL)matchString: (const char *)str usingSel: (SEL)comparator
{
	// Send a comparator message to each object in the list and compare
	// the return value to the string str. If there is a match, the 
	// search is aborted and YES is returned. Otherwise NO is returned.
	int  count = [self count];
	int  i;
	
	for (i=0; i<count; i++)
		if (strcmp ((const char *)[ [self objectAt: i] perform: comparator],
						 str) == 0)
			return YES;

	return NO;
}

@end
