//
//	MiscMailApp.m -- an API to NeXT's Mail.app
//		Modified by Don Yacktman Copyright (c) 1994 by Don Yacktman.
//				Version 1.1.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//

// Error checking note:  the speaker returns a non-zero value if the
// message couldn't be delivered for some reason.  If this happens we
// assume that we lost connection with the Mail.app (did it get quitted
// without us knowing, for example).  As such, we re-init (which ought
// to re-launch the Mail.app if it is gone) and then try to resend the
// message.  If that doesn't work, then we invalidate ourselves.  Next
// time around we'll be rebuilt and try connecting again...

#import <misckit/MiscMailApp.h>
#import <misckit/MiscDictionary.h>
#import <misckit/FREE.h>
#import <AppKit/AppKit.h>

@interface MiscMailApp(private)
+ _invalidateMailerForHost:aHost;
- (MiscMailApp *)_MiscRealInitForHostNamed:host;
- _invalidate;
@end

@implementation MiscMailApp

+ (MiscMailApp *)localMailer
{
	return [self mailerForHostNamed:nil];
}

static id localMailer = nil;
static id dictionary = nil;

+ (MiscMailApp *)mailerForHostNamed:aHost
{
	MiscMailApp *theMailer = nil;

	if (aHost && [[aHost stringValue] length]) {
		if (!dictionary) {
			dictionary = [[MiscDictionary alloc] init];
		}
		theMailer = (MiscMailApp *)[dictionary valueForKey:aHost];
		if (!theMailer) {
			theMailer = [[super alloc] _MiscRealInitForHostNamed:aHost];
			if (theMailer) [dictionary insertKey:aHost value:theMailer];
		}
	} else { // go local if no name given
		if (!localMailer) {
			localMailer = [[super alloc] _MiscRealInitForHostNamed:nil];
		}
		theMailer = localMailer;
	}
	return theMailer;
}

+ _invalidateMailerForHost:aHost
{
	if (aHost && [[aHost stringValue] length]) {
		id theMailer = [self mailerForHostNamed:aHost];
		[dictionary removeKey:aHost];
		[theMailer release];
	} else {
		[localMailer release];
		localMailer = nil;
	}
	return nil;
}

- _invalidate // called if the Speaker becomes unable to communicate...
{
	[[self class] _invalidateMailerForHost:hostName];
	return nil;
}

- init
{
	return [self error:"You cannot use -init on MiscMailApp.\n"];
}

+ alloc
{
	return [self error:"You cannot use +alloc on MiscMailApp.\n"];
}

#define TRY_PORT_CONNECT mailPort = NXPortFromName("MailSendDemo", \
		hostNamePointer); \
	if (PORT_NULL == mailPort)

- (MiscMailApp *)_MiscRealInitForHostNamed:host
{
	port_t mailPort;
	const char *hostNamePointer;

	if (!_initted) {
		if (![super init]) return nil;
	}
	FREE_OBJECT(hostName);
	FREE_OBJECT(theSpeaker);
	if (host && [host length]) {
		if (host != hostName)
			hostName = [MiscString newWithString:[[host stringValue] cString]];
		hostNamePointer = [[hostName stringValue] cString];
	} else {
		if (host != hostName)
			hostName = [MiscString newWithString:"localhost"];
		hostNamePointer = NULL;
	}
	// check to be sure that Mail.app is up and running first
	TRY_PORT_CONNECT { // launch mailer (or Mail.app) if we can't find the port
		// The GLOBAL dwrite "Mailer" is first checked for a valid mailer.
		// ***** change this to use the MiscDefault stuff when it comes out.
		const char *mailer = [[[NSUserDefaults standardUserDefaults] objectForKey:@"Mailer"] cString];
		mailPort = NXPortFromName([NSString stringWithCString:(mailer ? mailer : "Mail")], [NSString stringWithCString:hostNamePointer]);
		if (PORT_NULL == mailPort) { // Couldn't launch from dwrite,
			// so default to Mail.app.
			mailPort = NXPortFromName(@"Mail", [NSString stringWithCString:hostNamePointer]);
		}
		if (PORT_NULL != mailPort) port_deallocate(task_self(), mailPort);
	}
	TRY_PORT_CONNECT { // bail out if we still cannot connect
		[self release];
		return nil;
	}
	theSpeaker = [MiscMailSpeaker new];
	[theSpeaker setSendPort:mailPort];
	currentWindow = MISC_NO_MAIL_WINDOW; // no current window
	noclutter = NO;
	_initted = YES; // make sure the init method is not re-entered!
	return self;
}

- (void)dealloc
{
	FFREE_OBJECT(theSpeaker);
	FFREE_OBJECT(hostName);
	{ [super dealloc]; return; };
}

- hostname { return hostName; }
- (BOOL)noclutter { return noclutter; }
- setNoclutter:(BOOL)flag { noclutter = flag; return self; }

- sendMailTo:to subject:subject body:body
{
	return [self sendMailTo:to subject:subject body:body cc:nil andDeliver:NO];
}

- sendMailTo:to subject:subject body:body cc:cc andDeliver:(BOOL)flag
{
	if (![self openWindow]) return nil;
	[self setTo:to];
	[self setSubject:subject];
	[self setBody:body];
	[self setCc:cc];
	if (flag) [self deliver];
	return self;
}

- openWindow
{
	if (noclutter && (currentWindow >= 0)) [self deliver];
	if ([theSpeaker openSend:&currentWindow]) {
		if (![self _MiscRealInitForHostNamed:hostName] ||
				[theSpeaker openSend:&currentWindow]) {
			[self _invalidate];
			return nil;
		}
	}
	return self;
}

- deliver
{
	if (currentWindow >= 0) {
		if ([theSpeaker deliver:currentWindow]) { // error, so reconnect...
			currentWindow = MISC_NO_MAIL_WINDOW; // invalidate window
			if (![self _MiscRealInitForHostNamed:hostName]) return nil;
		}
	}
	currentWindow = MISC_NO_MAIL_WINDOW; // there is no longer a current window
	return self;
}


// if we get an error, we try to restart the mail app but invalidate the
// mail window since it probably means that the one we had went away for
// whatever reason.  (ie, if it doesn't exist, the handle is useless)
#define MESSAGE(x) 	if (sender && (currentWindow >= 0)) \
		if ([[sender stringValue] length]) \
			if ([theSpeaker x:[[sender stringValue] cString] inWindow:currentWindow]) { \
				currentWindow = MISC_NO_MAIL_WINDOW; \
				if (![self _MiscRealInitForHostNamed:hostName]) return nil; \
			} \
	return self

- setSubject:sender	{ MESSAGE(setSubject);	}
- setBody:sender	{ MESSAGE(setBody);		}
- setTo:sender		{ MESSAGE(setTo);		}
- setCc:sender		{ MESSAGE(setCc);		}

// to allow target/action for everything...
- openWindow:sender { return [self openWindow]; }
- deliver:sender { return [self deliver]; }

@end
