//
// MiscParseTableFile.m -- functions and categories to parse files of
//			the format taken by *.addresses, PB.project, and others.
//		Written by Don Yacktman Copyright (c) 1994 by Don Yacktman.
//				Version 1.0.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

// We use a recursive descent parser to accomplish the parsing and
// a recursive algorithm to write out the file contents.  It is
// important if you plan to write out a data structure that you
// use the right objects to build it.  The file is represented at
// the top level with a "MiscDictionary" object.  The keys are all
// NXAtoms (unique strings) and the values are all objects.  These
// objects can be MiscStrings, MiscLists, and MiscDictionaries.
// In the file, a MiscList is a comma separated list of strings,
// lists, or dictionaries surrounded by parenthesis ().  A MiscDictionary
// is a series of key/value pairs (with the values again as strings,
// lists, and dictionaries) surrounded by curly braces {}.

#import <misckit/misckit.h>
#import <misckit/MiscDictionary.h>

static int _MiscFileLine;
static int _MiscFileLineChar;
static char _MiscFileCurrentChar;
static BOOL _MiscFileEOF;

#define MISC_INDENTCHAR '\t'	// char to use for indenting
#define MISC_INDENTAMOUNT 1		// number of chars for one indent
static int _MiscFileIndent;

@interface MiscList(TableParse)
- parseFromASCIIStream:(NXStream *)aStream;
- writeASCIIToStream:(NXStream *)aStream;
@end

@interface MiscDictionary(TableParse)
- parseFromASCIIStream:(NXStream *)aStream;
- writeASCIIToStream:(NXStream *)aStream;
- writeInnerASCIIToStream:(NXStream *)aStream;
@end

@interface MiscString(TableParse)
- quotedParseFromASCIIStream:(NXStream *)aStream;
- plainParseFromASCIIStream:(NXStream *)aStream;
- parseFromASCIIStream:(NXStream *)aStream;
- writeASCIIToStream:(NXStream *)aStream;
@end

char _MiscGetCharFromStream(NXStream *aStream)
{	// Returns the next char off the stream
	char ret = _MiscFileCurrentChar;

	if (_MiscFileEOF) return '\0';
	_MiscFileCurrentChar = NXGetc(aStream);
	// detect and flag EOFs...
	if (NXAtEOS(aStream)) _MiscFileEOF = YES;
	// update the position in the file so we can report where errors occur
	_MiscFileLineChar++;
	if (ret == '\n') { // on a new line...
		_MiscFileLine++;
		_MiscFileLineChar = 0;
	}
	return ret;
}

char _MiscPeekCharFromStream(NXStream *aStream)
{	// Returns what the next char to come off the stream will be
	return _MiscFileCurrentChar;
}

void _MiscEatWhiteSpace(NXStream *aStream)
{	// Takes tabs, returns, newlines, and spaces off the stream.  This
	// is called more often than strictly necessary, but it should be
	// guaranteed to be called between fetching of each language element
	// so that any extraneous whitespace is ignored.
	char next = _MiscPeekCharFromStream(aStream);
	while ((next == '\r') || (next == '\t') ||
			(next == '\n') || (next == ' ')) {
		_MiscGetCharFromStream(aStream);
		next = _MiscPeekCharFromStream(aStream);
	}
}

void _MiscEatChar(NXStream *aStream, char aChar)
{	// Takes the char "aChar" off the stream or prints an error
	// if a different char is there.
	char theChar = _MiscGetCharFromStream(aStream);
	if (theChar != aChar) { // error
		fprintf(stderr, "Warning:  Illegal character in stream.\n    Expected '%c' but got '%c' on line %d, position %d.\n",
				aChar, theChar, _MiscFileLine, _MiscFileLineChar);
		// raise an exception here. *****
	}
}

int MiscWriteTableFile(MiscString *fileName, id tableRoot)
{	// Open a stream on the file and call MiscWriteTableStream()...
	NXStream *aStream;

	NS_DURING
	aStream = NXOpenMemory(NULL, 0, NX_WRITEONLY);
	MiscWriteTableStream(aStream, tableRoot);
	NXSaveToFile(aStream, [[fileName stringValue] cString]);
	NXClose(aStream);
	NS_HANDLER
	NS_ENDHANDLER
	return 0;
}

int MiscWriteTableStream(NXStream *aStream, id tableRoot)
{
	_MiscFileIndent = 0;
	[tableRoot writeInnerASCIIToStream:aStream];
	return 0;
}

id MiscParseTableFile(MiscString *fileName)
{	// Open a stream on the file and call MiscParseTableStream()...
	id ret = nil;
	NXStream *aStream;

	NS_DURING
	aStream = NXMapFile([[fileName stringValue] cString], NX_READONLY);
	if (aStream) {
		ret = MiscParseTableStream(aStream);
		NXClose(aStream);
	} else { // no stream == error.  Deal with it.  *****
		fprintf(stderr, "Unable to read table file \"%s\".\n",
				[[fileName stringValue] cString]);
	}
	NS_HANDLER
	NS_ENDHANDLER
	return ret;
}

id MiscParseTableStream(NXStream *aStream)
{
	id ret;
	// Prime the character getting routines.
	_MiscFileEOF = NO;
	_MiscFileCurrentChar = NXGetc(aStream);
	_MiscFileLine = 1;
	_MiscFileLineChar = 1;
	ret = [[MiscDictionary alloc] init];
	[ret parseFromASCIIStream:aStream];
	return ret;
}

id _MiscParseAssociatedObject(NXStream *aStream)
{	// Determine if next object on the stream is a list, dictionary, or
	// string and build/parse/return the appropriate object.

	char g1, g2, next = _MiscPeekCharFromStream(aStream);
	id ret = nil;

	g1 = ' '; g2 = ' ';
	switch (next) {
		case '{' /*}*/: { // Parse a Dictionary
			ret = [[MiscDictionary alloc] init];
			g1 = '{'; g2 = '}';
			break;
		}
		case '(' /*)*/: { // Parse a List
			ret = [[MiscList alloc] init];
			g1 = '('; g2 = ')';
			break;
		}
		default  : { // Parse a String
			ret = [[MiscString alloc] init];
			break;
		}
	}
	_MiscEatWhiteSpace(aStream);
	if (g1 != ' ') _MiscEatChar(aStream, g1);
	_MiscEatWhiteSpace(aStream);
	[ret parseFromASCIIStream:aStream];
	_MiscEatWhiteSpace(aStream);
	if (g2 != ' ') _MiscEatChar(aStream, g2);
	_MiscEatWhiteSpace(aStream);
	return ret;
}

void _MiscDoIndent(NXStream *aStream)
{	// Indents a certain amount.  An indent of x is x tabs.  Another
	// option is for an indent of x to be x spaces...
	int i;
	for (i=0; i<_MiscFileIndent; i++) {
		NXPutc(aStream, MISC_INDENTCHAR);
	}
}

@implementation MiscList(TableParse)

- parseFromASCIIStream:(NXStream *)aStream
{
	BOOL done = NO;

	_MiscEatWhiteSpace(aStream);
	if (_MiscPeekCharFromStream(aStream) == /*(*/')') return self;
	while (!done) {
		id newObj = _MiscParseAssociatedObject(aStream);
		[self addObject:newObj]; // add it to the end of the list
		_MiscEatWhiteSpace(aStream);
		if (_MiscPeekCharFromStream(aStream) != ',') {
			done = YES;
		} else {
			_MiscEatChar(aStream, ',');
			_MiscEatWhiteSpace(aStream);
			if ((_MiscPeekCharFromStream(aStream) == /*(*/')') ||
					(_MiscPeekCharFromStream(aStream) == ',')) {
				// ***** deal with error
				done = YES; // could probably do some recovery here
				// by just ignoring the problem and gobbling space...
				return self; // have error... null object.
			}
		}
	}
	return self;
}

- writeASCIIToStream:(NXStream *)aStream
{
	int i;
	NXPutc(aStream, '(');
	[[self objectAt:0] writeASCIIToStream:aStream];
	for (i=1; i<[self count]; i++) {
		NXPutc(aStream, ',');
		NXPutc(aStream, ' ');
		[[self objectAt:i] writeASCIIToStream:aStream];
	}
	NXPutc(aStream, ')');
	return self;
}

@end

@implementation MiscDictionary(TableParse)

- parseFromASCIIStream:(NXStream *)aStream
{
	id key, value;
	
	_MiscEatWhiteSpace(aStream);
	while ((_MiscPeekCharFromStream(aStream) != /*{*/ '}') &&
			(!_MiscFileEOF)) { 
		key = [[MiscString alloc] init];
		[key parseFromASCIIStream:aStream];
		_MiscEatWhiteSpace(aStream);
		// if no = part, assume that value == key object.
		if (_MiscPeekCharFromStream(aStream) == ';') {
			value = [key copy];
		} else {
			_MiscEatChar(aStream, '=');
			_MiscEatWhiteSpace(aStream);
			value = _MiscParseAssociatedObject(aStream);
		}
		[self insertKey:key value:value];
		_MiscEatWhiteSpace(aStream);
		_MiscEatChar(aStream, ';');
		_MiscEatWhiteSpace(aStream);
		[key release]; key = nil;
	}
	return self;
}

- writeASCIIToStream:(NXStream *)aStream
{
	NXPutc(aStream, '{');
	NXPutc(aStream, '\n');
	_MiscFileIndent += MISC_INDENTAMOUNT;
	[self writeInnerASCIIToStream:aStream];
	_MiscFileIndent -= MISC_INDENTAMOUNT;
	_MiscDoIndent(aStream);
	NXPutc(aStream, '}');
	return self;
}

- writeInnerASCIIToStream:(NXStream *)aStream
{
	NXHashState state = [self initState];
	const void *key;
	void *value;
	
	while ([self nextState:&state key:&key value:&value]) {
	// don't write = half if key==object.  Should only write key.
		_MiscDoIndent(aStream);
		[(id)key writeASCIIToStream:aStream];
		if ([(id)key compareTo:(id)value] != 0) {
			NXPutc(aStream, ' ');
			NXPutc(aStream, '=');
			NXPutc(aStream, ' ');
			[(id)value writeASCIIToStream:aStream];
		}
		NXPutc(aStream, ';');
		NXPutc(aStream, '\n');
	}
	return self;
}

@end

@implementation MiscString(TableParse)

- quotedParseFromASCIIStream:(NXStream *)aStream
{
	char next;
	_MiscEatChar(aStream, '"');
	next = _MiscPeekCharFromStream(aStream);
	while (next != '"') {
		if (next == '\\') { // backslash quoted parse
			_MiscEatChar(aStream, '\\');
			next = _MiscPeekCharFromStream(aStream);
			if (next == '0') { // decode octal or hex
				_MiscEatChar(aStream, '0');
				next = _MiscPeekCharFromStream(aStream);
				if ((next == 'x') || (next == 'X')) { // decode hex
					int theHex = 0;
					_MiscEatChar(aStream, next);
					while (NXIsXDigit(_MiscPeekCharFromStream(aStream))) {
						char theChar =
							NXToUpper(_MiscGetCharFromStream(aStream));
						theHex *= 16;
						if (theChar >= 'A') theHex += theChar - 'A' + 10;
						else theHex += theChar - '0';
					}
					[self addChar:(char)theHex];
				} else { // decode octal
					int theOctal = 0;
					while (NXIsDigit(_MiscPeekCharFromStream(aStream)) &&
							(_MiscPeekCharFromStream(aStream) < '8')) {
						char theChar = _MiscGetCharFromStream(aStream);
						theOctal *= 8;
						theOctal += theChar - '0';
					}
					[self addChar:(char)theOctal];
				}
			} else if (NXIsDigit(next)) { // decode octal; used to be decimal
				int theDecimal = 0;
				while (NXIsDigit(_MiscPeekCharFromStream(aStream))
						// The next line is for octal only...remove for dec.
						&& (_MiscPeekCharFromStream(aStream) < '8')) {
					char theChar = _MiscGetCharFromStream(aStream);
					//theDecimal *= 10;
					theDecimal *= 8; // looks like ANSI dictates octal here
					// so we will go with that.  And pray it works right!  
					theDecimal += theChar - '0';
				}
				[self addChar:(char)theDecimal];
			} else {
				switch (next) {
					case 't' : {
						[self addChar:'\t'];
						_MiscGetCharFromStream(aStream);
						break;
					}
					case 'n' : {
						[self addChar:'\n'];
						_MiscGetCharFromStream(aStream);
						break;
					}
					case 'r' : {
						[self addChar:'\r'];
						_MiscGetCharFromStream(aStream);
						break;
					}
					case '"' :
					case '\\' : {
						[self addChar:_MiscGetCharFromStream(aStream)];
						break;
					}
					default  : { // error: unrec. sequence.
						fprintf(stderr,
"Error:  Unknown backslash quote \"\\%c\" used at line %d, position %d.\n",
								next, _MiscFileLine, _MiscFileLineChar);
						break; // for now, ignore it.
					}
				}
			}
		} else { // regular character, add to end of string.
			[self addChar:_MiscGetCharFromStream(aStream)];
		}
		next = _MiscPeekCharFromStream(aStream);
	}
	_MiscEatChar(aStream, '"');
	_MiscEatWhiteSpace(aStream);
	return self;
}

- plainParseFromASCIIStream:(NXStream *)aStream
{
	char pk = _MiscPeekCharFromStream(aStream);
	//( allow any char but whitespace, ';', ',', ')' and '"' to be in string.
	while (NXIsGraph(pk) && (pk != ';') && (pk != /*(*/')') && (pk != ',') && (pk != '"')) {
		[self addChar:_MiscGetCharFromStream(aStream)];
		pk = _MiscPeekCharFromStream(aStream);
	}
	return self;
}

- parseFromASCIIStream:(NXStream *)aStream
{
	_MiscEatWhiteSpace(aStream);
	if (_MiscPeekCharFromStream(aStream) == '"')
		[self quotedParseFromASCIIStream:aStream];
	else [self plainParseFromASCIIStream:aStream];
	return self;
}

- writeASCIIToStream:(NXStream *)aStream
{
	// we quote on more strings than strictly need it.  We should only quote
	// strings with whitespace, ';', and '"' in them.
	if ([self isAllOfType:MISC_ALNUM]) {
		NXPrintf(aStream, "%s", [[self stringValue] cString]);
	} else {	// write out a quoted string.
		int i;
		NXPutc(aStream, '"');
		for (i=0; i<length; i++) {
			if (NXIsCntrl(buffer[i]) || (buffer[i] == '"') ||
					(buffer[i] == '\\') || (buffer[i] < 0)) {
				NXPutc(aStream, '\\');
				switch (buffer[i]) {
					case '"'  : {
						NXPutc(aStream, '"'); break;
					}
					case '\\' : {
						NXPutc(aStream, '\\'); break;
					}
					case '\n' : {
						NXPutc(aStream, 'n'); break;
					}
					case '\t' : {
						NXPutc(aStream, 't'); break;
					}
					case '\r' : {
						NXPutc(aStream, 'r'); break;
					}
					default   : { // write char as an octal number
						// donated fix for 8-bit cleanliness -- R. Todd Thomas
					    int octalNum = (int)buffer[i] + 256;
					    NXPrintf(aStream, "%o", octalNum);
						//NXPrintf(aStream, "0%o", (unsigned char)(buffer[i]));
						break;
					}
				}
			} else NXPutc(aStream, buffer[i]);
		}
		NXPutc(aStream, '"');
	}
	return self;
}

@end
