//
//	MiscSelectionMatrix -- a class to make the Matrix more functional
//		Written by Georg Tuparev (c) 1994, 1995 by Georg Tuparev, 
// 		EMBL & Academia Naturalis, Heidelberg, Germany.
//				Version 0.5.1 beta  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//		This is a free object!  Contact the author for the latest version.
//		Georg Tuparev, EMBL / BioComputing, Meyerhofstr. 1, 
//			D-69117 Heidelberg, Germany
//			Phone: +49 - 6221 - 387528
//			Fax:   +49 - 6221 - 387517
//		e-mail:  Tuparev@EMBL-Heidelberg.de
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import <misckit/MiscSelectionMatrix.h>
#import <objc/objc-runtime.h>

#define CLASS_VERSION 		0

@implementation MiscSelectionMatrix

static id localList = nil;

/*****************************************************************************
 *** Local methods
 *****************************************************************************/
- (void)freeSelectionList
{
	if (selectionOrder) {
		[selectionOrder empty];
		[selectionOrder release];
		selectionOrder = nil;
	}
}

- (void)prepareSelectionList
{
	if (selectionOrder) {
		[selectionOrder empty];
	}
	else {
		selectionOrder = [[List allocWithZone:[self zone]] initCount:10];
	}
}

- (void)syncSelection
{
	int i;
	BOOL changed;
	
	[super getSelectedCells:localList];
	
	for (i = 0; i < [localList count]; i++) {
		if ([selectionOrder indexOf:[localList objectAtIndex:i]] == NSNotFound) {
			[selectionOrder addObject:[localList objectAtIndex:i]];
		}
	}
	
	do {
		changed = NO;
		for (i = 0; i < [selectionOrder count]; i++) {
			if ([localList indexOfObject:[selectionOrder objectAt:i]]==NSNotFound) {
				[selectionOrder removeObject:[selectionOrder objectAt:i]];
				changed = YES;
			}
		}
	} while (changed);
	[localList removeAllObjects];
}

- (void)copySelList:fromList
{
	if (selectionOrder)
		[selectionOrder release];
	selectionOrder = [fromList copyWithZone:[self zone]];
}

/*****************************************************************************
 *** Overwritten methods
 *****************************************************************************/
+ (void)initialize
// Set the class version
{
#warning NSNameConversion:  Class names have changed. Check that the class name passed in is real.
	if (self == objc_lookUpClass([[MiscSelectionMatrix name] cString]))  {
		[self setVersion:CLASS_VERSION];
	}
	// Alloc the local list
	localList = [[List alloc] init];
	
	return;
}

// Designated initializer override from Matrix.  Sets up our selection stuff.
- initWithFrame:(NSRect)frm mode:(int)aMode prototype:cellId numberOfRows:(int)rowsHigh numberOfColumns:(int)colsWide
{
	[super initWithFrame:frm mode:aMode prototype:cellId numberOfRows:numRows numberOfColumns:numCols];
	
	// Sets up the selection list
	[self prepareSelectionList];
	
	return self;
}

// Designated initializer override from Matrix.  Sets up our selection stuff.
- initWithFrame:(NSRect)frm mode:(int)aMode cellClass:factoryId numberOfRows:(int)rowsHigh numberOfColumns:(int)colsWide
{
	[super initWithFrame:frm mode:aMode cellClass:factoryId numberOfRows:numRows numberOfColumns:numCols];
	
	// Sets up the selection list
	[self prepareSelectionList];
	
	return self;
}

- copyWithZone:(NSZone *)zone
{
	id obj = [super copyWithZone:zone];
	// Copy the selection list
	[obj copySelList:selectionOrder];
	return obj;
}

- (void)dealloc
{
	[self freeSelectionList];            // Frees the selection
	{ [super dealloc]; return; };
}

- (void)mouseDown:(NSEvent *)theEvent 
{
	// $$$ implement the Control Dragging stuff
	
	[super mouseDown:theEvent];
	
	[self syncSelection];
}

- (id)initWithCoder:(NSCoder *)aDecoder
{
	int classVersion;

	[super initWithCoder:aDecoder];
	
	classVersion = [aDecoder versionForClassName:[MiscSelectionMatrix name]];
	
	switch (classVersion)  {
		case 0:		// First version.
			selectionOrder = [[aDecoder decodeObject] retain];
			break;
		default:
			NSLog([NSString stringWithCString:"[%s read:] class version %d cannot read "
						"instances archived with version %d"], [[MiscSelectionMatrix name] cString], CLASS_VERSION, classVersion);
			[self prepareSelectionList];
			break;
	}
	return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
{
	[super encodeWithCoder:aCoder];
	[aCoder encodeObject:selectionOrder];
}

/*****************************************************************************
 *** New interface methods
 *****************************************************************************/
- setReorderActionEnabled:(BOOL)aFlag
{
	// $$$
	return self;
}

- (BOOL)reorderActionEnabled
{
	// $$$
	return NO;
}

- addSelectionForCell:aCell
{
	int cellNumb;
	
	if ([cellList indexOf:aCell] == NSNotFound)
		return nil;
		
	if (aCell && ([selectionOrder indexOf:aCell] == NSNotFound)) {
		if ( (cellNumb = [cellList indexOf:aCell]) != NSNotFound) {
			[self setSelectionFrom:cellNumb to:cellNumb anchor:cellNumb highlight:YES];
			[selectionOrder addObject:aCell];
		}
	}
	return aCell;
}

- addSelectionAt:(int)row :(int)col
{
	id theCell = [self cellAtRow:row column:col];
	
	return [self addSelectionForCell:theCell];
}

- addSelectionForRow:(int)row
{
	int i;
	int rows = 0;
	int cols = 0;
	id theCell;
	
	[self getNumberOfRows:&rows columns:&cols];
	if ((row >= 0) && (row < rows)) {
		for (i = row*cols; i < row*cols+cols; i++) {
			theCell = [cellList objectAt:i];
			[self addSelectionForCell:theCell];
		}
		return self;
	}
	else
		return nil;
}

- addSelectionForColumn:(int)col
{
	int i;
	int rows = 0;
	int cols = 0;
	id theCell;

	[self getNumberOfRows:&rows columns:&cols];
	if ((col >= 0) && (col < cols)) {
		for (i = col; i < rows*cols; i+=cols) {
			theCell = [cellList objectAt:i];
			[self addSelectionForCell:theCell];
		}
		return self;
	}
	else 
		return nil;
}

- addSelectionForCellWithTag:(int)aTag
{
	id theCell = [self cellWithTag:aTag];
	
	return [self addSelectionForCell:theCell];
}

- deselectCell:aCell
{
	int cellNumb;
	
	if ([cellList indexOf:aCell] == NSNotFound)
		return nil;
		
	if (aCell && ([selectionOrder indexOf:aCell] != NSNotFound)) {
		cellNumb = [cellList indexOf:aCell];
		[self setSelectionFrom:cellNumb to:cellNumb anchor:cellNumb highlight:NO];
		[selectionOrder removeObject:aCell];
	}
	return aCell;
}

- deselectCellAt:(int)row :(int)col
{
	id theCell = [self cellAtRow:row column:col];
	
	return [self deselectCell:theCell];
}

- deselectRow:(int)row
{
	int i;
	int rows = 0;
	int cols = 0;

	[self getNumberOfRows:&rows columns:&cols];
	if ((row >= 0) && (row < rows)) {
		for (i = 0; i < cols; i++) {
			[self deselectCellAt:row:i];
		}
		return self;
	}
	else
		return nil;
}

- deselectColumn:(int)col
{
	int i;
	int rows = 0;
	int cols = 0;

	[self getNumberOfRows:&rows columns:&cols];
	if ((col >= 0) && (col < cols)) {
		for (i = 0; i < rows; i++) {
			[self deselectCellAt:col:i];
		}
		return self;
	}
	else
		return nil;
}

- deselectCellWithTag:(int)aTag
{
	id theCell = [self cellWithTag:aTag];
	
	return [self deselectCell:theCell];
}

- getOrderedSelectedCells:(List *)aList
{
	int i;

	[self syncSelection];
	if (!aList) 
		aList = [[List alloc] init];
	for (i = 0; i < [selectionOrder count]; i++)
		[aList addObject:[selectionOrder objectAt:i]];
	return aList;
}


@end

