//
//  MiscSplitView.m
//
//  Written by Dwight D. Everhart, based on SideSplitView by Mary McNabb
//  of NeXT (from the ZooView mini-example).
//
//  Copyright (c) 1996 by Dwight D. Everhart.  All rights reserved.
//  Version 1.11; February 7, 1996
//
//      This notice may not be removed from this source code.
//
//  This object is included in the MiscKit by permission from the author
//  and its use is governed by the MiscKit license, found in the file
//  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//  for a list of all applicable permissions and restrictions.
//
//  Although this code started as SideSplitView, there is little that remains
//  of the original code.  Still, the author is grateful to Ms. McNabb for
//  providing a starting point for this code.  This is the distribution notice
//  that she put in her code:
//
//    You may freely copy, distribute, and reuse the code in this example.
//    NeXT disclaims any warranty of any kind, expressed or implied, as to its
//    fitness for any particular use.
//
//  Purpose:
//    This class implements a view which divides its frame, either vertically
//    or horizontally, into two or more adjacent areas in which are displayed
//    its subviews.  The user can change the proportion of the areas occupied
//    by the subviews relative to each other.  The class can also save and
//    restore its subview divisions to a string or a default.  The interface
//    is compatible with NXSplitView, and the divisions saving mechanism has
//    the same style of interface as Window's frame saving mechanism.
//
//  Requires:
//    MiscAppDefaults category
//    Dimple.tiff image (embedded into Dimple.m by Don Yacktman)
//

#import <misckit/MiscSplitView.h>
#import <misckit/MiscAppDefaults.h>

#include "Dimple_tiff.m"   /* contains Dimple.tiff */

// Retrace is 1/67 of sec * 10^6 micro secs per second.
#define RETRACE	(1000000 / 67)  // vertical retrace period in microsecs
#define MISC_SPLIT_VIEW_VERSION 1
#define DIVIDER_SIZE 8.0


#define nameToDefault(name,defaultName) \
   sprintf ((defaultName), "MiscSplitView Dividers %s", (name))


static NSImage *dimple;
static NSSize dimpleSize;
static HashTable *autosaveNames;



static void freeClassData (void)
{
   /*
     Don't free the dimple.  That'll generate an IOT. Something else frees it.
   */
   [autosaveNames release];
}



@implementation MiscSplitView



+ (void)initialize
{
   if (![super initialize]) return nil;

   if (self != [MiscSplitView class]) return;

   /* Set the class's version. */
   [self setVersion: MISC_SPLIT_VIEW_VERSION];

   /* Initialize random number generator for vertical retrace sleeping. */
   srandom (time (0));

   /* Get dimple image.  Modified by Don Yacktman--put tiff into a .m file */
   dimple = Dimple_tiff_fn();
   /* dimple = [NXImage findImageNamed: "Dimple"]; */
   dimpleSize = [dimple size];

   /* Create autosave name hash table. */
   autosaveNames = [[HashTable allocWithZone:[self zone]]
		    initKeyDesc: "@" valueDesc: "%"];
   atexit (freeClassData);
   return;
}



+ (BOOL) autosaveNameExists: (NXAtom) name
{
   NXHashState state = [autosaveNames initState];
   const void *key;
   void *value;

   while ([autosaveNames nextState: &state key: &key value: &value]) {
      if ((NXAtom) value == name) return YES;
   }

   return NO;
}



- initWithFrame:(NSRect)frameRect
{
   if (![super initWithFrame:frameRect]) return nil;
#error ViewConversion: '[NSView setFlipped:]' is obsolete; you must override 'isFlipped' instead of setting externally. However, [NSImage setFlipped:] is not obsolete. If that is what you are using here, no change is needed.
   [self setFlipped:YES];
   [self setAutoresizesSubviews:YES];
   return self;
}



- delegate
{
   return delegate;
}



- (void)setDelegate:(id)anObject
{
   delegate = anObject;
}



- (BOOL) isHorizontal
{
   return msvFlags.horizontal;
}



- setHorizontal: (BOOL) makeHorizontal
{
   if (msvFlags.horizontal != makeHorizontal) {
      msvFlags.horizontal = makeHorizontal;
      msvFlags.needsRetiling = YES;
   }

   return self;
}



- setDividersFromString: (const char *) data
{
   NSRect viewFrame;
   float newValue;
   NSView *view;
   int i, version;

   if (!data) return nil;

   /*
     Get our class version from the default in case
     we want to change its format later.
   */
   data += strspn (data, " \t");
   version = atoi (data);
   data += strspn (data, "0123456789");

   /*
     There's no need to set the origin values,
     since they will be set by resizeSubviews:.
   */
   for (i = 0; view = [subviews objectAt: i]; i++) {
      viewFrame = [view frame];
      data += strspn (data, " \t");
      newValue = atoi (data);
      if (newValue < 1) break;
      data += strspn (data, "0123456789");
      if (msvFlags.horizontal) {
	 viewFrame.size.width  = newValue;
      } else {
	 viewFrame.size.height = newValue;
      }
      [view setFrame:viewFrame];
   }

   [self resizeSubviewsWithOldSize:bounds.size];
   return self;
}



- saveDividersToString: (char *) string
{
   NSRect viewFrame;
   float value;
   NSView *view;
   int i;

   if (!string) return nil;

   sprintf (string, "%d ", [MiscSplitView version]);

   for (i = 0; view = [subviews objectAt: i]; i++) {
      viewFrame = [view frame];
      string = &string[strlen (string)];
      if (msvFlags.horizontal) {
	 value = viewFrame.size.width;
      } else {
	 value = viewFrame.size.height;
      }
      sprintf (string, "%.0f ", value);
   }

   return self;
}



- setDividersUsingName: (const char *) name
{
   const char *divString;
   char defaultName[64];

   if (!name) return nil;

   nameToDefault (name, defaultName);
   divString = [NSApp defaultValue: defaultName];
   if (!divString) return nil;

   [self setDividersFromString: divString];
   return self;
}



- saveDividersUsingName: (const char *) name
{
   char divString[[subviews count] * 8];
   char defaultName[64];

   if (!name) return nil;

   if ([self saveDividersToString: divString]) {
      nameToDefault (name, defaultName);
      [NSApp setDefault: defaultName to: divString];
      return self;
   } else {
      return nil;
   }
}



+ removeDividersUsingName: (const char *) name
{
   char defaultName[64];

   [autosaveNames removeKey: (const void *) NXUniqueString (name)];
   nameToDefault (name, defaultName);
#warning DefaultsConversion: [<NSUserDefaults> removeObjectForKey:...] used to be NXRemoveDefault([[[NSProcessInfo processInfo] processName] cString], defaultName). Defaults will be synchronized within 30 seconds after this change.  For immediate synchronization, call '-synchronize'. Also note that the first argument of NXRemoveDefault is now ignored; to write into a domain other than the apps default, see the NSUserDefaults API.
   [[NSUserDefaults standardUserDefaults] removeObjectForKey:[NSString stringWithCString:defaultName]];
   return self;
}



- (const char *) dividersAutosaveName
{
   return [autosaveNames valueForKey: (const char *) self];
}



- setDividersAutosaveName: (const char *) name
{
   NXAtom saveName;

   if (!name) {
      [autosaveNames removeKey: (const void *) self];
      return self;
   } else {
      saveName = NXUniqueString (name);
      if ([MiscSplitView autosaveNameExists: saveName]) {
	 return nil;
      } else {
	 [autosaveNames insertKey: (const void *) self
	  value: (void *) saveName];
	 [self setDividersUsingName: name];
	 return self;
      }
   }
}



- (float) dividerSize
{
   return DIVIDER_SIZE;
}



- (float)dividerThickness
{
   return [self dividerSize];
}



- (float) dividerWidth
{
   return [self dividerSize];
}



#warning RectConversion: drawRect:(NSRect)rects (used to be drawSelf:(const NXRect *)rects :(int)rectCount) no longer takes an array of rects
- (void)drawRect:(NSRect)rects
{
   float dividerSize;
   NSRect dividerFrame;
   NSView *view;
   int i, r;

   if (msvFlags.needsRetiling) [self resizeSubviewsWithOldSize:bounds.size];

   if (msvFlags.horizontal) {
      dividerSize = [self dividerWidth];
   } else {
      dividerSize = [self dividerThickness];
   }

   for (r = (rectCount == 1) ? 0 : 1; r < rectCount; r++) {
      for (i = 1; view = [subviews objectAt: i]; i++) {
	 dividerFrame = [view frame];
	 if (msvFlags.horizontal) {
	    dividerFrame.origin.x   -= dividerSize;
	    dividerFrame.size.width  = dividerSize;
	 } else {
	    dividerFrame.origin.y   -= dividerSize;
	    dividerFrame.size.height = dividerSize;
	 }

	 if (!NSIsEmptyRect(NSIntersectionRect(dividerFrame , &rects[r]))) {
	    [self drawDividerInRect:dividerFrame];
	 }
      }
   }
}



- (void)drawDividerInRect:(NSRect)aRect;
{
   NSPoint dimplePoint;

   PSsetgray (NSLightGray);
   NSRectFill(*aRect);

   /*
     The y coordinate is calculated differently
     because the view's coordinate system is flipped.
   */
   dimplePoint.x = floor (aRect->origin.x + (aRect->size.width  / 2.0) -
			  (dimpleSize.width  / 2.0));
   dimplePoint.y = floor (aRect->origin.y + (aRect->size.height / 2.0) +
			  (dimpleSize.height / 2.0));

   [dimple compositeToPoint:dimplePoint operation:NSCompositeSourceOver];
   return self;
}



- constrainDivider: (int) dividerIndex
  leftFrame:  (const NSRect *) leftFrame
  rightFrame: (const NSRect *) rightFrame
  min: (float *) min max: (float *) max
{
   float halfSize = [self dividerSize] / 2;
   float newMin, newMax;

   if (!(leftFrame && rightFrame)) return nil;

   /* Calculate how much the views can be resized. */
   if (msvFlags.horizontal) {
      *min = leftFrame->origin.x  + halfSize;
      *max = rightFrame->origin.x + rightFrame->size.width  - halfSize;
   } else {
      *min = leftFrame->origin.y  + halfSize;
      *max = rightFrame->origin.y + rightFrame->size.height - halfSize;
   }

   /* Ask the delegate to constrain the values. */
   if ((msvFlags.horizontal &&
	[delegate respondsToSelector:@selector (splitView:getMinX:maxX:ofSubviewAt:)]) ||
       (!msvFlags.horizontal &&
	[delegate respondsToSelector:@selector(splitView:constrainMinCoordinate:maxCoordinate:ofSubviewAt:)])) {

      newMin = *min;
      newMax = *max;

      if (msvFlags.horizontal) {
	 [delegate splitView: self getMinX: &newMin
	  maxX: &newMax ofSubviewAt: dividerIndex];
      } else {
	 [delegate splitView:self constrainMinCoordinate:&newMin maxCoordinate:&newMax ofSubviewAt:dividerIndex];
      }

      if ((newMin > *min) && (newMin <= *max)) *min = newMin;
      if ((newMax < *max) && (newMax >= *min)) *max = newMax;
      if (*max < *min) *max = *min;
   }

   return self;
}



- constrainFramesLeft: (NSRect *) leftFrame andRight: (NSRect *) rightFrame
  besideDivider: (int) dividerIndex
{
   float dividerSize = [self dividerSize];
   float halfSize = dividerSize / 2;
   float min, max, minEdge, maxEdge, newEdge;

   if (![self constrainDivider: dividerIndex leftFrame: leftFrame
	 rightFrame: rightFrame min: &min max: &max]) {
      return nil;
   }

   minEdge = min + halfSize;
   maxEdge = max + halfSize;
   newEdge = 0.0;

   if (msvFlags.horizontal) {
      if (minEdge > rightFrame->origin.x) {
	 newEdge = minEdge;
      } else if (maxEdge < rightFrame->origin.x) {
	 newEdge = maxEdge;
      }

      if (newEdge > 0.0) {
	 leftFrame->size.width =
	    newEdge - leftFrame->origin.x - dividerSize;
	 rightFrame->size.width = rightFrame->origin.x +
	    rightFrame->size.width - newEdge;
	 rightFrame->origin.x = newEdge;
      }
   } else {
      if (minEdge > rightFrame->origin.y) {
	 newEdge = minEdge;
      } else if (maxEdge < rightFrame->origin.y) {
	 newEdge = maxEdge;
      }

      if (newEdge > 0.0) {
	 leftFrame->size.height =
	    newEdge - leftFrame->origin.y - dividerSize;
	 rightFrame->size.height = rightFrame->origin.y +
	    rightFrame->size.height - newEdge;
	 rightFrame->origin.y = newEdge;
      }
   }

   return self;
}



- (void)adjustSubviews
{
   float dividerSize, oldSize, p;
   NSRect viewFrame, leftFrame, rightFrame;
   NSView *view, *leftView, *rightView;
   int lastView, i;

   if (msvFlags.horizontal) {
      dividerSize = [self dividerWidth];
   } else {
      dividerSize = [self dividerThickness];
   }

   oldSize = -dividerSize;
   for (i = 0; view = [subviews objectAt: i]; i++) {
      viewFrame = [view frame];
      if (msvFlags.horizontal) {
	 oldSize += viewFrame.size.width  + dividerSize;
      } else {
	 oldSize += viewFrame.size.height + dividerSize;
      }
   }

   lastView = i - 1;
   leftView = nil;
   p = msvFlags.horizontal ? bounds.origin.x : bounds.origin.y;

   for (i = 0; rightView = [subviews objectAt: i]; i++) {
      if (msvFlags.horizontal) {
	 rightFrame.origin.x    = p;
	 rightFrame.origin.y    = bounds.origin.y;
	 rightFrame.size.height = bounds.size.height;
      } else {
	 rightFrame.origin.x    = bounds.origin.x;
	 rightFrame.origin.y    = p;
	 rightFrame.size.width  = bounds.size.width;
      }

      if (i < lastView) {
	 viewFrame = [rightView frame];
	 if (msvFlags.horizontal) {
	    rightFrame.size.width =
	       rint (bounds.size.width  * (viewFrame.size.width  / oldSize));
	 } else {
	    rightFrame.size.height =
	       rint (bounds.size.height * (viewFrame.size.height / oldSize));
	 }
      } else {
	 if (msvFlags.horizontal) {
	    rightFrame.size.width  = bounds.size.width  - p;
	 } else {
	    rightFrame.size.height = bounds.size.height - p;
	 }
      }

      if (leftView) {
	 [self constrainFramesLeft: &leftFrame andRight: &rightFrame
	  besideDivider: i - 1];
	 [leftView setFrame:leftFrame];
	 [window invalidateCursorRectsForView:leftView];
      }

      if (msvFlags.horizontal) {
	 p = rightFrame.size.width  + dividerSize;
      } else {
	 p = rightFrame.size.height + dividerSize;
      }

      leftView  = rightView;
      leftFrame = rightFrame;
   }

   /* Set the final view if there was more than one. */
   if (lastView > 0) {
      [leftView setFrame:leftFrame];
      [window invalidateCursorRectsForView:leftView];
   }

   [self saveDividersUsingName: [self dividersAutosaveName]];
   msvFlags.needsRetiling = NO;
}



/* called when the view gets resized */

- (void)resizeSubviewsWithOldSize:(NSSize)oldSize
{
   if ([delegate respondsToSelector:@selector(splitView:resizeSubviewsWithOldSize:)]) {
      [delegate splitView:self resizeSubviewsWithOldSize:oldSize];
      [self saveDividersUsingName: [self dividersAutosaveName]];
      msvFlags.needsRetiling = NO;
   } else {
      [self adjustSubviews];
   }

#warning NotificationConversion: 'splitViewDidResizeSubviews:' is an NSSplitView notification method (used to be a delegate method); delegates of NSSplitView are automatically set to observe this notification; subclasses of NSSplitView do not automatically receive this notification
   if ([delegate respondsToSelector:@selector(splitViewDidResizeSubviews:)]) {
      [[NSNotificationCenter defaultCenter] postNotificationName:NSSplitViewDidResizeSubviewsNotification object:self];
   }
}



- (void)setAutoresizesSubviews:(BOOL)flag
{
   /* Don't allow anyone to disable autoresize for us. */
   [super setAutoresizesSubviews:YES];
}



- addSubview: aView
{
   msvFlags.needsRetiling = YES;
   return [super addSubview: aView];
}



- addSubview:aView positioned:(int)place relativeTo:otherView
{
   msvFlags.needsRetiling = YES;
   return [super addSubview:aView positioned:place relativeTo:otherView];
}



- (void)replaceSubview:(NSView *)oldView with:(NSView *)newView
{
   msvFlags.needsRetiling = YES;
   return [super replaceSubview:oldView with:newView];
}



#warning ViewConversion: 'acceptsFirstMouse:' (used to be 'acceptsFirstMouse') now takes the event as an arg
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
   return YES;
}



/* controls the movement of the divider as the user slides it with the mouse */

- (void)mouseDown:(NSEvent *)theEvent 
{
   NSPoint location = [theEvent locationInWindow];
   float dividerSize, halfSize;
   float min, max, last, new, p;
   NSEvent *loopEvent;
   NSRect leftFrame, rightFrame;
   NSView *leftView, *rightView;
   int i, oldMask;

   if (msvFlags.horizontal) {
      dividerSize = [self dividerWidth];
   } else {
      dividerSize = [self dividerThickness];
   }

   halfSize = dividerSize / 2;

   /* Figure out which divider is being moved. */
   location = [self convertPoint:location fromView:nil];
   for (i = 1; rightView = [subviews objectAt: i]; i++) {
      rightFrame = [rightView frame];
      if ((msvFlags.horizontal && (location.x < rightFrame.origin.x) &&
	   (location.x >= rightFrame.origin.x - dividerSize)) ||
	  (!msvFlags.horizontal && (location.y < rightFrame.origin.y) &&
	   (location.y >= rightFrame.origin.y - dividerSize))) {
	 i--;
	 break;
      }
   }

   if (!rightView) {[super mouseDown:theEvent]; return;}

   /* Set the limits of the new divider position. */
   leftView = [subviews objectAt: i];
   leftFrame = [leftView frame];
   [self constrainDivider: i leftFrame: &leftFrame
    rightFrame: &rightFrame min: &min max: &max];

   /* Prepare for drawing. */
   [self lockFocus];
   PSsetinstance (YES);
   PSsetgray (NSDarkGray);
   PSsetalpha (0.5);
   last = -1.0;
#error EventConversion: addToEventMask:NX_MOUSEDRAGGEDMASK: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
   oldMask = [window addToEventMask:NSLeftMouseDraggedMask];

   /*
     While the user is dragging, composite a slider outline in the new postion,
     watching for boundary conditions.  Exit the loop on mouse up.
   */
   for (loopEvent = theEvent;
	loopEvent && ([loopEvent type] != NSLeftMouseUp);
	loopEvent = [NSApp nextEventMatchingMask:NSLeftMouseUpMask | NSLeftMouseDraggedMask untilDate:[NSDate distantFuture] inMode:NSEventTrackingRunLoopMode dequeue:YES]) {
      location = [loopEvent locationInWindow];
      location = [self convertPoint:location fromView:nil];
      p = msvFlags.horizontal ? location.x : location.y;

      if (p < min) p = min;
      if (p > max) p = max;
      if (p == last) continue;
      last = p;
      p -= halfSize;

      PSnewinstance();
      if (msvFlags.horizontal) {
	 PScompositerect (p, bounds.origin.y,
			  dividerSize, bounds.size.height, NSCompositeSourceOver);
      } else {
	 PScompositerect (bounds.origin.x, p,
			  bounds.size.width, dividerSize, NSCompositeSourceOver);
      }

      /*
	To reduce flicker, wait some random
	amount of time within the retrace period.
      */
      [NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:(random() % RETRACE)/1000000.0]];
   }

   /* Finish drawing. */
   PSsetinstance (NO);
   [self unlockFocus];
#error EventConversion: setEventMask:oldMask: is obsolete; you no longer need to use the eventMask methods; for mouse moved events, see 'setAcceptsMouseMovedEvents:'
   [window setEventMask:oldMask];

   /* Resize the subviews. */
   [window disableFlushWindow];

   /* First the left. */
   leftFrame = [leftView frame];
   if (msvFlags.horizontal) {
      leftFrame.size.width  = last - halfSize - leftFrame.origin.x;
   } else {
      leftFrame.size.height = last - halfSize - leftFrame.origin.y;
   }
   [leftView setFrame:leftFrame];

   /* Then the right. */
   rightFrame = [rightView frame];
   new = last + halfSize;
   if (msvFlags.horizontal) {
      rightFrame.size.width =
	 rightFrame.origin.x + rightFrame.size.width - new;
      rightFrame.origin.x = new;
   } else {
      rightFrame.size.height =
	 rightFrame.origin.y + rightFrame.size.height - new;
      rightFrame.origin.y = new;
   }
   [rightView setFrame:rightFrame];

   /* Display the newly resized views. */
   [self display];
   [window enableFlushWindow];
   [window flushWindow];

   /* Update the default. */
   [self saveDividersUsingName: [self dividersAutosaveName]];

   /* Inform the delegate of the change. */
#warning NotificationConversion: 'splitViewDidResizeSubviews:' is an NSSplitView notification method (used to be a delegate method); delegates of NSSplitView are automatically set to observe this notification; subclasses of NSSplitView do not automatically receive this notification
   if ([delegate respondsToSelector:@selector(splitViewDidResizeSubviews:)]) {
      [[NSNotificationCenter defaultCenter] postNotificationName:NSSplitViewDidResizeSubviewsNotification object:self];
   }
}



- (void)encodeWithCoder:(NSCoder *)aCoder
{
   unsigned char horizontal = msvFlags.horizontal;

   if (![super encodeWithCoder:aCoder]) return nil;
   [aCoder encodeValueOfObjCType:"c" at:&horizontal];
   [aCoder encodeConditionalObject:delegate];
}



- (id)initWithCoder:(NSCoder *)aDecoder
{
   unsigned char horizontal;
   int version;

   if (![super initWithCoder:aDecoder]) return nil;

   version = [aDecoder versionForClassName:[[MiscSplitView class] name]];

   [aDecoder decodeValueOfObjCType:"c" at:&horizontal];
   msvFlags.horizontal = horizontal;

   if (version > 0) delegate = [[aDecoder decodeObject] retain];
   return self;
}



@end

