//
//	MiscStringNEXTSTEP.m
//		Written by Don Yacktman Copyright (c) 1993 by Don Yacktman.
//				Version 1.95  All rights reserved.
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

#import <misckit/MiscString.h>
#import <mach/mach_error.h>

@implementation MiscString(NEXTSTEP)

// This category includes methods to support various NEXTSTEP features,
// such as InterfaceBuilder (palettes), Distributed Objects (NXTransport),
// and object archiving.

- (id)initWithCoder:(NSCoder *)aDecoder
{
	int tempLength; char *tBuf;
	[super initWithCoder:aDecoder];
	
	// find out how much space we will need
	[aDecoder decodeValueOfObjCType:"i" at:&tempLength];
	
	// note that -allocateBuffer:fromZone: will set _length for
	// us -- it is important to not smash _length since it tracks
	// the actual size of the current buffer!  (See the notes for
	// -allocateBuffer:fromZone: for more info on why no smashing.)
	// Note that -allocateBuffer:fromZone: increments the buffer
	// size by one, so we must pass it decremented so that we unarchive
	// with a buffer of the same length as the original object.
	[self allocateBuffer:(tempLength-1) fromZone:[self zone]]; // make sure
	// we have a buffer and it is big enough.
	
	// Get the string off the stream and into the buffer
	[aDecoder decodeValueOfObjCType:"*" at:&tBuf];
	if (tBuf) strcpy(buffer, tBuf);
	if (!tBuf && buffer) buffer[0] = '\0'; // NULL string?
	
	// now make sure that we have the right length for the string stored.
	if (buffer) length = strlen(buffer);
	else length = 0;
	
	// clean up and blow this joint
	if (tBuf) free(tBuf);
	return self;
}

- (void)encodeWithCoder:(NSCoder *)aCoder
{
	[super encodeWithCoder:aCoder];
	// put the string and it's length on the stream.  The length is sort
	// of redundant, but this assures that the string really is like the
	// one that was archived -- if the stored instance had an oversize
	// buffer, then read will restore an oversize buffer.  Of course,
	// crud after the string won't be the same, so the new instance isn't
	// quite the same, but...
	[aCoder encodeValueOfObjCType:"i" at:&_length];
	[aCoder encodeValueOfObjCType:"*" at:&buffer];
}

// NXTransport protocol implementation (next three methods):
- (void)encodeWithCoder:(NSPortCoder *)aPortCoder
{
	[aPortCoder encodeValueOfObjCType:"i" at:&_length]; // no longer used, remains for compatability reasons
	[aPortCoder encodeValueOfObjCType:"i" at:&length];
	[aPortCoder encodeValueOfObjCType:"*" at:&buffer];
}

- (id)initWithCoder:(NSPortCoder *)aPortCoder
{
	int newLen;
	[self freeString];
	[aPortCoder decodeValueOfObjCType:"i" at:&newLen]; // no longer used, remains for compatability reasons
	// This is not necessary and causes a leak:  [self allocateBuffer:newLen];
	[aPortCoder decodeValueOfObjCType:"i" at:&length];
	[aPortCoder decodeValueOfObjCType:"*" at:&buffer];
	// Yes, I know I'm not supposed to fool with _length, but....
	_length = length + 1; // it needs to be set properly and this ought to
	// do the trick.  Note that the bycopy object doesn't necessarily have
	// a buffer as large as the original.  We don't really care since they
	// grow dynamically anyway and most strings passed through DO will end
	// up being treated as constants anyway, so the trimming down is OK.
	return self;
}

#warning DO Conversion: if freeAfterEncoding was true use autorelease instead.
- replacementObjectForPortCoder:(NSPortCoder *)aPortCoder
{
	if ([aPortCoder isBycopy]) {
		*flagp = NO; // object will not be freed (ie, copy,  not move)
		return self; //encode object (copy it)
	}
	*flagp = NO;
	// super will encode the proxy otherwise
#warning DO Conversion: if freeAfterEncoding was true use autorelease instead.
	return [super replacementObjectForPortCoder:[aPortCoder connection]];
}

// RTF support
- (BOOL)isRTFText
{
	if (![self cmp:"{\\rtf0" n:6]) return YES; //} rtf header/magic num
	if (![self cmp:"{\\\\rtf0" n:7]) return YES; //} happens in .address files
	return NO;
}

- (MiscString *)plainTextForRTF
{ // Donated by Darcy Brockbank and untested.  Let's hope it works!
	id text;
	vm_size_t allocSize = (strlen(buffer)+1) * sizeof(char);
	char * tmp;
	NXStream * stream;
	MiscString * other = nil;
	kern_return_t ret;

	if (![self isRTFText]) return [self copy];
	text = [[NSText alloc] init];
	[text setString:@""];
	ret = (vm_allocate(task_self(), (vm_address_t *)&tmp, allocSize, TRUE));
	if (ret != KERN_SUCCESS) {
   		mach_error("vm_allocate returned value of ", ret);
		printf("Exiting with error.\n");
	    	exit(-1);
	}
	strcpy(tmp,buffer);
	stream = NXOpenMemory(tmp,allocSize,NX_READONLY);
#error TextConversion: 'replaceCharactersInRange:withRTF...' (used to be readRichText:) takes an NSData instance as its second argument. stream must be converted to NSData
	[text replaceCharactersInRange:NSMakeRange(0, [[text text] length]) withRTF:stream];
	NXCloseMemory(stream,NX_TRUNCATEBUFFER);
    ret = vm_deallocate(task_self(), (vm_address_t)tmp, allocSize);
	if (ret != KERN_SUCCESS) {
		mach_error("vm_deallocate returned value of ",ret);
		printf("Exiting with error.\n");
		exit(-1);
	}
	if ([[text text] length]) {
		int len = [[text text] length];
		char otherBuf[len + 1];
#warning TextConversion: 'substringFromRange:' used to be 'getSubstring:start:length:'; make sure the range is valid.
		otherBuf = [[text substringFromRange:NSMakeRange(0, len)] cString];
		otherBuf[len]='\0';
		other = [[MiscString alloc] initString:otherBuf];
	}
	[text release];
	return other;	// nil if we were unsuccessful
}

// Interface Builder support
- (const char *)getInspectorClassName { return "MiscStringInspector"; }
- (NSImage *)getIBImage { return [NSImage imageNamed:@"MiscStringObj"]; }

- writeToStream:(NXStream *)aStream
{
	NXPrintf(aStream,"%s", [[self stringValue] cString]);
	return self;
}

@end
