/*========================= MiscTargetActionView.m =========================*/
//		Written by Dale Amon, copyright 1994 by Genesis Project, Ltd.
//				Version 0.8.  All rights reserved.
//
//		This notice may not be removed from this source code.
//
//	This object is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	

/* Abstract superclass that impliments a target action paradeigm for Views.
   This is not quite the same as the Control Class because there is no 
   "value". It allows the view to accept keyboard and mouse events and convert
   them into target/action messages (potentially using a responder chain).
   Note that events in the actionMask will only be recieved if the
   corresponding events are enabled in the window's event mask.

HISTORY
1-Aug-93   Dale Amon at GPL
	   Created.

NOTE:	- what should awake method do?
	- should setFirstResponder remember the state?
		- handle to resign as well?
		- automatically reset view attaches to a window?

*/

#import	<misckit/misckit.h>

#define ACCEPTED_EVENTS (NSLeftMouseDownMask | NSLeftMouseUpMask | \
			 NSRightMouseDownMask | NSRightMouseUpMask | \
			 NSMouseMovedMask | NSLeftMouseDraggedMask | \
			 NSRightMouseDraggedMask | NSMouseEnteredMask | \
			 NSMouseExitedMask | NSKeyDownMask | NSKeyUpMask | \
			 NSFlagsChangedMask)

@implementation MiscTargetActionView

/*===========================================================================*/
/* Define localized strings */
/*===========================================================================*/
#define FIRST_RESPOND [NSLocalizedStringFromTable(@"Accept first responder", @"MiscKit", "Debug message. View recieved acceptFirstResponder message.") cString]

#define FIRST_MOUSE [NSLocalizedStringFromTable(@"Accept first mouse", @"MiscKit", "Debug message. View received acceptFirtsMouse message.") cString]

#define FLAGS_CHANGED [NSLocalizedStringFromTable(@"Flags Changed", @"MiscKit", "Debug message. View recieved flagsChanged message.") cString]

#define KEY_DOWN [NSLocalizedStringFromTable(@"Key down", @"MiscKit", "Debug message. View recieved keyDown message.") cString]

#define KEY_UP [NSLocalizedStringFromTable(@"Key up", @"MiscKit", "Debug message. View recieved keyUp message.") cString]

#define LEFT_MOUSE_DOWN [NSLocalizedStringFromTable(@"Left Mouse down", @"MiscKit", "Debug message. View recieved mouseDown message.") cString]

#define LEFT_MOUSE_DRAG [NSLocalizedStringFromTable(@"Left Mouse Dragged", @"MiscKit", "Debug message. View recieved mouseDragged message.") cString]

#define MOUSE_ENTERED [NSLocalizedStringFromTable(@"Mouse Entered", @"MiscKit", "Debug message. View recieved mouseEntered message.") cString]

#define MOUSE_EXITED [NSLocalizedStringFromTable(@"Mouse Exited", @"MiscKit", "Debug message. View recieved mouseExited message.") cString]

#define MOUSE_MOVED [NSLocalizedStringFromTable(@"Mouse Moved", @"MiscKit", "Debug message. View recieved mouseMoved message..") cString]

#define LEFT_MOUSE_UP [NSLocalizedStringFromTable(@"Left Mouse Up", @"MiscKit", "Debug message. View recieved mouseUp message.") cString]

#define RIGHT_MOUSE_DOWN [NSLocalizedStringFromTable(@"Right Mouse down", @"MiscKit", "Debug message. View recieved rightMouseDown message.") cString]

#define RIGHT_MOUSE_DRAG [NSLocalizedStringFromTable(@"Right Mouse Dragged", @"MiscKit", "Debug message. View recieved rightMouseDragged message.") cString]

#define RIGHT_MOUSE_UP [NSLocalizedStringFromTable(@"Right Mouse Up", @"MiscKit", "Debug message. View recieved rightMouseUp message.") cString]

/*===========================================================================*/
/* Class methods */
/*===========================================================================*/
/* Initialize the class */

+ (void)initialize
  {	[MiscTargetActionView setVersion:MISC_TAVIEW_VERSION_ID];
	return;
  }

/*===========================================================================*/
/* Internal Instance methods */
/*===========================================================================*/
/* The print name is the class name. */

- (const char *) _pname
  {	const char *who = [[[self class] name] cString];
	return who;
  }

/*---------------------------------------------------------------------------*/
/* Print a message to the log in the format:
	imagename: message
*/

- (void) _logMsg: (const char *) what 
  {NSLog(@"%s: %s", [self _pname],what);}


/*===========================================================================*/
/* Init, alloc and free methods */
/*===========================================================================*/
/* DEFAULT INITIALIZER. Create a window of the specified size. Action mask
   is empty and there is no target action declared.
*/

- initWithFrame:(NSRect)theRect
  {	[super initWithFrame:theRect];
#error EventCoversion: NX_NULLEVENTMASK is obsolete.
	actionMask	= NX_NULLEVENTMASK;
	target		= nil;
	action		= (SEL)0;
	actionEvent	= NULL;
	debug		= NO;

	return self;
  }


/*===========================================================================*/
/* Action mask and first responder handling */
/*===========================================================================*/
/* if any action mask bits for keyboard is set, accept first responder */

- (BOOL)acceptsFirstResponder
  {	int temp = actionMask & 
		(NSKeyDownMask | NSKeyUpMask | NSFlagsChangedMask);
	if (debug) [self _logMsg: FIRST_RESPOND];
	return (temp ? YES : NO);
  }

#warning ViewConversion: 'acceptsFirstMouse:' (used to be 'acceptsFirstMouse') now takes the event as an arg
- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent 
  {	if (debug && firstMouse) [self _logMsg: FIRST_MOUSE];
	return firstMouse;
  }

- setFirstMouse: (BOOL) theMouse {firstMouse = theMouse; return self;}

/*---------------------------------------------------------------------------*/
/* This must be done AFTER this view becomes part of a view hierarchy.
   Tells the window to make us the first responder
*/

- setFirstResponder {return ([[self window] makeFirstResponder:self]);}

/*---------------------------------------------------------------------------*/
/* set and clear action mask bits. Protects against setting bits that we don't
   respond to. Logic equivalent to MOV, BIC & BIS.
*/
	
- (int)addToActionMask:(int)newActions
  {	int	oldActions = actionMask;
	actionMask |= (newActions & ACCEPTED_EVENTS);
	return oldActions;
  }

- (int)actionMask {return actionMask;}

- (int)removeFromActionMask:(int)removeActions;
  {	int	oldActions = actionMask;
	actionMask &= ~(removeActions & ACCEPTED_EVENTS);
	return oldActions;
  }

- (int)setActionMask:(int)newActions;
  {	int	oldActions = actionMask;
	actionMask = newActions & ACCEPTED_EVENTS;
	return oldActions;
  }


/*===========================================================================*/
/* Set up for the target and action paradeigm. Note that this differs from
   Control classes in that there is no value. This is only a smarter handler
   for the responder chain.
*/

- (void)setAction:(SEL)theAction	{action = theAction;
}
- (void)setTarget:theTarget		{target = theTarget;
}
- (SEL) action			{return action;}
- target			{return target;}

- (NSEvent *)actionEvent {return (const NXEvent *) actionEvent;}


/*===========================================================================*/
/* Override all the Responder event methods */
/*===========================================================================*/
/* If the action is enabled in the actionMask, the target action is sent.
   If the action is not enabled, the target does not accept the action, or
   returns nil, then the Responder version is called instead.

   The target can find out what event triggered the action by:
	[sender actionEvent];

   Note that the actionEvent is only available during the execution of
   the action method.

*/

- _chk:(NSEvent *)evn msk: (int) mask msg: (const char*)msg
  {	actionEvent = evn;
	if (debug) [self _logMsg: msg];
	if ((actionMask & mask) && action &&
	    [self tryToPerform: action with: target])
	   {actionEvent = NULL; return self;}
	else 
	   {actionEvent = NULL; return nil;}
  }
 
- (void)flagsChanged:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSFlagsChangedMask	msg: FLAGS_CHANGED	]) ?
	self : [super flagsChanged:evn];}
	
- (void)keyDown:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSKeyDownMask	msg: KEY_DOWN		]) ?
	self : [super keyDown:evn];}

- (void)keyUp:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSKeyUpMask		msg: KEY_UP		]) ?
	self : [super keyUp:evn];}

- (void)mouseDown:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSLeftMouseDownMask	msg: LEFT_MOUSE_DOWN	]) ?
	self : [super mouseDown:evn];}

- (void)mouseDragged:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSLeftMouseDraggedMask	msg: LEFT_MOUSE_DRAG	]) ?
	self : [super mouseDragged:evn];}

- (void)mouseEntered:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSMouseEnteredMask	msg: MOUSE_ENTERED	]) ?
	self : [super mouseEntered:evn];}

- (void)mouseExited:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSMouseExitedMask	msg: MOUSE_EXITED	]) ?
	self : [super mouseExited:evn];}

- (void)mouseMoved:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSMouseMovedMask	msg: MOUSE_MOVED	]) ?
	self : [super mouseMoved:evn];}

- (void)mouseUp:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSLeftMouseUpMask	msg: LEFT_MOUSE_UP	]) ?
	self : [super mouseUp:evn];}

- (void)rightMouseDown:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSRightMouseDownMask	msg: RIGHT_MOUSE_DOWN	]) ?
	self : [super rightMouseDown:evn];}

- (void)rightMouseDragged:(NSEvent *)evn;
  {return
    ([self _chk: evn msk: NSRightMouseDraggedMask	msg: RIGHT_MOUSE_DRAG	]) ?
	self : [super rightMouseDragged:evn];}

- (void)rightMouseUp:(NSEvent *)evn 
  {return
    ([self _chk: evn msk: NSRightMouseUpMask	msg: RIGHT_MOUSE_UP	]) ?
	self : [super rightMouseUp:evn];}


/*===========================================================================*/
/* Ivar methods */
/*===========================================================================*/
/* A debug flag is available for enabling debug printouts or other diagnostic
   code
*/

- setDebug: (BOOL) flg	{debug = flg; return self;}
- (BOOL) debug		{return debug;}


/*===========================================================================*/
/* Action methods */
/*===========================================================================*/

- toggleDebug: sender {[self setDebug: ~debug]; return self;}

/*===========================================================================*/
/* Archive methods */
/*===========================================================================*/

- (void)encodeWithCoder:(NSCoder *)aCoder
  {	[super encodeWithCoder:aCoder];
	[aCoder encodeValuesOfObjCTypes:"ii:i", &firstMouse, &actionMask, &action,
		       &debug];
	[aCoder encodeConditionalObject:target];
}

- (id)initWithCoder:(NSCoder *)aDecoder
  {	[super initWithCoder:aDecoder];
	[aDecoder decodeValuesOfObjCTypes:"ii:i", &firstMouse, &actionMask, &action,
		     &debug];
	target = [[aDecoder decodeObject] retain];
	return self;
  }

/* NOTE: may need to set first responder here? */

#error ArchiverConversion: put the contents of your 'awake' method at the end of your 'initWithCoder:' method instead
- awake
  {
	return self;
  }

@end
