/*
 *     File: 	MiscTimedEntry.m
 *  Version:	1.0
 *   Author:	James T. Romano
 *  Summary:	This class defines a timer that will call back a 
 *				target object with a selectable message.  The timer may be
 *				set up to go off once, repeat a N times, or repeat indefinately.
 *				Also, the timer may be run in another thread.
 *
 *	Revision History:	
 *		Dec 1994	james	created
 *		Dec	1994	steveq	treaked
 *
 *	Copyright 1994	James T. Romano	james@hbsd-im.telerate.com
 */
#import <misckit/MiscTimedEntry.h>
#import <objc/Protocol.h>
#import <mach/mach_error.h>
#import <mach/cthreads.h>

/*
 *	sent in msg.id, MSG_DIE indicates that waiting thread should return pronto
 */
#define MSG_DIE 0
#define MSG_LIVE 1
/*
 *	message passed to self to interupt a wait
 */
typedef struct {
	msg_header_t header;
	msg_type_t type;
} ThreadMsg;

@interface MiscTimedEntry	(PrivateMethods)
- (BOOL)validateSelf;
- waitForNextEvent;
- sampleActionMethod:sender;
@end


@implementation MiscTimedEntry

/*
 *	Unadorned init - use default values
 */
- init
{
	return [self initWithTarget:nil action:(SEL)0 interval:500 data:NULL];
}

- initWithTarget:aTarget action:(SEL)anAction interval:(int)ival data:(void *)data
{
	[super init];
	[super freeAfterRun:NO];
	target = aTarget;
	action = anAction;
	interval = ival;
	userdata = data;
	port = PORT_NULL;
	repeatCount = -1;
	return self;
}

/*
 * When this is called, it will check to see
 * that target and action are set, if either, or both
 * are nil, then this method will return nil, therefore
 * you won't be able to run the timed entry object
 * Until you set target and action.
 * This method returns self, therefore this is the unique identifier
 * For this timed entry, and you can use this to keep track
 * Of the timed entry.
 */
- (void)run
{
	int retval;

	if (repeatCount == 1) {			/* we allow a single timer with no target */
		if (target &&  ([self validateSelf] == NO) ) {
				return nil;
		}
	} else
		if ([self validateSelf] == NO)
			return nil;

	/*
	 *	Allocate a port
	 */
	retval = port_allocate(task_self(),&port);
	if (retval != KERN_SUCCESS) {
		mach_error("port_allocate",retval);
		return nil;
	}
	/*
	 *	wait & send msg
	 */
	while (repeatCount) {
		if ([self waitForNextEvent] != nil)
			[target perform:action withObject:self];		/* notify our client */
		else
			break;
		if (repeatCount > 0)				/* don't alter if == -1 */
			repeatCount--;
	}
	
	port_deallocate(task_self(), port);			/* no longer need this... */
	port = PORT_NULL;
}	

/*
 *	do the waiting on the timer... we may be interuptee by a 'free' message
 */
- waitForNextEvent
{
	ThreadMsg tm;
	int retval;

	tm.header.msg_id = MSG_LIVE;
	tm.header.msg_simple = TRUE;
	tm.header.msg_size = sizeof(ThreadMsg);
	tm.header.msg_remote_port = PORT_NULL;
	tm.header.msg_local_port = [self port];
	
	retval = msg_receive(&tm.header,RCV_TIMEOUT,interval);
	
	/* 
	 *	normally, msg_receive() will time out. If a message IS received,
	 *	it indicates that someone is trying to stop the timer. 
	 */
	if (retval == RCV_SUCCESS) {
		if (tm.header.msg_id == MSG_DIE) {
			/*
			 *	Once this thread has been told to die, it will
			 *	send a non specfic message (a message that has no meaning)
			 *	which signals the sender that this thread is terminating.
			 */
			tm.header.msg_local_port = PORT_NULL;
			tm.header.msg_remote_port = port;
			retval = msg_send(&tm.header,MSG_OPTION_NONE,0);
			if (retval != KERN_SUCCESS)
				mach_error("Thread msg_send",retval);
		} else
			[self error:"Unexpected message recieved on port"];
		return nil;				/* indicate we were stopped */
	}
	return self;
}

- stop
{
	ThreadMsg tm;
	int retval;

	/*
	 *	use 'port' ivar as flag to indicate active status
	 */
	if (port != PORT_NULL) {
		tm.header.msg_simple = TRUE;
		tm.header.msg_local_port = PORT_NULL;
		tm.header.msg_remote_port = port;
		tm.header.msg_size = sizeof(ThreadMsg);
		tm.header.msg_id = MSG_DIE;
		/*
		 *	message waiting thread to wake it
		 */
		retval = msg_send(&tm.header,MSG_OPTION_NONE,0);
		if (retval != KERN_SUCCESS) 
			mach_error("timer msg_send",retval);
		else {
			/*
			 * This is not totally necessary, but I wanted to make sure 
			 * that the thread is responding to the msg_send, so we'll 
			 * wait until the thread tells us that it's going to die.
			 */
			tm.header.msg_local_port = port;
			tm.header.msg_remote_port = PORT_NULL;
			retval = msg_receive(&tm.header,MSG_OPTION_NONE,0);
			if (retval != KERN_SUCCESS)
				mach_error("MainObj msg_receive",retval);
		}
	}
	return self;
}


- (void)dealloc
{
	if (port != PORT_NULL)
		[self stop];
	{ [super dealloc]; return; };
}


/*
 *	this is here to provide a simple control for comparison
 */
- sampleActionMethod:sender		{	return nil;	}

/*
 *	verify the target-action setup before we run to avoid problems
 */
- (BOOL)validateSelf
{
	struct objc_method_description *actionDesc, *sampleDesc;

	if (repeatCount == 0)
		return NO;	
	if ( (target == nil) || (action == (SEL)0) )
		return NO;
	if ([target respondsToSelector:action] == NO)
		return NO;
	/*
	 *	check signature of action method, doing it in the following way 
	 *	is cheap, easy, (and it works).
	 */
	actionDesc = [target descriptionForMethod:action];
	if (actionDesc == NULL)
		return NO;
	sampleDesc = [self descriptionForMethod:@selector(sampleActionMethod:)] ;
	if (strcmp(actionDesc->types,sampleDesc->types) != 0)
		return NO;

	return YES;
}

/* This is more OpenStep-ish; for 3.2 compat we don't return void...
- (void)setTarget:sender { 
	target = sender;
}
- (void)setAction:(SEL)anAction { 
	action = anAction;
}
*/
- (void)setTarget:sender { 
	target = sender;
}
- (void)setAction:(SEL)anAction { 
	action = anAction;
}
- (void)setInterval:(int)anInterval { 
	interval = anInterval;
}
- (void)setPort:(int)aPort {
	port = aPort;
}
- (void)setRepeatCount:(int)count {
	repeatCount = count;
}
- (void)setUserData:(void *)userData { 
	userdata = userData;
}
- (void *)userData {
	return userdata;
}
- target {
	return target;
}
- (SEL)action {
	return action;
}
- (int)interval {
	return interval;
}
- (int)port {
	return port;
}
- (int)repeatCount {
	return repeatCount;
}


@end
